\ CLOCKINT.FTH - Timebase on 68307 using Timer 1

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

PLEASE NOTE OUR PHONE NUMBER CHANGE


See TIMEBASE.FTH for user documentation.

This code is for a periodic interrupt that
a) Updates a millisecond counter
b) Runs the DO-TIMERS word in TIMEBASE.FTH

All time periods are in milliseconds. Note that on a 32 bit system,
time periods must be less than 2^31-1 milliseconds, (say 596 hours or
24 days), whereas on a 16 bit system, time periods must be less than
2^15-1 milliseconds, say 32 seconds.

TICKS           \ -- addr ; defined in TIMEBASE.FTH
A VALUE holding the timebase ticker in milliseconds. At power up,
this variable is set to zero, and after START-TIMERS is used it is
incremented and will wrap around.

START-CLOCK     \ --
Start the internal time clock, and enable interrupts. This word can
also be used to restart the periodic interrupt without re-initialising the
timer system.

STOP-CLOCK      \ --
Stop the internal time clock periodic interrupt

START-TIMERS    \ --
Call the word INIT-TIMERS from TIMEBASE.FTH, start internal time clock,
and enable interrupts.

STOP-TIMERS     \ --
Stop the periodic interrupt
))


\ ************************
\ Clock interrupt handling
\ ************************

\ Note that the peripheral interrupt vector is set up by the 68307 interrupt
\ driven serial code in DRIVERS\SER68307I.FTH

[undefined] ticks [if]
0 value ticks
( *G Get current clock value in milliseconds. )
[then]

proc clockisr
( *G TIMEBASE periodic clock interrupt service routine. )
\ first process the clock interrupt
  # 3  $l MBASE TER1 +  b. move,	\ reset interrupt
  # tick-ms  $l ' ticks >body  l. addi,	\ update clock

\ run the clock handler word
  { d0-7 a0-6 } -(a7) l. movem, \ save unsaved registers
  # 0 a3 l. movea,              \ A3 must be zero
  # int-init-u0 a5 l. movea,    \ interrupt user area
  ' do-timers w. bsr,           \ run timer handler (use NOOP for testing)
  (a7)+ { d0-7 a0-6 } l. movem, \ restore registers

\ Restore  
  rte,
end-code
clockisr  start-vector ResetPIVR $0A + 4 * + !	\ set Timer 1 vector


\ ********************
\ Clock initialisation
\ ********************

((
The period of the timer depends on:
1) The clock source: XTAL/1 or XTAL/16 defined by the timer mode
   register. We always use XTAL/16.
2) The prescaler, dividing by PS=1..256
3) The TRR reference count, TRR

The required divisors for PS and TRR are given by:
  PS.TRR = sysclock.ms/(16*1000) for periods defined in milliseconds.
))
interpreter
: genTMRx	\ sysclk ms -- TRR PS+ctl
( *G Given the system clock speed in Hz, and the the required timer )
( ** period in milliseconds, GENTMRX returns the required initial )
( ** value for the TRRx register, and the required value for the TMRx )
( ** register. The initial values assume and set a clock input of )
( ** XTAL/16. )
  #16000 */				\ gives PS.TRR
  dup $10000 /mod			\ split into PS and TRR
  over 0= if				\ but can't have TRR=0
    2drop $20000 /mod			\   so try again
  else
    rot drop
  endif
  8 lshift $01D or			\ merge TMRx bits
;
target

system-speed tick-ms genTMRx		\ calculate values
  equ initTMR1				\ initial TMR1 value
  equ initTRR1				\ initial TRR1 value

: Start-Clock   \ -- ; start the periodic interrupt
( *G Start the internal time clock, and enable interrupts. This word can )
( ** also be used to restart the periodic interrupt without re-initialising the )
( ** timer system. )
  initTRR1 MBASE TRR1 + w!		\ set reference count
  initTMR1 MBASE TMR1 + w!		\ initialise timer
  MBASE PICR + w@			\ get current peripheral interrupts
  $0FFF and				\ zero timer 1 portion
  $A000 or				\ set for level 2
  MBASE PICR + w!			\ and put back
  ei
;

: Stop-Clock    \ -- ; disable periodic interrupt
( *G Stop the internal time clock periodic interrupt. )
  MBASE TMR1 + w@
  $FFEF and				\ reset bit 4, ORI
  MBASE TMR1 + w!
;

: start-timers  \ -- ; Start internal time clock
( *G Call the word INIT-TIMERS from TIMEBASE.FTH, and start the internal )
( ** time clock. )
  init-timers                           \ initialise time base system
  start-clock
;

: Stop-Timers   \ -- ; disable timer system
( *G Stop the periodic interrupt. )
  Stop-Clock
;

