\ CLOCKINT340.FTH - Timebase on 68340 System Integration Module

((
        Copyright (c) 1999
        MicroProcessor Engineering
        133 Hill Lane
        Southampton SO15 5AF
        England

        tel: +44 2380 631441
        fax: +44 2380 339691
        net: mpe@mpeltd.demon.co.uk
             tech-support@mpeltd.demon.co.uk
        web: www.mpeltd.demon.co.uk

See TIMEBASE.FTH for user documentation.

This code is for a periodic interrupt that
a) Updates a millisecond counter
b) Runs the DO-TIMERS word in TIMEBASE.FTH

All time periods are in milliseconds. Note that on a 32 bit system,
time periods must be less than 2^31-1 milliseconds, (say 596 hours or
24 days), whereas on a 16 bit system, time periods must be less than
2^15-1 milliseconds, say 32 seconds.

TICKS           \ -- addr ; defined in TIMEBASE.FTH
A VALUE holding the timebase ticker in milliseconds. At power up,
this variable is set to zero, and after START-TIMERS is used it is
incremented and will wrap around.

START-CLOCK     \ --
Start the internal time clock, and enable interrupts. This word can
also be used to restart the periodic interrupt without re-initialising the
timer system.

STOP-CLOCK      \ --
Stop the internal time clock periodic interrupt

START-TIMERS    \ --
Call the word INIT-TIMERS from TIMEBASE.FTH, start internal time clock,
and enable interrupts.

STOP-TIMERS     \ --
Stop the periodic interrupt
))


\ ***********************
\ Clock hardware handling
\ ***********************

MCR &PICR + equ PICR            \ timer control
MCR &PITR + equ PITR            \ timer period

$FF equ clock-mod               \ modulus of clock
$00 equ clock-prescale          \ 0 or 1
#31 equ clock-update            \ adjust this for your own hardware
                                \ it should be the number of milliseconds
                                \ of the clock interrupt period
$80 equ clock-vector            \ vector used for clock interrupt
1   equ clock-irl
clock-irl 8 <<n  clock-vector or
    equ init-picr               \ initial value of PICR
clock-prescale 8 <<n  clock-mod or
    equ init-pitr               \ initial value of PITR
0   equ term-pitr               \ stop value of PITR

[undefined] ticks [if]
0 value ticks
[then]

\ SBD 11th June 2002
\ Added re-entrancy guard because emit with timer on crashes
variable in-timer				\ re-entrancy guard

proc clockisr
\ first process the clock interrupt
  # clock-update  $l ' ticks >body  l. addi,	\ update clock
  in-timer $l tst,
  eq, if,
    \ run the clock handler word
    1 # in-timer $l move,
    { d0-7 a0-6 } -(a7) l. movem,	\ save unsaved registers, D0 already saved
    # 0 a3 l. movea,			\ A3 must be zero
    # int-init-u0 a5 l. movea,		\ interrupt user area
    ' do-timers w. bsr,			\ run timer handler (use NOOP for testing)
    (a7)+ { d0-7 a0-6 } l. movem,	\ restore registers
    0 # in-timer $l move,
  then,

\ Restore  
  rte,
end-code
clockisr  start-vector clock-vector 4 * + !     \ set vector

: Start-Clock   \ -- ; start the periodic interrupt
  0 in-timer !
  init-picr picr w!                     \ int 80h, level 1
  init-pitr pitr w!                     \ $FF, no prescale
  ei
;

: Stop-Clock    \ -- ; disable periodic interrupt
  term-pitr pitr w!
  0 in-timer !
;

: start-timers  \ -- ; Start internal time clock
  init-timers                           \ initialise time base system
  start-clock
;

: Stop-Timers   \ -- ; disable timer system
  Stop-Clock
;


