\ 68xxx code file

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

PLEASE NOTE OUR PHONE NUMBER CHANGE


20031216 SFP004 Added DOCGEN commentary for v6.2
20030109 SBD003 Corrected M+ and made it faster.
20001907 SFP002 Improved performance of EXECUTE
20000630 SFP001 Converted to v6.1 code model
		N.B. USER variables have a different structure to v6.0

Target Model: Subroutine Threaded Code with VFX optimisation
  A7 = return stack
  A6 = data stack
  A5 = user pointer
  A4 = locals pointer
  A3 = 0
  A2 = scratch
  A1 = scratch
  A0 = scratch

  D7 = TOS
  D6..D2 = scratch
  D1, D0 = scratch and optimiser temporary
))

only forth definitions
decimal

\ **********
\ *! code68k
\ **********
\ *T 68xxx code definitions
\ *S Register usage
\ *P On 68xxx CPUs the following register usage is the default:
\ *E   A7 = return stack
\ **   A6 = data stack
\ **   A5 = user pointer
\ **   A4 = locals pointer
\ **   A3 = 0
\ **   A2 = scratch
\ **   A1 = scratch
\ **   A0 = scratch
\ ** 
\ **   D7 = TOS
\ **   D6..D2 = scratch
\ **   D1, D0 = scratch and optimiser temporary
\ *P The VFX optimiser reserves D0 and D1 for internal operations.
\ ** CODE definitions must use D7 as TOS with NOS pointed to by
\ ** A6 as a full descending stack. D2..D6 and A0..A2 are
\ ** free for use by CODE definitions and need not be preserved or
\ ** restored. You should assume that any register can be affected
\ ** by other words.


\ ***********************************
\ *S Logical and relational operators
\ ***********************************

CODE AND        \ x1 x2 -- x3
\ *G Perform a logical AND between the top two stack items and retain
\ ** the result in top of stack.
  (A6)+ D7 AND,   next,
END-CODE

CODE OR         \ x1 x2 -- x3
\ *G Perform a logical OR between the top two stack items and retain
\ ** the result in top of stack.
  (A6)+ D7 OR,   next,
END-CODE

CODE XOR        \ x1 x2 -- x3
\ *G Perform a logical XOR between the top two stack items and retain
\ ** the result in top of stack.
  (A6)+ d0 move,
  d0 D7 EOR,
  next,
END-CODE

((
CODE NOT        \ x -- x'
  D7 NOT,   next,
END-CODE
))

CODE INVERT     \ x -- x'
\ *G Perform a bitwise NOT on the top stack item and retain result.
  D7 NOT,   next,
END-CODE

CODE MIN        \ n1 n2 -- n1|n2 
\ *G Given two data stack items preserve only the smaller.
  (A6)+ D0 MOVE,  D0 D7 CMP,   MINX BLT,        \ tos-nos < 0 == tos < nos
  D0 D7 MOVE,
L: MINX
  next,
END-CODE

CODE MAX        \ n1 n2 -- n1|n2
\ *G Given two data stack items preserve only the larger.
  (A6)+ D0 MOVE,   D0 D7 CMP,   MAXX BGT,       \ tos-nos > 0 == tos > nos
  D0 D7 MOVE,
L: MAXX
  next,
END-CODE

CODE 0=         \ x -- flag
\ *G Compare the top stack item with 0 and return TRUE if equals.
  d7 TST,   D7 SEQ,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE 0<>        \ x -- flag
\ *G Compare the top stack item with 0 and return TRUE if not-equal.
  d7 TST,   D7 SNE,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE 0<         \ x -- flag
\ *G Return TRUE if the top of stack is less-than-zero.
  d7 TST,   D7 SMI,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE 0>         \ x -- flag
\ *G Return TRUE if the top of stack is greater-than-zero.
  d7 TST,   D7 SGT,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE =          \ x1 x2 -- flag
\ *G Return TRUE if the two topmost stack items are equal.
 (a6)+ d7 cmp,   D7 SEQ,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE <>         \ x1 x2 -- flag
\ *G Return TRUE if the two topmost stack items are different.
  (A6)+ D7 CMP,   D7 SNE,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE <          \ x1 x2 -- flag
\ *G Return TRUE if n1 is less than n2.
  (A6)+ d7 CMP,   D7 SGT,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE >          \ x1 x2 -- flag
\ *G Return TRUE if n1 is greater than n2.
  (A6)+ d7 CMP,   D7 SLT,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE <=         \ x1 x2 -- flag
\ *G Return TRUE if n1 is less than or equal to n2.
  (A6)+ d7 CMP,   D7 SGE,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE >=         \ x1 x2 -- flag
\ *G Return TRUE if n1 is greater than or equal to n2.
  (A6)+ d7 CMP,   D7 SLE,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE U>         \ u2 u2 -- flag
\ *G An UNSIGNED version of >.
  (A6)+ d0 move,   d7 d0 cmp,   D7 SHI,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE U<         \ u1 u2 -- flag
\ *G An UNSIGNED version of <.
  (A6)+ D7 CMP,   D7 SHI,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE DU<        \ ud1 ud2 -- flag
\ *G Returns true if ud1 (unsigned double) is less than ud2. 
  (a6)+ d2 move,                                        \ get d2 low
  (a6)+ d1 move,                                        \ get d1 high
  (a6)+ d0 move,                                        \ get d1 low
  d2 d0 sub,                                            \ d1l - d2l
  d7 d1 subx,                                           \ d1h - d2h
  d7 scs,                                               \ set tos
  d7 w. ext,
  d7 l. ext,
  next,
END-CODE

CODE D0<        \ d -- flag
\ *G Returns true if signed double d is less than zero.
  (a6)+ d0 move,                                        \ lose low
  d7 tst,
  d7 smi,
  d7 w. ext,
  d7 l. ext,
  next,
END-CODE

CODE D0=        \ xd -- flag
\ *G Returns true if xd is 0.
  (a6)+ d0 move,                                        \ get low
  d0 d7 or, eq,
if,
  TRUE  # d7 moveq,
else,
  FALSE # d7 moveq,
endif,
  next,
END-CODE

CODE D=         \ xd1 xd2 -- flag
\ *G Return TRUE if the two double numbers are equal.
  (a6)+ d2 move,                        \ get d2 low
  (a6)+ d1 move,                        \ get d1 high
  (a6)+ d0 move,                        \ get d1 low
  d0 d2 sub,                            \ d2l - d1l
  d1 d7 subx,                           \ d2h - d1h
  d2 d7 or,                             \ combine hi and low order
  d7 seq,                               \ set tos true if result is 0
  d7 w. ext,
  d7 l. ext,
  next,
END-CODE

CODE D<         \ d1 d2 -- flag
\ *G Return TRUE if the double number d1 is < the double number d2.
  (a6)+ d2 move,                        \ get d2l
  (a6)+ d1 move,                        \ get d1h
  (a6)+ d0 move,                        \ get d1l
  d2 d0 sub,                            \ d1l - d2l
  d7 d1 subx,                           \ d1h - d2h
  d1 l. tst,                            \ test bit 63, effectively
  d7 slt,                               \ set tos appropriately
  d7 w. ext,
  d7 l. ext,
  next,
END-CODE

CODE DMAX       \ d1 d2 -- d3 ; d3=max of d1/d2
\ *G Return the maximum double number from the two supplied.
  (a6)+ d1 move,                        \ get d2 low
  (a6)+ d0 move,                        \ get d1 high
  d0 d7 cmp,
  DMAX1 bgt,
  DMAX2 blt,
  (a6) d1 l. cmp,                       \ check low
  DMAX2 blt,
L: DMAX1
  d1 (a6) move,                         \ return d2 is highest
  next,
L: DMAX2
  d0 d7 move,                           \ return d1 is highest
  next,
END-CODE

CODE DMIN       \ d1 d2 -- d3 ; d3=min of d1/d2
\ *G Return the minimum double number from the two supplied.
  (a6)+ d1 move,                        \ get d2 low
  (a6)+ d0 move,                        \ get d1 high
  d0 d7 cmp,
  DMIN1 blt,
  DMIN2 bgt,
  (a6) d1 cmp,                          \ check low
  DMAX2 bgt,
L: DMIN1
  d1 (a6) move,                         \ return d2 is highest
  next,
L: DMIN2
  d0 d7 move,                           \ return d1 is highest
  next,
END-CODE

CODE WITHIN?    \ n1 n2 n3 -- flag
\ *G Return TRUE if N1 is within the range N2..N3.
\ ** This word uses signed arithmetic.
  d7 d0 move,                           \ d2=x, d1=a, d0=b
  (A6)+ D1 MOVE,
  (A6)+ D2 MOVE,
  d7 CLR,
  D1 D2 CMP,   WITH1 BLT,               \ x-a
  D0 D2 CMP,   WITH1 BGT,               \ x-b
  1 # d7 SUBQ,
L: WITH1
  next,
END-CODE

CODE WITHIN     \ n1|u1 n2|u2 n3|u3 -- flag
\ *G The ANS version of WITHIN?.
\ ** This word uses unsigned arithmetic, so that signed compares are 
\ ** treated as existing on a number circle.
  d7 d0 move,                           \ d2=x, d1=a, d0=b
  (a6)+ d1 move,
  (a6)+ d2 move,
  d1 d0 l. sub,
  d1 d2 l. sub,
  d0 d2 l. sub,
  d2 d2 l. subx,
  d2 d7 l. move,
  next,
END-CODE

((
CODE <<         \ x1 u -- x1<<u                                                 
  (a6)+ d0 l. move,
  d7 d0 l. lsl,
  d0 d7 l. move,
  next,
END-CODE

CODE >>         \ x1 u -- x1>>u                                                 
  (a6)+ d0 l. move,
  d7 d0 l. lsr,
  d0 d7 l. move,
  next,
END-CODE
))

CODE LSHIFT     \ x1 u -- x1<<u
\ *G Logically shift X1 by U bits left.
  (a6)+ d0 l. move,
  d7 d0 l. lsl,
  d0 d7 l. move,
  next,
END-CODE

CODE RSHIFT     \ x1 u -- x1>>u
\ *G Logically shift X1 by U bits right.
  (a6)+ d0 l. move,
  d7 d0 l. lsr,
  d0 d7 l. move,
  next,
END-CODE


\ ***************
\ *S Control flow
\ ***************

((
CODE EXIT       \ -- ; R: next-sys --
  (a7)+ a0 movea,                       \ discard first level RTS
  next,                                 \ do second level RTS
END-CODE
))

CODE EXECUTE    \ xt --
\ *G Execute the code described by the XT. This is a Forth equivalent
\ ** to an assembler JSR/CALL instruction.
  d7 a0 movea,                          \ move CFA
  (a6)+ d7 move,                        \ get new TOS
  (a0) jmp,                             \ execute CFA
END-CODE

internal

CODE BRANCH     \ --
\ *G The run time action of unconditional branches compiled on the target. 
l: bran1
  (a7)+ a0 movea,                       \ get pointer for branch
  (a0) a1 movea,                        \ get destination from pointer
  (a1) jmp,                             \ do BRANCH
END-CODE

CODE ?BRANCH    \ n --
\ *G The run time action of conditional branches compiled on the target. 
l: 0bran1
  d7 TST,   0bran2 beq,                 \ test TOS
  (a6)+ d7 move,                        \ get new TOS
  (a7)+ a0 movea,                       \ get BRANCH ADDR
  # 4 a0 addq,                          \ skip past BRANCH addr
  (a0) jmp,                             \ jump back
l: 0bran2
  (a6)+ d7 move,                        \ get new TOS
  bran1 bra,                            \ carry out BRANCH
END-CODE

CODE (OF)       \ n1 n2 --  n1|n1 n2
\ *G The run time action of OF compiled on the target. 
  (a6) d7 cmp,   pof1 beq,              \ compare TOS and NOS
  (a6)+ d7 move,                        \ different: n1 becomes TOS
  (a7)+ a0 movea,                       \ get BRANCH address
  (a0) a1 movea,  (a1) jmp,             \ and take BRANCH
l: pof1
  # 4 a6 addq,   (a6)+ d7 move,         \ drop two items
  (a7)+ a0 movea,                       \ get BRANCH addr
  # 4 a0 addq,   (a0) jmp,              \ skip past BRANCH addr
END-CODE

CODE (LOOP)     \ --
\ *G The run time action of LOOP compiled on the target. 
  (a7)+ a0 movea,                       \ get BRANCH addr
  1 # (A7) ADDQ,   ploop1 BVC,          \ no overflow, take branch
  $0C # A7 W. ADDA,                     \ drop 3 items from return stack
  # 4 a0 addq,  (a0) jmp,               \ skip branch
l: ploop1
  (a0) a1 movea,  (a1) jmp,             \ take BRANCH
END-CODE

CODE (+LOOP)    \ n --
\ *G The run time action of +LOOP compiled on the target. 
  (a7)+ a0 movea,                       \ get BRANCH addr
  d7 (A7) ADD,   pploop1 BVC,           \ no overflow, take branch
  (a6)+ d7 move,                        \ get new TOS
  $0C # A7 W. ADDA,                     \ drop 3 items from return stack
  # 4 a0 addq,  (a0) jmp,               \ skip over target address
l: pploop1
  (a6)+ d7 move,                        \ get new TOS
  (a0) a1 movea,  (a1) jmp,             \ take branch
END-CODE

CODE (DO)       \ limit index --
\ *G The run time action of DO compiled on the target. 
  (A6)+ D1 MOVE,                        \ get limit
L: PDO
  (a7)+ a0 movea,                       \ get BRANCH address
  (A0) -(A7) MOVE,   $80000000 # D1 ADD,        \ push LEAVE addr on ret. stk
  D1 -(A7) MOVE,   D1 D7 SUB,   D7 -(A7) MOVE,  \ then also limit, index
  (a6)+ d7 move,                        \ update TOS
  4 # a0 addq,  (a0) jmp,               \ continue
END-CODE

CODE (?DO)      \ limit index --
\ *G The run time action of ?DO compiled on the target. 
  (A6)+ D1 MOVE,   D1 D7 CMP,   pdo bne,        \ take DO ?
  (a6)+ d7 move,                        \ update TOS
  (a7)+ a0 movea,                       \ get BRANCH addr
  (a0) a1 movea,                        \ and take LEAVE branch
  (a1) jmp,
END-CODE

external

CODE LEAVE	\ -- ; N.B. now non-immediate
\ *G Remove the current DO..LOOP parameters and jump to the
\ ** end of the DO..LOOP structure.
L: PLEAVE
  $0C # A7 W. ADDA,                     \ drop 3 items from return stack
  (a7)+ a0 movea,  (a0) jmp,            \ take LEAVE branch
END-CODE

CODE ?LEAVE	\ flag -- ; N.B. now non-immediate
\ *G If flag is non-zero, remove the current DO..LOOP parameters
\ ** and jump to the end of the DO..LOOP structure.
  d7 TST,   PQLEAVE1 BNE,               \ test TOS
  (a6)+ d7 move,   next,                \ 0, just update TOS
l: pqleave1                              
  (a6)+ d7 move,                        \ non-zero, update TOS
  pleave bra,                           \ join LEAVE
END-CODE

CODE I          \ -- n
\ *G Return the current index of the inner-most DO..LOOP.
  d7 -(a6) move,                        \ save old TOS
  (a7)+ a0 movea,                       \ get RTS addr
  (A7) D7 MOVE,   4 D(A7) D7 ADD,       \ form index
  (a0) jmp,                             \ jump to return addr.
END-CODE

CODE J          \ -- n
\ *G Return the current index of the second DO..LOOP.
  d7 -(a6) move,                        \ save old TOS
  (a7)+ a0 movea,                       \ get RTS addr
  $0C D(A7) D7 MOVE,   $10 D(A7) D7 ADD,        \ form index
  (a0) jmp,                             \ jump to return addr.
END-CODE

CODE UNLOOP     \ -- ; R: loop-sys --
\ *G Remove the DO..LOOP control parameters from the return stack.
  (a7)+ a0 movea,                       \ preserve return address
  $0C # A7 W. ADDA,                     \ drop 3 items from return stack
  (a0) jmp,                             \ jump to return address
END-CODE


\ *************
\ *S Arithmetic
\ *************
\ *P Note that two sets of the words UM* M* and * exist which are
\ ** selected by the equate CPU32?. 
(( no longer needed
CODE LIT        \ -- x
  d7 -(a6) move,                                        \ save TOS
  (a7)+ a0 movea,                                       \ get new one
  (a0) d7 move,                                         \
  # 4 a0 adda,                                          \ get new RETURN address
  (a0) jmp,                                             \ JUMP to it
END-CODE
))

cpu32? [if]                                                                            

CODE UM*        \ u1 u2 -- ud
\ *G Perform unsigned-multiply between two numbers and return double
\ ** result.
  (a6) d0: d7 l. mulu,                  \ perform multiply
  d7 (a6) l. move,   d0 d7 l. move,     \ order the stack
  next,   end-code

CODE *          \ n1 n2 -- n3
\ *G Standard signed multiply. N3 = n1 * n2.
  (a6)+ d7 l. muls,                     \ perform multiply
  next,   end-code

CODE M*         \ n1 n2 -- d3
\ *G Signed multiply yielding double result.
  (a6) d0: d7 l. muls,                  \ perform multiply
  d7 (a6) l. move,   d0 d7 l. move,     \ order the stack
  next,   end-code

[else]

CODE UM*        \ u1 u2 -- ud ; MUST preserve D5
\ *G The unsigned 32 * 32 -> 64 bit multiply primitive for CPUs
\ ** without the long multiply instructions.
  d7 d6 move,
  # $ffff0000 d6 andi,                  \ use fast version if top 16 bits 0
  eq, if,
    (a6) w. tst,  um*fast beq,          \ are both 0?
  endif,

  d7 -(a6) move,                                \ easier in memory
  D4 CLR,                                       \ used for fast 0
  6 D(A6) D0 W. MOVE,   2 D(A6) D0 W. MULU,     \ d0 = u1.l * u2.l              
  4 D(A6) D1 W. MOVE,   (A6)    D1 W. MULU,     \ d1 = u1.h * u2.h              
  4 D(A6) D2 W. MOVE,   2 D(A6) D2 W. MULU,     \ d2 = u1.h * u2.l              
  6 D(A6) D3 W. MOVE,   (A6)    D3 W. MULU,     \ d3 = u1.l * u2.h              
  D1 (A6) MOVE,   D0 4 D(A6) MOVE,
  (A6) D1 W. MOVE,   2 D(A6) D0 L. MOVE,
  D2 D0 L. ADD,   D4 D1 W. ADDX,                \ used as ADDX # 0
  D3 D0 L. ADD,   D4 D1 W. ADDX,
  D0 2 D(A6) L. MOVE,   D1 (A6) W. MOVE,
  (a6)+ d7 move,
  mul-exit bra,

l: um*fast
  2 D(A6) D7 W. MULU,                           \ multiply low words            
  d7 (a6) move,
  # 0 d7 moveq,                                 \ clear high word

l: mul-exit
  next,
END-CODE

: *             \ n1 n2 -- n3
\ *G Standard signed multiply. N3 = n1 * n2.
  um* drop
;

CODE M*         \ n1 n2 -- d ; relies on UM* leaving D5 alone
\ *G Signed multiply yielding double result.
  d7 d5 l. move,
  (a6) d1 move,  d1 d5 eor,             \ form sign of result
  d7 l. tst,                            \ form |n2|
  mi, if,  d7 l. neg,  endif,
  d1 l. tst,                            \ form |n1|
  mi, if,  (a6) neg,  endif,
  ' um* w. bsr,
  d5 l. tst,
  mi, if,
    D2 CLR,   (A6) D1 MOVE,             \ d in d7:d1
    D7 NOT,   D1 NOT,
    1 # D1 ADDQ,   D2 D7 ADDX,
    D1 (A6) MOVE,
  endif,
  next,
END-CODE

[then]


\ *P The division words select different forms according to the
\ ** setting of CPU32?.

cpu32? [if]                                                                            

CODE UM/MOD   \ ud u -- urem uquot                                              
\ *G Perform unsigned division of double number UD by single number U
\ ** and return remainder and quotient.
  (A6)+ D0 MOVE,                        \ d7=divisor                              
  (A6)+ D1 MOVE,                        \ d0:d1=dividend                          
  d7 d0: d1 l. divu,                    \ divide
  trapv,                                \ overflow check
  d0 -(a6) l. move,                     \ push remainder                          
  d1 d7 l. move,                        \ copy quotient                           
  next,                                                                         
end-code                                                                        

CODE FM/MOD     \ d1 n2 -- rem quot ; floored division
\ *G Perform a signed division of double number D1 by single number N2
\ ** and return remainder and quotient using floored division. See the
\ ** ANS Forth specification for more details of floored division.
  (a6)+ d0 move,
  (a6)+ d1 move,                        \ d0:d1=dividend
L: FM/MOD1
  d0 d3 move,
  d7 d3 eor,                            \ d3=sign of quotient
  d7 d4 move,                           \ d4=sign of divisor
  mi, if,  d7 neg,  endif,              \ form |divisor|
  d0 tst,                               \ form |dividend|
  mi, if,
    # 0 d5 moveq,                       \ d5=0 for fast addition
    d0 not,
    d1 not,                             \ invert dividend
    # 1 d1 addq,
    d5 d0 addx,                         \ and add 1
  endif,

  d7 d0: d1 l. divu,                    \ perform division
  trapv,

l: fm/mod2
  d4 tst,                               \ remainder takes sign of divisor
  mi, if,  d0 neg,  endif,
  d3 tst,                               \ set sign of quotient
  mi, if,
    d1 neg,
    d0 tst,                             \ quot -> -ve, remainder non zero?
    ne, if,
      # 1 d1 subq,                      \ quot := quot - 1
      d0 d4 sub,                        \ rem := divisor - rem
      d4 d0 move,
    endif,
  endif,
  d0 -(a6) move,                        \ return remainder
  d1 d7 move,                           \ return quotient
  next,   
END-CODE

CODE SM/REM     \ d1 n2 -- rem quot ; symmetric division
\ *G Perform a signed division of double number D1 by single number N2
\ ** and return remainder and quotient using symmetric (normal) division.
  (a6)+ d0 move,
  (a6)+ d1 move,                        \ d0:d1=dividend
  d7 d0: d1 l. divs,                    \ divide
  trapv,
  d0 -(a6) move,                        \ return remainder
  d1 d7 move,                           \ return quotient
  next,   
END-CODE

code /mod       \ n1 n2 -- rem quot ; 6.1.0240
\ *G Signed division of N1 by N2 single-precision yielding remainder
\ ** and quotient.
  (a6)+ d1 move,                        \ d1=dividend
  d7 d0: d1 l. divsl,                   \ divide
  trapv,
  d0 -(a6) move,                        \ return remainder
  d1 d7 move,                           \ return quotient
  next,   
end-code

code /          \ n1 n2 -- quot ; 6.1.0230
\ *G Standard signed division operator. n3 = n1/n2.
  (a6)+ d1 move,                        \ d1=dividend
  d7 d0: d1 l. divsl,                   \ divide
  trapv,
  d1 d7 move,                           \ return quotient
  next,   
end-code

code mod        \ n1 n2 -- rem ; 6.1.1890
\ *G Return remainder of division of N1 by N2. n3 = n1 mod n2.
  (a6)+ d1 move,                        \ d1=dividend
  d7 d0: d1 l. divsl,                   \ divide
  trapv,
  d0 d7 move,                           \ return remainder
  next,   
end-code

[else]

asmcode
l: udiv64/32	\ d0:d1 / d7 ; 64/32 unsigned divide -> d1=quot, d0=rem
		\ N.B. D3 must be preserved
  d0 d7 l. cmp,   udiv64/32-oflo bls,   \ check for overflow                      \ MSD011
  
  d3 -(a7) l. move,                     \ preserve D3

  d0 l. tst,                            \ can use chip instruction if d0=0, d7.h=0
  udiv64/32-slow w. bne,
  d7 d3 l. move,
  # $FFFF0000 d3 l. andi,
  udiv64/32-slow w. bne,
  d1 d3 l. move,			\ check for overflow d1:h d7.l
  d3 swap,
  d3 d7 w. cmp,   udiv64/32-fast bhi,

l: udiv64/32-slow
  $20 # d2 moveq,                       \ set loop counter
  begin,
    d1 d1 l. add,  d0 d0 l. addx,	\ 64 bit left shift
    d3 d3 l. subx,			\ save carry flag
    d7 d0 l. sub,			\ trial subtraction, carry clear?
    cc, if,                             \ y:
      1 # d1 l. addq,			\   increment quotient
    else,                               \ n:
      d3 d3 l. or,                      \   carry from shift set?
      ne, if,                           \   y:
        # 1 d1 l. addq,                 \     increment quotient
      else,                             \   n:
        d7 d0 l. add,                   \     undo subtraction
      endif,
    endif,
    1 # d2 l. subq,
  eq, until,

  (a7)+ d3 l. move,                     \ restore D3
  rts,

l: udiv64/32-fast
  d7 d1 w. divu,                        \ machine division
  d1 d0 l. move,                        \ copy result
  # $FFFF d1 l. andi,                   \ extract quotient
  # 8 d0 l. lsr,  # 8 d0 l. lsr,        \ extract remainder
  
  (a7)+ d3 l. move,                     \ restore D3
  rts,

l: udiv64/32-oflo
  d0 l. clr,                            \ set remainder
  # $FFFFFFFF d1 l. move,               \ set quotient
  rts,
end-code

CODE UM/MOD     \ ud u -- urem uquot
  (A6)+ D0 MOVE,                                      \ d7=divisor
  (A6)+ D1 MOVE,                                      \ d0:d1=dividend
  UDIV64/32 BSR,
  D0 -(A6) MOVE,                                      \ remainder
  D1 d7 MOVE,                                         \ quotient
  next,
END-CODE

CODE FM/MOD     \ d1 n2 -- rem quot ; symmetric division
  (a6)+ d0 move,
  (a6)+ d1 move,                        \ d0:d1=dividend
L: FM/MOD1
  d0 d3 move,
  d7 d3 eor,                            \ d3=sign of quotient
  d7 d4 move,                           \ d4=sign of divisor
  mi, if,  d7 neg,  endif,              \ form |divisor|
  d0 tst,                               \ form |dividend|
  mi, if,
    # 0 d5 moveq,                       \ d5=0 for fast addition
    d0 not,
    d1 not,                             \ invert dividend
    # 1 d1 addq,
    d5 d0 addx,                         \ and add 1
  endif,
  UDIV64/32 w. bsr,                     \ call divide; d0=rem, d1=quot
  d4 tst,                               \ remainder takes sign of divisor
  mi, if,  d0 neg,  endif,
  d3 tst,                               \ set sign of quotient
  mi, if,
    d1 neg,
    d0 tst,                             \ quot -> -ve, remainder non zero?
    ne, if,
      # 1 d1 subq,                      \ quot := quot - 1
      d0 d4 sub,                        \ rem := divisor - rem
      d4 d0 move,
    endif,
  endif,
  d0 -(a6) move,                        \ return remainder
  d1 d7 move,                           \ return quotient
  next,   
END-CODE

CODE SM/REM     \ d1 n2 -- rem quot ; symmetric division
  (a6)+ d0 move,
  (a6)+ d1 move,                        \ d0:d1=dividend
L: sm/rem1
  d0 d3 move,
  d7 d3 eor,                            \ d3=sign of quotient
  d0 d4 move,                           \ d4=sign of remainder/dividend
  d7 tst,
  mi, if,  d7 neg,  endif,              \ form |divisor|
  d0 tst,                               \ form |dividend|
  mi, if,
    # 0 d5 moveq,                       \ d5=0 for fast addition
    d0 not,
    d1 not,                             \ invert dividend
    # 1 d1 addq,
    d5 d0 addx,                         \ and add 1
  endif,
  UDIV64/32 w. bsr,                     \ call divide; d0=rem, d1=quot
  d4 tst,                               \ remainder takes sign of dividend
  mi, if,  d0 neg,  endif,
  d3 tst,                               \ set sign of quotient
  mi, if,  d1 neg,  endif,
  d0 -(a6) move,                        \ return remainder
  d1 d7 move,                           \ return quotient
  next,   
END-CODE

: /mod          \ n1 n2 -- rem quot ; 6.1.0240
  >r s>d r> sm/rem
;

: /             \ n1 n2 -- n3
  >r s>d r> sm/rem nip
;

: MOD           \ n1 n2 -- n3
  >r s>d r> sm/rem drop
;

[endif]

((
: m/mod         \ d n -- rem quot
  sm/rem
;
))

: */MOD         \ n1 n2 n3 -- rem quot
\ *G Multiply n1 by n2 to give a double precision result, and then
\ ** divide it by n3 returning the remainder and quotient. The point
\ ** of this operation is to avoid loss of precision.
  >r m* r> sm/rem
;

: */            \ n1 n2 n3 -- n4
\ *G Multiply n1 by n2 to give a double precision result, and then
\ ** divide it by n3 returning the quotient. The point
\ ** of this operation is to avoid loss of precision.
  >r m* r> sm/rem nip
;

: M/            \ d n1 -- n2
\ *G Signed divide of a double by a single integer.
  sm/rem nip
;

: MU/MOD        \ d n  --  rem d#quot
\ *G Perform an unsigned divide of a double by a single, returning
\ ** a single remainder and a double quotient.
  >r  0  r@  um/mod  r>  swap  >r  um/mod  r>
;

internal
CODE TNEGATE    \ t1 -- -t1 ; negate value of triple cell number
\ *G Negate a triple (3 cell) number.
  4 d(a6) neg,
  (a6) negx,
  d7 negx,
  next,
END-CODE
external

: M*/           \ d1 n1 +n2 -- d2
\ *G Multiply double d1 by single n1 to give a triple precision
\ ** result, and then divide it by n2 returning the quotient.
\ ** The point of this operation is to avoid loss of precision.
  ?dup 0= if				\ check for divide by 0
    xor 0< >r				\ form sign of result
    drop  $FFFFFFFF $7FFFFFFF		\ from double maxint
    r> if  dnegate  endif		\ apply sign
    exit
  endif
  >r
  rot over m*				\ -- d1h n1 t0lm        R: -- +n2
  2swap m*				\ -- t0lm t1mh          R: -- +n2
  rot rot				\ -- t0l t1h t0m t1m    R: -- +n2
  0 tuck d+				\ -- t0l t1h tzm carry  R: -- +n2
  rot +					\ -- tl tm th           R: -- +n2
  dup 0< if				\ negative triple
    tnegate
    r@ mu/mod				\ -- tl r d0            R: -- +n2
    2drop				\ -- tl r               R: -- +n2
    r>					\ -- tl r +n2
    mu/mod				\ -- r d2
    dnegate				\ -- r -d2
    rot drop				\ -- -d2
  else					\ positive triple
    r@ mu/mod				\ -- tl r d0            R: -- +n2
    2drop				\ -- tl r               R: -- +n2
    r>					\ -- tl r +n2
    mu/mod				\ -- r d2
    rot drop				\ -- d2
  then
;

CODE M+         \ d1|ud1 n -- d2|ud2
\ *G Add double d1 to sign extended single n to form double d2.
   d7 d1 move,				\ sign extend n in D7:D1
   d7 smi,
   cpu32? if( d7 l. extb, )else( d7 w. ext,  d7 l. ext, )endif
   (a6)+ d0 move,                        \ get d1 high
   d1 (a6) add,                          \ SBD n1 + d1l and push
   d0 d7 addx,                           \ d2h + sex to TOS
   next,
END-CODE

CODE D+         \ d1 d2 -- d3
\ *G Add two double precision integers.
  (A6)+ D1 MOVE,   (A6)+ D2 MOVE,       \ d7:d1 + d2:(a7)
  (a6)+ D1 ADD,   D2 D7 ADDX,           \ -> d7:d1
  D1 -(A6) MOVE,   
  next,
END-CODE

CODE D-         \ d1 d2 -- d3 
\ *G Subtract two double precision integers. D3=D1-D2.
  (A6)+ D0 move,			\ D0 = d2.l
  (A6)+ D1 move,			\ D1 = d1.h
  D0 (A6) sub,				\ d1.l - d2.l
  D7 D1 subx,				\ d1.h - d2.h
  D1 D7 move,
  next,
END-CODE

CODE 1+         \ n1|u1 -- n2|u2
\ *G Add one to top-of stack.
  1 # d7 ADDQ,   next,
END-CODE

CODE 2+         \ n1|u1 -- n2|u2
\ *G Add two to top-of stack.
  2 # d7 ADDQ,   next,
END-CODE

CODE 4+         \ n1|u1 -- n2|u2
\ *G Add four to top-of stack.
  4 # d7 ADDQ,   next,
END-CODE

CODE 1-         \ n1|u1 -- n2|u2
\ *G Subtract one from top-of stack.
  1 # d7 subq,   next,
END-CODE

CODE 2-         \ n1|u1 -- n2|u2
\ *G Subtract two from top-of stack.
  2 # d7 subq,   next,
END-CODE

CODE 4-         \ n1|u1 -- n2|u2
\ *G Subtract four from top-of stack.
  4 # d7 subq,   next,
END-CODE

CODE 2*         \ x1 -- x2
\ *G Signed multiply top of stack by 2.
  1 # D7 ASL,   next,
END-CODE

CODE 4*         \ x1 -- x2
\ *G Signed multiply top of stack by 4.
  2 # D7 ASL,   next,
END-CODE

code 2/         \ x1 -- x2
\ *G Signed divide top of stack by 2 by arithmetic right shift.
  1 # d7 asr,   next,
END-CODE

code U2/        \ x1 -- x2
\ *G Unsigned divide top of stack by 2 by logical right shift.
  1 # d7 lsr,   next,
END-CODE

code 4/         \ x1 -- x2
\ *G Signed divide top of stack by 4 by arithmetic right shift.
  2 # d7 asr,   next,
END-CODE

code U4/        \ x1 -- x2
\ *G Unsigned divide top of stack by 4 by logical right shift.
  2 # d7 lsr,   next,
END-CODE

CODE +          \ n1|u1 n2|u2 -- n3|u3
\ *G Add two single precision integer numbers.
  (A6)+ d7 add,   next,
END-CODE

CODE -          \ n1|u1 n2|u2 -- n3|u3
\ *G Subtract two single precision integer numbers. N3|u3=n1|u1-n2|u2.
  (a6)+ d0 move,   d7 d0 SUB,   d0 d7 move,   next,
END-CODE

CODE NEGATE     \ n1 -- n2
\ *G Negate a single precision integer number.
L: NEG1         d7 NEG,   next,
END-CODE

CODE DNEGATE    \ d1 -- d2 
\ *G Negate a double number.
L: DNEG1
  D2 CLR,   (A6) D1 MOVE,                               \ d in d7:d1
  D7 NOT,   D1 NOT,   1 # D1 ADDQ,   D2 D7 ADDX,
  D1 (A6) MOVE,   
  next,
END-CODE

CODE ?NEGATE    \ n1 flag -- n1|n2
\ *G If flag is negative, then negate n1.
  d7 tst,
  mi, if,  (a6)+ d7 move,   d7 neg,   next,   endif,
  (a6)+ d7 move,   next,
END-CODE

CODE ?DNEGATE   \ d1 flag -- d1|d2
\ *G If flag is negative, then negate d1.
  d7 tst,
  mi, if,  (a6)+ d7 move,   dneg1 bra,  endif,
  (a6)+ d7 move,   next,
END-CODE

CODE ABS        \ n -- u
\ *G If n is negative, return its positive equivalent (absolute value).
  d7 tst,
  mi, if,  d7 neg,  endif,
  next,
END-CODE

CODE DABS       \ d -- ud
\ *G If d is negative, return its positive equivalent (absolute value).
  d7 tst,   dneg1 bmi,   next,
END-CODE

CODE D2*        \ xd1 -- xd2
\ *G Multiply the given double number by two.
  (a6)+ d0 move,                                        \ get low
  1 # d0 asl,
  1 # d7 roxl,
  d0 -(a6) move,
  next,
END-CODE

CODE D2/        \ xd1 -- xd2
\ *G Divide the given double number by two.
  (a6)+ d0 move,                                        \ get low
  1 # d7 asr,
  1 # d0 roxr,
  d0 -(a6) move,
  next,
END-CODE


\ *********************
\ *S Stack manipulation
\ *********************

CODE NIP        \ x1 x2 -- x2
\ *G Dispose of the second item on the data stack.
  # 4 a6 addq,    next,
END-CODE

CODE TUCK       \ x1 x2 -- x2 x1 x2
\ *G Insert a copy of the top data stack item underneath the current
\ ** second item.
  (A6)+ D1 MOVE,
  D7 -(A6) MOVE,   D1 -(A6) MOVE,
  next,
END-CODE

CODE PICK       \ xu .. x0 u -- xu .. x0 xu
\ *G Get a copy of the Nth data stack item and place on top of stack.
\ ** 0 PICK is equivalent to DUP.
  2 # D7 ASL,                           \ scale by 4
  0 D(A6, D7.L) d7 MOVE,
  next,
END-CODE

CODE ROLL       \ xu xu-1 .. x0 u -- xu-1 .. x0 xu
\ *G Rotate the order of the top N stack items by one place such that
\ ** the current top of stack becomes the second item and the Nth item
\ ** becomes TOS. See also ROT.
  D7 D0 MOVE,   rollx bmi,  rollx beq,
  2 # D0 ASL,   0 D(A6, D0.L) D7 MOVE,
L: ROLL1
  -4 D(A6, D0.L) 0 D(A6, D0.L) MOVE,
  4 # D0 SUBQ,   ROLL1 BNE,

  4 # A6 ADDQ,
  D7 D0 MOVE,
  next,

L: ROLLX
  (a6)+ d7 move,   next,
END-CODE

CODE ROT        \ x1 x2 x3 -- x2 x3 x1
\ *G ROTate the positions of the top three stack items such that the 
\ ** current top of stack becomes the second item. See also ROLL.
  d7 d0 move,                           \ save n3
  4 D(A6) d7 MOVE,                      \ get n1 into TOS
  (A6) 4 D(A6) MOVE,   d0 (a6) move,    \ move n2, n3
  next,
END-CODE

CODE -ROT       \ x1 x2 x3 -- x3 x1 x2
\ *G The inverse of ROT.
  d7 d0 move,                                           \ save n3
  (a6) d7 move,   4 d(a6) (a6) move,                    \ move n2, n1
  d0 4 d(a6) move,
  next,
END-CODE

CODE >R         \ x -- ; R: -- x
\ *G Push the current top item of the data stack onto the top of the
\ ** return stack.
  (a7)+ a0 movea,                                       \ get RTS address
  d7 -(A7) MOVE,  (a6)+ d7 move,                        \ stash TOS on return stack
  (a0) jmp,                                             \ jump to RTS address
END-CODE

CODE R>         \ -- x ; R: x --
\ *G Pop the top item from the return stack to the data stack.
  (a7)+ a0 movea,                                       \ get RTS address
  d7 -(a6) move,   (a7)+ d7 move,                       \ get real return stack value
  (a0) jmp,                                             \ jump to RTS address
END-CODE

CODE R@         \ -- x ; R: x -- x 
\ *G Copy the top item from the return stack to the data stack.
  (a7)+ a0 movea,                                       \ get RTS address
  d7 -(a6) move,   (A7) d7 MOVE,                        \ get real return stack value
  (a0) jmp,                                             \ jump to RTS address
END-CODE

CODE 2>R        \ x1 x2 -- ; R: -- x1 x2
\ *G Transfer the two top data stack items to the return stack.
  (a7)+ a0 movea,                                       \ get return address
  (a6)+ d0 move,                                        \ x1 into d0
  d0 -(a7) move,                                        \ x1 to stack
  d7 -(a7) move,                                        \ x2 to stack
  (a6)+ d7 move,                                        \ pop tos
  (a0) jmp,                                             \ jump to return address
END-CODE

CODE 2R>        \ -- x1 x2 ; R: x1 x2 --
\ *G Transfer the top two return stack items to the data stack. 
  (a7)+ a0 movea,                                       \ get return address
  d7 -(a6) move,                                        \ push tos
  (a7)+ d7 move,                                        \ retrieve x2
  (a7)+ d0 move,                                        \ retrieve x1
  d0 -(a6) move,                                        \ push x1 onto stack
  (a0) jmp,                                             \ jump to return address
END-CODE

CODE 2R@        \ -- x1 x2 ; R: x1 x2 -- x1 x2
\ *G Copy the top two return stack items to the data stack.
  (a7)+ a0 movea,                                       \ get return address
  d7 -(a6) move,                                        \ push TOS
  (a7)+ d7 move,                                        \ retrieve x2
  (a7) d0 move,                                         \ retrieve x1
  d7 -(a7) move,                                        \ push x2 back again
  d0 -(a6) move,                                        \ push x1 onto stack
  (a0) jmp,                                             \ jump to return address
END-CODE

CODE 2ROT       \ x1 x2 x3 x4 x5 x6 -- x3 x4 x5 x6 x1 x2
\ *G Perform ROT operation on three double numbers.
  (a6)+ { d0-4 } l. movem,                              \ x5-x1 into d0-4
  { d0-2 } -(a6) l. movem,                              \ d0-2 into x3-5
  d7 -(a6) move,                                        \ now x6
  d4 -(a6) move,                                        \ now x1
  d3 d7 move,                                           \ now x2 to TOS
  next,   
END-CODE

CODE DROP       \ x --
\ *G Lose the top data stack item and promote NOS to TOS.
  (a6)+ d7 move,   next,
END-CODE

CODE 2DROP      \ x1 x2 --
\ *G Discard the top two data stack items.
  # 4 a6 addq,   (a6)+ d7 move,   next,
END-CODE

CODE SWAP       \ x1 x2 -- x2 x1
\ *G Exchange the top two data stack items.
  (a6) d0 move,   d7 (a6) move,   d0 d7 move,   next,
END-CODE

CODE 2SWAP      \ x1 x2 x3 x4 -- x3 x4 x1 x2
\ *G Exchange the top two cell-pairs on the data stack.
  d7 D0 MOVE,   (A6) D1 MOVE,                           \ save d0=n4 d1=n3
  4 D(A6) d7 MOVE,   8 D(A6) (A6) MOVE,                 \ move n1 n2
  D0 4 D(A6) MOVE,   D1 8 D(A6) MOVE,                   \ move n4 n3
  next,
END-CODE

CODE DUP        \ x -- x x
\ *G DUPlicate the top stack item.
  d7 -(A6) MOVE,   next,
END-CODE

CODE ?DUP       \ x -- | x
\ *G DUPplicate the top stack item only if it non-zero.
  d7 tst,
  ne, if,  d7 -(a6) move,  endif,
  next,
END-CODE

CODE 2DUP       \ x1 x2 -- x1 x2 x1 x2
\ *G DUPlicate the top cell-pair on the data stack.
  (a6) d0 move,                         \ d0=n1 d7=n2
  d7 -(a6) move,   d0 -(a6) move,
  next,
END-CODE

CODE OVER       \ x1 x2 -- x1 x2 x1
\ *G Copy NOS to a new top-of-stack item.
  d7 -(a6) move,   4 d(a6) d7 move,   next,
END-CODE

CODE 2OVER      \ x1 x2 x3 x4 -- x1 x2 x3 x4 x1 x2
\ *G Similar to OVER but works with cell-pairs rather than cell items.
  d7 -(a6) move,                                        \ put TOS into deep stack
  $0C d(a6) d0 move,   $08 d(a6) d7 move,
  d0 -(a6) move,   next,
END-CODE

CODE SP@        \ -- x
\ *G Get the current address value of the data-stack pointer.
  d7 -(a6) move,   a6 d7 move,   next,
END-CODE

CODE SP!        \ x --
\ *G Set the current address value of the data-stack pointer.
  d7 a6 movea,  (a6)+ d7 move,   next,
END-CODE

CODE RP@        \ -- x
\ *G Get the current address value of the return-stack pointer.
  (a7)+ a0 movea,
  d7 -(a6) move,   a7 d7 move,
  (a0) jmp,
END-CODE

CODE RP!        \ x --
\ *G Set the current address value of the return-stack pointer.
  (a7)+ a0 movea,
  d7 a7 movea,   (a6)+ d7 move,
  (a0) jmp,
END-CODE


\ **********
\ *S Strings
\ **********

CODE COUNT      \ c-addr1 -- c-addr2' u
\ *G Given the address of a counted string in memory this word will
\ ** return the address of the first character and the length in
\ ** characters of the string.
  D7 A0 MOVEA,                                          \ need an address reg.
  1 # D7 ADDQ,   d7 -(a6) move,                         \ return ADDR'
  D7 CLR,   (A0) D7 B. MOVE,                            \ get length
  next,
END-CODE

CODE /STRING    \ c-addr1 u1 n -- c-addr2 u2
\ *G Modify a string address and length to remove the first N characters
\ ** from the string.
  d7 4 d(a6) add,   d7 (a6) sub,                        \ update addr & len
  (a6)+ d7 move,   next,                                \ pop it
END-CODE

CODE SKIP       \ c-addr1 u1 char -- c-addr2 u2
\ *G Modify the string description by skipping over leading occurrences of
\ ** 'char'.
  (A6)+ D1 MOVE,  (A6)+ A0 MOVEA,                       \ d7=char, d1=len, a0=adr1
  1 # D1 ADDQ,
L: SK1
  1 # D1 SUBQ,   DONE BEQ,
  (A0) D7 B. CMP,
  DONE BNE,
  1 # A0 ADDQ,   SK1 BRA,
L: DONE
  A0 -(A6) MOVE,   D1 D7 MOVE,
  next,
END-CODE

CODE SCAN       \ c-addr1 u1 char -- c-addr2 u2
\ *G Look for first occurrence of CHAR in string and return new string.
\ ** C-addr2/u2 describe the string with CHAR as the first character.
  (A6)+ D1 MOVE,  (A6)+ A0 MOVEA,
  1 # D1 ADDQ,
L: SC1
  1 # D1 SUBQ,   DONE BEQ,
  (A0) D7 B. CMP,
  DONE BEQ,
  1 # A0 ADDQ,   SC1 BRA,
END-CODE

CODE S=         \ c-addr1 c-addr2 u -- flag
\ *G Compare two same-length strings/memory blocks, returning TRUE if
\ ** they are identical.
  (A6)+ A0 MOVEA,  (A6)+ A1 MOVEA,
  D7 TST,   S=X BEQ,
L: S=1
  (A0)+ (A1)+ B. CMPM,
  S=X BNE,
  1 # D7 SUBQ,   S=1 BNE,
L: S=X
  D7 SEQ,
  CPU32? [if] D7 L. extb, [else] D7 W. EXT,  D7 L. EXT, [then]                   
  next,
END-CODE

(( \ Use this version when bringing up the target
: COMPARE       \ c-addr1 u1 c-addr2 u2 -- n ; compare strings
  2swap                                                 \ a2 l2 a1 l1 --
  rot                                                   \ a2 a1 l1 l2 --
  2dup - >r                                             \ a2 a1 l1 l2 --
  min 0                                                 \ a2 a1  l1/l2 0
  ?do
    swap 1+  swap 1+                                    \ step to next address
    over 1- c@ over 1- c@ <>                            \ a2 a1 a2<>a1 --
    ?leave                                              \ until different
  loop                                                  \ or until done
  swap 1- c@  swap 1- c@  -                             \ last two chars
  ?dup if                                               \ see if same
    0<                                                  \ not same, return indicator
    ( if  +1  else  -1  then )  1 or
    r> drop
  else
    r> dup if                                           \ strings =, see if same length
      0<                                                \ not, return indicator
      ( if  -1  else  +1  then )  1 or
    then
  then
;
))

CODE COMPARE    \ c-addr1 len1 c-addr2 len2 -- n
\ *G Compare two strings. The return result is 0 for a match or can be
\ ** -ve/+ve indicating string differences.
\ ** If the two strings are identical, n is zero. If the two strings
\ ** are identical up to the length of the shorter string, n is
\ ** minus-one (-1) if u1 is less than u2 and one (1) otherwise.
\ ** If the two strings are not identical up to the length of the
\ ** shorter string, n is minus-one (-1) if the first non-matching
\ ** character in the string specified by c-addr1 u1 has a lesser
\ ** numeric value than the corresponding character in the string
\ ** specified by c-addr2 u2 and one (1) otherwise. 

\ collect all the parameters
                                        \ len2 in D7 (tos)
   (a6)+ a0 l. movea,                   \ addr2 in A0
   (a6)+ d1 l. move,                    \ len1 in D1
   (a6)+ a1 l. movea,                   \ addr1 in A1
\ find the length to compare (smallest)
   d1 d2 l. move,   d1 d7 l. cmp,
   cs, if,  d7 d2 l. move,  endif,
\ compare byte by byte
   d3 l. clr,
   d2 l. tst,   compare3 beq,
l: compare1
        (a0)+ (a1)+ b. cmpm,
        compare2 bne,
        # 1 d2 l. subq,
        compare1 bne,
\ length to compare is 0, or both strings are the same so far
l: compare3
     d1 d7 l. cmp,
\ return flag -1, 0, +1
l: compare2
   ne, if,
     cs, if,  # 1 d3 l. addq,  else,  1 # d3 l. subq,  endif,
   endif,
   d3 d7 l. move,
   next,
END-CODE

internal
: (search)   { c-addr1 u1 c-addr2 u2 -- c-addr3 u3 flag }
  c-addr1 u1 u2 - 1+ bounds     \ Search for first char of $2 in $1.  Note that
  ?do                           \ we only check LEN($1)-LEN($2)+1 chars of $1.
    i c@  c-addr2 c@  =         \ Compare next char of $1 with first char of $2.
    if
      i c-addr2 u2 s=           \ Is the whole of $2 present?
      if
        i u1 i c-addr1 - - true \ Yes - build return stack parameters...
        unloop exit             \ ...and bail out.
      then
    then
  loop
  c-addr1 u1 false            \ End of loop and first char not found - no match.
;
external

: SEARCH	\ c-addr1 u1 c-addr2 u2  --  c-addr3 u3 flag
\ *G Search the string c-addr1/u1 for the string c-addr2/u2. If a match	
\ ** is found return c-addr3/u3, the address of the start of the match	
\ ** and the number of characters remaining in c-addr1/u1, plus flag f	
\ ** set to true. If no match was found return c-addr1/u1 and f=0.	
  2 pick over <         		\ Is $1 shorter than $2?
  if  2drop false exit  then   		\ Yes - $2 *can't* be in $1.
  dup 0 <=              		\ Is $2 zero length?
  if  2drop true exit  then    		\ Yes - we have found it at the start of $1.
  2 pick 0 <=           		\ Is $1 zero length?
  if  2drop false exit  then   		\ Yes - string not found.
  (search)
;

CODE CMOVE      \ c-addr1 c-addr2 u --
\ *G Copy U bytes of memory forwards from C-ADDR1 to C-ADDR2.
  (A6)+ A0 MOVEA,
  (A6)+ A1 MOVEA,
  D7 TST,   CMOV2 BEQ,
L: CMOV1
  (A1)+ (A0)+ B. MOVE,   1 # D7 SUBQ,  CMOV1 BNE,
L: CMOV2
  (a6)+ d7 move,   next,
END-CODE

CODE CMOVE>     \ c-addr1 c-addr2 u --
\ *G As CMOVE but working in the opposite direction,
\ ** copying the last character in the string first.
  (A6)+ A0 MOVEA,   D7 A0 ADDA,
  (A6)+ A1 MOVEA,   D7 A1 ADDA,
  D7 TST,   CMV>2 BEQ,
L: CMV>1
  -(A1) -(A0) B. MOVE,   1 # D7 SUBQ,  CMV>1 BNE,
L: CMV>2
  (a6)+ d7 move,   next,
END-CODE

internal

CODE (")        \ -- a-addr ; return address of string, skip over it
\ *G Return the address of a counted string that is inline after the 
\ ** CALLING word, and adjust the CALLING word's return address to 
\ ** step over the inline string. See the definition of (.") for
\ ** an example.
  (a7)+ a0 movea,                                       \ get RTS address
  d7 -(a6) move,                                        \ save TOS
  (A7) A2 MOVEA,   A2 d7 MOVE,                          \ save address
  D0 CLR,   (A2) D0 B. MOVE,                            \ get count byte
  2 # D0 ADDQ,
  $FFFE # D0 W. ANDI,                                   \ 1 for c.b. 1 to force even
  D0 (A7) ADD,                                          \ update second level RTS address
  (a0) jmp,                                             \ jump to first level RTS address
END-CODE

external

CODE UPPER      \ c-addr u --
\ *G Convert the ASCII string described to upper-case. This operation
\ ** happens in place.
  (a6)+ a0 movea,                                       \ a0=addr, d7=len
  d7 tst,   upper3 beq,                                 \ test for 0 length

l: upper1
  (a0) d0 b. move,                                      \ get character
    $60 # d0 b. cmp,   upper2 bls,                      \ check lower bound
    $7A # d0 b. cmp,   upper2 bhi,                      \ check upper bound
    $DF # d0 b. and,                                    \ mask character
l: upper2
  d0 (a0)+ b. move,                                     \ restore character
  # 1 d7 subq,   upper1 bne,

l: upper3
  (a6)+ d7 move,   next,
END-CODE


\ *******************
\ *S Memory operators
\ *******************

CODE ON         \ a-addr --
\ *G Given the address of a CELL this will set its contents to TRUE (-1).
  d7 A0 MOVEA,   -1 # (a0) MOVE,
  (a6)+ d7 move,   next,
END-CODE

CODE OFF        \ a-addr --
\ *G Given the address of a CELL this will set its contents to FALSE (0).
  d7 A0 MOVEA,  (A0) CLR,
  (a6)+ d7 move,   next,
END-CODE

CODE +!         \ n|u a-addr --
\ *G Add N to the CELL at memory address ADDR.
  d7 a0 movea,   (a6)+ d0 move,
  d0 (a0) add,
  (a6)+ d7 move,   next,
END-CODE

CODE c+!        \ b addr --                                                     
\ *G Add N to the character (byte) at memory address ADDR.
  d7 a0 movea,   (a6)+ d0 move,                                                 
  d0 (a0) b. add,                                                               
  (a6)+ d7 move,   next,                                                        
end-code                                                                        

CODE INCR       \ a-addr --
\ *G Increment the data cell at a-addr by one.
  d7 a0 movea,   # 1 (a0) l. addq,
  (a6)+ d7 move,   next,
END-CODE

CODE DECR       \ a-addr --
\ *G Decrement the data cell at a-addr by one.
  d7 a0 movea,   # 1 (a0) l. subq,
  (a6)+ d7 move,   next,
END-CODE

CODE 2@         \ a-addr -- x1 x2
\ *G Fetch and return the two CELLS from memory ADDR and ADDR+sizeof(CELL).
\ ** The cell at the lower address is on the top of the stack.
  d7 a0 movea,
  4 d(a0) -(a6) move,   (a0) d7 move,
  next,
END-CODE

CODE @          \ a-addr -- x
\ *G Fetch and return the CELL at memory ADDR.
  d7 A0 MOVEA,   (A0) d7 MOVE,
  next,
END-CODE

CODE W@         \ a-addr -- w
\ *G Fetch and 0 extend the word (16 bit) at memory ADDR.
  d7 A0 MOVEA,   # 0 d7 moveq,   (A0) D7 W. MOVE,
  next,
END-CODE

CODE C@         \ c-addr -- char
\ *G Fetch and 0 extend the character at memory ADDR and return.
  d7 a0 movea,   0 # d7 moveq,   (A0) D7 B. MOVE,
  next,
END-CODE

CODE 2!         \ x1 x2 a-addr --
\ *G Store the two CELLS x1 and x2 at memory ADDR.
\ ** X2 is stored at ADDR and X1 is stored at ADDR+CELL.
  d7 a0 movea,
  (a6)+ (a0) move,   (a6)+ 4 d(a0) move,
  (a6)+ d7 move,   next,
END-CODE

CODE !          \ x a-addr --
\ *G Store the CELL quantity N at memory ADDR.
  d7 A0 MOVEA,   (a6)+ (a0) move,
  (a6)+ d7 move,   next,
END-CODE

CODE W!         \ w a-addr --
\ *G Store the word (16 bit) quantity w at memory ADDR.
  d7 a0 movea,   (a6)+ d0 move,   d0 (a0) w. move,
  (a6)+ d7 move,   next,
END-CODE

CODE C!         \ char c-addr --
\ *G Store the character CHAR at memory C-ADDR.
  d7 a0 movea,   (a6)+ d0 move,   d0 (a0) b. move,
  (a6)+ d7 move,   next,
END-CODE

CODE FILL       \ c-addr u char --
\ *G Fill LEN bytes of memory starting at ADDR with the byte information
\ ** specified as CHAR.
  (A6)+ D1 MOVE,   (A6)+ A0 MOVEA,                      \ d7=char, d1=count, a0=addr
  D1 TST,   FIL2 BEQ,
L: FIL1
  D7 (A0)+ B. MOVE,   # 1 d1 subq,   fil1 bne,
L: FIL2
  (a6)+ d7 move,  next,
END-CODE

CODE TEST-BIT   \ u c-addr -- flag
\ *G AND the mask with the contents of addr and return the result.
\ ** Byte operation.
  d7 A0 MOVEA,                                          \ address
  # 0 d7 moveq,   (a0) d7 b. move,                      \ get byte
  (a6)+ d7 and,   next,                                 \ apply mask
END-CODE

CODE SET-BIT    \ u c-addr --
\ *G Apply the mask ORred with the contents of c-addr.
\ ** Byte operation.
  d7 A0 MOVEA,   (A6)+ D0 MOVE,                         \ get address and mask
  d0 (a0) b. or,                                        \ apply mask
  (a6)+ d7 move,   next,
END-CODE

CODE RESET-BIT  \ u c-addr --
\ *G Apply the mask inverted and ANDed with the contents of c-addr.
\ ** Byte operation.
  d7 A0 MOVEA,   (A6)+ D0 MOVE,                         \ get address and mask
  d0 not,   d0 (a0) b. and,                             \ apply mask
  (a6)+ d7 move,   next,
END-CODE

CODE TOGGLE-BIT \ u c-addr --
\ *G Invert the bits at c-addr specified by the mask. Byte operation.
  d7 A0 MOVEA,   (A6)+ D0 MOVE,                         \ get address and mask
  d0 (a0) b. eor,                                       \ apply mask
  (a6)+ d7 move,   Next,
END-CODE


\ **********************
\ *S Miscellaneous words
\ **********************

internal
: (SEARCH-WORDLIST)     ( c-addr u ^^nfa -- 0 | xt 1 | xt -1 )
  begin
    @ dup
  while
    dup c@ $20 and 0= if
      dup c@ $1F and			\ -- c-addr u ^nfa nfalen ;
      2 pick = if			\ Are the names the same *length*?
        dup 1+ 3 pick 3 pick s= if	\ Are they the same *name*?
          nip nip
          dup name> swap
          c@ $40 and 0=			\ check for immediacy
	  if  -1  else  1  then
          exit
        then
      then
    then
    n>link
  repeat
  drop 2drop 0
;
external

: SEARCH-WORDLIST       ( c-addr u wid  --  0|xt 1|xt -1  )
\ *G Search the given wordlist for a definition. If the definition is
\ ** not found then 0 is returned, otherwise the XT of the definition
\ ** is returned along with a non-zero code. A -ve code indicates a
\ ** "normal" definition and a +ve code indicates an IMMEDIATE word.
  dup 0= if
    nip nip
  else
    over 3 pick c@ + over @ 1- and      \ -- c-addr u wid thread# ;
    1+ cells +                          \ -- c-addr u ^nfa ;
    (search-wordlist)
  then
;

CODE NAME>      \ nfa -- xt
\ *G Move a pointer from an NFA to the CFA or "XT" in ANS parlance.
  d7 a0 movea,                                          \ get nfa in address reg.
  d7 l. clr,   (a0) d7 b. move,                         \ fetch count byte
  $1f # d7 w. andi,                                     \ strip top 3 bits
  a0 d7 add,                                            \ add base address
  2 # d7 addq,                                          \ add 1 for c.b., 1 to even
  $FFFE # d7 w. andi,                                   \ force even boundary
  next,                                                 \ add to base address
END-CODE

CODE >NAME      \ xt -- nfa
\ *G Move a pointer from an XT back to the NFA or name-pointer.
\ ** If the original pointer was not an XT or if the definition in 
\ ** question has no name header in the dictionary the returned pointer
\ ** will be useless. Care should be taken when manipulating or scanning
\ ** the Forth dictionary in this way.
  d7 a0 movea,                                          \ form absolute address
l: to-name-1
  -(a0) b. tst,   to-name-1 bpl,                        \ find byte, top bit set
  a0 d7 move,                                           \ put back address
  next,
END-CODE

: >BODY         \ xt -- a-addr
\ *G Move a pointer from a CFA or "XT" to the definition BODY. This
\ ** should only be used with children of CREATE. E.g. if FOOBAR is
\ ** defined with CREATE foobar, then the phrase ' foobar >body 
\ ** would yield the same result as executing foobar. 
  6 + @                                                                         
;                                                                               

internal

CODE DIGIT      \ char n -- 0|n true
\ *G If the ascii value CHAR can be treated as a digit for a number
\ ** within the radix N then return the digit and a TRUE flag, otherwise
\ ** return FALSE.
  (A6)+ D1 MOVE,                                        \ base in D7, char in D1
  $30 # D1 W. SUBI,   DIG2 BMI,                         \ fail if char < '0'
  $0A # D1 W. CMPI,   DIG1 BLT,                         \ ok if result < 0A
  07 # D1 SUBQ,                                         \ convert 'A'..'F'
  $0A # D1 W. CMPI,   DIG2 BLT,                         \ fail if result < 0A
L: DIG1
  D7 D1 CMP,   DIG2 BGE,                                \ check digit in range of base
  -1 # D7 MOVEQ,   D1 -(A6) MOVE,                       \ return digit and true
  next,
L: DIG2
  # 0 d7 moveq,   next,                                 \ return false
END-CODE

CODE UPC        \ char -- char'
\ *G Convert a character to upper case.
  $60 # D7 B. CMP,   UPCX BLS,
  $7A # D7 B. CMP,   UPCX BHI,
  $DF # D7 B. AND,
L: UPCX
  next,
END-CODE

external

CODE S>D        \ n -- d
\ *G Convert a single number to a double one.
  d7 -(a6) move,   D7 smi,
  cpu32? [if] d7 l. extb, [else] d7 w. ext,  d7 l. ext, [then]                   
  next,
END-CODE

CODE D>S        \ d -- n
\ *G Convert a double number to a single.
  (a6)+ d7 move,                                        \ low to TOS
  next,
END-CODE

CODE NOOP       \ -- 
\ *G A NOOP, null instruction.                                          )
  next,
END-CODE


\ **********************
\ *S Portability helpers
\ **********************
\ *P Using these words will make code easier to port between
\ ** 16, 32 and 64 bit targets.

aligning? 1 =	\ 16 bit alignment checks
aligning? 2 =	\ 32 bit alignment handled in KERNELxx.FTH
or [if]
CODE ALIGNED  	\ addr -- a-addr  
\ *G Given an address pointer this word will return the next ALIGNED
\ ** address subject to system wide alignment restrictions. Note that
\ ** this definition is only necessary in the code file for CPUs that
\ ** require alignment, but do not require alignment to a cell
\ ** boundary.
  1 # d7 addq,
  # $FFFFFFFE d7 andi,
  next,
END-CODE
[then]

CODE CELL+      \ a-addr1 -- a-addr2
\ *G Add the size of a CELL to the top-of stack.
  4 # d7 addq,
  next,
END-CODE

CODE CELLS      \ n1 -- n2
\ *G Return the size in address units of N1 cells in memory.
  2 # d7 asl,
  next,
END-CODE

CODE CELL-      \ a-addr1 -- a-addr2
\ *G Decrement an address by the size of a cell.
  4 # d7 subq,
  next,
END-CODE

CODE CELL       \ -- n
\ *G Return the size in address units of one CELL.
  d7 -(a6) move,                \ save TOS
  # 4 d7  moveq,                \ return cell size
  next,
END-CODE

CODE CHAR+      \ c-addr1 -- c-addr2
\ *G Increment an address by the size of a character.
  1 # d7 addq,
  next,
END-CODE

CODE CHARS      \ n1 -- n2
\ *G Return size in address units of N1 characters.
  next,
END-CODE


\ ********************
\ *S Runtime for VALUE
\ ********************

internal

CODE VAL!	\ n -- ; store value address in-line
\ *G Store n at the inline address following this word.
  (a7)+ a1 movea,               \ get inline address
  (a1)+ a0 movea,		\ get data address (in-line)
  d7 (a0) move,			\ & write data from TOS
  (a6)+ d7 move,		\ pop TOS
  (a1) jmp,
END-CODE

CODE VAL@	\ -- n ; read value data address in-line
\ *G Read n from the inline address following this word.
  (a7)+ a1 movea,               \ get inline address
  d7 -(a6) move,		\ push TOS
  (a1)+ a0 movea,		\ get data address (in-line)
  (a0) d7 move,			\ finally read data into TOS
  (a1) jmp,
END-CODE

(( replaced by LIT
CODE VAL	\ -- a-addr ; read address of value offset in-line
  (a7)+ a1 movea,               \ get inline address
  d7 -(a6) move,		\ push TOS
  (a1)+ d7 move,		\ get data address (in-line)
  (a1) jmp,
END-CODE
))
external


\ *************************************
\ *S Defining words and runtime support
\ *************************************

: COMPILE,      \ xt -- ; compile call to xt - may be optimised
\ *G Compile the word specified by xt into the current definition.
  $4EB9 w,  ,
;

ASMCODE
L: DODOES
  d7 -(a6) move,
  (a7)+ a0 movea,
  (a7)+ a1 movea,
  (a1) d7 move,
  (a0) jmp,
L: DOCREATE
  d7 -(a6) move,
  (a7)+ a0 movea,
  (a0) d7 move,
  next,
END-CODE


: CONSTANT      \ x "<spaces>name" -- ; Exec: --  x
\ *G Create a new CONSTANT called "name" which has the value "x". When 
\ ** "NAME" is executed the value is returned on the top-of-stack.7
  CREATE  -4 allot ,  ;CODE
  d7 -(a6) move,                                        \ save old TOS
  (a7)+ a0 movea,                                       \ get PFA
  (a0) d7 move,                                         \ get constant value
  next,
END-CODE

: VARIABLE      \ "<spaces>name" -- ; Exec: -- a-addr
\ *G Create a new variable called "name". When "Name" is executed the 
\ ** address of the data-cell is returned for use with @ and ! operators.
  CREATE  0 ,  ;CODE
  d7 -(a6) move,                                        \ save old TOS
  (a7)+ a0 movea,                                       \ get PFA
  (a0) d7 move,                                         \ get variable address
  next,
END-CODE

: USER          \ u "<spaces>name" -- ; Exec: -- addr
\ *G Create a new USER variable called "name". The 'u' parameter
\ ** specifies the index into the user-area table at which to place the 
\ ** data.								
\ ** USER variables are located in a separate area of memory for each	
\ ** task or interrupt. Use in the form: "$400 USER TaskData".
  CONSTANT  ;CODE
  d7  -(a6)  l. move,			\ save old TOS
  (a7)+  a0  l. movea,			\ jsr douser, offset
  (a0)  d7  l. move,			\ get offset
  a5  d7  l. add,			\ add user pointer to offset
  next,
end-code


target-only

: :             \ C: "<spaces>name" -- colon-sys ; Exec: i*x -- j*x ; R: -- nest-sys 
\ *G Begin a new definition called "name".
  ?exec  !csp  ( current @ context ! )
  create smudge  #-10 allot docolon, [compile] ]
;
\ IMMEDIATE                                             \                       ***JL002***

host&target


target-only

: :NONAME       \ C: -- colon-sys ; Exec: i*x -- i*x ; R: -- nest-sys
\ *G Begin a new code definition which does not have a name. After the
\ ** definition is complete the semi-colon operator returns the XT of 
\ ** newly compiled code on the stack.
  align here cell+                                      \ return XT
  !csp                                                  \ compiler stack security
  here last ! 0 ,                                       \ lay null header to fool SMUDGE
  ]                                                     \ turn compiler on
;

host&target

internal
: (;CODE)       \ -- ; R: a-addr --
\ *G Performed at compile time by ;CODE and DOES>. Patch the last
\ ** word defined (by CREATE) to have the run time actions that follows
\ ** immediately after (;CODE).
  r> latest name> 2+ !
;
external

: DOES>         \ C: colon-sys1 -- colon-sys2 ; Run: -- ; R: nest-sys --
\ *G Begin definition of the runtime-action of a child of a defining word.
\ ** You should not use RECURSE after DOES>.
  compile (;code)
  dodoes compile,			\ dodoes $l jsr,
;
IMMEDIATE


internal

: DOCREATE,     \ --
\ *G Compile the run time action of CREATE. 
  $4EB9 w, docreate ,  here 4+ ,	\ jsr docreate, addr
;

: CRASH         \ -- ; used as action of DEFER
\ *G The default action of a DEFERed word. This will simply THROW a 
\ ** code back to the system.
  $C throw
;

external

: DEFER         \ Comp: "<spaces>name" -- ; Run: i*x -- j*x
\ *G Creates a new DEFERed word. A default action, CRASH, is
\ ** automatically assigned.
  CREATE  ['] CRASH ,  ;CODE
  (a7)+ a1 movea,
  (a1) a0 movea,
  (a0) a1 movea,
  (a1) jmp,
END-CODE

: 2CONSTANT     \ Comp: x1 x2 "<spaces>name" -- ; Run: -- x1 x2
\ *G A two-cell equivalent to CONSTANT.
  create
    cell negate allot , ,
  ;code
  d7 -(a6) move,                                        \ push tos
  (a7)+ a0 movea,                                       \ get PFA
  4 d(a0) -(a6) move,                                   \ push low cell
  (a0) d7 move,                                         \ get high cell
  next,
END-CODE

: 2VARIABLE     \ Comp: "<spaces>name" -- ; Run: -- a-addr
\ *G A two-cell equivalent to VARIABLE.
  create
    0 , 0 ,
  ;code
  d7 -(a6) move,                                        \ save old TOS
  (a7)+ a0 movea,                                       \ get PFA
  (a0) d7 move,                                         \ get variable address
  next,
END-CODE

: FIELD         \ size x "<spaces>name" -- size+x ; Exec: addr -- addr+x
\ *G Create a new field within a structure definition of size n bytes.
  CREATE  -4 allot ,  ;CODE
  (a7)+ a0 movea,                                       \ get PFA
  (a0) d7 add,                                          \ get constant value
  next,
END-CODE


\ ************************
\ *S Structure compilation
\ ************************
\ *P These words define high level branches. They are used by the structure
\ ** words such as IF and AGAIN.

hex

internal

: >mark         \ -- addr ; mark start of forward branch
\ *G Mark the start of a forward branch. HIGH LEVEL CONSTRUCTS ONLY.
  here  0 ,  
;

: >resolve      \ addr -- ; resolve absolute target of forward branch
\ *G Resolve absolute target of forward branch. HIGH LEVEL CONSTRUCTS ONLY.
\ here over 2+ - swap !                 \ relative
  here swap !  ;                        \ absolute

: <mark         \ -- addr ; mark start (destination) of backward branch
\ *G Mark the start (destination) of a backward branch.
\ ** HIGH LEVEL CONSTRUCTS ONLY.
  here  
;

: <resolve      \ addr ; resolve (at branch point) backward branch
\ *G Mark the start (destination) of a backward branch.
\ ** HIGH LEVEL CONSTRUCTS ONLY.
\  here swap 2+ -  ,  		        \ relative
   ,                    		\ absolute for target
;

synonym >c_res_branch >resolve	\ addr -- ; fix up forward referenced branch
\ *G See >RESOLVE.
synonym c_mrk_branch< >mark	\ -- addr ; mark destination of backward branch
\ *G See >MARK


\ **********************
\ *S Branch constructors
\ **********************
\ *P Used when compiling code on the target.

: c_branch<     \ addr -- ; lay BRANCH instructions
\ *G Lay the code for BRANCH.
  compile branch ,
;

: c_?branch<    \ addr -- ; lay ?BRANCH instructions
\ *G Lay the code for ?BRANCH.
  compile ?branch ,
;

: c_branch>     \ -- addr ; forward referenced branch
\ *G Lay the code for a forward referenced unconditional branch.
  compile branch  >mark
;

: c_?branch>    \ -- addr ; forward referenced branch
\ *G Lay the code for a forward referenced conditional branch.
  compile ?branch  >mark
;


\ ***************************
\ *S Main structure compilers
\ ***************************

: DOCOLON,      \ --
\ *G Compile the runtime entry code required by colon definitions.
;

: c_lit		\ lit --
\ *G Compile the code for a LITERAL.
  $2D07 w,				\ move .l  d7 -(a6)
  $2E3C w, ,				\ move .l  # xxxx d7
;

: c_drop        \ --
\ *G Compile the code for DROP.
  $2E1E w,				\ move .l  (a6)+ d7
;

: c_exit        \ --
\ *G Compile the code for EXIT.
  $4E75 w,
;

: c_do          \ C: -- do-sys ; Run: n1|u1 n2|u2 -- ; R: -- loop-sys 
\ *G Compile the code for DO.
  compile (do) >mark <mark
;

: c_?DO         \ C: -- do-sys ; Run: n1|u1 n2|u2 -- ; R: -- | loop-sys
\ *G Compile the code for ?DO.
  compile (?do) >mark <mark
;

: c_LOOP        \ C: do-sys -- ; Run: -- ; R: loop-sys1 -- | loop-sys2
\ *G Compile the code for LOOP.
  compile (loop) <resolve >resolve
;

: c_+LOOP       \ C: do-sys -- ; Run: -- ; R: loop-sys1 -- | loop-sys2
\ *G Compile the code for +LOOP.
  compile (+loop) <resolve >resolve
;

variable NextCaseTarg	\ -- addr
\ *G Holds the entry point of the current CASE statement.

: c_case        \ -- addr
\ *G Compile the code for CASE.
  NextCaseTarg @  <mark NextCaseTarg !
;
: c_case        \ -- addr
  <mark
;

: c_OF          \ C: -- of-sys ; Run: x1 x2 -- | x1
\ *G Compile the code for OF.
  compile (of) >mark
;

: c_?OF         \ C: -- of-sys ; Run: flag --
\ *G Compile the code for ?OF.
  c_?branch>
;

: c_ENDOF	\ C: case-sys1 of-sys -- case-sys2 ; Run: --
\ *G Compile the code for ENDOF.
  c_branch> swap >c_res_branch
;

: FIX-EXITS     \ n1..nn --
\ *G Compile the code to resolve the forward branches at the end
\ ** of a CASE structure.
  begin
    sp@ csp @ <>
  while
    >c_res_branch
  repeat
;

: c_ENDCASE     \ C: case-sys -- ; Run: x --
\ *G Compile the code for ENDCASE.
  c_drop  fix-exits  Nextcasetarg !
;

FullCase? [if]
: c_END-CASE    \ C: case-sys -- ; Run: --
\ *G Compile the code for END-CASE.
  fix-exits  NextCaseTarg !
;

: c_NEXTCASE    \ C: case-sys -- ; Run: x --
\ *G Compile the code for NEXTCASE.
  c_drop  NextCaseTarg @ c_branch<  fix-exits  NextCaseTarg !
;
[then]
external

((
: exchange  \ addr1 addr2 -- ; swap contents
  over @ over @   \ a1 a2 (a1) (a2) -- 
  -rot            \ a1 (a2) a2 (a1)
  swap !  swap !
;
))


\ ******
\ *> ###
\ ******

