\ 68307 Umbilical Forth control file for CMS MicroModule board

((
Copyright (c) 2000, 2003
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

PLEASE NOTE OUR PHONE NUMBER CHANGE


20000630 SFP001 Converted to v6.1 code model
		N.B. USER variables have a different structure to v6.0

Target Model: Subroutine Threaded Code with VFX optimisation
  A7 = return stack
  A6 = data stack
  A5 = user pointer
  A4 = locals pointer
  A3 = 0
  A2 = scratch
  A1 = scratch
  A0 = scratch

  D7 = TOS
  D6..D2 = scratch
  D1, D0 = scratch and optimiser temporary
))

Only Forth Definitions


\ ******************************
\ Define the default directories
\ ******************************

"" ..\common" setmacro CommonDir	\ where the common code lives
"" ."           setmacro CpuDir		\ where the CPU specific code lives
"" .\Hardware\CMS68307" setmacro HWDir	\ where board specific code lives
"" ..\Examples" setmacro ExampleDir	\ Examples
"" ..\MyApp"    setmacro AppDir		\ where application code lives


\ ****************************************************************
\ Add any files that extend the host, for example assembler macros
\ ****************************************************************

  Include %CpuDir%\Macros.fth		\ Compiler/Assembler macros


\ **********************************************************
\ If you want the compiler log to go to a file, name it here
\ **********************************************************

\ File: PROG.log


\ **************************************
\ Start and set up the cross compilation
\ **************************************

CROSS-COMPILE

  no-heads                      \ No heads on target
  No-log                        \ No output log
  Rommed                        \ Split ROM/RAM system
  Align-Even                    \ Code is 16b aligned
  M68000                        \ Chip has 68000 core
  +xrefs                        \ enable cross references
  no inlining			\ disable inlining
\  normal inlining		\ allow inlining

Only Forth Definitions          \ default search order

  0         Equ False
  False Not Equ True


\ ********************
\ Target Configuration
\ ********************


\ Do we need interactive debugging over the XTL?
 1 equ umbilical?		\ true if debugging over XTL
   $40 Equ max-depth		\   max depth for UF data stack before UF barfs
   0 equ SPI?			\   true if XTL over SPI port

\ Does the initialised data area actually need to be initialised?
 0 equ init-idata?

\ Kernel components
 0 equ tasking?			\ true if multitasker needed
   6 cells equ tcb-size		\   for internal consistency check
   0 equ event-handler?		\   true to include event handler
   0 equ message-handler?	\   true to include message handler
   0 equ semaphores?		\ true to include semaphores
 0 equ default-traps?		\ default trap handles
 0 equ timebase?		\ true for TIMEBASE code
 0 equ softfp?			\ true for software floating point
 0 equ FullCase?		\ true to include ?OF END-CASE NEXTCASE extensions
 $0000 equ sizeofheap		\ 0=no heap, nz=size of heap
   0 equ heap-diags?		\   true to include diagnostic code

\ Clock, serial and ticker rates
#7372500 equ system-speed	\ System clock rate in HZ.
#9600    equ console-speed	\ Serial port speed in BPS.
#10 equ tick-ms			\ TIMEBASE tick in ms 

\ version numbers
 char 6 equ mpe-rel		\ x in Vx.yz
 char 1 equ mpe-ver		\ y in Vx.yz 
 char 0 equ usrver		\ z in Vx.yz

\ define stack and user area sizes
$0100 equ UP-SIZE		\ size of each task's user area
$0100 equ SP-SIZE		\ size of each task's data stack
$0200 equ RP-SIZE		\ size of each task's return stack
up-size rp-size + sp-size +
  equ task-size			\ size of TASK data area
UP-SIZE equ INTRAM		\ space used by interrupt page


\ **********
\ Memory map
\ **********

\ e27512 8bit Define-Emulator		\ define EPROM emulator
$000000 $00FFFF cdata section CMS307UF	\ primary ROM area
\ $10000 in-emulator
$180000 $183FFF idata section PROGd	\ initialised data
$184000 $187FFF udata section PROGu	\ unitialised data

interpreter
: prog	cms307uf  ;
target


PROG PROGD PROGU CDATA			\ use Code for HERE , and so on

\ define memory to the startup code
$100000 equ MBASE                       \ base address of 68307 SIM registers
MBASE #12 rshift  $0 or  equ initMBAR	\ initial value for MBAR, ignore FC2..0
$180000 equ RAMBASE			\ base of physical RAM
$187FFF equ RAMTOP			\ top of physical RAM
$008000 equ RAMSIZE			\ size of physical RAM


\ **********************************
\ USER area and Multi tasker equates
\ **********************************
\ Assume stacks grow down: user area, sp stack, rp-stack
0                 equ TASK-U0		\ initial offset of user area
up-size sp-size + equ TASK-S0		\ initial offset of data stack
task-s0 rp-size + equ TASK-R0		\ initial offset of return stack

intram    reserve equ INT-INIT-U0	\ base of interrupt area
task-size reserve equ INIT-T0		\ base of task area
  init-t0 task-u0 + equ INIT-U0
  init-t0 task-r0 + equ INIT-R0		\ top of main return stack
  init-t0 task-s0 + equ INIT-S0		\ top of main data stack


\ ************
\ Kernel files
\ ************

  include %CpuDir%\sfr307		\ 68307 special function registers
  include %HWDir%\init307		\ CMS 68307 startup code and vector table
  Include %CpuDir%\Min68K		\ minimal Umbilical Forth kernel
  include %CommonDir%\UFkernel		\ high level kernel definitions
  include %CpuDir%\drivers\68307i	\ interrupt driven 68307 UART drivers
  include %CpuDir%\Int68k		\ interrupt handlers
Umbilical? [if]
  Synonym wait-byte  (serkey)		\ Say which XTL drivers to use
  Synonym send-byte  (seremit)
  Synonym Wait-Byte? (serkey?)
  Synonym Init-XTL   Init-Ser
  Include %CommonDir%\messages		\ Equates only
  Include %CommonDir%\Targend		\ Message driver
[Endif]

default-traps? [if]
  Include %CpuDir%\Traps		\ default exception handlers
[then]

tasking? [if]
  include %CpuDir%\multi68K		\ multi-tasker, MUST be before TIMEBASE
[then]

timebase? [if]
  include %CommonDir%\timebase		\ time base common code
  include %CpuDir%\ClockInt307		\ timer tick
[then]

SIZEOFHEAP [if]
  include %CommonDir%\heap32		\ memory allocation set
[then]

softfp? [if]
  include %CpuDir%\softfp		\ floating point
  include %CommonDir%\softcom		\ common floating point code
[then]


\ *************
\ End of kernel
\ *************

((
: .CPU          \ -- ; sign on
  ." MPE 68307 ANS "
  [ tasking?   ] [if] ." Multi-Tasking " [else] ." Single-Task " [then]
  [ Umbilical? ] [if] ." Umbilical "     [else] ." Production "  [then]
  ." Forth"
;
))


\ **************************
\ Load application code here
\ **************************

\ include AppDir\MyApp                  \ you can use a macro for APPDIR


\ *******************
\ Load libraries here
\ *******************

libraries
  include %CommonDIR%\UFlib		\ high level library
\  include %CpuDir%\lib68k		\ CPU specific library
end-libs


\ **************************************
\ Define what happens at target power up
\ **************************************

Umbilical? [if]
  make-turnkey run-umbilical            \ in TARGEND.FTH
  "" COM1" console-speed serial         \ specified required target link
  umbilical-forth                       \ start interactive session
[else]
  Make-turnkey Run                      \ application start up word
[then]

finis

