\ FS332 Forth Systeme 68332 SBC control file

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

PLEASE NOTE OUR PHONE NUMBER CHANGE
))

only forth definitions


\ ***********************************************************
\ Uncomment these lines if you want to use the MPE test suite
\ ***********************************************************

\ for 5-1-x tests...
\ also c-c definitions
\ required-symbol --)-later
\ previous definitions
\ ...for 5-1-x tests


\ ******************************
\ Define the default directories
\ ******************************

"" ..\common"   setmacro CommonDir	\ where the common code lives
"" ."           setmacro CpuDir		\ where the CPU specific code lives
"" .\Hardware\Fs332" setmacro HWDir	\ where board specific code lives
"" ..\Examples" setmacro ExampleDir	\ Examples
"" ..\MyApp"    setmacro AppDir		\ where application code lives
"" c:\products\vfsfp\tests" setmacro TestDir	\ where test code lives



\ *********************************************************
\ Turn on the cross compiler and define CPU and log options
\ *********************************************************

include %cpudir%\macros         \ compiler / assembler macros

\ file: PROG.log

CROSS-COMPILE

only forth definitions          \ default search order

  no-log			\ uncomment to suppress output log
  rommed                        \ split ROM/RAM target
  interactive                   \ enter interactive mode at end
  +xrefs                        \ enable cross references
  align-even                    \ code is 16bit aligned
  CPU32				\ CPU32 core
  normal inlining		\ allow inlining

0         equ false
false not equ true


\ ****************
\ Configure target
\ ****************

\ Kernel components
 1 equ DEFAULT-TRAPS?		\ true if default trap handler required
 1 equ tasking?			\ true if multitasker needed
   6 cells equ tcb-size		\   for internal consistency check
   1 equ event-handler?		\   true to include event handler
   1 equ message-handler?	\   true to include message handler
   1 equ semaphores?		\ true to include semaphores
 0 equ timebase?		\ true for TIMEBASE code
 0 equ softfp?			\ true for software floating point
 0 equ FullCase?		\ true to include ?OF END-CASE NEXTCASE extensions
 0 equ target-locals?		\ true if target local variable sources needed
 1 equ romforth?		\ true for ROMForth handler
 0 equ blocks?			\ true if BLOCK needed
 $000 equ sizeofheap		\ 0=no heap, nz=size of heap
   0 equ heap-diags?		\   true to include diagnostic code
 0 equ paged?			\ true if ROM or RAM is paged/banked
 0 equ MPE-SET?			\ compatibility with MPE v5 targets
 0 equ ENVIRONMENT?		\ true if ANS ENVIRONMENT system required

\ Clock, serial and ticker rates
#16670000 equ system-speed	\ System clock rate in HZ.
#9600     equ console-speed	\ Serial port speed in BPS.
#10 equ tick-ms			\ TIMEBASE tick in ms 

\ version numbers
 char 6 equ mpe-rel		\ x in Vx.yz
 char 1 equ mpe-ver		\ y in Vx.yz 
 char 0 equ usrver		\ z in Vx.yz

\ define stack and user area sizes
$0100 equ UP-SIZE		\ size of each task's user area
$0100 equ SP-SIZE		\ size of each task's data stack
$0200 equ RP-SIZE		\ size of each task's return stack
up-size rp-size + sp-size +
  equ task-size			\ size of TASK data area
UP-SIZE equ INTRAM		\ space used by interrupt page

$0100 equ TIB-LEN		\ terminal i/p buffer length


\ ********************
\ Define memory layout
\ ********************

\  e27010 8bit define-emulator
$000000 $00FFFF cdata section FS332
\  $00000 in-emulator
$800000 $807FFF idata section PROGd
$808000 $80FFFF udata section PROGu

interpreter
: prog	fs332  ;
target


PROG PROGd PROGu  CDATA                 \ use Code for HERE , and so on


\ **********************************
\ USER area and Multi tasker equates
\ **********************************
\ Assume stacks grow down: user area, sp stack, rp-stack
0                 equ TASK-U0		\ initial offset of user area
up-size sp-size + equ TASK-S0		\ initial offset of data stack
task-s0 rp-size + equ TASK-R0		\ initial offset of return stack

intram    reserve equ INT-INIT-U0	\ base of interrupt area
task-size reserve equ INIT-T0		\ base of task area
  init-t0 task-u0 + equ INIT-U0
  init-t0 task-r0 + equ INIT-R0		\ top of main return stack
  init-t0 task-s0 + equ INIT-S0		\ top of main data stack
tib-len   reserve equ INIT-TIB		\ base of TIB

PROGd  sec-top 1+ equ UNUSED-TOP  PROG	\ top of memory for UNUSED
CDATA


\ ********************************
\ ROM PowerForth section if needed
\ ********************************

romforth? [if]
  $810000 $810FFF udata section PROGr
  sec-base equ APPL-ROM                 \ start of application space
  sec-top  equ APPL-ROM-END             \ end of application space
  APPL-ROM-END equ EM			\ in case it isn't set later
  0 equ APPL-FLASH

PROG PROGd PROGu  CDATA                 \ use Code for HERE , and so on
[then]


\ ************
\ Kernel files
\ ************

  include %CpuDir%\sfr332		\ 68332 special function registers
  include %HWDIR%\init332		\ 68332 startup code and vector table
  include %CpuDir%\code68k		\ low level kernel definitions
  include %CommonDir%\kernel		\ high level kernel definitions
  include %CpuDir%\drivers\68332i	\ UART drivers
  include %CpuDir%\int68k		\ exception handlers
  include %CommonDir%\devtools		\ DUMP .S etc development tools
  include %CommonDir%\voctools		\ ORDER VOCS etc
  include %CommonDir%\methods		\ target support for methods
  include %CpuDir%\local		\ local variables

default-traps? [if]
  include %CpuDir%\traps		\ default exception handlers
[then]

tasking? [if]
  include %CpuDir%\multi68K		\ multi-tasker, MUST be before TIMEBASE
[then]

timebase? [if]
  include %CommonDir%\timebase		\ time base common code
  include %CpuDir%\ClockInt332		\ timer tick
[then]

environment? [if]
  include %CommonDir%\environ		\ ENVIRONMENT?
[then]

SIZEOFHEAP [if]
  include %CommonDir%\heap32		\ memory allocation set
[then]

softfp? [if]
  include %CpuDir%\Sfp68k		\ 68k floating point primitives
\  include %CommonDir%\sfp32hi		\ floating point primitives - replace with coded version
  include %CommonDir%\sfp32com		\ common floating point code
[then]

romforth? [if]
tasking? [if]
  Synonym Ser-Key (mser-key)
[else]
  Synonym Ser-Key (ser-key)
[then]
  Synonym Ser-Key? (ser-key?)
  Synonym Ser-Emit (ser-emit)
  include %CommonDir%\RomForth\link	\ appl. rom link
  include %CommonDir%\RomForth\iodef	\ link i/o
  include %CommonDir%\RomForth\filetran	\ ascii file uploader
  include %CommonDir%\RomForth\xmodem	\ XMODEM downloader
  include %CommonDir%\RomForth\intelhex	\ Intel Hex downloader
  include %CommonDir%\RomForth\textfile	\ XSHELL textfile support
\ include %CommonDir%\RomForth\blocks	\ XSHELL blocks support
[then]

mpe-set? [if]
  include %CpuDir%\mpe_supp		\ MPE v5 compatibility word set
[then]


\ *************
\ End of kernel
\ *************

internal
: .CPU		\ -- ; display CPU type
  ." MPE 68332 ANS ROM PowerForth v6.20"
;
external


\ *************************
\ Add application code here
\ *************************

  include %CpuDir%\drivers\FS_led	\ FS332 LED drivers


\ ********************************************
\ The following code can be included if wanted
\ ********************************************

\ include %ExampleDir%\tester		\ John Hayes' ANS test engine
\ include %ExampleDir%\5-1-a		\ MPE test engine
\ include %ExampleDir%\5-1-b		\ MPE tests

\ include %ExampleDir%\orbits		\ Sun-Earth-Moon simulation
\ include %ExampleDir%\bench		\ Sieve of Eratosthenes
\ include %ExampleDir%\str_loc		\ structures and local variables
\ include %ExampleDir%\checksum		\ checksum / crc
\ include %ExampleDir%\buildno		\ build number

\ include %CommonDir%\widinfo
\ include %CommonDir%\taskinfo

\ include %TestDir%\mpe\TestCore	\ MPE test harness, load TEST32.FTH interactively
\ include %ExampleDir%\testtask
\ include %ExampleDir%\testheap


libraries
  include %CommonDir%\library
end-libs

romforth? [if]
  idata sec-end rp !  cdata		\ set top of application RAM area
[then]


\ XREF DUP                              \ where is DUP used
\ XREF-ALL                              \ full cross reference
\ XREF-UNUSED                           \ find unused words

FINIS

