\ 68332 Umbilical Forth control file for Forth Systeme SBC

((
        Copyright (c) 1999, 2003
        MicroProcessor Engineering
        133 Hill Lane
        Southampton SO15 5AF
        England

        tel: +44 (0)2380 631441
        fax: +44 (0)2380 339691
        net: mpe@mpeltd.demon.co.uk
             tech-support@mpeltd.demon.co.uk
        web: http://www.mpeltd.demon.co.uk
))

Only Forth Definitions


\ ******************************
\ Define the default directories
\ ******************************

"" ..\common"   setmacro CommonDir	\ where the common code lives
"" ."           setmacro CpuDir		\ where the CPU specific code lives
"" .\Hardware\Fs332" setmacro HWDir	\ where board specific code lives
"" ..\Examples" setmacro ExampleDir	\ Examples
"" ..\MyApp"    setmacro AppDir		\ where application code lives


\ ****************************************************************
\ Add any files that extend the host, for example assembler macros
\ ****************************************************************

  Include %CpuDir%\Macros.fth		\ Compiler/Assembler macros


\ **********************************************************
\ If you want the compiler log to go to a file, name it here
\ **********************************************************

\ File: PROG.log


\ **************************************
\ Start and set up the cross compilation
\ **************************************

CROSS-COMPILE

  no-heads                      \ No heads on target
  No-log                        \ No output log
  Rommed                        \ Split ROM/RAM system
  Align-Even                    \ Code is 16b aligned
  CPU32                         \ Chip has CPU32 core
  +xrefs                        \ enable cross references
  no inlining			\ disable inlining
\  normal inlining		\ allow inlining

Only Forth Definitions          \ default search order

  0 Equ False
  -1 Equ True


\ ********************
\ Target Configuration
\ ********************

\ Do we need interactive debugging over the XTL?
 1 equ umbilical?		\ true if debugging over XTL
   $40 Equ max-depth		\   max depth for UF data stack before UF barfs
   0 equ SPI?			\   true if XTL over SPI port

\ Does the initialised data area actually need to be initialised?
 0 equ init-idata?

\ Kernel components
 0 equ tasking?			\ true if multitasker needed
   6 cells equ tcb-size		\   for internal consistency check
   0 equ event-handler?		\   true to include event handler
   0 equ message-handler?	\   true to include message handler
   0 equ semaphores?		\ true to include semaphores
 0 equ default-traps?		\ default trap handles
 0 equ timebase?		\ true for TIMEBASE code
 0 equ softfp?			\ true for software floating point
 0 equ FullCase?		\ true to include ?OF END-CASE NEXTCASE extensions
 $0000 equ sizeofheap		\ 0=no heap, nz=size of heap
   0 equ heap-diags?		\   true to include diagnostic code

\ Clock, serial and ticker rates
#16670000 equ system-speed	\ System clock rate in HZ.
#9600     equ console-speed	\ Serial port speed in BPS.
#10 equ tick-ms			\ TIMEBASE tick in ms 

\ version numbers
 char 6 equ mpe-rel		\ x in Vx.yz
 char 1 equ mpe-ver		\ y in Vx.yz 
 char 0 equ usrver		\ z in Vx.yz

\ define stack and user area sizes
$0100 equ UP-SIZE		\ size of each task's user area
$0100 equ SP-SIZE		\ size of each task's data stack
$0200 equ RP-SIZE		\ size of each task's return stack
up-size rp-size + sp-size +
  equ task-size			\ size of TASK data area
UP-SIZE equ INTRAM		\ space used by interrupt page


\ **********
\ Memory map
\ **********

\ e27512 8bit Define-Emulator	\ define EPROM emulator
$000000 $00FFFF cdata section FS332UF
\ $0 in-emulator
$800000 $807FFF idata section PROGd
$808000 $80FFFF udata section PROGu

interpreter
: prog	fs332uf  ;
target

PROG PROGd PROGu CDATA		\ use Code for HERE , and so on


\ **********************************
\ USER area and Multi tasker equates
\ **********************************
\ Assume stacks grow down: user area, sp stack, rp-stack
\ These equates calculate how much RAM is needed for the USER area,
\ Return stack and data stack of each task, plus an area used
\ by Forth interrupts that share a common USER (and optionally)
\ common stacks. This can be done to reduce the overall amount
\ of RAM needed for interrupt stacks. Since interrupts can occur
\ in any task, switching stacks for high level interrupts will
\ reduce the total RAM requirements.
0                 equ TASK-U0		\ initial offset of user area
up-size sp-size + equ TASK-S0		\ initial offset of data stack
task-s0 rp-size + equ TASK-R0		\ initial offset of return stack

intram    reserve equ INT-INIT-U0	\ base of interrupt area
task-size reserve equ INIT-T0		\ base of task area
  init-t0 task-u0 + equ INIT-U0
  init-t0 task-r0 + equ INIT-R0		\ top of main return stack
  init-t0 task-s0 + equ INIT-S0		\ top of main data stack


\ ************
\ Kernel files
\ ************

  include %CpuDir%\sfr332		\ 68332 special function registers
  include %HWDIR%\init332		\ 68332 startup code and vector table
  Include %CpuDir%\Min68K		\ minimal Umbilical Forth kernel
  include %CommonDir%\UFkernel		\ high level kernel definitions
  include %CpuDir%\drivers\68332i	\ interrupt driven 68332 UART drivers
  include %CpuDir%\Int68k		\ interrupt handlers
Umbilical? [if]
  Synonym wait-byte  (ser-key)		\ Say which XTL drivers to use
  Synonym send-byte  (ser-emit)
  Synonym Wait-Byte? (ser-key?)
  Synonym Init-XTL   Init-Ser
  Include %CommonDir%\messages		\ Equates only
  Include %CommonDir%\Targend		\ Message driver
[Endif]

default-traps? [if]
  Include %CpuDir%\Traps		\ default exception handlers
[then]

tasking? [if]
  include %CpuDir%\multi68K		\ multi-tasker, MUST be before TIMEBASE
[then]

timebase? [if]
  include %CommonDir%\timebase		\ time base common code
  include %CpuDir%\ClockInt332		\ timer tick
[then]

SIZEOFHEAP [if]
  include %CommonDir%\heap32		\ memory allocation set
[then]

softfp? [if]
  include %CpuDir%\softfp		\ floating point
  include %CommonDir%\softcom		\ common floating point code
[then]


\ *************
\ End of kernel
\ *************

((
: .CPU          \ -- ; sign on
  ." MPE 68332 ANS "
  [ tasking?   ] [if] ." Multi-Tasking " [else] ." Single-Task " [then]
  [ Umbilical? ] [if] ." Umbilical "     [else] ." Production "  [then]
  ." Forth"
;
))


\ **************************
\ Load application code here
\ **************************

\ include AppDir\MyApp                  \ you can use a macro for APPDIR


\ *******************
\ Load libraries here
\ *******************

libraries
  include %CommonDIR%\UFlib		\ high level library
\  include %CpuDir%\lib68k		\ CPU specific library
end-libs


\ **************************************
\ Define what happens at target power up
\ **************************************

Umbilical? [if]
  make-turnkey run-umbilical            \ in TARGEND.FTH
  "" COM1" console-speed serial         \ specified required target link
  umbilical-forth                       \ start interactive session
[else]
  Make-turnkey Run                      \ application start up word
[then]

finis

