\ 68307 interrupt RX, polled TX, serial communication drivers

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

PLEASE NOTE OUR PHONE NUMBER CHANGE

20000630 SFP002 Converted to v6.1 generic I/O model
19980318 MSD001 Adopt a new baud rate generation scheme.
))


only forth definitions
hex


\ ****************************
\ Baud rate and initialisation
\ ****************************

\ Table entries are BPS , SYS-VALUE1 , SYS-VALUE2 and are terminated by a zero
\ BPS value.  In this instance, SYS-VALUE1 is the UART's baud rate "set" number
\ mask ($80=set2,$00=set1) and SYS-VALUE2 is the RESETUCSR byte value.

host
create baud-table
  #00050 , $00 , $00 ,          \ Baud rate set one.
  #00110 , $00 , $11 ,
  #00135 , $00 , $22 ,
  #00200 , $00 , $33 ,
  #00300 , $00 , $44 ,
  #00600 , $00 , $55 ,
  #01200 , $00 , $66 ,
  #01050 , $00 , $77 ,
  #02400 , $00 , $88 ,
  #04800 , $00 , $99 ,
  #07200 , $00 , $AA ,
  #09600 , $00 , $BB ,
  #38400 , $00 , $CC ,
  #00075 , $80 , $00 ,          \ Baud rate set two.
  #00110 , $80 , $11 ,
  #00135 , $80 , $22 ,
  #00150 , $80 , $33 ,
  #00300 , $80 , $44 ,
  #00600 , $80 , $55 ,
  #01200 , $80 , $66 ,
  #02000 , $80 , $77 ,
  #02400 , $80 , $88 ,
  #04800 , $80 , $99 ,
  #01800 , $80 , $AA ,
  #09600 , $80 , $BB ,
  #19200 , $80 , $CC ,
  #0 ,                          \ Terminator.
target

\ This next word takes the system clock rate (in HZ) and the required serial
\ comms rate (in BPS) and returns a system-dependant number of parameters that
\ should be sufficent to effect the baud rate selection by initialising the
\ correct hardware.
\
\ It will traverse the baud rate table looking for a best match to the required
\ speed.  Note that it "pre-scales" the required baud rate. If, for example,
\ you ask for 9600 bps but are running a system clock at twice the value used
\ to calculate the SFR values in the documentation, then the search will be
\ looking for a match for 4800 bps.

interpreter
: gen-baud-rate	\ clk-rate bps-rate -- SetMask ResetUCSR
  #3686400 8 >>n * swap 8 >>n /         \ -- (scaled)speed ;
  baud-table baud-table                 \ -- speed *best *current ;
  begin
    dup @(h)                            \ -- speed *best *current current ;
  while
    over @(h) 3 pick - abs              \ -- speed *best *current bdiff ;
    over @(h) 4 pick - abs              \ -- speed *best *current bdiff cdiff ;
    >= if
      nip dup                           \ Ditch *BEST and use *CURRENT instead.
    endif
    3 cells+                            \ Step on to next entry.
  repeat
  drop nip                              \ -- *best ;
  cell+ dup @(h) swap cell+ @(h)
;
target

system-speed console-speed gen-baud-rate
  equ ResetUCSR
  equ SetMask


\ Using the Timer as the baud rate generator:
\   SCALE =  system_clock  /  ( 16 * 2 * (  baudrate  )  )
\   SCALEH = high8bit( SCALE )
\   SCALEL = low8bits( SCALE )
\
\   example :-
\     system_clock = 7.372Mhz, baudrate = 9600
\     SCALE = 7372000 / ( 16 * 2 * ( baudrate ) ) = 24
\     SCALEH = 0, SCALEL = 24

\ use these 3 lines when using the timer as the baud rate generator
\ $DD equ ResetUCSR                                     \ timer as baud rate generator
\ $00 equ SCALEH                                        \ baud rate and prescaler high (9600)
\ $18 equ SCALEL                                        \  "     "   "       "    low  (9600)

\ use these lines when using the integral baud rate generator                   \ MSD001
\ remember to take into account the system clock differences upon integral baud rates \ MSD001

$10 equ ResetMRP                                        \ mode register pointer = UMR1
$40 SetMask or equ ResetACR                             \ Required set, no CTS int. \ MSD001
$13 equ ResetUMR1                                       \ 8 data, no parity, RxRDY int.
$07 equ ResetUMR2                                       \ 1 stop bit
$20 equ ResetRX                                         \ reset receiver
$30 equ ResetTX                                         \ reset transmitter
$01 equ EnableRX                                        \ enable receiver
$04 equ EnableTX                                        \ enable transmitter
$003C equ PBUARTMask                                    \ RX,TX,RTS,CTS enabled on Port B

$02 equ ResetUIMR                                       \ unmask RX int.
$4C equ UARTvector                                      \ UART int. vector number
$8898 equ ResetPICR                                     \ UART interrupt priority (001)
$40 equ ResetPIVR                                       \ Base of peripheral vectors - top 4 bits


internal

code INIT-SER	\ -- ; initialise 68307 UART
  # MBASE  a0  l. movea,                                \ set pointer to SIM registers
  # ResetMRP  UCR d(a0)  b. move,                       \ reset mode register pointer
  # ResetACR  UACR d(a0)  b. move,                      \ select baud rate set
  # ResetUCSR  UCSR d(a0)  b. move,                     \ select baud rate generator

\ only use 2 lines below if using the timer for baud rate generation
\  # SCALEH  UBG1 d(a0)  b. move,                       \ set timer prescaler
\  # SCALEL  UBG2 d(a0)  b. move,                       \

  # 0  UIMR d(a0)  b. move,                             \ mask all UART interrupts
  # ResetUMR1  UMR1 d(a0)  b. move,                     \ set MODE register 1
  # ResetUMR2  UMR2 d(a0)  b. move,                     \ set MODE register 2
  # ResetRX  UCR d(a0)  b. move,                        \ reset receiver
  # ResetTX  UCR d(a0)  b. move,                        \ reset transmitter
  # EnableRX  UCR d(a0)  b. move,                       \ enable receiver
  # EnableTX  UCR d(a0)  b. move,                       \ enable transmitter
  # PBUARTMask  PBCNT d(a0)  w. ori,                    \ enable RX,TX,RTX and CTS functions on Port B

\ interrupt specific initialisation
  # ResetUIMR  UIMR d(a0)  b. move,                     \ unmask RX interrupt
  # UARTvector  UIVR d(a0)  b. move,                    \ set UART interrupt vector number
  # ResetPICR  PICR d(a0)  w. move,                     \ set UART interrupt priority
  # ResetPIVR  PIVR d(a0)  b. move,                     \ write PIVR (forces perph. interrupts to be processed)

  d0 move<sr,				                \ enable all interrupts inc. level 1
  # 0F8FF d0 w. andi,			                \ insert 000 in bits 10..8, all enabled
  d0 move>sr,				                \ and put back in status register

  next,
end-code


\ *****************
\ Low level drivers
\ *****************

code (SEREMIT)	\ char --
l: pemit1
  # 2 $l USR MBASE + b. btst,		                \ test UART TX status
  pemit1 beq,                                           \ wait until ready
  d7 $l UTB MBASE + b. move,		                \ transmit character
  (a6)+ d7 move,			                \ restore TOS
  next,
end-code

variable RX-AVAIL                                       \ non-zero when character available
variable RX-CHAR                                        \ last character read

asmcode
l: UART-RX-ISR
  d0 -(a7) l. move,			                \ save registers used
  # 0 d0 l. moveq,			                \ clear receiving register
  $l URB MBASE + d0 b. move,	                        \ fetch it
  d0 $l rx-char l. move,		                \ store it
  # -1 $l rx-avail l. move,		                \ set character available flag
  (a7)+ d0 l. move,			                \ restore registers used
  rte,
end-code

uart-rx-isr  start-vector UARTvector 4 * +  !

code (SERKEY)	\ -- char
  d7 -(a6) move,			                \ save old TOS
l: pkey1
  $l rx-avail d0 l. move,   pkey2 bne,	                \ test UART RX status
     # $00000008  $l SCR  l. ori,                       \ turn CPU off
     # $02000 stop,			                \ permit all interrupts while asleep
     pkey1 bra,
l: pkey2
  $l rx-char d7 l. move,		                \ fetch it
  # 0 $l rx-avail l. move,		                \ reset char. avail. flag
  next,
end-code

code (SERKEY?)	\ -- flag 
  d7 -(a6) move,			                \ save old TOS
  $l rx-avail d7 l. move,		                \ test UART RX status
  next,
end-code

external


\ ***********************
\ Generic i/o assignments
\ ***********************

internal

tasking? [if]
: (mserkey)	\ -- char ; tasks wile waiting
  begin
    (serkey?) 0=
   while
    pause
  repeat
  (serkey)
;
[then]

: (serTYPE)	\ c-addr len --
  bounds
  ?do  i c@ (seremit)  loop
;

: (serCR)		\ --
  $0D (seremit)  $0A (seremit)
;

external

cdata
create Console	\ -- addr ; OUT managed by upper driver
tasking? [if]
  ' (mserkey) ,
[else]
  ' (serkey) ,		\ -- char ; receive char
[then]
  ' (serkey?) ,		\ -- flag ; check receive char
  ' (seremit) ,		\ -- char ; display char
  ' (sertype) ,		\ caddr len -- ; display string
  ' (sercr) ,		\ -- ; display new line

decimal

