\ 68332 interrupt RX, polled TX, serial communication drivers

((
        Copyright (c) 1988-1998
        MicroProcessor Engineering
        133 Hill Lane
        Southampton SO15 5AF
        England

        tel: +44 2380 631441
        fax: +44 2380 339691
        net: mpe@mpeltd.demon.co.uk
             tech-support@mpeltd.demon.co.uk
        web: www.mpeltd.demon.co.uk
))

hex

\ --------------------------------------------------------------------------

\ This next word takes the system clock rate (in HZ) and the required serial
\ comms rate (in BPS) and returns a system-dependant number of parameters that
\ should be sufficent to effect the baud rate selection by initialising the
\ correct hardware.

interpreter
: gen-baud-rate	\ clk-rate bps-rate -- sys-scbr
  #32 * /
;
target

system-speed console-speed gen-baud-rate equ sys-scbr

\ ------------------------------------------------------------------------------

\ serial comms receive interrupt.
\ It is always a good idea to use an interrupt to buffer received serial data.
\ n.b. unlike most uarts, the 68332 can send and receive data packets of
\ more than 8 bits.  Therefore characters received by this interrupt routine
\ are stored as 16-bit numbers.
\ See SIM manual, section 5.

$40     equ sci-int-vector#		\ use this exception vector

$40     equ rx-buf-size

rx-buf-size buffer: rx-buffer		\ input buffer
cell buffer: rx-in			\ input ponter
cell buffer: rx-out			\ output pointer

\ ------------------------------------------------------------------------------

asmcode
l: sci-int				\ interrupt entry point
  { a0-1 } -(a6) l. movem,		\ save registers

  # QMCR a0 l. movea, 			\ point to QSM block
  &SCSR d(a0) a1 w. movea,		\ read status register to enable
					\ interrupt acknowledge
  $l rx-in a1 l. movea,			\ a1 = pointer to character destination
  &SCDR d(a0) (a1)+ w. move,		\ read & store character
					\ (clears status bits)
  # rx-buffer rx-buf-size +		\ end of buffer + 1
  a1 l. cmpa,				\ make sure pointer wraps around
  eq, if,
    # rx-buffer $l rx-in l. move,	\ to start of buffer
  else,
    a1 $l rx-in l. move,		\ or to next character
  endif,

  (a6)+ { a0-1 } l. movem,		\ restore registers
  rte,
end-code
sci-int start-vector sci-int-vector# 4 * + !	\ set exception vector


\ ------------------------------------------------------------------------------

\ serial communications interface initialisation
\ See 68332 manual, section 5, p5-57 et seq.

code init-ser				\ initialise SCI & and set baud rate
  # rx-buffer $l rx-in	l. move,	\ reset buffer pointers
  # rx-buffer $l rx-out l. move,

  # QMCR a0 l. movea,			\ point to QSM
  # $008F (a0) w. move, 		\ QSM enable, supervisor-only module,
					\ interrupt arbitration level 14
  # $7 &QILR  d(a0) b. move, 	        \ sci interrupt level 5
  # sci-int-vector# &QIVR d(a0) b. move, \ set int vector in QSM
  # $0000 &SCCR1 d(a0) w. move, 	\ clear &SCCR1 to disable everything
  # sys-scbr &SCCR0 d(a0) w. move, 	\ &SCCR0: set to required rate,
					\ 8 data bits, no parity
  d1 w. move<sr,                        \ get SR and mask
  # $F8FF d1 w. and,
  # $0500 d1 w. or,
  d1 w. move>sr,                        \ put it back

  # $0002C &SCCR1 d(a0) w. move, 	\ enable transmitter, receiver,
					\ and receiver interrupts
  next,
end-code

\ ------------------------------------------------------------------------------

code (ser-emit?)	\ -- flag ; check to see if free to transmit a character
  d7 -(a6) l. move,			\ save old TOS
  # QMCR a0 l. movea,			\ point to QSM block
  &SCSR d(a0) d2 w. move,		\ fetch status word
  d7 l. clr,				\ set false flag
  # $0100 d2 w. andi, ne,		\ test TDRE
  if,	# 1 d7 l. subq,  endif,	\ set true flag
  next,
end-code

\ ------------------------------------------------------------------------------

\ multi-tasking (EMIT)
\ n.b. there is rarely any practical benefit in having interrupt-driven
\ character transmission, as in most situations one would have to poll to
\ see if the transmit buffer was empty rather than the uart transmitter.

: (ser-emit)	\ char --
  begin
    (ser-emit?)                         \ pause until free to transmit
  until
  [ QMCR &SCDR + ] literal w!           \ transmit character
;

\ ------------------------------------------------------------------------------

code (ser-key?)	\ -- flag
  d7 -(a6) move,		        \ save old TOS
  d7 l. clr,
  $l rx-in d0 l. move,	                \ a character has been received if
  $l rx-out d0 l. cmp,   ne, if,        \ input & output pointers are
    # true d7 l. moveq,		        \ different
  endif,
  next,
end-code

code (ser-key@)	\ -- char
  d7 -(a6) move,			\ save old TOS
  $l rx-out a1 l. movea,		\ pointer to received character
  d7 l. clr,
  (a1)+ d7 w. move,			\ retrieve character
  # $00FF d7 w. andi,                   \ remove rubbish
  # rx-buffer rx-buf-size +		\ end of buffer + 1
  a1 l. cmpa, eq,			\ make sure pointer wraps around
  if,    # rx-buffer $l rx-out l. move, \ to start of buffer
  else,  a1 $l rx-out l. move,
  endif,
  next,
end-code

code tosleep	\ -- ; put CPU to sleep until next interrupt
  # $02000 stop,                        \ permit all interrupts while asleep
  next,
end-code

: (ser-key)	\ -- n
  begin
    (ser-key?) 0=
   while
    tosleep
  repeat
  (ser-key@)
;


\ ***********************
\ Generic i/o assignments
\ ***********************

internal

tasking? [if]
: (mser-key)	\ -- char ; tasks wile waiting
  begin
    (ser-key?) 0=
   while
    pause
  repeat
  (ser-key@)
;
[then]

: (ser-TYPE)	\ c-addr len --
  bounds
  ?do  i c@ (ser-emit)  loop
;

: (ser-CR)	\ --
  $0D (ser-emit)  $0A (ser-emit)
;

external

cdata
create Console	\ -- addr ; OUT managed by upper driver
tasking? [if]
  ' (mser-key) ,
[else]
  ' (ser-key) ,		\ -- char ; receive char
[then]
  ' (ser-key?) ,	\ -- flag ; check receive char
  ' (ser-emit) ,	\ -- char ; display char
  ' (ser-type) ,	\ caddr len -- ; display string
  ' (ser-cr) ,		\ -- ; display new line

decimal

