\ Forth Systeme 68332 startup code

((
Copyright (c) 2000, 2003
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

PLEASE NOTE OUR PHONE NUMBER CHANGE

20031216 SFP003 Added DOCGEN commentary for v6.2
20000703 SFP002 Converted to v6.1 model
19981019 MSD001 Converted for optimised model (A3 == 0).
))


\ ==========
\ *! init68k
\ *T Generic start up
\ ==========
\ *P The file 68K\HARDWARE\FS332\INIT332.FTH contains the generic
\ ** vector table, Forth start up code and Forth initialisation
\ ** tables. It must be the first file compiled that generates code
\ ** unless you are compiling operating system specific code, such
\ ** as with an O/S that requires a header.

\ *P This file is an example for a 68332. See the CTL files in
\ ** CONFIGS folder to find examples of other initialisation 
\ ** files.

\ *S Exception Vectors
\ *P At power up, the CPU branches by loading the stack pointer
\ ** and PC from absolute address 0. By default, the PC is
\ ** loaded with the address START332, which contains CPU
\ ** specific startup code for chip selects and the external
\ ** bus. This code falls through to label ECLD, which
\ ** is the start of the Forth initialisation code.

\ *P By default a 1kb area at START-VECTOR is reserved for the
\ ** exception table.


\ ------------------------------------------------------------------------------
\       start up table and entry points
\ ------------------------------------------------------------------------------

l: START-VECTOR                                         \ start up vector
  INIT-S0 ,                                             \ top of data stack
  0 ,                                                   \ PC entry point, filled in later

here $0400 8 - erase                                    \ clear vector area

start-vector $0400 + Corg                               \ reserve space for ROM vectors

L: CLD1                                                 \ place holder for xt of word to run
  0 ,
L: WRM1                                                 \ place holder for xt of word to run
  0 ,

\ ------------------------------------------------------------------------------

asmcode
l: start332
\ keep all ints disabled for now, can enable then using seti during initialisation
 # $2700 sr w. move,            \ supervisor mode, only int level 7

\ SIM module
 # $FFFFFA00 a0 l. movea,       \ point to SIM module
 # 6 $20 d(a0) w. move,         \ disable watchdog, set bus monitor
 # $4000 4 d(a0) w. ori,        \ set up SYNCR X bit to 16.67 MHz

\ Set up boot eprom
 # $0006 $48 d(a0) w. move,     \ set eprom size 7 for 1024, 6 for a 27512
 # $6CB1 $4A d(a0) w. move,     \ boot eprom options SJC

\ RAM CS uses CS3
 # $8007 $58 d(a0) w. move,     \ set ram base to 80 0000 size 1M
 # $7831 $5A d(a0) w. move,     \ set ram select configuration SJC

\ FLASH uses CS4
 # $A007 $5C d(a0) w. move,     \ base of A0 0000 size 1M
 # $7CB0 $5E d(a0) w. move,

\ PIT uses CS5
 # $F000 $60 d(a0) w. move,     \ PIT maps at F0 0000 for 2K
 # $7FF0 $62 d(a0) w. move,     \ set cs configuration

\ RTC uses CS10
 # $D000 $74 d(a0) w. move,     \ RTC maps at D0 0000 for 2K
 # $5CB0 $76 d(a0) w. move,     \ set cs configuration


\ interrupt autovector generation uses CS7
 # $FA   $47 d(a0) b. move,     \ set CS6 and CS7 as 8bit ports
 # $FFF9 $68 d(a0) w. move,     \ chip select base register, size 8K
 # $4801 $6A d(a0) w. move,     \ read only, upper byte, cpu space, avec on

\ CSEXT uses CS6  (needs to be changed if 16 bit boot used)
 # $FA01 $64 d(a0) w. move,     \ FA 0000 for 8K
 # $7DF0 $66 d(a0) w. move,
\ # $7CB0 $66 d(a0) w. move,     \ 2 ws

\ Join main Forth entry point

L: ECLD
  # init-s0 a6 l. movea,                \ set data stack
  # init-r0 a7 l. movea,                \ set return stack
  # init-u0 a5 l. movea,                \ set user pointer
                                        \ A4 = locals pointer, self setting     
  # 0 a3 l. movea,                      \ the system needs A3=0                 

  $l cld1 a1 l. movea,                  \ get address of word to run            
  (a1) jsr,                             \ run it                                
  ecld bra,                             \ should never happen, but ...

L: EWRM
  # init-s0 a6 l. movea,                \ set data stack
  # init-r0 a7 l. movea,                \ set return stack
  # init-u0 a5 l. movea,                \ set user pointer
                                        \ A4 = locals pointer, self setting     
  # 0 a3 l. movea,                      \ the system needs A3=0                 

  $l wrm1 a1 l. movea,                  \ get address of word to run            
  (a1) jsr,                             \ run it                                
  ewrm bra,                             \ should never happen, but ...
end-code

start332 start-vector 4 + !             \ patch up entry vector
ecld wrm1 !                             \ default if never set


\ ------------------------------------------------------------------------------
\       reset values for user and system variables
\ ------------------------------------------------------------------------------

[undefined] sp-guard [if]
0 equ sp-guard
[then]

[undefined] umbilical? [if]
\ initial values of user variables
L: USER-RESET
  init-s0 tos-cached? sp-guard + cells - ,	\ s0
  init-r0 ,				\ r0
  0 ,  0 ,				\ #tib, 'tib
  0 ,  0 ,                              \ >in, out
  $0A ,  0 ,                            \ base, hld

  \ initial values of system variables
L: INIT-FENCE 0 ,                       \ fence
L: INIT-DP 0 ,                          \ dp
L: INIT-VOC-LINK 0 ,                    \ voc-link
[then]


\ ======
\ *> ###
\ ======

