\ 68000 Interrupt handlers

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


20000630 SFP005 Brought up to version 6.1, and added [I and I]
19990310 SFP004 Fixed DI - perhaps we should fix SETI instead
19981019 MSD003 Remove use of A4 since that is now used as the locals pointer.
         MSD003 I don't know why A4 was used instead of, for example, A0.  To
         MSD003 facilitate a change to a *different* register, I've chosen to
         MSD003 replace all references to A4 with A2, since A2 hasn't been
         MSD003 used in here at all.
19980409 MSD002 Apply COLDFIRE modifications.
19980311 MSD001 Added EI and DI to make the 68K word-set compatible with
         MSD001 other processors.
))

\ =========
\ *! int68k
\ *T Interrupt handlers
\ =========
\ *P The file 68K\INT68K.FTH contains generic ARM interrupt
\ ** handlers in the v6.2 style.

\ *P The file 68K\TRAPS.FTH may be compiled after this file.
\ ** TRAPS.FTH provides diagnostic information and other
\ ** code such as divide by zero handling.


only forth definitions
decimal

Is-ColdFire? [if]
  #15 cells equ INT-STACK       \ # bytes to stash registers for interrupts
[then]


\ ***********************
\ *S Interrupt management
\ ***********************

code SETI	\ x -- ; set interrupt level x
\ *G Global set interrupts to level n, where x=$0n00.
Is-ColdFire? [if]
  # $0700 d7 and,                       \ mask interupt bits    \ MSD002
  d1 w. move<sr,                        \ get SR and mask       \ MSD002
  # $F8FF d1 and,                       \ MSD002
  d7 d1 or,                             \ MSD002
[else]
  # $0700 d7 w. and,                    \ mask interupt bits
  d1 w. move<sr,                        \ get SR and mask
  # $F8FF d1 w. and,
  d7 d1 w. or,
[then]
  d1 w. move>sr,                        \ put it back
  (a6)+ d7 move,   next,
end-code

: ei		\ -- ; Enable interrupts
\ *G Global enable interrupts to level 0.
  0 seti
;

: di		\ -- ; Disable interrupts
\ *G Global disable interrupts to level 7.
  $0700 seti
;

code SAVE-INT	\ -- x ; save interrupt status
\ *G Geturn interrupt status and then disable interrupts.
\ ** Use [I and I] for new code.
  d7 -(a6) move,                        \ save old TOS
  d0 l. clr,                            
  d0 w. move<sr,                        \ get SR
  d0 d7 move,                           \ put on stack. no ints
Is-ColdFire? [if]
  # $0700 d0 or,
[else]
  # $0700 d0 w. or,
[then]
  d0 w. move>sr,			\ put back to SR
  next,
end-code

code RESTORE-INT	\ x -- ; restore interrupt
\ *G Restore state returned by SAVE-INT.
\ ** Use [I and I] for new code.
  d7 w. move>sr,                        \ restore SR
  (a6)+ d7 move,                        \ restore TOS
  next,
end-code

heads? [if]
[undefined] [I [if]
code [I		\ R: -- x ; save SR on return stack and disable interrupts
\ *G Preserve I/F status on return stack, disable interrupts.
\ ** The state is restored by I].
  (a7)+  a0  l. movea,			\ return address
  sr  -(a7)  move,
  # $0700 sr ori,
  (a0)  jmp,
end-code

code I]		\ R: -- x ; restore SR from return stack
\ *G Restore interrupt status saved by [I from the return stack.
  (a7)+  a0  l. movea,			\ return address
  (a7)+  sr  move,			\ restore status
  (a0)  jmp,
end-code
[then]
[then]


\ *****************************
\ *S Interrupt service routines
\ *****************************
\ *P Each predefined ISR has four components:
\ *B A DEFERred word into which you ASSIGN the required action.
\ *B A code fragment which references the DEFERred word and RETI.
\ *B An assembler entry point.
\ *B A vector assignment

asmcode
\ switch user page, and call forth
l: next-int
  # int-init-u0 a5 movea,               \ point to user area
  (a2) jmp,                             \ go run specific handler \ MSD003
end-code

code RETI	\ --
\ *G Return from high level interrupt.
l: [hreti]                              \ return from interrupt
  (a7)+ d0 move,                        \ throw away return address
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    (a7) { d0-7 a0-6 } l. movem,        \ restore registers     \ MSD002
    # INT-STACK a7 l. adda,                                                     \ MSD002
  )else(                                                                        \ MSD002
    (a7)+ { D0-7 a0-6 } l. movem,       \ restore registers
  )endif                                                                        \ MSD002
  rte,
end-code


defer BUSERR-ISR	\ --
\ *G The deferred word for a bus errot

l: buserr-int
\ *G The code fragment which references BUSERR-ISR.
 ] buserr-isr reti [                    \ action and return

proc buserr-int0
\ *G The assembly level entry point for the bus error handler.
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # buserr-int a2 l. movea,             \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

buserr-int0 start-vector 2 4 * + !	\ set vector


\ *P The predefined DEFERred words are:
\ *E BUSERR-ISR ADDRERR-ISR ILLOP-ISR DIVZERO-ISR
\ ** CHKINST-ISR TRAPVINST-ISR PRIVVIOL-ISR TRACE-ISR
\ ** OPCODE1010-ISR OPCODE1111-ISR PHANTOM-ISR 
\ ** AVECn-ISR where n=1..7
\ *P All these words have their associated other three
\ ** components.


defer ADDRERR-ISR

l: addrerr-int
 ] addrerr-isr reti [                   \ action and return

proc addrerr-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # addrerr-int a2 l. movea,            \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

addrerr-int0 start-vector 3 4 * + !     \ set vector


defer ILLOP-ISR

l: illop-int
 ] illop-isr reti [                     \ action and return

proc illop-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # illop-int a2 l. movea,              \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

illop-int0 start-vector 4 4 * + !       \ set vector


defer DIVZERO-ISR

l: divzero-int
 ] divzero-isr reti [                   \ action and return

proc divzero-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # divzero-int a2 l. movea,            \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

divzero-int0 start-vector 5 4 * + !     \ set vector


defer CHKINST-ISR

l: chkinst-int
 ] chkinst-isr reti [                   \ action and return

proc chkinst-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # chkinst-int a2 l. movea,            \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

chkinst-int0 start-vector 6 4 * + !     \ set vector


defer TRAPVINST-ISR

l: trapvinst-int
 ] trapvinst-isr reti [                 \ action and return

proc trapvinst-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # trapvinst-int a2 l. movea,          \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

trapvinst-int0 start-vector 7 4 * + !   \ set vector


defer PRIVVIOL-ISR

l: privviol-int
 ] privviol-isr reti [                  \ action and return

proc privviol-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # privviol-int a2 l. movea,           \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

privviol-int0 start-vector 8 4 * + !    \ set vector


defer TRACE-ISR

l: trace-int
 ] trace-isr reti [                     \ action and return

proc trace-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # trace-int a2 l. movea,              \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

trace-int0 start-vector 9 4 * + !       \ set vector


defer OPCODE1010-ISR

l: opcode1010-int
 ] opcode1010-isr reti [                \ action and return

proc opcode1010-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # opcode1010-int a2 l. movea,         \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

opcode1010-int0 start-vector 10 4 * + ! \ set vector


defer OPCODE1111-ISR

l: opcode1111-int
 ] opcode1111-isr reti [                \ action and return

proc opcode1111-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # opcode1111-int a2 l. movea,         \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

opcode1111-int0 start-vector 11 4 * + ! \ set vector


defer PHANTOM-ISR

l: phantom-int
 ] phantom-isr reti [                   \ action and return

proc phantom-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # phantom-int a2 l. movea,            \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

phantom-int0 start-vector 24 4 * + !    \ set vector


defer AVEC1-ISR

l: avec1-int
 ] avec1-isr reti [                     \ action and return

proc avec1-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # avec1-int a2 l. movea,              \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

avec1-int0 start-vector 25 4 * + !      \ set vector


defer AVEC2-ISR

l: avec2-int
 ] avec2-isr reti [                     \ action and return

proc avec2-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # avec2-int a2 l. movea,              \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

avec2-int0 start-vector 26 4 * + !      \ set vector


defer AVEC3-ISR

l: avec3-int
 ] avec3-isr reti [                     \ action and return

proc avec3-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # avec3-int a2 l. movea,              \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

avec3-int0 start-vector 27 4 * + !      \ set vector


defer AVEC4-ISR

l: avec4-int
 ] avec4-isr reti [                     \ action and return

proc avec4-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # avec4-int a2 l. movea,              \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

avec4-int0 start-vector 28 4 * + !      \ set vector


defer AVEC5-ISR

l: avec5-int
 ] avec5-isr reti [                     \ action and return

proc avec5-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # avec5-int a2 l. movea,              \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

avec5-int0 start-vector 29 4 * + !      \ set vector


defer AVEC6-ISR

l: avec6-int
 ] avec6-isr reti [                     \ action and return

proc avec6-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # avec6-int a2 l. movea,              \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

avec6-int0 start-vector 30 4 * + !      \ set vector


defer AVEC7-ISR

l: avec7-int
 ] avec7-isr reti [                     \ action and return

proc avec7-int0
  Is-ColdFire?                                                                  \ MSD002
  if(                                                                           \ MSD002
    # INT-STACK a7 l. suba,             \ predecrement stack ptr \ MSD002
    { d0-7 a0-6 } (a7) l. movem,        \ save all registers    \ MSD002                        \ MSD002
  )else(                                                                        \ MSD002
    { d0-7 a0-6 } -(a7) l. movem,       \ save all registers
  )endif                                                                        \ MSD002
  # avec7-int a2 l. movea,              \ set up new IP         \ MSD003
  next-int w. bra,                      \ point to int, and go
end-code

avec7-int0 start-vector 31 4 * + !      \ set vector


\ ******
\ *> ###
\ ******

