\ 68xxx Multi-Tasker

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

PLEASE NOTE OUR PHONE NUMBER CHANGE

20031216 SFP002 Added DOCGEN documentation.
		Added SLEEPER and HIS.
20000630 SFP001 Rewritten for v6.1 model


*************************
TCB data structure layout
*************************

cell	LINK	link to next task
cell	SSP	Saved Stack Pointer
cell	STAT	BIT 0   1 = running, 0 = halted
                BIT 1   1 = message pending
                BIT 2   1 = event triggered
                BIT 3   1 = event handler run
	        others  1 = set to run task, available to user
cell	TASK    Task # that sent message here
cell	MESG    Message address
cell	EVNTw   CFA of word run as event handler

This structure is allocated at the start of the USER area. Consequently
the TCB of the current task is given by UP.


********************
Resource definitions
********************
 
68xxx/Coldfire Target Model: Subroutine Threaded Code
  A7 = return stack
  A6 = data stack
  A5 = user pointer
  A4 = locals pointer (when target code converted)
  A3 = zero
  A2 = scratch
  A1 = scratch
  A0 = scratch

  D7 = TOS
  D6 = scratch
  D5 = scratch
  D4 = scratch
  D3 = scratch
  D2 = scratch
  D1 = reserved for compiler (holds status flags)
  D0 = reserved for compiler (Shuffle register)
))

\ ===========
\ *! multi68k
\ *T 68xxx multitasker
\ ===========
\ *P The 68xxx multitasker follows the model introduced
\ ** with the v6.1 compilers. A few extensions are also
\ ** provided.


\ *************
\ *S Configuration - normally performed earlier
\ *************

[undefined] test-multi? [if]
0 equ test-multi?	\ true to compile test code
\ *G If not previously defined, TEST-MULTI? is set to zero
\ ** and test code is not compiled.
[then]


\ ****************************
\ *S TCB data structure layout
\ ****************************
\ *E cell       LINK    link to next task
\ ** cell       SSP     Saved Stack Pointer
\ ** cell       STAT    BIT 0   1 = running, 0 = halted
\ **                    BIT 1   1 = message pending
\ **                    BIT 2   1 = event triggered
\ **                    BIT 3   1 = event handler run
\ ** 	        others  1 = set to run task, available to user
\ ** cell       TASK    Task # that sent message here
\ ** cell       MESG    Message address
\ ** cell       EVNTw   CFA of word run as event handler
\ *P This structure is allocated at the start of the USER area. Consequently
\ ** the TCB of the current task is given by UP.

struct /TCB	\ -- size
\ *G The structure used by the code that matches the
\ ** description above.
  ptr tcb.link		\ link to next task ; MUST BE FIRST
  ptr tcb.ssp	        \ Saved Stack Pointer
  int tcb.status	\ status word
  int tcb.msrc		\ message source
  int tcb.mesg		\ message
  ptr tcb.event		\ xt of word which is event handler
end-struct

\ bit masks for the status cell
$0001 equ run-mask  run-mask invert equ ~run-mask	\ running
  0 equ run-bit#
$0002 equ msg-mask  msg-mask invert equ ~msg-mask	\ message available
  1 equ msg-bit#
$0004 equ trg-mask  trg-mask invert equ ~trg-mask	\ event triggered
  2 equ trg-bit#
$0008 equ evt-mask  evt-mask invert equ ~evt-mask	\ event run
  3 equ evt-bit#


\ ******************
\ Consistency checks
\ ******************

/tcb tcb-size <> [if]
  cr ." Control file and tasker disagree as to TCB size"  abort
[endif]


\ ***************************
\ *S Task handling primitives
\ ***************************

init-u0 constant main	\ -- addr ; tcb of main task
\ *G Returns the base address of the main task's USER area.
l: main-link
  0 ,					\ end of task chain

0 value multi?	\ -- flag ; true if tasker enabled
\ *G Returns true if the tasker is enabled.

: single	\ -- ; disable scheduler
\ *G Disable scheduler.
  0 to multi?
;

: multi		\ -- ; enable scheduler
\ *G Enable scheduler.
  -1 to multi?
;

Is-ColdFire? [if]
  #15 cells equ MULT-STACK		\ number of cells to save on stack
[then]

CODE pause      \ -- ; the scheduler
\ *G The software scheduler itself.
  $l ' multi? >body  d0 l. move,	\ do nothing if scheduler disabled
  eq, if,  rts,  endif,

\
\ save previously running task ip, rp, up, sp
\ can use with interrupts if we stack everything
\
Is-ColdFire? [if]
  # MULTI-STACK a7 l. suba,             \ adjust stack pointer
  { d0-7 a0-6 } (a7) l. movem,          \ save registers
[else]
  { d0-7 a0-6 } -(a7) l. movem,         \ stack the lot
[then]
  a7  0 tcb.ssp d(a5)  l. move,		\ save SP in TCB

\
\ select next task to run
\
  begin,
    0 tcb.link d(a5) a5 l. movea,	\ get next User Area
    0 tcb.status d(a5) d0 l. move,	\ get its status
  ne, until,

\
\ run selected task - sp, up, rp, ip
\
  0 tcb.ssp d(a5)  a7 l. movea,		\ restore saved stack pointer
Is-ColdFire? [if]
  { d0-7 a0-6 } (a7) l. movem,		\ restore registers
  # MULTI-STACK a7 l. adda,		\ adjust stack pointer
[else]
  (a7)+  { d0-7 a0-6 }  l. movem,       \ restore registers
[then]

event-handler? [if]	\ if user wants the event handler
\
\ event handler exit
\
  0 tcb.status d(a5)  d0  l. move,	\ look at status
  # trg-bit#  d0  bclr,			\ inspect event trigger bit (and clear it)
  ne, if,
    # evt-bit#  d0  bset,		\ set event bit
    d0  0 tcb.status d(a5)  l. move,	\ restore status
    0 tcb.event d(a5)  a0  l. movea,	\ run event handler
    (a0)  jmp,
  endif,
[then]

  next,
end-code

code status	\ -- task-status
\ *G Returns the current task's status cell, but with the run
\ ** bit masked out.
  d7 -(a6) l. move,			\ save TOS
  0 tcb.status d(a5)  d7  l. move,	\ get status
  # run-bit#  d7  bclr,			\ clear running bit
  next,
end-code

CODE restart    \ task -- ; mark task TCB as running
\ *G Sets the RUN bit in the task's status cell.
  d7 a0 l. movea,
  # run-bit#  3 tcb.status d(a0)  bset,	\ set task running bit (in low byte)
  (a6)+ d7 move,
  next,
END-CODE

CODE halt       \ task# -- ; reset running bit in TCB
\ *G Clears the RUN bit in the task's status cell.
  d7 a0 l. movea,
  # run-bit#  3 tcb.status d(a0)  bclr,	\ clear task running bit (in low byte)
  (a6)+ d7  l. move,
  next,
END-CODE

: stop          \ -- ; halt oneself
\ *G HALT's the current task, and executes PAUSE.
  self halt  pause
;


\ **************
\ *S Event handling
\ **************
\ *P Event handling is only compiled if the equate
\ ** EVENT-HANDLER? is set non-zero in the control
\ ** file.

event-handler? [if]

: set-event	\ task -- ; set event trigger in task TCB
\ *G Set the event trigger in task TCB.
  tcb.status dup @  trg-mask or  swap !
;

: event?	\ task -- flag ; true if task had event
\ *G Returns true if true if task has received an event
\ ** trigger which has not been cleared yet.
  tcb.status @  trg-mask and 0<>
;

: clr-event-run	\ -- ; reset own EVENT_RUN flag
\ *G Reset the current task's EVENT_RUN flag.
  up@ tcb.status dup @  ~evt-mask and  swap !
;

: to-event      \ xt task -- ; define action of a task
\ *G Sets XT as the event handler for the task.
  tcb.event !				\ set into TCB of task
;
[then]


\ ****************
\ *S Message handling
\ ****************
\ *P Message handling is only compiled if the equate
\ ** MESSAGE-HANDLER? is set non-zero in the control
\ ** file.

message-handler? [if]
: msg?		\ task -- flag ; true if task has message
\ *G Returns true if task has received a message.
  tcb.status @ msg-mask and 0<>
;

internal
: (send-msg)	\ mesg task -- ; store message in task
  self over tcb.msrc !			\ set source
  dup
  tcb.status dup @  msg-mask or  swap !	\ set message bit
  tcb.mesg !				\ set message
;

: (get-message)	\ -- mesg task
  up@ tcb.mesg @			\ message
  up@ tcb.msrc @			
  up@ tcb.status
  dup @  ~msg-mask and  swap !		\ reset message bit
;
external

: send-message  \ addr task -- ; send message to task
\ *G Send a message to a task.
  begin  dup msg?               	\ if task already has a message
    while  pause                	\ wait
  repeat                        	\ and try again
  (send-msg)				\ send the message
;

: get-message   \ -- addr task ; wait for any message
\ *G Wait for any message and return the message and the task it
\ ** came from.
  begin  self msg? 0=           	\ check for a pending message
    while  pause                	\ wait for it
  repeat                        	\ try again
  (get-message)                 	\ read it
;
[then]

event-handler? message-handler? and [if]
: wait-event/msg        \ -- ; wait for message or event trigger
\ *G Wait for a message or an event trigger.
  begin
    self msg?  self event?  or 0=
   while
    pause
  repeat
;
[then]


\ *************************
\ *S Task structure management
\ *************************

internal

(( Extract from scheduler save
  { d0-7 a0-6 } -(a7) l. movem,         \ stack the lot
  a7  0 tcb.ssp d(a5)  l. move,		\ save SP in TCB
))
code init-task	\ xt task -- ; Initialise a task stack
\ *G Initialise a task's stack before running it and
\ ** set it to execute the word whose XT is given.
  d7 a0 l. movea,			\ A0 = TCB
  a3  0 tcb.status d(a0)  l. move,	\ clear task status
  d7 a1 l. movea,			\ A1 = new RSP/SSP
  # task-r0 a1 l. adda,
  a1  r0-offset d(a0)  l. move,		\ set new R0

  (a6)+ -(a1) l. move,			\ save entry point, new PC
  d7 a2 l. movea,			\ save A6, new PSP
  # task-s0 a2 l. adda,
  a2  -(a1)  l. move,
  # 4  a2  suba,			\ and set new S0 according to rules of SP@ SP!
  a2  s0-offset d(a0)  l. move,
  d7 -(a1) l. move,			\ save A5, new UP
  a3 -(a1) l. move,			\ save A4, new LP
  a3 -(a1) l. move,			\ save A3, zero register
  a3 -(a1) l. move,			\ save A2
  a3 -(a1) l. move,			\ save A1
  a3 -(a1) l. move,			\ save A0
  a3 -(a1) l. move,			\ save D7
  a3 -(a1) l. move,			\ save D6
  a3 -(a1) l. move,			\ save D5
  a3 -(a1) l. move,			\ save D4
  a3 -(a1) l. move,			\ save D3
  a3 -(a1) l. move,			\ save D2
  a3 -(a1) l. move,			\ save D1
  a3 -(a1) l. move,			\ save D0
  a1  0 tcb.ssp d(a0)  l. move,		\ save SSP

  (a6)+ d7 move,
  next,
end-code

: add-task      \ task -- ; insert into list
\ *G Add the task to the list of tasks after the current task.
  self tcb.link @			\ save task currently pointed to
  over tcb.link !			\ and make new task point to it
  self tcb.link !			\ current task points to new task
;

: sub-task      \ task -- ; remove task from chain
\ *G Remove the task from the task list.
\ look for task which points to required task
  self					\ head of chain ; -- target current
  begin  2dup tcb.link @ <>		\ while curr. does not point to target
    while  tcb.link @			\ step through chain
           dup self =			\ but if task not in list, check for end
           if  2drop  exit  endif       \ and get out
  repeat                                \ -- target points-to-it
\ set task that points to target to point to task pointed to by the target
  swap tcb.link @			\ task pointed to by target
  swap tcb.link !			\ set into link field
;

external
    
: initiate	\ xt task -- ; start task from scratch and run it
\ *G Start the given task executing the word whose XT is given, e.g.
\ *C   ['] <name> <task> INITIATE
  tuck init-task			\ reset TCB
  dup add-task                          \ add to chain
  run-mask swap tcb.status !		\ mark as running
  pause					\ let it run
;

: sleeper	\ xt task -- ; start task from scratch, but leave it HALTed
( *G Use in the form: )
( *E   ['] <action> <taskname> SLEEPER )
( *P to put a task on the active task list, but as if HALTed. SLEEPER )
( ** allows you to make a task ready for waking up later, perhaps )
( ** by another task. This avoids having to put STOP as the first )
( ** word in a task. Note that SLEEPER does not call PAUSE. See also )
( ** INITIATE. )
  tuck init-task			\ reset TCB
  dup add-task                          \ add to chain
  tcb.status off			\ mark as idle
;

: terminate	\ task -- ; stop task, and remove from list
\ *G Stop a task, and remove it from the list.
  sub-task pause			\ halt task, and remove from chain
;

: init-multi    \ -- ; initialisation with multi-tasking
\ *G Initialise the multitasker and start it. If tasking
\ ** is selected by setting the equate TASKING? in the control
\ ** file, KERNEL62.FTH will automatically run this word.
\ ** Make sure that your initialisation code includes
\ ** INIT-MULTI or your code will crash.
  single				\ tasker disabled
  main main tcb.link !			\ initialise MAIN to point to itself
  run-mask main tcb.status !		\ and reset status field
  multi					\ mark tasker enabled
;

: his		\ task uservar -- addr ; produce address of user var in another task
\ *G Given a task id and a USER variable, returns the address of that variable
\ ** in the given task. This word is used to set up USER variables in other
\ ** tasks.
  self - +
;


\ **********
\ *S semaphores
\ **********
\ *P The semaphore code is only compiled if the equate
\ ** SEMAPHORES? is set non-zero in the control
\ ** file.

semaphores? [if]

\ *P A SEMAPHORE is an extended variable used for signalling between tasks,
\ ** and for resource allocation.
\ ** The counter field is used as a count of the number of times the
\ ** resource may be used, and the arbiter field contains the TCB of the
\ ** task that last gained access. This field can be used for priority
\ ** arbitration and deadlock detection/arbitration.

interpreter
: semaphore     \ -- ; -- addr [child]
  idata create
    0 ,  0 ,				\ count and arbiter fields
;
target

heads? [if]
: semaphore	\ -- ; -- addr [child]
\ *G Creates a semaphore which returns its address at runtime.
\ ** Use in the form:
\ *C Semaphore <name>
  create
    0 ,  0 ,				\ count and arbiter fields
;
[then]

: signal        \ addr --
\ *G SIGNAL increments the counter field of a semaphore, indicating
\ ** either that another item has been allocated to the resource, or
\ ** that it is available for use again, 0 indicating in use by a task
  [I					\ must be interrupt safe
    dup incr  cell+ off			\ inc. counter, release
  I]
;

: request       \ sem -- ; get access to resource, wait if count = 0
\ ** REQUEST waits until the counter field of a semaphore is non-zero,
\ ** and then decrements the counter field by one.
\ ** This allows the semaphore to be used as a COUNTED semaphore. For
\ ** example a character buffer may be used where the semaphore counter
\ ** shows the number of available characters.
\ ** Alternatively the semaphore may be used purely to share resources.
\ ** The semaphore is initialised to one. The first task to REQUEST it
\ ** gains access, and all other tasks must wait until the accessing task
\ ** SIGNALs that it has finished with the resource.
  begin
    [I  dup @ 0=			\ n.b test and set
   while
    I]  pause				\ operations must be
  repeat                                \ non-interruptible and indivisible
  dup decr				\ got it, decrement counter
  self swap cell+ !			\ mark resource as mine
  I]					\ re-enable interrupts
;

[then]


\ ***************
\ *S TASK and START:
\ **************
\ *P TASK <name> builds a named task user area.
\ ** The action of a task is assigned and the task started
\ ** by the word INITIATE
\ *C   ['] <action> <task> INITIATE

\ *P START: is used inside a colon definition. The code before START: is the
\ ** task's initialisation, performed by the current task. The code after
\ ** START: up to the closing ; is the action of the task. For example:

\ *E   TASK FOO
\ **   : RUN-FOO
\ **     ...
\ **     FOO START:
\ **     ...
\ **     begin ... pause again
\ **   ;

\ *P All tasks must run in an endless loop, except for initialisation
\ ** code.
\ ** When RUN-FOO is executed, the code after START: is set up as the action
\ ** of task FOO and started. RUN-FOO then exits.

\ *P If you want to perform additional actions after starting the task, you
\ ** should use INITIATE to start the task.

variable task-chain	\ -- addr
\ *G Anchors list of all tasks created by TASK and friends.
  main-link task-chain !		\ MAIN already defined

interpreter	\ host version
: task		\ -- ; -- task ; TASK <name> builds a task
\ *G Note that the cross-interpreter's version of TASK has
\ ** been modified from v6.2 onwards to leave the current
\ ** section as CDATA.
  udata
  task-size reserve task-u0 + constant	\ build pointer ; SFP006
  cdata
  here  task-chain @ ,  task-chain !	\ link task 
;
target

heads? [if]	\ target version
: task		\ -- ; -- task ; TASK <name> builds a task
\ *G Creates a new task and data area, returning the address
\ ** of the user area at run time. The task is also linked into
\ ** the task chain anchored by TASK-CHAIN.
  here task-size allot constant		\ build pointer
;
[then]

: start:	\ task -- ; exits from caller
\ *G Used inside a colon definition. The code following
\ ** START: up to the ending semi-colon forms the action
\ ** of the task. The word containing START: finishes at
\ ** START:.
  r> swap initiate
;


\ ***************
\ *S Debugging tools
\ ***************

: .task		\ task --
\ *G Display task's name if it has one.
  task-chain
  begin					\ -- task link
    @ dup
   while				\ -- task link
    2dup cell - @ = if
      nip  #6 - >name .name space	\ 68K specific
      exit
    endif
  repeat
  drop .lword ."  is not in TASK chain"
;

: .tasks	\ task -- ; display all task names
\ *G Display all the tasks anchored by TASK-CHAIN.
  task-chain
  begin
    @ dup
   while
    dup #12 - >name .name space		\ ARM specific
  repeat
  drop
;

: .running	\ --
\ *G Display running tasks.
  main
  begin
    cr dup .task
    tcb.link @
    dup main =
  until
  drop
;


\ ********
\ all done
\ ********

decimal


\ *********
\ Test code
\ *********

test-multi? [if]

task task1	\ -- addr ; build task area

50000 value rate1	\ -- n ; number of idle iterations

: action1	\ -- ; action of task1
  hex  console opvec !
  begin
    [char] * emit
    rate1 0 do pause loop
  again
;

task task2	\ -- addr ; build task area

200000 value rate2	\ -- n ; number of idle iterations

: run-task2	\ -- ; example of using START:
  task2 start:
    console opvec !
    begin
      [char] / emit
      rate2 0 do pause loop
    again
;

[then]

\ ======
\ *> ###
\ ======

decimal

