\ 68xxx Multi-Tasker

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

PLEASE NOTE OUR PHONE NUMBER CHANGE

20000630 SFP001 Rewritten for v6.1 model
SM219 use clr, instead of loading with A3 to zero registers & memory.

*************************
TCB data structure layout
*************************

cell	LINK	link to next task
cell	SSP	Saved Stack Pointer
cell	STAT	BIT 0   1 = running, 0 = halted
                BIT 1   1 = message pending
                BIT 2   1 = event triggered
                BIT 3   1 = event handler run
	        others  1 = set to run task, available to user
cell	TASK    Task # that sent message here
cell	MESG    Message address
cell	EVNTw   CFA of word run as event handler

This structure is allocated at the start of the USER area. Consequently
the TCB of the current task is given by UP.


********************
Resource definitions
********************
 
68xxx/Coldfire Target Model: Subroutine Threaded Code
  A7 = return stack
  A6 = data stack
  A5 = user pointer
  A4 = locals pointer (when target code converted)
  A3 = zero
  A2 = scratch
  A1 = scratch
  A0 = scratch

  D7 = TOS
  D6 = scratch
  D5 = scratch
  D4 = scratch
  D3 = scratch
  D2 = scratch
  D1 = reserved for compiler (holds status flags)
  D0 = reserved for compiler (Shuffle register)
))

\ MODIFICATIONS
\  SM226 Steve Doughton 8th Jan 2003
\	   A few places where LSB not used to index tcb.status corrected:
\	   pause ( clearing schedule bit ), clr-event-run, noop 	 
\      Operations on tcb.status in event handler now made atomic using
\	   Read-Modify-Write instructions instead of reading into a register
\      ( interupt could occur and modify status between read into register,
\		and write from register )
\	   set-event re-written as CODE definition. New word set-msg-bit
\	   sets message bit atomically, used by (send-msg).


\ *************************
\ define TCB area structure
\ *************************

struct /TCB	\ -- size
  ptr tcb.link		\ link to next task ; MUST BE FIRST
  ptr tcb.ssp	        \ Saved Stack Pointer
  int tcb.status	\ status word
  int tcb.msrc		\ message source
  int tcb.mesg		\ message
  ptr tcb.event		\ xt of word which is event handler
  int tcb.parent	\ parent of task
  int tcb.cousin	\ cousin of task
end-struct

\ bit masks for the status cell
$0001 equ run-mask  run-mask invert equ ~run-mask	\ running
  0 equ run-bit#
$0002 equ msg-mask  msg-mask invert equ ~msg-mask	\ message available
  1 equ msg-bit#
$0004 equ trg-mask  trg-mask invert equ ~trg-mask	\ event triggered
  2 equ trg-bit#
$0008 equ evt-mask  evt-mask invert equ ~evt-mask	\ event run
  3 equ evt-bit#
$0010 equ sch-mask sch-mask invert equ ~sch-mask
  4 equ sch-bit#



\ ******************
\ Consistency checks
\ ******************

/tcb tcb-size <> [if]
  cr ." Control file and tasker disagree as to TCB size"  abort
[endif]


\ ************************
\ Task handling primitives
\ ************************

init-u0 constant main	\ -- addr ; tcb of main task

0 value multi?	\ -- flag ; true if tasker enabled

3 EQU LSB	\ LSB of status byte is at offset 3 when addresses as a byte

: single	\ -- ; disable scheduler
  0 to multi?
;

: multi		\ -- ; enable scheduler
  -1 to multi?
;

Is-ColdFire? [if]
  #15 cells equ MULT-STACK		\ number of cells to save on stack
[then]

CODE pause      \ -- ; the scheduler
  # $55 module-base $27 + $L b. move,       \ SM71: SWSR, Service Watchdog
  # $AA module-base $27 + $L b. move,       \ SM71: SWSR, Service Watchdog
  $l ' multi? >body  d0 l. move,	\ do nothing if scheduler disabled
  eq, if,  rts,  endif,

\ SBD 2nd Feb 2001
\	Clear sch-bit 
\   The schedule bit indicates that Pause has been called
\	It is set by SET-SCHEDULE prior to calling code that may
\   or may not call pause. When this code returns ?schedule is called
\	which only calls pause if the schedule bit is still set.
\	This mechanism prevents a task from hogging the scheduler but 
\	does not yield more than necessary
\
  # sch-bit#  LSB tcb.status d(a5)  bclr,  \ SM226			

\
\ save previously running task ip, rp, up, sp
\ can use with interrupts if we stack everything
\
Is-ColdFire? [if]
  # MULTI-STACK a7 l. suba,             \ adjust stack pointer
  { d0-7 a0-6 } (a7) l. movem,          \ save registers
[else]
  { d0-7 a0-6 } -(a7) l. movem,         \ stack the lot
[then]
  a7  0 tcb.ssp d(a5)  l. move,		\ save SP in TCB

\
\ select next task to run
\
  begin,
    0 tcb.link d(a5) a5 l. movea,	\ get next User Area
    0 tcb.status d(a5) d0 l. move,	\ get its status
  ne, until,

\
\ run selected task - sp, up, rp, ip
\
  0 tcb.ssp d(a5)  a7 l. movea,		\ restore saved stack pointer
Is-ColdFire? [if]
  { d0-7 a0-6 } (a7) l. movem,		\ restore registers
  # MULTI-STACK a7 l. adda,		\ adjust stack pointer
[else]
  (a7)+  { d0-7 a0-6 }  l. movem,       \ restore registers
[then]

event-handler? [if]	\ if user wants the event handler
\
\ event handler exit
\
\  0 tcb.status d(a5)  d0  l. move,	\ look at status
\  # trg-bit#  d0  bclr,			\ inspect event trigger bit (and clear it)
  # trg-bit#  LSB tcb.status d(a5)  bclr,			\ SM226 inspect event trigger bit (and clear it)
  ne, if,
\    # evt-bit#  d0  bset,		\ set event bit
\    d0  0 tcb.status d(a5)  l. move,	\ restore status
    # evt-bit#  LSB tcb.status d(a5)  bset,		\ SM226 set event bit
    0 tcb.event d(a5)  a0  l. movea,	\ run event handler
    (a0)  jmp,
  endif,
[then]

  next,
end-code

code status	\ -- task-status
  d7 -(a6) l. move,			\ save TOS
  0 tcb.status d(a5)  d7  l. move,	\ get status
  # run-bit#  d7  bclr,			\ clear running bit
  next,
end-code

(( Moved to MULTI68K.HIDEN.FTH
CODE restart    \ task -- ; mark task TCB as running
  d7 a0 l. movea,
  # run-bit#  LSB ( **SBD** ) tcb.status d(a0)  bset,	\ set task running bit (in low byte)
  (a6)+ d7 move,
  next,
END-CODE
))

CODE halt       \ task# -- ; reset running bit in TCB
  d7 a0 l. movea,
  # run-bit#  LSB ( **SBD** ) tcb.status d(a0)  bclr,	\ clear task running bit (in low byte)
  (a6)+ d7  l. move,
  next,
END-CODE

: stop          \ -- ; halt oneself
  self halt  pause
;


\ **************
\ Event handling
\ **************

event-handler? [if]

\ : set-event	\ task -- ; set event trigger in task TCB
\  tcb.status dup @  trg-mask or  swap !
\ ;				   

\ SM226 Make atomic
CODE set-event	\ task -- ; set event trigger in task TCB
  d7 a0 l. movea, 
  # trg-bit#  LSB tcb.status d(a0)  bset,	\ SM226 set event trigger bit (in low byte)
  (a6)+ d7  l. move,
  next,
END-CODE


: event?	\ task -- flag ; true if task had event
  tcb.status @  trg-mask and 0<>
;

((
\ What the hell is up@ ??? Where is it defined or documented???
: clr-event-run	\ -- ; reset own EVENT_RUN flag
  up@ tcb.status dup @  ~evt-mask and  swap !
;
))

CODE clr-event-run	\ -- ; reset own EVENT_RUN flag
  # evt-bit#  LSB tcb.status d(a5)  bclr,	\ SM226 clear task running bit (in low byte)
  next,
END-CODE

CODE noop	\ -- ; noop should reset own EVENT_RUN flag too
  # evt-bit#  LSB tcb.status d(a5)  bclr,	\ SM226 clear task running bit (in low byte)
  next,
END-CODE



: to-event      \ xt task -- ; define action of a task
  tcb.event !				\ set into TCB of task
;
[then]


\ ****************
\ Message handling
\ ****************

message-handler? [if]
: msg?		\ task -- flag ; true if task has message
  tcb.status @ msg-mask and 0<>
;

\ SM226 Make atomic
CODE set-msg-bit	\ task -- ; SM226:set message bit in task TCB
  d7 a0 l. movea, 
  # msg-bit#  LSB tcb.status d(a0)  bset,	\ SM226 set msg bit (in low byte)
  (a6)+ d7  l. move,
  next,
END-CODE

internal
: (send-msg)	\ mesg task -- ; store message in task
  self over tcb.msrc !			\ set source
  dup
\  tcb.status dup @  msg-mask or  swap !	\ set message bit
  set-msg-bit				\ SM226 Set msg bit, atomic operation
  tcb.mesg !				\ set message
;

: (get-message)	\ -- mesg task
  up@ tcb.mesg @			\ message
  up@ tcb.msrc @			
  up@ tcb.status
  dup @  ~msg-mask and  swap !		\ reset message bit
;
external

: send-message  \ addr task -- ; send message to task
  begin  dup msg?               	\ if task already has a message
    while  pause                	\ wait
  repeat                        	\ and try again
  (send-msg)				\ send the message
;

: get-message   \ -- addr task ; wait for any message
  begin  self msg? 0=           	\ check for a pending message
    while  pause                	\ wait for it
  repeat                        	\ try again
  (get-message)                 	\ read it
;
[then]

event-handler? message-handler? and [if]
: wait-event/msg        \ -- ; wait for message or event trigger
  begin
    self msg?  self event?  or 0=
   while
    pause
  repeat
;
[then]


\ *************************
\ Task structure management
\ *************************

internal

(( Extract from scheduler save
  { d0-7 a0-6 } -(a7) l. movem,         \ stack the lot
  a7  0 tcb.ssp d(a5)  l. move,		\ save SP in TCB
))
code init-task	\ xt task -- ; Initialise a task stack
  d7 a0 l. movea,			\ A0 = TCB
\  a3  0 tcb.status d(a0)  l. move,	\ clear task status
  0 tcb.status d(a0)  l. clr,	\ SM219 clear task status
  d7 a1 l. movea,			\ A1 = new RSP/SSP
  # task-r0 a1 l. adda,
  a1  r0-offset d(a0)  l. move,		\ set new R0

  (a6)+ -(a1) l. move,			\ save entry point, new PC
  d7 a2 l. movea,			\ save A6, new PSP
  # task-s0 a2 l. adda,
  a2  -(a1)  l. move,
  # 4  a2  suba,			\ and set new S0 according to rules of SP@ SP!
  a2  s0-offset d(a0)  l. move,
  d7 -(a1) l. move,			\ save A5, new UP
(( SM219 
  a3 -(a1) l. move,			\ save A4, new LP
  a3 -(a1) l. move,			\ save A3, zero register
  a3 -(a1) l. move,			\ save A2
  a3 -(a1) l. move,			\ save A1
  a3 -(a1) l. move,			\ save A0
  a3 -(a1) l. move,			\ save D7
  a3 -(a1) l. move,			\ save D6
  a3 -(a1) l. move,			\ save D5
  a3 -(a1) l. move,			\ save D4
  a3 -(a1) l. move,			\ save D3
  a3 -(a1) l. move,			\ save D2
  a3 -(a1) l. move,			\ save D1
  a3 -(a1) l. move,			\ save D0
))
  \ SM219... 
  -(a1) l. clr,			\ clear A4, new LP
  -(a1) l. clr,			\ clear A3, zero register
  -(a1) l. clr,			\ clear A2
  -(a1) l. clr,			\ clear A1
  -(a1) l. clr,			\ clear A0
  -(a1) l. clr,			\ clear D7
  -(a1) l. clr,			\ clear D6
  -(a1) l. clr,			\ clear D5
  -(a1) l. clr,			\ clear D4
  -(a1) l. clr,			\ clear D3
  -(a1) l. clr,			\ clear D2
  -(a1) l. clr,			\ clear D1
  -(a1) l. clr,			\ clear D0
  \ ...SM219
  a1  0 tcb.ssp d(a0)  l. move,		\ save SSP

  (a6)+ d7 move,
  next,
end-code

: add-task      \ task -- ; insert into list
  self tcb.link @			\ save task currently pointed to
  over tcb.link !			\ and make new task point to it
  self tcb.link !			\ current task points to new task
;

: sub-task      \ task -- ; remove task from chain
\ look for task which points to required task
  self					\ head of chain ; -- target current
  begin  2dup tcb.link @ <>		\ while curr. does not point to target
    while  tcb.link @			\ step through chain
           dup self =			\ but if task not in list, check for end
           if  2drop  exit  endif       \ and get out
  repeat                                \ -- target points-to-it
\ set task that points to target to point to task pointed to by the target
  swap tcb.link @			\ task pointed to by target
  swap tcb.link !			\ set into link field
;

external

(( See MULTI68K.HIDEN.FTH    
: initiate	\ xt task -- ; start task from scratch
  tuck init-task			\ reset TCB
  dup add-task                          \ add to chain
  run-mask swap tcb.status !		\ mark as running
  pause					\ let it run
;
))
: terminate	\ task -- ; stop task, and remove from list
  sub-task pause			\ halt task, and remove from chain
;

((
: init-multi    \ -- ; initialisation with multi-tasking
  single				\ tasker disabled
  main main tcb.link !			\ initialise MAIN to point to itself
  run-mask main tcb.status !		\ and reset status field
  multi					\ mark tasker enabled
;
))

\ **********
\ semaphores
\ **********

semaphores? [if]

\ A SEMAPHORE is an extended variable used for signalling between tasks,
\ and for resource allocation.
\ The counter field is used as a count of the number of times the
\ resource may be used, and the arbiter field contains the TCB of the
\ task that last gained access. This field can be used for priority
\ arbitration and deadlock detection/arbitration.

interpreter
: semaphore     \ -- ; -- addr [child]
  idata create
    0 ,  0 ,				\ count and arbiter fields
;
target

heads? [if]
: semaphore	\ -- ; -- addr [child]
  create
    0 ,  0 ,				\ count and arbiter fields
;
[then]

\ SIGNAL increments the counter field of a semaphore, indicating
\ either a) that another item has been allocated to the resource, or
\        b) it is available for use again, 0 indicating in use by a task
: signal        \ addr -- ; increment counter field of semaphore,
                \ so making it available for use by a task
  [I					\ must be interrupt safe
    dup incr  cell+ off			\ inc. counter, release
  I]
;

\ REQUEST waits until the counter field of a semaphore is non-zero,
\ and then decrements the counter field by one.
\ This allows the semaphore to be used as a COUNTED semaphore. For
\ example a character buffer may be used where the semaphore counter
\ shows the number of available characters.
\ Alternatively the semaphore may be used purely to share resources.
\ The semaphore is initialised to one. The first task to REQUEST it
\ gains access, and all other tasks must wait until the accessing task
\ SIGNALs that it has finished with the resource.
: request       \ sem -- ; get access to resource, wait if count = 0
  begin
    [I  dup @ 0=			\ n.b test and set
   while
    I]  pause				\ operations must be
  repeat                                \ non-interruptible and indivisible
  dup decr				\ got it, decrement counter
  self swap cell+ !			\ mark resource as mine
  I]					\ re-enable interrupts
;

[then]


\ ***************
\ TASK and START:
\ **************

((
TASK <name> builds a named task user area.
The action of a task is assigned and the task started
by the word INITIATE
  ['] <action> <task> INITIATE

START: is used inside a colon definition. The code before START: is the
task's initialisation, performed by the current task. The code after
START: up to the closing ; is the action of the task. For example:

  TASK FOO
  : RUN-FOO
    ...
    FOO START:
    ...
    begin ... pause again
  ;

All tasks must run in an endless loop, except for initialisation
code.

When RUN-FOO is executed, the code after START: is set up as the action
of task FOO and started. RUN-FOO then exits.

If you want to perform additional actions after starting the task, you
should use INITIATE to start the task.
))

interpreter	\ host version
: task		\ -- ; -- task ; TASK <name> builds a task
  udata task-size reserve constant	\ build pointer
  cdata
;
target

heads? [if]	\ target version
: task		\ -- ; -- task ; TASK <name> builds a task
  here task-size allot constant		\ build pointer
;
[then]

: start:	\ task -- ; exits from caller
  r> swap initiate
;


\ ********
\ all done
\ ********

decimal
pto


\ *********
\ Test code
\ *********


task task1

5000 value rate1

: action1
  hex  console opvec !
  begin
    [char] * emit
    rate1 0 do pause loop
  again
;

task task2

20000 value rate2

: run-task2
  task2 start:
    console opvec !
    begin
      [char] / emit
      rate2 0 do pause loop
    again
;




