\ 68000 Exception handlers

((
Copyright (c) 1991-2002
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====
Update divide by 0 and TRAPV handlers to decode the instruction. The
VFX compiler can use registers other than D0 and D1 and the DIVL
instructions.


Change history
==============
20000703 SFP010 Added CPU32 code courtesy of Hiden Analytical
19990312 SFP010 Removed ColdFire code (not flagged)
19990310 SFP009 Added the spurious interrupt handler
19980501 MSD008 Modified the 68010+ format to COLDFIRE requirements.
19980501 MSD007 I checked that the MSD006 change didn't break the traps code
         MSD007 on the 68332 (it was OK) and the 68307 (it wasn't).  The '307
         MSD007 trap code seemed to be stumbling over the "1 .R" when
         MSD007 displaying the register numbers.  I replaced this with
         MSD007 the computationally simpler "[CHAR] 0 + EMIT".
19980501 MSD006 Altered the way that SYSTEM-ACTION is called.  The old way,
         MSD006 using "$ ' SYSTEM-ACTION JSR," didn't work on the COLDFIRE,
         MSD006 so we've changed it to "$L ' SYSTEM-ACTION W. BSR,"
19980429 MSD005 Changed the "#12 >>", in the 68010+ version of .STACK_FRAME,
         MSD005 to be "4/ 4/ 4/ 4/ 4/ 4/" because CHIP doesn't have ">>".
19980424 MSD004 Altered the test for the processor type.  The old code merely
         MSD004 used "M68000? M68008? OR" to detect the processor type and
         MSD004 switched the expected stack layout on that.  However, the
         MSD004 flags are set for the specific processor selected and all
         MSD004 processors *below*.  Hence, when using COLDFIRE or CPU32, for
         MSD004 instance, the M68000? and M68008? flags are *still* set.  The
         MSD004 above ORed test therefore fails.  All I've done here is to
         MSD004 switch the test to say "M68010? NOT".
19980409 MSD003 Apply COLDFIRE changes.
19980302 MSD002 Hacked into shape for subroutine-threaded compiler.  These
         MSD002 changes have *not* been "flagged" in the source.
19980217 MSD001 Replaced reference to EWRM with a jump to the address at CLD1.
199209?? ?????? Register dump prettied.
199109?? IRE??? Register dump added by Mike Bell of IRE.
))


\ ------------------------------------------------------------------------------
\       common interrupt routine
\ ------------------------------------------------------------------------------

\
\ The common interrupt routine is called by all exception handlers
\ that cause a warm restart. These include the bus error and divide by 0
\ exceptions. The assembly language handler is responsible for placing
\ the exception vector number on the stack.

\ The 68xxx processors can push up to 16 words of information on
\ to the stack. In addition, the interrupt routine saves the 16 main
\ registers. On entry these are copied for later use by any debuggers.

decimal

#96 buffer: stack_frame                         \ 16 registers, 16 words

\ ------------------------------------------------------------------------------

[undefined] spaces [if]
  : Spaces        ( n -- )
    0 Max 0 ?Do  $20 emit  Loop
  ;
[then]

[undefined] .word [if]
  : .word           \ n -- ; print n in hex - 16b
    4 0
    do  dup $0C 0 do 2/ loop              \ top 4 bits to bottom
        $0F and  dup 9 >                  \ mask, check 0..9,A..F
        if  7 +  endif                    \  was A..F
        $030 + emit                       \ convert to ASCII
        2* 2* 2* 2*                       \ get next 4 bits
    loop                                  \
    drop  ;                               \ drop n
[then]

[undefined] .lword [if]
  : .lword          \ n -- ; print n in hex - 32b
    8 0
    do  dup $01C 0 do 2/ loop             \ top 4 bits to bottom
        $0F and  dup 9 >                  \ mask, check 0..9,A..F
        if  7 +  endif                    \  was A..F
        $030 + emit                       \ convert to ASCII
        2* 2* 2* 2*                       \ get next 4 bits
    loop                                  \
    drop  ;                               \ drop n
[then]

[undefined] .r [if] : .r drop . ; [then]

\ ------------------------------------------------------------------------------

M68010? not [if]        \ M68010? returns true for 68010 and above CPUs, and
                        \ so is false for 68008 and 68000 only
  cr ." Compiling 68000/68008 stack frame display code" cr

  : .STACK_FRAME  ( u -- ; vector number on stack )
    stack_frame
    #16 0                                       \ display data registers
    do  i 4 mod 0=                              \ 4 registers per line
        if  cr  then
        i 8 <                                   \ r0..7 = D0..7, r8..f = A0..7
        if  ." D"  else  ." A"  then            \ display register type
        i 7 and  [char] 0 + emit  ." : "        \ display register number     
        dup @ .lword  3 spaces  4+              \ display register value
    loop
    over dup 2 =  swap 3 =  or                  \ if  BUS or ADDRESS error
    if  cr ." EI: " dup w@ .word  2+            \ additional info
          cr ." FA: " dup @ .lword  4+          \ address that caused the fault
        cr ." OP: " dup w@ .word  2+            \ opcode being executed
    endif
    cr ." SR: " dup w@ .word  2+                \ display SR
    cr ." PC: " @ .lword                        \ display PC
  ;

[else]
  cpu32? 0= [if]
  cr ." Compiling 68010+ stack frame display code" cr
  : .STACK_FRAME        \ --
    stack_frame                                 \ start address of saved stack frame
    #16 0                                       \ display data registers
    do  i 4 mod 0=                              \ 4 registers per line
        if  cr  then
        i 8 <                                   \ r0..7 = D0..7, r8..f = A0..7
        if  ." D"  else  ." A"  then            \ display register type
        i 7 and  1 .r ." : "                    \ display register number
        dup @ .lword  3 spaces  4+              \ display register value
    loop
      cr ." SR: " dup w@ .word  2+
      cr ." PC: " dup @ .lword  4+
      cr ." Format " dup w@  dup 4/ 4/ 4/ 4/ 4/ 4/ . ." (dec)"            
      cr ." Vector " #4095 and 4/ . ." (dec)"   2+
      cr  8 0                                   \ display 680x0 stack frame
      do  ." S" i  1 .r ." : "
          dup w@ .word  2 spaces  2+
      loop
    drop                                        \ discard stack frame address
  ;
  [else] 
  cr ." Compiling CPU32 stack frame display code" cr
  -short-branches
  : .stack_frame		\ -- ; display the saved stack frame
    stack_frame				\ start address of saved stack frame
    16 0					\ display data registers
    do  i 4 mod 0=			\ 4 registers per line
        if  cr  endif
        i 8 <				\ r0..7 = D0..7, r8..f = A0..7
        if  ." D"  else  ." A"  endif	\ display register type
        i 7 and  1 .r ." : "		\ display register number
        dup @ .lword  3 spaces  4+	\ display register value
    loop
    cr ." SR: " dup w@ .word  2+
    cr ." PC: " dup @ .lword  4+
    cr ." Vector " dup w@ dup 4095 and 4/ . ." (dec)"   
    cr ." Format " 12 rshift dup . ." (dec) "
    case
       0 of  ." Normal 4 word stack frame "
             2+ 12
             endof
       2 of  ." Normal 6 word stack frame"
             2+
             cr ." Faulted Instruction PC: " dup @ .lword  4+
             10
         endof
      12 of  ." Bus Error Stack Frame"
             2+ cr ." Faulted Address: " dup @ .lword  4+
             dup 12 + W@ 14 rshift 	 \ get SSW TP & MV bits
             case
               0 of  cr ." normal BERR"
  		     cr ." DBUF: " dup @ .lword  4+
  		     cr ." Current Instruction PC: " dup @ .lword  4+
		     4 swap
		 endof
	       1 of  cr ." BERR on MOVEM"
  		     cr ." DBUF: " dup @ .lword  4+
  		     cr ." Current Instruction PC: " dup @ .lword  4+
		     4 swap
		 endof
	       2 of  cr ." BERR during exception"
  		     cr ." Pre-execption SR: " dup w@ .word  2+
  		     cr ." Faulted Vector " dup w@ dup 4095 and 4/ . ." (dec)"   
  		     cr ." Faulted Format " 12 rshift dup . ." (dec) " 
		     0= if
		       6 swap
		     else
  		       2+ cr ." Faulted Instruction PC: " dup @ .lword  2+
		       4 swap
		     then
		     2+
		 endof
		      cr ." Funny BERR"
		      8 -rot
	     endcase
	     cr ." Transfer Count: " dup w@ .word  2+
	     cr ." Special Status Word: " dup w@ .word  2+
	     swap
	 endof
    endcase
    cr   0				\ display 680x0 stack frame
    do  ." S" i  1 .r ." : "
      dup w@ .word  2 spaces  2+
    loop
    drop				\ discard stack frame address
  ;
  +short-branches
  [then]
[then]

\ ------------------------------------------------------------------------------

-short-branches
: SYSTEM-ACTION         ( u u  --  ; two copies of vector number on stack  )
  console opvec !  console ipvec !	\ define I/O vectors
  decimal
  cr cr ." Exception: "
  case
    #02 of  ." Bus error"                 endof
    #03 of  ." Address error "            endof
    #04 of  ." Illegal instruction"       endof
    #05 of  ." Divide by zero"            endof
    #06 of  ." CHK, CHK2 trap"            endof
    #07 of  ." TRAPcc, TRAPV trap"        endof
    #08 of  ." Privilege violation"       endof
    #09 of  ." Trace"                     endof
    #10 of  ." Line A"                    endof
    #11 of  ." Line F"                    endof
    #12 of  ." Hardware Breakpoint"       endof
    #13 of  ." Coprocessor Protocol"      endof
    #14 of  ." Format error/Unintialised" endof
    #15 of  ." Format error/Unintialised" endof
            ." Unknown"
  endcase
  .stack_frame
;
+short-branches

proc system-int0	\ Common entry point ( TOS = trap # )
  a7 a0 l. movea,
  # stack_frame a1 l. movea,            \ Copy registers, SR, PC, and up to
  # #22 d0 w. move,                     \ 8 additional words from stack frame
  begin,
    (a0)+ (a1)+ l. move,                \ to buffer stack_frame.
    # 1 d0 w. subq,                                                            
  eq, until,                                                                    
  d7 -(a6) l. move,                     \ duplicate TOS
  # int-init-u0 a5 l. movea,            \ point to user area
  $l ' SYSTEM-ACTION w. bsr,                                                    
l: die                                  \ Termination code                      
\  die # a1 l. movea,                    \ was "$ EWRM jsr,".                    
\  (a1) jmp,
  $ die jmp,
  $ ECLD jmp,                           \ restart the system
end-code

\ ------------------------------------------------------------------------------

proc trap2	\ BUS ERROR
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #02 d7 l. move,                     \ push trap number
  system-int0 bra,                      \ join common code
end-code
trap2   start-vector #02 4 * + !        \ set vector

proc trap3	\ ADDRESS ERROR
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #03 d7 l. move,
  system-int0 bra,
end-code
trap3   start-vector #03 4 * + !

proc trap4	\ ILLEGAL INSTRUCTION
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #04 d7 l. move,
  system-int0 bra,
end-code
trap4   start-vector #04 4 * + !

cpu32? 0= [if]

proc trap5	\ DIVIDE BY ZERO
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #05 d7 l. move,
  system-int0 bra,
end-code
trap5   start-vector #05 4 * + !

[else]

$FFFFFFFF equ minint                    \ most negative integer
$7FFFFFFF equ maxint                    \ most positive integer

\ If a non-fatal divide by zero is required, use the code below.
\ N.B. This is for CPU32 processors only.
: trap5-isr     \ D0 D1 faultpc -- D0' D1' ; PC used to identify instruction
                \ resets saved result to 
  @  dup $0800 and                      \ bit 12 set if signed instruction
  if
    $0400 and                           \ bit 11 set if 64 bit dividend
    if
      drop                              \ sign is from D0
    else
      nip                               \ sign is from D1
    then
    0<                                  \ form sign and return it
    if  0 minint  else  0 maxint  endif \ return remainder and quotient
  else
    drop  2drop                         \ discard PC, D1, D0
    0 minint                            \ return max unsigned integer
  endif
;

proc trap5	\ TRAPcc, TRAPV INSTRUCTION - only used by DIVS/UL long
  { d0-7 a0-6 } -(a7) l. movem,         \ save all registers except stack pointer (15 regs)
  # 0 -(a6) l. move,
  d0 -(a6) l. move,                     \ save required registers D0 D1
  d1 -(a6) l. move,
  #15 4 * 8 + d(A7) d7 l. move,         \ skip saved regs, offset 8 of 6 word stack frame is fault PC
  ' trap5-isr w. bsr,                   \ call handler
  d7 d1 l. move,                        \ get D1 from TOS
  (a6)+ d0 l. move,                     \ get D0 from NOS
  # 4 a6 l. addq,                       \ remove dummy 0 from data stack
  # 8 a7 l. addq,                       \ step over D0 and D1 return values
  (a7)+ { d2-7 a0-6 } l. movem,         \ restore all other registers
  rte,
end-code
trap5   start-vector #05 4 * + !

[then]

proc trap6	\ CHK, CHK2 INSTRUCTION
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #06 d7 l. move,
  system-int0 w. bra,
end-code
trap6   start-vector #06 4 * + !

cpu32? [if]     \ TRAPV, used for divide overflow checking

: trap7-isr     \ D7 D0 D1 faultpc -- D0' D1' ; PC used to identify instruction
                \ resets saved result to 
  4 - @                                 \ DIVS/U .L instructions are 4 bytes long
  dup $0800 and                         \ bit 12 set if signed instruction
  if
    $0400 and                           \ bit 11 set if 64 bit dividend
    if
      drop                              \ sign is from D7/D0
    else
      nip                               \ sign is from D7/D1
    then
    xor 0<                              \ form sign and return it
    if  0 minint  else  0 maxint  endif \ return remainder and quotient
  else
    2drop 2drop                         \ discard parameters
    0 minint                            \ return max unsigned integer
  endif
;

proc trap7	\ TRAPcc, TRAPV INSTRUCTION - only used bu DIVS/UL long
  { d0-7 a0-6 } -(a7) l. movem,         \ save all registers except stack pointer (15 regs)
  # 0 -(a6) l. move,
  d7 -(a6) l. move,                     \ save required registers D7 D0 D1
  d0 -(a6) l. move,
  d1 -(a6) l. move,
  #15 4 * 8 + d(A7) d7 l. move,         \ skip saved regs, offset 8 of 6 word stack frame.is fault PC
  ' trap7-isr w. bsr,                   \ call handler
  d7 d1 l. move,                        \ get D1 from TOS
  (a6)+ d0 l. move,                     \ get D0 from NOS
  # 4 a6 l. addq,                       \ remove dummy 0 from data stack
  # 8 a7 l. addq,                       \ step over D0 and D1 return values
  (a7)+ { d2-7 a0-6 } l. movem,         \ restore all other registers
  rte,
end-code
trap7   start-vector #07 4 * + !

[else]

proc trap7	\ TRAP, TRAPV INSTRUCTION
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #07 d7 l. move,
  system-int0 w. bra,
end-code
trap7   start-vector #07 4 * + !
[then]

proc trap8	\ PRIVILEGE VIOLATION
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #08 d7 l. move,
  system-int0 w. bra,
end-code
trap8   start-vector #08 4 * + !

proc trap9	\ TRACE
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #09 d7 l. move,
  system-int0 w. bra,
end-code
trap9   start-vector #09 4 * + !

proc trap10	\ A-LINE EMULATOR
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #10 d7 l. move,
  system-int0 w. bra,
end-code
trap10  start-vector #10 4 * + !

proc trap11	\ F-LINE EMULATOR
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #11 d7 l. move,
  system-int0 w. bra,
end-code
trap11  start-vector #11 4 * + !

proc trap12	\ RESERVED
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #12 d7 l. move,
  system-int0 w. bra,
end-code
trap12  start-vector #12 4 * + !

proc trap13	\ CO-PROCESSOR PROTOCOL VIOLATION
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #13 d7 l. move,
  system-int0 w. bra,
end-code
trap13  start-vector #13 4 * + !

proc trap14	\ FORMAT ERROR
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #14 d7 l. move,
  system-int0 w. bra,
end-code
trap14  start-vector #14 4 * + !

proc trap15	\ UNINITIALIZED INTERRUPT
  { d0-7 a0-7 } -(a7) l. movem,         \ save all registers
  # #15 d7 l. move,
  system-int0 w. bra,
end-code
trap15  start-vector #15 4 * + !

proc spurious-int	\ Spurious INTERRUPT    SFP009 ...
  rte,
end-code
spurious-int  start-vector #24 4 * + !  \ ... SFP009


