
\ ***************************************************************
\ Module:       extent.fth
\               Contains file extension handing functions
\ Author:       Jeremy Moller
\ Organisation: CEM Systems Ltd, Belfast
\ Copyright:    1999
\ ******************************************************************

\ These function handle file extensions

: annex-file { l_blocks l_fid | l_buff l_start l_last_extent -- t/f }
    FILE-EXTENSION allocate if
        drop
        false exit
    endif
    -> l_buff
    l_blocks l_fid new-extent-needed -> l_blocks
    l_blocks 0> not if      			\ do we actually need any extra blocks ?
        TRUE exit
    endif
    drive-control request
    l_fid fcb-addr ->file-expansion @ if        \ check allowed extension
        l_fid l_buff ->ext-fid !                \ save fid
        l_fid fcb-addr ->file-expansion @ l_blocks max      \ allocate required or minimum space
    \   cr ." Getting " dup .decimal
        l_buff ->ext-size !                 	\ save size
        l_buff next-extent? -> l_last_extent
        l_buff ->ext-size @ need-blocks -> l_start
        l_start l_buff ->ext-start !
        l_start -1 <> if
        \   ." blocks @ " l_start .
            l_buff add-to-existing-extent? if
            \   cr ." add to extent "
                l_buff extend-extent    	\ add to existing extent
                l_fid fcb-addr update-fcb   	\ now update FLD on disk
            else
                l_last_extent if
                \   cr ." new extent "
                    l_buff set-extent       	\ create new extent
                    l_fid fcb-addr update-fcb   \ now update FLD on disk
                else
                \   ." indirect extents "
                    l_buff l_fid need-indirect-extent dup 0= if
                        l_start l_buff ->ext-size @ free-blocks \ failed
                    endif
                endif
            endif
        else
\           ." no free blocks "
            false
        endif
    else
    \   ." Not allowed to increase "
        false
    endif
    drive-control signal
    l_buff free drop ;

\ RETURNS 0 if no space
: next-extent? { l_buff | l_found -- t/f }
    FALSE -> l_found
    NO-OF-EXTENTS 0 do
        i l_buff ->ext-no !
        l_buff fcb-free-extent? if  		\ found an unused extent
            TRUE -> l_found
            leave
        endif
    loop
    l_found ;

: set-extent { l_buff | l_ext l_fcb -- }
    l_buff ->ext-no @ file-extents * -> l_ext   \ get extent offset
    l_buff ->ext-fid @ fcb-addr -> l_fcb
    l_buff ->ext-start @ l_fcb l_ext + ->extent-start-0 !
    l_buff ->ext-size @  l_fcb l_ext + ->extent-size-0 !
    l_buff ->ext-size @  l_fcb ->file-blocks +! \ read existing file size
    lseconds @  l_fcb ->file-last-mod !
    ;

: extend-extent { l_buff | l_fcb l_ext -- }
    l_buff ->ext-no @ 1- file-extents * -> l_ext \ get extent offset
    l_buff ->ext-fid @ fcb-addr -> l_fcb
    l_buff ->ext-size @ l_fcb l_ext + ->extent-size-0 +!
    l_buff ->ext-size @ l_fcb ->file-blocks +!  \ read existing file size
    lseconds @ l_fcb ->file-last-mod !                          \ 
    ;

: fcb-free-extent? { l_buff | l_fcb l_ext -- t/f }
    l_buff ->ext-no @ file-extents * -> l_ext   \ get extent offset
    l_buff ->ext-fid @ fcb-addr -> l_fcb
    l_fcb l_ext + ->extent-start-0 @        	\ -- fcb, ext, start
    l_fcb l_ext + ->extent-size-0 @ or 0= ;

: add-to-existing-extent? { l_buff | -- t/f }   \ last block in this extent
    l_buff end-of-existing-extent
    l_buff ->ext-start @            		\ proposed start of next extent
    = ;                             		\ does it match ?

: end-of-existing-extent { l_buff | l_fcb l_ext -- end_block }  \ last block in this extent
    l_buff ->ext-no @ 1-  file-extents * -> l_ext \ get extent offset
    l_buff ->ext-fid @ fcb-addr -> l_fcb
    l_fcb l_ext + ->extent-start-0 @        	\ -- fcb, ext, start
    l_fcb l_ext + ->extent-size-0 @ + ;     	\ end of last extent


\ Copy the FCB version to the FLD on the disk
: update-fcb { l_fcb | -- t/f }
    fld-update request
    l_fcb ->fld-no @ fld-d-block        	\ disk block
    l_fcb                               	\ source
    l_fcb ->fld-no @ fld-m-block        	\ start
    dup FLD-REC +                       	\ end
    write-p-blk
    fld-update signal
    ;

: need-extent? ( blocks fid -- t/f )    \ return TRUE if more space needed
    new-extent-needed 0> ;

\ a +ve block is number of blocks required
: new-extent-needed { l_blocks l_fid | -- blocks }
    l_fid get-r/w-b-ptr     		\ -- current-block
    l_blocks +              		\ new expected end
    l_fid fcb-addr ->file-blocks @ - ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

