
\ ***************************************************************
\ Module:       ide.fth
\               Contains IDE compact flash device driver
\ Author:       Jeremy Moller
\ Organisation: CEM Systems Ltd, Belfast
\ Copyright:    1999
\ ******************************************************************


\ compact drive

CSEXT IDE-TF-REG +      equ IDE-DATA
CSEXT IDE-TF-REG + 1+   equ IDE-FEATURES
CSEXT IDE-TF-REG + 1+   equ IDE-ERROR
CSEXT IDE-TF-REG + 2+   equ IDE-SEC-CNT
CSEXT IDE-TF-REG + 3 +  equ IDE-LBA-0
CSEXT IDE-TF-REG + 4 +  equ IDE-LBA-1
CSEXT IDE-TF-REG + 5 +  equ IDE-LBA-2
CSEXT IDE-TF-REG + 6 +  equ IDE-LBA-3
CSEXT IDE-TF-REG + 6 +  equ IDE-DRIVE
CSEXT IDE-TF-REG + 7 +  equ IDE-COMMAND
CSEXT IDE-TF-REG + 7 +  equ IDE-STATUS
CSEXT ide-misc-reg +    equ IDE-ALT-STATUS
CSEXT ide-misc-reg + 1+ equ IDE-ADR-REG

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\ command codes

$90 equ IDE-DIAG
$EC equ IDE-ID
$E4 equ IDE-RD-BUFF
$20 equ IDE-RD-SEC
$E8 equ IDE-WD-BUFF
$30 equ IDE-WD-SEC
$C0 equ IDE-ER-SEC
$50 equ IDE-FR-SEC
$87 equ IDE-TRANS-SEC
\ -----------------------------

\ status register bits
$80 equ IDE-BUSY    	\ ide drive busy
$40 equ IDE-RDY     	\ ide drive ready
$20 equ IDE-WF      	\ ide write fault
$10 equ CF-RDY      	\ compact flash ready
$08 equ IDE-DRQ     	\ requesting data
$01 equ ST-IDE-ERROR    \ last command caused error

\ ------------------------------

$00 equ master-ide
$10 equ slave-ide
$e0 equ lba-master
$f0 equ lba-slave

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
#256 equ    MAX-SEC/OPER
BLOCK_SIZE MAX-SEC/OPER * equ MAX-SEC-DATA

variable disk-error

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ configure super io chip
: init-ide ( -- )
    drive-access signal 	\ set access semaphore
    ;

: drive-diag ( -- n )
    ide-diag IDE-COMMAND w!
    IDE-ERROR c@ ;

: drive-status ( -- )
    IDE-STATUS c@ cr ." ------------------"
    dup ide-busy and if cr ." ide drive busy " endif
    dup ide-rdy and if  cr ." ide drive ready " endif
    dup ide-wf and if   cr ." ide write fault " endif
    dup cf-rdy and if   cr ." compact flash ready " endif
    dup ide-drq and if  cr ." requesting data " endif   
    dup st-ide-error and if cr ." last command error "
        IDE-ERROR c@ .byte
    endif
    drop ;

: show-drive-details ( addr -- )
    decimal
                    cr ." ------------------------------"
    dup #27 2* +    cr ."     Model " #40 type
    dup #10 2* +    cr ." Serial no " #20 type
    dup #23 2* +    cr ."  Firmware " #8 type
    dup #58 2* + w@ $10000 *
    over #57 2* + w@ + #512 *
                    cr ."      Size " . ." Bytes"
                    cr ." ------------------------------"
    free drop ;

: show-drive-id ( addr -- )
    decimal
    dup #01 2* + w@ cr ." Cylinders " .
    dup #03 2* + w@ cr ."     Heads " .
    dup #05 2* + w@ cr ." Bytes/sec " .
    dup #06 2* + w@ cr ." Sec/Track " .
    dup #07 2* +  @ cr ."   Sectors " .
    dup #10 2* +    cr ." Serial no " #20 type
    dup #20 2* + w@ cr ."  Buf type " .
    dup #21 2* + w@ cr ."  Buf size " #512 * .
    dup #23 2* +    cr ."  Firmware " #8 type
    dup #27 2* +    cr ."     Model " #40 type
    dup #54 2* + w@ cr ."   No. Cyl " .
    dup #55 2* + w@ cr ." No. heads " .
    dup #56 2* + w@ cr ." No Sec/Tk " .
    dup #58 2* + w@ $10000 *
    over #57 2* + w@ + cr ."      LBAs " .
    dup #61 2* + w@ $10000 *
    over #60 2* + w@ + cr ." Addr LBAs " .
    drop ;

: show-w/r-info ( addr -- )
    decimal
    dup  w@ cr ." Cylinders " .
    dup 2 + c@ cr ."     Heads " .
    dup 3 + c@ cr ." Sector " .
    dup 4 + w@ $100 * 
    over 6 + c@  + cr ." lba " .
    dup $13 +  c@ cr ."   erased " .
    dup $18 + w@ $100 * 
    over $1a + c@  + cr ." hot count " .
    free drop ;


: bswap ( n -- n1 )
    dup 8 lshift swap 8 rshift or
    $0ffff and ;

: drive-id ( -- )
    (drive-id)
    ?dup if
        dup show-drive-id
        free drop
    else
        cr ." Error reading drive "
    endif ;     


: (drive-id) ( -- buff )
    BLOCK_SIZE allocate if  		\ grab buffer
        ." Can't allocate "
    else
        drive-access request
        MASTER-IDE sel-drive
        IDE-ID send/wait-cmd        	\ command to drive
        ide-data-rdy if
            i/o-16bit               	\ select 16 bit mode
            dup BLOCK_SIZE bounds do
                ide-data w@             \ read from HD
                a-h-b
                bswap
                i w!
            2 +loop
            i/o-8bit                	\ select 8 bit mode
        else
            free drop                   \ buffer not required
            ." requested data not ready "
            false
        endif
        drive-access signal
    endif ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: write-block ( l_lba l_buff -- )
        1 write-blocks ;
        
: write-blocks ( l_lba l_buff l_no-blocks )
        rot swap
        drive-access request
        disk-led-on
        set-t/f-reg                     \ waits till free, then sets register
        IDE-WD-SEC send/wait-cmd        \ command to drive
        ide-data-rdy if
            i/o-16bit               	\ select 16 bit mode
            BLOCK_SIZE IDE-DATA f-block-tx
            i/o-8bit                	\ select 8 bit mode
            true
        else
            drop
            false
        endif
        drive-access signal
        disk-led-off
 ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: erase-block ( l_lba -- t/f )
        BLOCK_SIZE allocate if
            drop
            false exit
        endif
        dup BLOCK_SIZE erase
        tuck 1 write-blocks 
        swap free drop ;
        
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: read-f-block ( l_lba l_buff -- t/f )
        drive-access request
        (read-f-block)
        drive-access signal ;
        
: (read-f-block) ( l_lba l_buff -- t/f )
        swap
        disk-rd-led-on
        01 set-t/f-reg                  \ set task file registers
        IDE-RD-SEC send/wait-cmd    	\ issue command to drive and wait
        ide-data-rdy if
            i/o-16bit           	\ select 16 bit mode
            BLOCK_SIZE IDE-DATA f-block-xfer
            i/o-8bit            	\ select 8 bit mode
            true
        else
            drop    			\ lose buffer address
            false
        endif
        disk-rd-led-off
        ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: i/o-8bit ( -- )
    02 _abwcr set-bit ;         \ select 8 bit mode

: i/o-16bit ( -- )
    02 _abwcr reset-bit ;       \ select 16 bit mode

: ide-data-rdy ( -- t/f )
    ide-status c@
    dup IDE-DRQ and                     \ ready for data
    swap ST-IDE-ERROR and 0= and ;      \ AND not error

: ide-drive-rdy? ( -- t/f )
        ide-status c@   
        CF-RDY IDE-RDY or = ;

#150 equ IDE-TIMEOUT
IDE-TIMEOUT #015 - equ IDE-TIMEOUT-W

: wait-ide-rdy { | l_timer l_tpause -- }
        IDE-TIMEOUT later -> l_timer
        IDE-TIMEOUT-W later -> l_tpause
        begin
            ide-drive-rdy? 0=
            l_timer timedout? 0= and
        while
            l_tpause timedout? if
                pause
            endif
        repeat ;


: send/wait-cmd ( cmd -- )
        IDE-COMMAND c!
        wait-ide-cmd ;

: wait-ide-cmd { | l_timer l_tpause -- }
        IDE-TIMEOUT later -> l_timer
        IDE-TIMEOUT-W later -> l_tpause
        begin
            ide-status c@ IDE-BUSY and  \ wait unit done
            l_timer timedout? 0= and
        while
            l_tpause timedout? if
                pause
            endif
        repeat ;

: sel-drive ( n -- )
    ide-drive c! ;


: rd-drive ( -- n )
    ide-data w@ ;

: wd-drive ( n -- )
    ide-data w! ;

: set-t/f-reg ( lba sec -- )     \ set task file registers
        wait-ide-rdy            \ wait till drive free
        ide-sec-cnt c!
        $100 /mod $100 /mod $100 /mod $100 mod
        $0f and LBA-MASTER or ide-lba-3 c!
        ide-lba-2 c!
        ide-lba-1 c!
        ide-lba-0 c! ;


