\ ***************************************************************
\ Module:       string.fth
\               String handling for the CEM file system
\ Author:       Jeremy Moller
\ Organisation: CEM Systems Ltd, Belfast
\ Copyright:    1999
\ ******************************************************************

\ This returns a counted string left of a delimit char
: left-string { l_*name l_len l_delimit | -- next-addr next-len }
    l_*name l_len l_delimit scan        \ find first \ 
    nip l_len swap -
    l_*name swap ;

\ This returns a counted string between two delimit chars
: extract-string { l_*name l_len l_delimit | -- next-addr next-len }
    l_*name l_len l_delimit skip        \ find first \ 
    2dup                                \ 
    l_delimit scan                      \ find next 
    nip - ;

\ This returns the next counted string after the next delimit char
: jump-string ( l_*name l_len l_delimit -- next-addr next-len )
    -rot 2 pick skip                    \ find first \ 
    rot scan                            \ find next 
    ;

: string> { *name len l_format | l_value l_started -- value }
    0 -> l_value
    false -> l_started
    *name len bounds ?do
        i c@ l_format digit if          \ found ASCII
            l_value l_format * + -> l_value
            true -> l_started           \ mark start found
        else
            l_started if                \ non decimal & start found
                leave
            endif
        endif
    loop
    l_value ;

\ This converts a ascii counted string to decimal
: string>decimal ( *name len -- value )
    #10 string> ;

\ This converts a ascii counted string to hexadecimal
: string>hex ( *name len -- value )
    #16 string> ;
    

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\ converts a number to a hex string in buffer adding leading zeros

: hex>string { l_n l_*buff l_len | -- }
    l_*buff l_len ascii 0 fill
    l_len 1- -> l_len
    begin
        l_n 0<>
        l_len 0< not and 
    while
        l_n #16 /mod -> l_n
        hex>ascii l_*buff l_len + c!
        l_len 1- -> l_len
    repeat
    ;

: decimal>string  (  n  --  *name len )
    base @ decimal swap
    s>d 0
    >r swap over dabs <# #s rot sign #>
    r> over - spaces
    rot base ! ;

