
\ ***************************************************************
\ Module:       tests.fth
\               Contains disgnostic functions
\ Author:       Jeremy Moller
\ Organisation: CEM Systems Ltd, Belfast
\ Copyright:    1999
\ ******************************************************************


\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: df-v { | l_*buff l_free l_bitm l_base -- }
    base @ -> l_base decimal
    cr ." Disk size " DISK_SIZE #2048 / 0 .r ." Mb"
    cr
    BLOCK_SIZE allocate if
        drop ." Alloc error " exit
    endif
    -> l_*buff
    0 -> l_free
    USAGE_BLOCK0 USAGE_TABLE_SIZE bounds ?do
        i l_*buff read-hd-block drop                \ read in block
        BLOCK_SIZE 0 do
            l_*buff i + @ -> l_bitm
            l_bitm -1 <> if \ area free
                #32 0 do
                    l_bitm 2 mod 1 xor l_free + -> l_free
                    l_bitm  1 rshift -> l_bitm
                loop
            endif
        cell +loop
    loop
    l_*buff free drop
    l_free 2/ -> l_free
    l_free 0 .r ." Kb, "
    l_free #1024 /mod 0 .r ." ." 0 .r ." Mb, "
    l_free #200 * DISK_SIZE / 0 .r ." % free"
    cr
    l_base base ! ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ display all open files
: list-fcb ( -- )
    active-fcbs
    cr
    ." FCB   FLD Op T   Blk   loc0  loc1  loc2  loc3  loc4  loc5  loc6  loc7  exp" 
    begin
        @ ?dup
    while
        dup cr  .hex space
        dup ->fld-no @ .
        dup ->open-tables @ .
        dup show-file-details
    repeat
    cr
;

\ display all open file handle details
: list-otb ( -- )
    active-otbs
    begin
        @ ?dup
    while
        cr
        dup ." OTB " .hex
        dup ->fcb-number @ dup ." FCB " .hex
        ->fld-no @ ." FLD " .decimal
        dup ->r/w-data-ptr @ ."  R/W pointer " .
    repeat
    cr ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: fdump { l_*name | l_handle l_*buff -- }
    l_*name count open-file dup 0= abort" Can't open "
    -> l_handle
    BLOCK_SIZE allocate abort" Can't allocate "
    -> l_*buff
    begin
        l_*buff BLOCK_SIZE 0 f-fill
        l_handle get-f-r/w-ptr 
        l_*buff BLOCK_SIZE 2/ l_handle read_f if
            cr cr
            ." Block " dup BLOCK_SIZE /mod .decimal ." offset " .decimal ." Ptr " .decimal
            l_*buff BLOCK_SIZE 2/ dump
            cr ." Press any to continue or D to end " cr
            key UPC case
                ascii D of true endof
                ascii S of 0 l_handle lseek_f if ." Moving to start" endif false endof
                ascii E of -513 l_handle lseek_f if ." Moving to last 512 bytes " endif false endof
                false swap
            endcase
        else
            drop true
        endif
    until
    l_*buff free drop
    l_handle close-file drop
    ." Done " ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
#240000 equ TAIL_TIMEOUT        \ four minutes

\ Show an ascii file
: tail-f ( *name -- )
    count (tail-f) ;
    
: (tail-f) { l_*name l_len | l_handle l_*buff l_key-wait -- }
    l_*name l_len open-file dup 0= abort" Can't open "
    -> l_handle
    BLOCK_SIZE allocate abort" Can't allocate "
    -> l_*buff
    cr
    TAIL_TIMEOUT later u_tail-timer !
    begin
        l_*buff BLOCK_SIZE l_handle read_f ?dup if
            l_*buff over bounds ?do
                i c@ emit
            loop
            BLOCK_SIZE = if
                begin
                    key?
                    u_tail-timer @ expired or
                until
                key? if
                    key upc case
                        ACR of true endof
                        ascii E of
                            cr ." ----- Skip ----- " cr cr
                            #-512 l_handle lseek_f drop false
                        endof
                        false swap
                    endcase
                    TAIL_TIMEOUT later u_tail-timer !   \ reset timer
                else
                    false
                endif
            else
                false
            endif
        else
            #250 later -> l_key-wait
            begin
                #25 ms
                key? l_key-wait expired or
            until
            key? if
                TAIL_TIMEOUT later u_tail-timer !
                key upc ACR =
            else
                false
            endif
        endif
        u_tail-timer @ expired or
    until
    l_*buff free drop
    l_handle close-file drop
    cr cr ." --- Done ------ " ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ list directory
\ -------------

: ls { | l_buff l_fld-buff }
    BLOCK_SIZE allocate abort" Alloc error "
    -> l_buff
    FLD-REC allocate abort" Alloc error "
    -> l_fld-buff
    working-dir-blk l_buff read-hd-block 0= abort" Error reading "
    0
    cr ." Directory listing " pwd
    cr ." FLD Name       Blk  Loc0  Loc1  Loc2  Loc3  Loc4  Loc5  Loc6  Loc7  Loc8  Loc9 Expn"
    cr ." ---------------------------------------------------------------------------------"
    begin
        dup l_buff + ->file-name @ dup 0<> swap -1 <> and if
            cr
            dup l_buff + ->file-number @ 3 .r
            dup l_buff + ->file-name count path-length over - spaces type
            dup l_buff + ->file-number @ l_fld-buff read-in-fld drop
            l_fld-buff show-file-details
        endif
        dir-struct +
        dup BLOCK_SIZE >=
    until
    drop cr
    l_fld-buff free drop
    l_buff free drop ;

: show-file-details ( buff -- )
    base @ >r decimal
    dup ->file-blocks @ 5 .r space
    dup ->extent-start-0 @ 5 .r space
    dup ->extent-start-1 @ 5 .r space
    dup ->extent-start-2 @ 5 .r space
    dup ->extent-start-3 @ 5 .r space
    dup ->extent-start-4 @ 5 .r space
    dup ->extent-start-5 @ 5 .r space
    dup ->extent-start-6 @ 5 .r space
    dup ->extent-start-7 @ 5 .r space
    dup ->extent-start-8 @ 5 .r space
    dup ->extent-start-9 @ 5 .r space
    cr #8 spaces
    dup ->indirect-block @ 4 .r space
    dup ->file-expansion @ 4 .r space
    2 spaces
    dup ->extent-size-0 @ 5 .r space
    dup ->extent-size-1 @ 5 .r space
    dup ->extent-size-2 @ 5 .r space
    dup ->extent-size-3 @ 5 .r space
    dup ->extent-size-4 @ 5 .r space
    dup ->extent-size-5 @ 5 .r space
    dup ->extent-size-6 @ 5 .r space
    dup ->extent-size-7 @ 5 .r space
    dup ->extent-size-8 @ 5 .r space
    dup ->extent-size-9 @ 5 .r space
    drop
    r> base ! ;
    
: (show-ftype) ( code -- )
    case
        FILE-FTYPE of ."  F" endof
        DIR-FTYPE  of ."  D" endof
        SYS-FTYPE  of ."  S" endof
    endcase ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ display a directory listing in UNIX format
: list { | l_buff l_fld-buff -- }
    BLOCK_SIZE allocate abort" Alloc error "
    -> l_buff
    FLD-REC allocate abort" Alloc error "
    -> l_fld-buff
    working-dir-blk l_buff read-hd-block 0= if
        ." Error reading "
        l_buff free drop exit
    endif
    0
    begin
        dup l_buff + ->file-name @ dup 0<> swap -1 <> and if
            cr
            dup l_buff + ->file-number @ l_fld-buff read-in-fld drop
            l_fld-buff show-file-details2
            dup l_buff + ->file-name count space type
                
        endif
        dir-struct +
        dup BLOCK_SIZE >=
    until
    drop cr
    l_fld-buff free drop
    l_buff free drop ;


: show-file-details2 ( buff -- )
    base @ >r decimal
    dup ->file-type w@ (show-ftype2)
    dup ->fld-no @ 4 .r space                           \ show fld
    ." S9040    " ." group   "
    dup ->file-size @ 8 .r space
    dup ->file-last-mod @ (show-time2)
    drop
    r> base ! ;

: (show-ftype2) ( code -- )
    case
        FILE-FTYPE of ." -rw-rw-rw-" endof
        DIR-FTYPE  of ." drw-rw-rw-" endof
        SYS-FTYPE  of ." -r-s------" endof
    endcase ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ display the current working directory
: pwd ( -- )
    *cwd @ ?dup if
        $.
    else
        DIR-DELIMIT emit
    endif
     ;


