
\ ***************************************************************
\ Module:	dir-parse.fth
\		Contains filename parseing functions
\ Author:	Jeremy Moller
\ Organisation:	CEM Systems Ltd, Belfast
\ Copyright:	1999
\ ******************************************************************


\ returns part of pathname
: get-path ( *name len -- *name2 len2 )
  DIR-DELIMIT extract-string ;

\ Returns address of next part of pathname
: get-next-path ( *name len -- next-addr next-len )
  DIR-DELIMIT jump-string ;

\ Returns TRUE if looking for a file and FALSE for a directory
: looking-file? ( *name len -- t/f )
  get-next-path nip 0= ;


: from-root? ( *name len -- t/f )
  0<> if
    c@ DIR-DELIMIT =
  else
    drop false
  endif ;

: find-file? { l_dir-blk l_*name l_len | l_*buff -- fn }
  BLOCK_SIZE allocate if
    drop 0 exit
  endif
  -> l_*buff
  l_dir-blk l_*buff 0 BLOCK_SIZE read-p-blk if	\ read in directory name block
    l_*name l_len get-path dup if
      l_*buff get-fid
    else
      \	." No path "
    endif
  else
    ." DIR read error "
    0
  endif
  l_*buff free drop  ;

\ -----------------------------------------------------

: get-fid { l_*name l_q-len l_dir-buff | l_f-num  l_wild -- file-num }
  false -> l_f-num
  l_*name s" *" s= if				\ wild card
    l_*name l_q-len ascii * skip -> l_q-len -> l_*name
    true -> l_wild
  else
    false -> l_wild
  endif
  begin
    l_wild if					\ wild card
      l_dir-buff ->file-name count l_*name l_q-len search nip nip 0<>
    else
      l_dir-buff ->file-name count l_*name l_q-len compare 0= \ read from directory list
    endif
    dup if					\ found
      l_dir-buff ->file-number @
\     cr ." Found" dup .word space
      -> l_f-num
    endif
    l_dir-buff DIR-STRUCT + -> l_dir-buff	\ next directory entry
    l_dir-buff ->file-number @ true = or	\ test last entry
  until
  l_f-num ;
		
: no-of-paths ( *name len -- n )
  false -rot					\ returned if too many paths
  PATH-LIMIT 0 do
    2dup get-path nip 0= if
      rot drop					\ lose bad outcome
      i -rot					\ put number of steps
      leave
    endif
    get-next-path
  loop
  2drop ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: working-dir-blk ( -- addr )
  dir-block @ ;

: set-working-dir ( addr -- )
  dir-block ! ;

: set-default-dir ( -- )
  ROOT_DIR_BLOCK set-working-dir
  *cwd @ free drop *cwd off
;

	
: step-next-dir ( *name len buff -- *name1 len1 dir-code )
  -rot get-next-path
  rot ->extent-start-0 @
  dup set-working-dir ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: get_fn { l_*name l_len | l_dir l_fld l_*fld -- fld-num }
  l_*name l_len from-root? if
    set-default-dir
  endif
  working-dir-blk -> l_dir
  FLD-REC allocate if
    drop FALSE exit
  endif
  -> l_*fld
  l_*name l_len no-of-paths 1 swap do
\ ." Path " i .
    l_dir l_*name l_len find-file? ?dup if \ read dir blk ; -- fld-num
      -> l_fld
      l_fld l_*fld read-in-fld drop
      l_*fld ->file-type w@
      case				\ read FLD and return 
        FILE-FTYPE of
\         ." file "
          i 1 <> if			\ not the end of path, but a file name 
            0 leave			\ return failure code
          else
	    l_fld leave
	  endif
        endof
        DIR-FTYPE of
\	  ." dir "
	  i 1 = if			\ last path section
            l_fld negate  leave		\ returns -ve FID number
	  else
	    l_*name l_len l_*fld step-next-dir
	    -> l_dir -> l_len -> l_*name
	  endif
        endof
\	  ." strange "
	  drop l_fld leave		\ probably a SYS file, or FLD missing !
      endcase
    else
\     ." not found "
      0 leave				\ file not found
    endif
  -1 +loop
  l_*fld free drop
;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\


: find-next-fld { | l_fld l_*buff -- fld-no }
  -1 -> l_fld
\ cr ." Finding next FLD entry "
  BLOCK_SIZE allocate if
    0 exit
  endif
  -> l_*buff
  FLD0_BLOCK FLD-SIZE bounds ?do	\ search all FLD blocks
\   cr ." block " i .
    i l_*buff read-f-blk 0= if
      leave
    endif
    BLOCK_SIZE 0 ?do
\     i FLD-REC / .
      i l_*buff + ->file-type w@ 0= if
        j FLD0_BLOCK - FLD/BLOCK * i FLD-REC / +
	-> l_fld
	leave
      endif
    FLD-REC +loop
    l_fld -1 <> if
      leave
    endif
  loop
  l_*buff free drop
  l_fld ;


\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
((
1) find first path F-NUM in root dir
2) lookup descriptor fo that F-NUM
3) if it's file, open it
  3a) check for existing FCB
  3a1) NO, allocate FCB
  3a2) YES, use existing FCB
  3b) create task open table
4) if it's a Dir
  4a) ensure applied name not expecting a file.
  4b) get that dir table
  4c) get next path and repeat above.
))

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: break-dir ( *name len -- )
  begin
    DIR-DELIMIT skip			\ find first \ 
    dup
   while
    2dup DIR-DELIMIT scan
    2swap
    2 pick - type cr
  repeat
  2drop ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\ used for PWD function

: inc-str-len ( N -- )
  *cwd @ c@ +	*cwd @ c! ;

: add-delimit-str ( -- )
  DIR-DELIMIT *cwd @ count + c!
  1 inc-str-len
;

: clr-dir-str ( -- )
  0 *cwd @ c! ;

: append-dir-str ( *name len -- )
  tuck
  *cwd @ count + swap cmove	\ cat strings together
  inc-str-len 			\ increase length
;							
	
: move-back-dir ( -- )
  *cwd @ count
  1- 0 swap ?do
    dup i + c@ DIR-DELIMIT = if
      i *cwd @ c!
      leave
    endif
  -1 +loop
  drop ;

: back-dir? ( *name -- t/f )
  s" .." s= ;

: remove-dir-hdr ( *name len -- '*name 'len )
  dup 3 min /string		\ short new string
;


