
\ ***************************************************************
\ Module:	file-main.fth
\		Contains the main generic file functions
\ Author:	Jeremy Moller
\ Organisation:	CEM Systems Ltd, Belfast
\ Copyright:	1999
\ ******************************************************************


\ To open a file with counted string with file and pathname
\ ENTRY:	counted string
\ EXIT:		File id number, used for all read write and close functions
: open-file ( *name len -- FID )
  get_fn dup if				\ -- fld
    open_fn 
  else
    \ ." file not found "
  endif
 ;

: open_fn ( fn -- FID )
  drive-control request 
  create-fcb dup if 			\ -- fcb,
    create-otb dup if			\ -- fcb, fid
      tuck ->fcb-number !		\ put fcb number in open table
      $0 over set-f-r/w-ptr		\ reset r/w ptr
    else
      ." no spare OTBs"
      nip				\ lose fcb
    endif
  else
    ." no spare FCBs "
  endif
  drive-control signal
;

: close-file ( fid -- t/f )
  dup ->valid-otb @ VALID_OTB <> if
    drop  0 exit
  endif
  sync_vol
  drive-control request
  dup ->fcb-number @ ?dup if		\ get FCB number from open table
    swap release-otb			\ -- fcb
    dup dec-fcb-open if			\ all links removed
      release-fcb
      -1
    else				\ other process has file open
      drop				\ lose FCB
      -2				\ 	File not completely closed
    endif	
  else
    drop
    false				\ error closing
  endif
  drive-control signal 
;

: file-not-exist? ( *name len -- t/f )
  get_fn 0= ;

\ -------------------------------------------
\ split Dir path and filename, returns file name part
: spllt-path&file { l_*name l_len | l_*tname l_tlen -- name len }
  begin
    l_*name l_len DIR-DELIMIT skip -> l_tlen -> l_*tname
    l_*tname l_tlen DIR-DELIMIT scan -> l_len -> l_*name
    l_len 0=
  until
  l_*tname l_tlen ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
: create-file { l_*name l_len l_size_b l_extent_b | l_*tname l_tlen -- t/f }
  l_*name l_len file-not-exist?		\ name does not exist
  l_len PATH-STRING <= and if		\ AND name not TOO long
    l_*name l_len spllt-path&file -> l_tlen -> l_*tname
    l_len l_tlen - ?dup if		\ any path ?
      l_*name swap ch-dir 0= if		\ change to that
        FALSE exit			\ couldn't
      endif
    endif
    l_tlen PATH-LENGTH <= if
      l_*tname l_tlen l_size_b l_extent_b (create-file)	\ now create file
    else
      FALSE
    endif
  else
    FALSE
  endif ;
	

: (create-file) { l_*name l_len l_size l_extent | l_start l_*buff -- t/f }
  BLOCK_SIZE allocate if
    drop exit
  endif
  -> l_*buff
  l_size adjust-blocks -> l_size
  l_extent (adjust-blocks) -> l_extent
  drive-control request 
  l_size need-blocks -> l_start		\ find data space
  find-next-fld dup -1 = if
    ." No more file FLDs"		\ find next fld
    drive-control signal 
    l_*buff free drop	
    drop  false exit
  endif
  dup l_*name l_len working-dir-blk add-to-dir	\ find & record filename
  0= if
    drive-control signal 
    l_*buff free drop	
    ." No more directory space "
    drop  false exit
  endif
  l_*buff over 0 FILE-FTYPE l_size l_extent l_start (create-fld)	
  l_*buff rewrite-fld			\ store FLD
  drive-control signal
  l_*buff free drop	
;

: create-dir ( l_*name l_len -- f/f )
  2dup file-not-exist? if
    (create-dir)
  else
    2drop  false
  endif
;

: (create-dir) { l_*name l_len | l_start l_fld l_prev-fld l_*buff -- t/f }
  s" ." get_fn dup 0< if
    negate
  endif
  -> l_prev-fld
  BLOCK_SIZE allocate if
    drop exit
  endif
  -> l_*buff
  drive-control request 
  DIR_BLOCK_SIZE need-blocks -> l_start	\ find data space
  find-next-fld dup -1 = if
    ." No more dir FLDs"		\ find next fld
    l_*buff free drop
    drive-control signal 
    drop false exit
  endif
  -> l_fld

  l_fld l_*name l_len working-dir-blk add-to-dir	\ find & record filename
  0= abort" No more directory space "

  l_*buff l_fld BLOCK_SIZE DIR-FTYPE DIR_BLOCK_SIZE 0 l_start (create-fld)
  l_fld l_*buff rewrite-fld drop	\ store FLD

  l_*buff BLOCK_SIZE -1 f-fill		\ mark dir as unused
  l_fld s" ." 		l_*buff write-filename
  l_prev-fld s" .."	l_*buff dir-struct + write-filename
  l_start l_*buff write-f-blk		\ clear this dir

  drive-control signal
  l_*buff free drop
  sync_vol
;

: rename_file { *old-name old-len *new-name new-len | l_fld -- t/f }
  *new-name new-len no-of-paths 1 =
  *old-name old-len no-of-paths 1 = and 0= if
    false exit
  endif
  *new-name new-len get_fn 0<>		\ new name exist ?
  *old-name old-len get_fn -> l_fld 	\ OR old name not exists ?
  l_fld 0= or if
    false exit
  endif
  drive-control request 
  l_fld working-dir-blk remove-file-dir if	\ remove name from dir list
    l_fld *new-name new-len working-dir-blk add-to-dir
  else
    false
  endif
  drive-control signal ;

: delete-files { l_*name l_len | -- t/f }
  l_*name s" *" s= if			\ wild card
    begin
      l_*name l_len delete-file
    0= until
    true
  else
    l_*name l_len delete-file
  endif ;

: delete-file ( name len -- t/f )
  get_fn ?dup if
    (delete-file)
  else
    false
  endif ;

: (delete-file) { l_fn | l_*buff l_size -- t/f }
  FLD-REC allocate if
    0 exit
  endif
  -> l_*buff
  0 -> l_size
  drive-control request 
\  cr ." delete dir " dup .decimal
  l_fn working-dir-blk remove-file-dir drop	\ remove directory entry
  l_fn l_*buff read-in-fld drop		\ load FLD from disk
  NO-OF-EXTENTS 0 ?do
    i l_*buff file-extent dup l_size + -> l_size
    2dup or if
      \ cr ." delete extent " i . 2dup .decimal .decimal
      free-blocks			\ remove bitmap entry
    else
      2drop leave			\ all extents found
    endif
  loop
  l_*buff ->indirect-block @ ?dup if
    l_size delete-indir-extent
  endif
\ cr ." delete fld "
  l_*buff FLD-REC erase
  l_fn l_*buff rewrite-fld
  l_*buff free drop
  drive-control signal ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: cdir ( *name -- )
  count ch-dir ?load-ok ;
	
: need-dir-$buffer ( -- )
  *cwd @ 0= if				\ no buffer defined
    PATH-STRING allocate if
      drop 2drop drop	false exit
    endif
    *cwd !
    clr-dir-str				\ initialise string length
  endif ;

:  ch-dir ( *name len -- t/f )
   2dup s" /" compare 0= if		\ root dir ?
     2drop
     set-default-dir
     TRUE
   else
     (ch-dir)	\ NO, find file
   endif ;

: (ch-dir) ( *name len -- t/f )
  2dup get_fn dup 0< if
    drive-control request
    negate cdir-buffer read-in-fld drop
    cdir-buffer ->extent-start-0 @ set-working-dir
    drive-control signal
    need-dir-$buffer			\ allocate a string name buffer
    2dup from-root? if			\ new string defined from root ?
      clr-dir-str
    else
      over back-dir? if			\ move to previous dir
        move-back-dir			\ adjust stored string
        remove-dir-hdr			\ remove header from new string
      endif
      add-delimit-str			\ put dir delimiter after existing string
    endif
    append-dir-str			\ add new string to existing dir
    true
  else
    drop 2drop
    set-default-dir
    FALSE
  endif ;
	
	
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\ PTR = 0 	set to start
\ PTR = -1 set to end of written data
\ PTR = +ve from start of file
\ PTR = -ve from end of data

: lseek_f ( ptr fid -- t/f )
  dup ->valid-otb @ VALID_OTB <> if
    2drop  0 exit
  endif
  dup fcb-addr ->file-blocks @ 0<> if	\ see if exists
    over 0< if				\ test ptr
      dup fid-file-size 1+ rot + dup 0< if
	drop  0
      endif
      swap	 			\ set PTR to end of data area
    endif
    tuck set-f-r/w-ptr			\ set new pointer
    r/w-ptr-inrange?			\ test it
  else
    2drop  false
  endif ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ ENTRY: l_buf is destination buffer,
\ ENTRY: l_len is amount of data (bytes)
\ ENTRY: l_fid is file id, (open table pointer)

: read_f { l_buf l_len l_fid | l_read-size l_start-buff -- len }
  l_fid ->valid-otb @ VALID_OTB <> if
    0 exit
  endif
  l_buf -> l_start-buff
  l_len l_fid required-blocks 0 ?do					
    l_fid r/w-block			\ get physical block to hd read from pointer
    l_fid r-ptr-inrange? 0= if 		\ test in read range of stored data
      drop  leave			\ outside read range, return 0 bytes read
    else				\ -- block
      l_buf
      l_fid l_len file-read-area -> l_read-size
      l_read-size 0= if			\ work out which part of block to read
        drop 2drop  leave
      endif
      l_read-size read-d-block 0= if	\ read block from disk
        leave				\ error reading			
      endif
      l_len l_read-size - -> l_len	\ dec length
      l_read-size l_buf + -> l_buf	\ inc buffer pointer
      l_read-size l_fid inc-f-r/w-ptr	\ move logical disk pointer
    endif
  loop
  l_buf l_start-buff - ;		\ amount actually read

: file-read-area { l_fid l_len | l_*c-pos -- b-start size }
  l_fid get-f-r/w-ptr -> l_*c-pos
  l_fid get-r/w-pb-ptr			\ get read start offset
  l_len l_*c-pos + 			\ new end ptr
  l_fid fid-file-size min		\ end of file ptr
  l_*c-pos -				\ max amount required to read
  BLOCK_SIZE min dup BLOCK_SIZE = if	\ only if full size
    over -				\ less the start pos = Max this read
  endif
  2dup + BLOCK_SIZE > if		\ does the read straddle
    drop				\ lose size
    BLOCK_SIZE over -
  endif	
;


\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
: write_f  { l_buf l_len l_fid | l_size l_start l_req-blk -- len }
  l_fid ->valid-otb @ VALID_OTB <> if
    0 exit
  endif
  l_buf -> l_start
  l_len l_fid required-blocks	-> l_req-blk	\ get number of blocks required
  l_req-blk l_fid need-extent? if	\ need more space for this file
    l_req-blk l_fid annex-file 0= if	\ YES, try and grab some
      0 exit				\ couldn't, so leave
    endif
  endif
  l_req-blk 0 do	
    l_fid r/w-block dup 0< if		\ get block to write to from pointer
      drop leave			\ error with allocation
    endif
    l_buf				\ buffer
    l_fid get-r/w-pb-ptr		\ start pointer
    dup l_len + BLOCK_SIZE min		\ end pointer
    2dup swap - -> l_size		\ size
    write-hd-block2 0= if		\ write block
      leave
    endif
    l_len l_size - -> l_len		\ dec length
    l_size l_buf + -> l_buf		\ inc buffer pointer
    l_size l_fid inc-f-r/w-ptr		\ move logical disk pointer
  loop
  l_fid update-size&time drop		\ update FCB and FLD last mod date & file size
  l_buf l_start - ;			\ number of blocks written

: update-size&time { l_fid | l_size l_fcb -- t/f }
  l_fid ->valid-otb @ VALID_OTB <> if
    0 exit
  endif
  l_fid get-f-r/w-ptr -> l_size 
  l_fid fcb-addr -> l_fcb
  l_fcb ->file-size @ l_size < if
    l_size l_fcb ->file-size !
  endif
  lseconds @ l_fcb ->file-last-mod !
  l_fcb update-fcb			\ change on disk
;		

: zero-file ( l_fid -- t/f )
  dup ->valid-otb @ VALID_OTB <> if
    drop 0 exit
  endif
  fcb-addr				\ get fcb number
  0 over ->file-size !
  lseconds @ over ->file-last-mod !
  update-fcb				\ change on disk
;		

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: (adjust-blocks) ( len -- blocks )
  BLOCK_SIZE /mod swap 0<> if	\ round up to complete blocks
    1+
  endif ;

: adjust-blocks ( len -- blocks )
  (adjust-blocks)
  ?dup 0= if				\ ensure always at least 1 block
    1
  endif ;

: required-blocks ( len fid -- blocks )
  get-r/w-pb-ptr + adjust-blocks ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

