
\ ***************************************************************
\ Module:	init-disk.fth
\		Contains disk format and initialisation functions
\ Author:	Jeremy Moller
\ Organisation:	CEM Systems Ltd, Belfast
\ Copyright:	1999
\ ******************************************************************


 0 equ DISK_OK
-1 equ DISK_WRONG_FORMAT
-2 equ DISK_ERROR


\ this is done at boot up, to read the disk format details
: get-disk-info ( -- )
  drive-control signal 			\ initialise semaphore
  fld-update signal			\ initialise semaphore
  init-fcbs				\ setup the FILE CONTROL BLOCK list
  init-otbs				\ setup the OPEN TABLE CONTROL BLOCK list
  inited-disk
;

: inited-disk ( -- )
  (get-disk-info) ?dup if
    dup DISK_WRONG_FORMAT = if
      cr ." Do you wish to CEM Format this disk? (y/n)"
      y/n? if
	cr ." Initialising CF disk"
	init-vol
      else
	cr ." Replace CF disk "
	begin pause again
      endif
    endif
    DISK_ERROR = if
      cr ." DISK ERROR, Can't continue "
      begin pause again
    endif
  endif ;
	
\ RETURNS, 0 OK, -1 wrong format, -2 read error
: (get-disk-info) ( -- oc )
  cr ." Checking disk "
  ROOT_BLOCK cdir-buffer read-f-block 0= if
    ."                                 Failed"
    DISK_ERROR
    exit
  endif
  ." format"
  cdir-buffer ->CEM_FORMAT @ CEM_FORMAT = if
    cdir-buffer ->USAGE_BLOCK0 @ to USAGE_BLOCK0
    cdir-buffer ->vol-size @ dup to DISK_SIZE 		\ number of blocks per disk
    BLOCKS/BIT_TABLE / 		 to USAGE_TABLE_SIZE	\ size of bit map table
    cdir-buffer ->FLD0_BLOCK @ 	 to FLD0_BLOCK		\ start of fld table
    MAX_FILES FLD/BLOCK / 	 to FLD-SIZE
    cdir-buffer ->DATA_BLOCK0 @  to DATA_BLOCK0		\ start of files
    set-default-dir					\ set-default-dir
    DISK_OK
    ."                           OKay"
  else
    DISK_WRONG_FORMAT
    ."                 NOT CEM format"
  endif
;

: Test-disk-speed { l_*dest | l_time -- }
  cr ." Testing Disk performance "
  ticks -> l_time
  ." W1 "
  DISK_SIZE 2 - l_*dest write-block if
    ." W2 "
    DISK_SIZE 2 - l_*dest write-block 0= if
      ." error"
    endif
  else
    ." error"
  endif
  ticks l_time -  . ." mS" ;

: show-drive ( -- )
  Disk_size		cr ." Disk size            " .
  ROOT_DIR_BLOCK	cr ." Root starts          " .
  usage_block0 	cr ." Bit map starts       " .
  FLD0_BLOCK		cr ." FLD starts           " .
  DATA_BLOCK0		cr ." Data area starts     " .
  USAGE_TABLE_SIZE cr ." Bit map table size   " .
  FLD-SIZE		cr ." File descriptor size " .
  cr ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
	
: init-vol ( -- )
  (drive-id) dup			\ read drive
  active-fcbs @ 0= and if		\ files not open
    release-all-cache
    dup init-root-block
    free drop
    show-drive 
    init-root-dir			\ setup root directory on disk
    init-vol-bitmap			\ setup usage bit map on disk
    init-fld				\ setup File List Descriptor on disk
    init-fcbs				\ setup the FILE CONTROL BLOCK list
    init-otbs				\ setup the OPEN TABLE CONTROL BLOCK list
  else
    0= if 
      cr ." Error reading disk "
    else
      cr ." Files still open "
    endif
  endif ;

: init-root-dir ( -- )
  cdir-buffer
  dup BLOCK_SIZE $0ff fill
  02 s" ." 2 pick dir-struct * 4 pick + write-filename 
  00 s" bitmap.sys" 2 pick dir-struct * 4 pick + write-filename 
  01 s" flist.sys" 2 pick dir-struct * 4 pick + write-filename
  working-dir-blk over write-block 0= abort" error writing"
  drop ;

SWAP_DISK_DATA 0= [if]

: a-h-b ( n -- n1 )
;

[else]

code a-h-b ( n -- n1 )
	sub.l er1 er1
	mov.b tos.l r0l
	rotxr.b r0l rotxl.b r1l
	rotxr.b r0l rotxl.b r1l
	rotxr.b r0l rotxl.b r1l
	rotxr.b r0l rotxl.b r1l
	rotxr.b r0l rotxl.b r1l
	rotxr.b r0l rotxl.b r1l
	rotxr.b r0l rotxl.b r1l
	rotxr.b r0l rotxl.b r1l
	mov.b r1l tos.l
	next,
end-code

[endif]

: init-root-block { l_buff | l_dest }
  block_size allocate abort" ALLOC ERROR "
  -> l_dest 
  l_dest BLOCK_SIZE erase
  USAGE_BLOCK to USAGE_BLOCK0
  USAGE_BLOCK0 l_dest ->USAGE_BLOCK0 !	\ start of bit map table
  l_buff #58 2* + w@ 
  $10000 *
  l_buff #57 2* + w@ a-h-b +
  dup to DISK_SIZE
  dup l_dest ->vol-size !		\ number of blocks per disk
  BLOCKS/BIT_TABLE / 			\ size of bit map table
  dup to USAGE_TABLE_SIZE
  USAGE_BLOCK0 +
  dup l_dest ->FLD0_BLOCK !		\ start of fld table
  dup to FLD0_BLOCK
  MAX_FILES FLD/BLOCK /
  dup to FLD-SIZE
  +
  dup to DATA_BLOCK0
  l_dest ->DATA_BLOCK0 !		\ start of files
  lseconds @ 
  l_dest ->init-time !			\ time of 
  CEM_FORMAT l_dest ->CEM_FORMAT !
  ROOT_BLOCK l_dest write-block 0= abort" Write error "
  l_dest free abort" Error free "
;

: init-vol-bitmap ( -- )
  USAGE_BLOCK0 USAGE_TABLE_SIZE bounds ?do
    i erase-block 0=
    if  cr ." Error ERASE " leave  endif
  loop
  grab/free off				\ flag to grab blocks
  0 DATA_BLOCK0 block-order allocate-blocks \ mark the first few blocks as used
;

: init-fld ( -- )
  FLD0_BLOCK FLD-SIZE bounds ?do	\ clear all FLD blocks
    i erase-block 0=
    if  cr ." Error ERASE " leave  endif	
  loop
  Block_size allocate abort" ALLOC error "
  dup BLOCK_SIZE erase
  dup 0 USAGE_TABLE_SIZE BLOCK_SIZE * SYS-FTYPE USAGE_TABLE_SIZE 0 usage_block0 create-fld
  dup 1 FLD-SIZE BLOCK_SIZE * SYS-FTYPE FLD-SIZE 0 FLD0_BLOCK create-fld
  dup 2 BLOCK_SIZE DIR-FTYPE 1 0 ROOT_DIR_BLOCK create-fld
  FLD0_BLOCK over write-block drop
  free drop
;


\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: add-to-dir { l_fld l_*name l_len l_block | l_success -- }
  false -> l_success
  BLOCK_SIZE allocate abort" Alloc error "
  l_block over read-f-blk drop
  dup BLOCK_SIZE bounds ?do			\ find next directory slot
    i ->file-name @ dup 0= swap -1 = or if	\ free or unused slot
      l_fld l_*name l_len i write-filename
      l_block over write-f-blk -> l_success
      leave
    endif
  dir-struct +loop
  free drop
  l_success ;

: remove-file-dir { l_fld l_block | l_success -- }
  false -> l_success
  BLOCK_SIZE allocate abort" Alloc error "
  l_block over read-f-blk drop
  dup BLOCK_SIZE bounds ?do		\ find next directory slot
    i ->file-number @ l_fld = if	\ 
      i dir-struct bounds ?do		\ now clear the entry
	i off					
      cell +loop
      l_block over write-f-blk drop
      true -> l_success
      leave
    endif
  dir-struct +loop
  free drop
  l_success ;

: write-filename { fld *name len buff -- }
  *name len get-path -> len -> *name
  fld 	buff ->file-number !
  len		buff ->file-name c!
  *name	buff ->file-name 1+ len cmove
;

: create-fld { l_buff l_fld l_bytes l_type l_blocks l_expan l_start -- }
  l_fld fld-m-block l_buff +
  l_fld l_bytes l_type l_blocks l_expan l_start (create-fld) ;

: (create-fld) { l_buff l_fld l_bytes l_type l_blocks l_expan l_start -- }
  l_buff FLD-REC erase
  l_blocks BLOCK_SIZE * l_bytes < abort" Size definition error "
  l_fld		l_buff ->fld-no !
  l_blocks	l_buff ->file-blocks !
  l_bytes 	l_buff ->file-size !
  l_type	l_buff ->file-type w!
  l_expan	l_buff ->file-expansion !
  l_start	l_buff ->extent-start-0 !
  l_blocks	l_buff ->extent-size-0 !
  lseconds @	l_buff ->file-last-mod !
;


: file-extent { l_ext-no l_fld | -- block size }
  l_ext-no cells 2* -> l_ext-no
  l_fld ->extent-start-0 l_ext-no + @
  l_fld ->extent-size-0 l_ext-no + @ ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

