\ Parallel port driver for I2C

((
Copyright (c) 1999
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
fax: +44 (0)2380 339691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk
))


\ Parallel Port----------------------- PCF8574
\ Data0  (pin 2) ------------------ SCL (pin 14) ------ To Other
\ STROBE (pin 1)------------------- SDA (pin 15) ------ I2C Devices


\ ----------------------------------------------------------------------------
\ Parallel port addresses
\ ----------------------------------------------------------------------------

$378 constant scl_port	 	\ scl port address - parallel data port address
$37A constant sda_port	 	\ sda port address - parallel control port address
$001 constant sda_bit	        \ address of sda bit
$001 constant scl_bit	        \ address of scl bit
		
\ ----------------------------------------------------------------------------
\ VFX Harness
\ ----------------------------------------------------------------------------

extern DWORD PASCAL GetTickCount( void );

: ms		\ time --
  GetTickCount +
  begin dup GetTickCount - 0< until 
  drop
;

variable <delay>
  200 <delay> !		\ gives 40uS delay on P166Mhz CPU

: delay		\ -- ; short wait
  <delay> @ 0 
  ?do  noop  loop
;

code pc@	\ port# -- u8
	mov	dx, bx
	in	al, dx
	mov	bl, al
	and	ebx, 255
	next,
end-code

code pc!	\ u8 port# --
	mov	edx, ebx
	mov	eax, 0 [ebp]
	mov	ebx, 4 [ebp]
	add	ebp, 8
	out	dx, al
	next,
end-code

\ ----------------------------------------------------------------------------
\ Target specific definitions for SDA!, SDA@, SCL!
\ ----------------------------------------------------------------------------

: scl@	        \ -- b
  scl_port pc@
;

: scl!          \ b -- ; store byte to printer data port
  scl_port pc!
;

: sda@          \ -- b ; read state of SDA bit - inversion required
  sda_port pc@ 
;

: sda!        	\ b -- ; store byte to printer control port
  sda_port pc!
;

\ ----------------------------------------------------------------------------
\ Definitions to set SDA and SCL high and low, but perserves other 
\ bits on parallel port by reading and masking
\ ----------------------------------------------------------------------------

: sda_low 	\ --       	
  sda@ sda_bit OR SDA! 		\		; set SDA LOW - inversion required	     
;			      

: sda_high  	\ --        
  sda@ sda_bit OR scl_bit XOR sda!	\		; set SDA HIGH - inversion required
;			

: scl_low 	\ -- 	
  scl@ scl_bit OR scl_bit XOR scl!	\		; set SCL LOW 
;

: scl_high	\ -- 		
  delay    			\	        ; delay
  scl@ scl_bit OR scl!		\	        ; set SCL high
  delay    			\	        ; delay
;

: read_sda	\ -- sda
  sda@ sda_bit AND sda_bit XOR  \ sda-flag	; reads sda and inverts for parallel port
;

