\ Parallel port driver for I2C

((
Copyright (c) 1999
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
fax: +44 (0)2380 339691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk


Parallel Port----------------------- PCF8574
Data0  (pin 2) ------------------ SCL (pin 14) ------ To Other
STROBE (pin 1)------------------- SDA (pin 15) ------ I2C Devices
))


\ *! i2c_vfx
\ *T VFX Forth for Windows I2C driver
\ ===============
\ *S Introduction
\ ===============
\ *P The I2C interface is a software bit-banging system
\ ** using the PC printer port. The
\ ** initialisation code sets up these bits as open
\ ** drain outputs.

\ *P The driver code in I2VFXDRV.FTH must be compiled
\ ** before the generic drive code in I2CBASE.FTH and the
\ ** device specific files (e.g. AT24C512.FTH) are
\ ** compiled.

\ *P The code reads and writes the printer port directly.
\ ** If you are using Windows ME, NT, 2000, or XP or any
\ ** other version that prohibits direct port access from
\ ** applications, you will have to install the port access
\ ** driver supplied with VFX Forth. Installation instructions
\ ** for this are in the manual.

0 constant TestI2C?	\ -- n
\ *G Set this constant non-zero to include the test code
\ ** at the end of the file. This equate also controls
\ ** the compilation of I2C test code in other files.


\ --------------------------
\ *S Parallel port addresses
\ --------------------------

$378 value parallel_port
\ *G Returns the parallel port address.

: lpt1:		( -- port )  $378 to parallel_port  ; 
\ *G Select LPT1 at $0378.
: lpt2:		( -- port )  $278 to parallel_port  ; 
\ *G Select LPT2 at $0278.

: scl_port	\ -- port
\ *G SCL port I/O address - data port
  parallel_port
;

: sda_port 	\ -- port
\ *G SDA port I/O address - control port
  parallel_port 2 +
;

$001 constant scl_bit  	\ -- mask ; scl bit mask
\ *G SCL bit mask for SCL port
0 constant #SCLshift	\ -- n
\ *G number of bits to shift right to put the SCL bit
\ ** in the lsb.
$001 constant sda_bit	\ -- mask sda bit mask
\ *G SDA bit mask for SDA port
0 constant #SDAshift	\ -- n
\ *G number of bits to shift right to put the SDA bit
\ ** in the lsb.

		
\ --------------
\ *S VFX Harness
\ --------------

extern: DWORD PASCAL GetTickCount( void );

[undefined] ms [if]
[undefined] pause [if]
: ms		\ time --
\ *G wait for time milliseconds.
  GetTickCount +
  begin dup GetTickCount - 0< until 
  drop
;
[else]
: ms		\ time --
  GetTickCount +
  begin pause dup GetTickCount - 0< until 
  drop
;
[then]
[then]

variable <delay>  200 <delay> !	\ -- addr ;
\ *G Tune this value for your hardware and O/S.
\ ** The default value of 200 gives 40uS delay on a P166Mhz CPU.

[-sin
: I2Cdelay		\ -- ; short wait
\ *G Wait for a quarter I2C bit time.
  <delay> @ 0 
  ?do  noop  loop
;
sin]

code pc@	\ port# -- u8
\ *G Read a byte from an I/O port.
  mov	dx, bx
  in	al, dx
  mov	bl, al
  and	ebx, 255
  next,
end-code

code pc!	\ u8 port# --
\ *G Write a byte to an I/O port.
  mov	edx, ebx
  mov	eax, 0 [ebp]
  mov	ebx, 4 [ebp]
  add	ebp, 8
  out	dx, al
  next,
end-code

: scl@	        \ -- b
\ *G Read the port containing the SCL line.
  scl_port pc@
;

: scl!          \ b -- ; store byte to printer data port
\ *G Write the port containing the SCL line.
  scl_port pc!
;

: sda@          \ -- b ; read state of SDA bit - inversion required
\ *G Read the port containing the SDA line.
  sda_port pc@ 
;

: sda!        	\ b -- ; store byte to printer control port
\ *G Write the port containing the SDA line.
  sda_port pc!
;

\ ===================
\ *S I2C bit handling
\ ===================
\ *P This code provides words to set SDA and SCL high and low,
\ ** but preserves other bits on the parallel port by reading
\ ** and masking.

: sda_low 	\ --       	
\ *G Set SDA low.
  sda@ sda_bit OR sda! 			\ set SDA LOW - inversion required	     
;			      

: sda_high  	\ --
\ *G Set SDA high.
  sda@ sda_bit OR scl_bit XOR sda!	\ set SDA HIGH - inversion required
;			

: read_sda	\ -- sda ; return 0/1.
\ *G Read the state of the SDA line into bit 0.
  sda@ sda_bit AND sda_bit XOR  	\ read SDA - inversion required
  #SDAshift rshift			\ put in lsb.
;

: write_sda	\ sda --
\ *G Set/Clear the SDA line.
  if  sda_high  else  sda_low  endif
;

: scl_low 	\ --
\ *G Set SCL low. A quarter cycle delay is performed
\ ** before and after the transition.
  I2Cdelay
  scl@ scl_bit OR scl_bit XOR scl!
  I2Cdelay
;

: scl_high	\ -- 		
\ *G Set SCL high. A quarter cycle delay is performed
\ ** before and after the transition.
  I2Cdelay    			\	        ; delay
  scl@ scl_bit OR scl!		\	        ; set SCL high
  I2Cdelay    			\	        ; delay
;


\ ======
\ *> ###
\ ======

