\ Keypad read routine for hex matrix keypad

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============


Keypad layout
=============
        C1      C2      C3      C4
      -----   -----   -----   -----   
R1:  |  1  | |  2  | |  3  | |  C  |  
      -----   -----   -----   -----   
      -----   -----   -----   -----   
R2:  |  4  | |  5  | |  6  | |  D  |  
      -----   -----   -----   -----   
      -----   -----   -----   -----   
R3:  |  7  | |  8  | |  9  | |  E  |  
      -----   -----   -----   -----   
      -----   -----   -----   -----   
R4:  |  A  | |  0  | |  B  | |  F  |  
      -----   -----   -----   -----  

Principle of operation
======================

The keypad is connected to an 8-bit I/O port. The top four bits are outputs
(active low) connected to the four row lines of the keypad. The bottom four
bits are inputs connected to the column lines of the keypad.

The keypad is scanned by setting a row output low, and then checking to
see if one of the column lines has been pulled low.

To converting this code for a different keypad or port layout is simple
1) Modify the key table KEYTABLE
2) Modify the row/column write read routine CHECKROW
))


decimal

\ ************************
\ Key management variables
\ ************************

-1 constant nokey       \ no key pressed indicator

variable currkey        \ -1 if no key pressed, otherwise key code
  nokey currkey !
variable lastkey
  nokey lastkey !


\ ************************************************************
\ Key code table, contains a mask for the 8 bit i/o port, plus
\ code values for each condition. This allows us to perform
\ some rollover detection.
\ ************************************************************

#17 constant rowsize    \ size of key table row entry

create keytable
\       mask     0/8     1/9     2/A     3/B     4/C     5/D     6/E     7/F
( R1 )  $EF c,   $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $0C c,
                 $FF c,  $FF c,  $FF c,  $03 c,  $FF c,  $02 c,  $01 c,  $FF c,
( R2 )  $DF c,   $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $0D c,
                 $FF c,  $FF c,  $FF c,  $06 c,  $FF c,  $05 c,  $04 c,  $FF c,
( R3 )  $BF c,   $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $0E c,
                 $FF c,  $FF c,  $FF c,  $09 c,  $FF c,  $08 c,  $07 c,  $FF c,
( R4 )  $7F c,   $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $0F c,
                 $FF c,  $FF c,  $FF c,  $0B c,  $FF c,  $00 c,  $0A c,  $FF c,


\ *********************
\ Key handling routines
\ *********************

: checkrow      \ mask -- state ; check status of given row ; code this if it is easy
  p1 p!                                 \ write mask to output port
  p1 p@ $0F and                         \ read and mask input bits
;

: checkpad      \ -- keystate ; check keypad, returning state
\  service-watchdog                      \ kick watchdog if necessary
\  timer-update                          \ update timers as necessary
\  minutes&seconds                       \ update clock display as necessary
  $FF                                   \ initial value
  [ keytable rowsize 4 * + ] literal    \ end of table
  keytable                              \ start of table
  do
    drop                                \ discard dummy key
    i c@c checkrow i + 1+ c@c           \ check row, extract keycode
    dup $FF <> ?leave                   \ get out if valid keycode
  rowsize +loop                         \ step to next row
  dup $FF =                             \ convert invalid code to -1
  if  drop  nokey  endif
;

: read-keypad   \ -- ; check keypad state
  checkpad                              \ get first key state
  begin
    checkpad dup rot =                  \ keep going until two values the same
  until
  dup lastkey @ =                       \ if it's the same as the last key
  if
    drop                                \  ignore it
  else
    dup lastkey !  currkey !            \ otherwise accept it as valid
  endif
;

: wait-for-key  \ -- ; wait for key
  nokey currkey !
  begin
    read-keypad  currkey @ nokey <>
  until
;


\ *****************************************
\ Application words similar to KEY and KEY?
\ *****************************************

: keypad        \ -- key ; wait for key, and clear it when found
  wait-for-key  currkey @
  nokey currkey !
;

: keypad?       \ -- flag ; true when a key has been pressed
  currkey @ nokey =
  if
    read-keypad  currkey @ nokey <>
  else
    true
  endif
;
  
decimal

