\ LOCALCOM.FTH - high-level ANS local variables extensions

((
Copyright (c) 2001
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
19981007 MSD003 The code that searched for the closing curly brace in LOCALS|
         MSD003 used the dirty trick of "BL WORD W@".  This worked fine on
         MSD003 the H8/3048F and 68K chips because they are *big*endian.
         MSD003 However, the LH77790A (ARM7) chip is *little* endian and this
         MSD003 trick didn't work anymore...
19980428 MSD002 Altered the code to only include the environment stuff if
         MSD002 ENVIRONMENT is defined.  This will mean that this file
         MSD002 doesn't have to know the name of the "want environment?"
         MSD002 switch that I'm about to add to all the control files.
19980427 MSD001 Apply Harry's fix to LVAR where a C-ADDR U was COUNTed.
))


\ ===========
\ *! localcom
\ *T Target local variables
\ ===========
\ *P The file COMMON\LOCALCOM.FTH implements the compilation of
\ ** the ANS Forth LOCALS| syntax for compilation on the host.
\ ** Compilation of this file requires CPU dependent support, usually 
\ ** called LOCAL.FTH in the %CpuDir% directory, and MPE standard
\ ** control files will compile these files if the equate TARGET-LOCALS?
\ ** is set non-zero in the control file.

\ *P Note that this file is only provided for full ANS compliance.
\ ** The MPE extended local variable syntax is provided by the
\ ** cross compiler, and is much more powerful and more readable.

[defined] environment [if]
  also environment definitions
  #16  constant #LOCALS		\ -- true ; max number of locals allowed 
  TRUE constant LOCALS          \ -- true ; extension present
  TRUE constant LOCALS-EXT      \ -- true ; extension present
  previous definitions
[then]                                                                          \ MSD002

\ ------------------------------------------------------------------------------
\       local variable compiler for target
\ ------------------------------------------------------------------------------

vocabulary LOCALVARS			\ in temp area for local vars

internal

variable TDP                            \ temp dp for local use
variable OLDDP                          \ saved current dp
variable OLDLAST                        \ saved last nfa
variable OLDCURRENT                     \ saved current
variable LVCOUNT                        \ current local variable
variable lvcountmax                     \ total offset required
variable LOCAL-OFFSET                   \ offset used for local dictionary
$400 local-offset !                     \ declare how far above dp to lay temp words

create LOCAL-MARKER  ", $local$"

: >TEMP-DICT	\ -- ; switch in temp dictionary
  dp @ olddp !                          \ first save current dp
  tdp @ dp !                            \ use temp dp
  last @ oldlast !                      \ save last
  current @ oldcurrent !                \ save vocabs
  ['] localvars >body @ current !
;

: >REAL-DICT	\ -- ; switch out temp dictionary
  dp @ tdp !                            \ save temp dp
  olddp @ dp !                          \ use real current dp
  oldlast @ last !                      \ restore last
  oldcurrent @ current !                \ restore vocabs
;

: START-LOCALS	\ --
  dp @ local-offset @ + tdp !           \ create new dp
  also localvars
  >temp-dict
[defined] $create [if]
  local-marker $create
[else]
  local-marker count makeheader docreate,
[then]
  >real-dict
;

: FORGET-LOCALS	\ --
  >temp-dict
  local-marker $forget
  >real-dict
  tdp off
  lvcount off
  previous
;

[defined] MAKEHEADER [if]
: LVAR		\ c-addr u -- ; build temp local var
  makeheader docreate,
  immediate                             \ and mark as immediate
  lvcount @ ,                           \ lay offset
Does>
  @
  OperatorType @off                     \ any prefixed operator ???
  case
    DEFAULTOP of  ['] lv@ compile, ,  endof	\ lay function and offset
    STOREOP   of  ['] lv! compile, ,  endof	\ lay function and offset
    ADDROP    of  ['] lv  compile, ,  endof	\ lay function and offset
                   bad-method
  endcase
;
[else]
: LVAR		\ c-addr u -- ; build temp local var
  align                                 \ force DP to even address
  2dup upper                            \ convert to upper case
  2dup current @ search-wordlist        \ see if name exists    \ MSD001
  if  
    ." Local variable "
    >name CR .name 
    ." is redefined "                   \ yes, say so
  then
  here cell+ place			\ copy string to HERE+4
  header docreate,
  immediate                             \ and mark as immediate
  lvcount @ ,                           \ lay offset
Does>
  @
  OperatorType @off                     \ any prefixed operator ???
  case
    DEFAULTOP of  ['] lv@ compile, ,  endof	\ lay function and offset
    STOREOP   of  ['] lv! compile, ,  endof	\ lay function and offset
    ADDROP    of  ['] lv  compile, ,  endof	\ lay function and offset
                   bad-method
  endcase
;
[then]

: MAKELVAR	\ c-addr u -- ; build temp local var
  >temp-dict
  lvar
  >real-dict
;

: DROP-LOCALS	\ -- ; cleanup code for local stack vars
  lvcount @                             \ only compile is we had lvars
  if  compile rellvs  endif             \ number of vars to lose
;

: CLEANUP-LOCALS	\ -- ; cleanup code for semicolon
  lvcount @                             \ only compile is we had lvars
  if  forget-locals  endif              \ now lose local definitions
;

external


: (LOCAL)	\ Comp: c-addr u -- ; Exec: -- x ; define local var
\ *G When executed during compilation, defines a local variable
\ ** whose name is given by c-addr/u. If u is zero, c-addr is ignored
\ ** and compilation of local variables is assumed to finish.
\ ** When the word containing the local variable executes, the local
\ ** variable is initialised from the stack. When the
\ ** local variable executes, its value is returned. The local variable
\ ** may be written to by preceding its name with TO. This word is
\ ** provided for the construction of user-defined local variable notations.
\ ** This word is only provided for ANS compatibility, and locals created
\ ** by it cannot be optimised by the VFX code generator.
  lvcount @ 0= if
    start-locals                        \ initialise locals
    [ 2 cells ] literal lvcount !	\ 2 cells used for frame management
    compile makelvs
    here -rot  0 ,                      \ count will be fixed up later
  endif
  ?dup if                               \ OK not last
    makelvar                            \ build temp local variable
    cell lvcount +!                     \ bump for next
  else                                  \ is last local
    drop
    lvcount @ 2 cells -                 \ discount frame management in patchup of MAKELVS
    u4/ over w!                         \ fix up two *words* after MAKELVS
    0 swap 2+ w!                        \ #args #locals
  endif
;

: LOCALS|	\ "name...name |" --
\ *G Create named local variables <name1> to <namen>. At run time
\ ** the stack effect is ( xn..x1 -- ), such that <name1> is
\ ** initialised with x1 and <namen> is initialised with xn.
\ ** Note that this means that the order of declaration is the reverse
\ ** of the order used in stack comments!
\ ** When referenced, a local variable returns its value. To write
\ ** to a local, precede its name with TO.
\ *P In the example below, a and b are named inputs.
\ *E : foo      \ a b --
\ **   locals| b a |
\ **   a b +  cr .
\ **   a b *  cr .
\ ** ;
  lvcount @ abort" previous LOCALS| not finished"
  >in @
  0
  begin
    bl word                                                                     \ MSD003
    dup     c@ 1 =                                                              \ MSD003
    swap 1+ c@ [char] | =                                                       \ MSD003
    and not                                                                     \ MSD003
   while
    cell +
  repeat
  lvcountmax !
  >in !
  begin
    bl parse
    over c@ ascii | -
    over 1 - or
  while
    (local)
  repeat
  2drop
  0 0 (local)
;
IMMEDIATE


\ ======
\ *> ###
\ ======

