\ ROMFORTH.MAN - ROM PowerForth generic documentation

\ ===========
\ *! romforth
\ *T ROM PowerForth utilities
\ ===========

\ *S Introduction
\ *P Supplied as source in the ROMFORTH directory are utilities to:
\ *B compile source code on your target board from the cross-compiler IDE
\ *B upload a binary image from your target to your PC
\ *B download a binary image to your target from your PC.

\ *P Note that the target source code supplied with cross compiler
\ ** versions 6.02 onwards is incompatible with code supplied for
\ ** previous versions of the cross compiler.

\ *P These utilities can be used to generate an EPROM that has all
\ ** the tools required to develop an application, or can be used during
\ ** development to transfer modules to and from your PC. All the code
\ ** is designed to be used with the MPE development environment, AIDE.
\ ** The code will also work with other compatible terminal emulators.

\ *P Users who wish to distribute ROMs containing the ROM PowerForth
\ ** utilities should contact MPE for details of the OEM licence,
\ ** which includes documentation on disc of the Forth kernel
\ ** and the ROM PowerForth utilities.

\ *S Compiling text files
\ *P Source text files can be compiled from the host PC onto the
\ ** target system. This saves time in not having to cross-compile
\ ** the entire source if a small modification is made. The utilities
\ ** permit text file to be split into pages for better layout when
\ ** printed. An ASCII Form Feed character (decimal 12) separates one
\ ** page from another.

\ *N The required files
\ *P To compile text files from your target board, cross-compile the
\ ** files IODEF.FTH and TEXTFILE.FTH.

\ *N Compiling a specified text file
\ *P To compile all or part of a specified text file onto your target,
\ ** use GET or INCLUDE in the form:
\ *C   INCLUDE <filename>
\ *P This compiles the file <filename> into the target's dictionary.
\ ** AIDE's internal file server must be enabled (in the console window
\ ** configuration), and will be triggered automatically.

\ *S Downloading a binary image
\ *P A binary image can be downloaded from the target to your host PC.
\ ** Two utilities are provided:
\ *B Intel hex download
\ *B XMODEM download
\ *P For both utilities the cross-compiler IDE or a suitable
\ ** communications package will be required.

\ *N XMODEM binary image download
\ *P Binary images can be downloaded to your PC using the XMODEM protocol.
\ *H Required files
\ *P To use this utility you must cross-compile the file COMMON\XMODEMTXRX.FTH.
\ *H Using the XMODEM binary download utility
\ *P To download a binary image from the target system to your PC,
\ ** use BIN-DOWN in the form:
\ *C   addr #bytes BIN-DOWN
\ *P where addr is the start address and #bytes is the number of bytes
\ ** to down-load starting from addr. For example,
\ *C   1200 400 BIN-DOWN
\ *P sends the area of memory from 1200 to 1599 to your host PC.
\ ** AIDE's internal file server must be enabled (in the console window
\ ** configuration), and will be triggered.

\ *N Intel hex download
\ *P Binary images can be downloaded to your PC using the Intel hex format.
\ ** Because this format only uses the standard printable character and CR/LF,
\ ** the data can be captured by very simple tools.
\ *H Required files
\ *P To use this utility you must cross-compile the file INTELHEX.FTH.
\ *H Using the Hex download utility
\ *P To download a binary image from the target system to your PC,
\ ** use HEX-DOWN in the form:
\ *C   addr #bytes HEX-DOWN
\ *P where addr is the start address and #bytes is the number of bytes
\ ** to down-load starting from addr. For example,
\ *C   1200 400 HEX-DOWN
\ *P sends the area of memory from 1200 to 1599 to your host PC.
\ ** In AIDE, turn on console logging to receive the file. In other
\ ** packages this may be referred to as file capture.

\ *S ROM PowerForth
\ *P ROM PowerForth can be used to generate a stand-alone Forth system.
\ ** With these utilities, you can generate an EPROM that contains an
\ ** interactive Forth with the ability to develop an application.
\ ** Note: A licence is required to distribute open Forth systems.
\ ** Contact MPE for more details.

\ *N Hardware requirements
\ *P To develop an application using ROM PowerForth, your board requires an area which:
\ *B is always EPROM
\ *B is always RAM
\ *B is RAM for development and EPROM for application
\ *N EPROM/Flash area
\ *P The area that is always EPROM contains the development kernel.
\ *N RAM area
\ *P The area that is always RAM is used for variables and all changeable data.
\ *N RAM/EPROM area
\ *P This area is used to develop your application. Therefore, it must
\ ** be RAM while developing. Once your application is developed, the
\ ** application's image must be saved into battery-backed RAM or
\ ** EPROM or Flash. Therefore, this area must have the ability to be alterable
\ ** but also non-volatile.

\ *N Types of board
\ *P The type of board that can be used to develop using ROM PowerForth is restricted to:
\ *B three site boards
\ *B two site boards with battery backed RAM
\ *B two site boards with socket converter.
\ *H Three site boards
\ *P The three areas are provided by three memory sockets:
\ *B EPROM holding development kernel
\ *B RAM which holds the variables and changeable data
\ *B EPROM or RAM which is selectable by a link on the board
\ *H Two site boards with battery backed RAM
\ *P The three areas are provided by two sockets:
\ *B EPROM holding the development kernel
\ *B battery-backed RAM which is split into two areas.
\ *H Two site boards with socket converter
\ *P On many boards, there is unused space in the EPROM as ROM PowerForth
\ ** occupies less than 32k bytes of memory. Therefore, a header board
\ ** can be made which converts one socket into two. For example,
\ ** if the socket normally takes a 27512 EPROM, a board can be
\ ** made which has a 32k EPROM with the ROM PowerForth 
\ ** development kernel and 32k bytes of RAM. To access
\ ** the RAM, the write line is attached to a suitable point
\ ** on the main board with a fly lead. After the application
\ ** has been developed, the two images are combined back into
\ ** a single EPROM.

\ *N Making your application turnkey
\ *P Once your application has been developed, it needs to be made
\ ** turnkey so that it is always available. The application can be
\ ** made semi-permanent by compiling into battery-backed RAM in
\ ** the RAM/EPROM area. Alternatively, it can be copied into an
\ ** EPROM if the board allows.

\ *H Configuring a turnkey application
\ *P The word SETUP takes the address of the word passed to it 
\ ** and marks this in the RAM/EPROM header as the address of
\ ** the word to be run at power-up. If a value of zero is passed
\ ** to SETUP, the interactive Forth kernel will be run at power up.
\ *P For example, the word JOB is to be run at power-up. Therefore
\ ** you type, 
\ *C   ' JOB SETUP

\ *N Discarding the application RAM area
\ *P The application can be discarded by typing:
\ *C   0 ROM !

\ *N Changing the application RAM start address
\ *P The constant ROM returns the start address of the application
\ ** RAM area. If the address of this area is to be changed, the
\ ** EPROM must be modified. To do this, the cell value in
\ ** ROM must be changed.

\ *N AIDE file server protocols
\ *P AIDE's file server must be enabled for automatic file handling.
\ ** Details of the protocols used should be obtained from this
\ ** manual and the source code in the COMMON\ROMFORTH directory.


\ ======
\ *> ###
\ ======
