\ ROM Forth serial i/o definitions         

((
Copyright (c) 1991..2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

20000530 SFP002 Converted to v6.1 targets  
19990510 SFP001 Updated for AIDE/PowerTerm
First Release: 29/04/91

Changes:        See RELEASE.TXT for full details
))


\ *> romforth
\ *S IODEF.FTH
\ *P Before compiling this file, synonyms may need to be defined
\ ** for SER-EMIT SER-KEY? and SER-KEY. Add these in the control file
\ ** before compiling IODEF.FTH. 

\ *P IODEF.FTH provides equates and protocol primitives for AIDE.


\ ***********************
\ File transfer constants
\ ***********************

Hex

\ 
\ characters important in these protocols
\ 
$001 equ START
$006 equ ACK
$015 equ NAK

\ 
\ constants sent as triggers to host end of comms link
\ 
$001 equ ^a             \ use
$003 equ ^c             \ clear screen
$005 equ ^e             \ error
$006 equ ^f             \ $from
$012 equ ^R             \ block read command code	not in AIDE yet
$015 equ ^u             \ $using for screen files	not in AIDE yet
$017 equ ^W             \ block write			not in AIDE yet
$018 equ ^x		\ start UP-LOAD from host
$019 equ ^Y		\ Xmodem to host

\ to be added
			\ host logging on
			\ host logging off


\ ***************
\ *N AIDE support
\ ***************

variable disk-error     \ -- addr ; set non-zero on error
\ *G This variable is set true when a transfer error occurs.

internal

decimal

: wait-ack      \ -- ; wait for ACK character
\ *G This word waits for the host to send an ACK at the end of
\ ** part of a transfer.
\ ** INTERNAL.
  BEGIN  ser-key ACK =  UNTIL
;

: wait-ack/nack	\ -- t/f ; true for NACK
\ *G This waits for either a NACK or an ACK from the host
\ ** and leaves true or false on stack.
\ ** INTERNAL.
  0
  begin
    drop                                \ drop old flag
    ser-key dup NAK = tuck              \ -- t/f char t/f
    swap ACK =  or                      \ -- t/f t/f'
  until  ;

: send-block#   \ n -- ; send block number to server
\ *G Sends a single length number as two bytes 00-FF,
\ ** low byte first.
\ ** INTERNAL.
  0 #256 um/mod swap            \ split into high & low bytes
  ser-emit ser-emit             \ send block no., low byte first
  wait-ack                      \ wait for ack
;

: synch-to-host \ -- ; sync host to us
\ *G Waits for a START (0x01) character, flushes the input,
\ ** and sends an ACK.
\ ** INTERNAL.
  begin  ser-key START = until	\ receive bytes until start code
  BEGIN
    ser-key?                    \ while there are keys still
  WHILE
    ser-key drop		\ discard them
  REPEAT
  ACK ser-emit                  \ acknowledge as end of synch.
;

: send$         \ $addr ^char --
\ This word sends a trigger character and a counted string
\ to the host, then sends an ACK, and waits for an ACK or NACK.
\ ** INTERNAL.
  ser-emit                      \ send code
  synch-to-host                 \ synchronise
  dup 1+ swap                   \ addr+1 as base for reading
  c@ dup send-block#            \ send length of string
  0                             \ length - 0
  ?do dup i + c@                \ get char at i+$addr
     ser-emit                   \ send char
  loop
  drop                          \ drop spare address
  ACK ser-emit                  \ send ack to complete transfer
  Wait-Ack/Nack Disk-error !    \ get status response
; 

external


\ ****************
\ *S Miscellaneous
\ ****************

: cls           \ --
\ *G Clear PowerTerm screen
  ^c ser-emit
;


\ ======
\ *> ###
\ ======

decimal

