\ ROMFORTH application linkage tools       

((
Copyright (c) 1991..2002
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

20000530 SFP001 Converted to v6.1 targets

Changes:	See RELEASE.TXT for full details
))


\ ===========
\ *> romforth
\ *S Application Extensions
\ ===========
\ *P ROM PowerForth can be used to generate a stand-alone Forth system.
\ ** With these utilities, you can generate an EPROM that contains an
\ ** interactive Forth with the ability to develop an application.
\ ** Note: A licence is required to distribute open Forth systems.
\ ** Contact MPE for more details.

\ *P The code in COMMON\ROMFORTH\LINK.FTH provides the facilities
\ ** to link the extension area into the ROM PowerForth kernel.

hex 

appl-rom equ ROM	\ start of ROM/RAM area
\ *G Define the address of a ROM/RAM area in which applications
\ ** are to be developed. This area is RAM during development.
\ ** When development is complete, the word SETUP is executed
\ ** to initialise the memory area.
\ ** The compiled image is then downloaded to the PC using the
\ ** tools in the ROMFORTH directory. The image is then programmed
\ ** into EPROM, which can then replace the RAM.
\ ** If the word RELINK is added to COLD, the application EPROM
\ ** will be automatically linked in when the target powers up.

harvard? 0= [if]
variable rp    EM rp !			\ top of RAM area
\ *G If your processor has separate CODE and DATA address spaces,
\ ** it has a HARVARD architecture. This word is only
\ ** compiled for Harvard targets such as 8051s.
\ ** The working RAM area has the interactive RAM dictionary at
\ ** the bottom and the application buffer space at the top.
\ ** Define the initial value of RP to be the highest available
\ ** free location in the RAM area. Application variable and
\ ** buffer space will then be allocated downwards from this
\ ** address. When creating an application to be ROMmed,
\ ** download the code for BUFFER and the redefined VARIABLE
\ ** first of all.
[then]


hex

: setup         \ xt|0 -- ; sets up for ROM generation
\ *G Use XT as the xt of the word to be run when the extension
\ ** is found.
  forth definitions  $055aa rom !c      \ set marker
  [ rom cell+ ] literal !c		\ set CFA to execute
  latest [ rom 2 cells + ] literal !c	\ set last NFA
  rp @  [ rom 3 cells + ] literal !c	\ set top of RAM
  dp @   [ rom 4 cells + ] literal !c	\ set end of dictionary
  forth-wordlist >#threads              \ start of FORTH threads
  [ rom 5 cells + ] literal		\ storage area
[ harvard? ] [if]
  over @ cells cell+ cmove->c		\ copy threads
[else]
  over @ cells cell+ cmove		\ copy threads
[then]
;

internal

: use-setup-data        \ --
\ *G Apply the data in the setup area.
  [ rom 5 cells + ] literal		\ storage area
  forth-wordlist >#threads              \ start of FORTH threads
  over @c cells cell+ cmovec		\ copy threads
  [ rom 3 cells + ] literal @c rp !	\ top of RAM
  [ rom 2 cells + ] literal @c last !	\ last word defined
  [ rom 1 cells + ] literal @c ?dup	\ if CFA set
  if  execute  endif  ;                 \  execute it

: link-rom      \ --
\ *G Link in the application ROM area.
  use-setup-data                        \ copy data from header
  Cr ." Application ROM installed"  ;

hex

: link-ram      \ -- ; link in application RAM area
\ *G Link in the application RAM area.
  0AA55 rom !c  rom @c 0AA55 =          \ see if RAM present
  if  [ rom #32 cells + ] literal dup	\ yes, reserve first 32 cells
      dp !  fence !                     \ and reset pointers
      Cr ." Application RAM installed"  \ tell user
  endif  ;

: link-nv-ram   \ -- ; link in for NV-RAMs
\ *G Link in the application NV-RAM area.
  [ rom 4 cells + ] literal @c
  dup dp ! fence !                      \ end of dictionary
  use-setup-data                        \ copy data from header
  Cr ." Application NV-RAM installed" ;

external

: relink        \ -- ; re-link application if there
\ *G Link in the application ROM/RAM/NV-RAM area.
[ appl-flash ] [if]
  appl-flash @c $55AA = if		\ copy to RAM area
    appl-flash rom			\ src, dest
    [ appl-rom-end appl-rom - ] literal \ size
    cmovec
  endif
[then]
  rom @c 055aa =                        \ see if ROM present
  if  0AA55 rom !c                      \ ROM or NV-RAM ?
      rom @c 0AA55 =                    \ true if NV-RAM
      055AA rom !c                      \ restore header
      if  link-nv-ram                   \ link in NV-RAM
      else  link-rom                    \ link in ROM area
      endif                             \
  else  link-ram                        \ no, attempt RAM area
  endif  ;


\ ********
\ all done
\ ********

\ ======
\ *> ###
\ ======

decimal

