((
	Testing for Forth words

	Adapted from 'Testing Toolkit' by P. Koopman Jr.
	- Forth Dimensions XII,3

	Author:	J Lee
	Date started:	5/10/90

	History:
	V1.000	13/12/91	J Lee
		Original release.
 		This wordset provides a means of testing programs in a
		formalised fashion. A word is executed surrounded by a
		formal stack definition.

		DS( specified_inputs --) <WORD> (-- required_outputs )DS

		This version works with MF3.xx.
	V1.001	21/2/92		J Lee
		Modified to work in COLON definitions during Cross
		Compilation.

	V1.100	13/11/95	J Lee
		Modified for EPIC Forth token compiler.

        V1.200  17/04/98        M Davis
                Ensure that line numbers come out in *decimal*.
))


\ Cross compiler extensions -- J Lee ..21/02/92

host-compilation
c-c only forth also c-c also c-c definitions

cc/c

: --)  cc/c
  textfileline general-literal
  target-sym --)-later compile-ref
\  target-sym try-later compile-ref
;

c-c only forth also c-c definitions
target-compilation


\ Try Ds( --) (-- Reset-Stack Check-Element )Ds .. J Lee -- 05/10/90
\

\ variables
variable Stack1		\ stack depth marker
variable Stack2		\   "	"	  "
variable Stack3		\   "	"	  "

\ variable Test-Word	\ CFA of word under test
variable Test-Line#		\ Line# of word under test

variable Left-Loop	\ flag for early loop exit


\ hex

((
: Try  ( -- <wordname> )
  ' dup test-word !
  execute
;
))

((
: Try-Later  ( -- ; looks ahead in code stream )
  r@ @ test-word !
;
))

: Ds(  ( -- sentinel )
  $0bad1				\ sentinel (check for stack corruption)
  depth stack1 !  		\ intial depth
;

((
: --)  ( .. -- .. )
  depth stack2 !  		\ pre execution depth
  try  				\ execute word
;
))

: --)-later  ( .. line# -- .. )
  test-line# !
  depth stack2 !  		\ pre execution depth
;

: (--  ( .. -- .. )
  depth stack3 !  		\ post execution depth
;

: Reset-Stack  ( -- ; reset data stack )
\  s0 @ sp!								\ this would be nice
  begin  depth  while  drop  repeat		\ long winded but effective
;

((
variable roll-initial
variable roll-current
variable roll-return

\ yuk!!!!!!!!
: roll  ( xu xu-1 .. x0 u -- xu-1 .. x0 xu ; grab element from stack )
  dup roll-initial !  roll-current !
  begin
    roll-current @
  while
    >r  -1 roll-current +!
  repeat
  roll-return !
  roll-initial @ roll-current !
  begin
    roll-current @
  while
    r>  -1 roll-current +!
  repeat
  roll-return @
;
))
  
: Check-Element	 ( element# -- t/f )
  dup					\ only do for non-zero elements
  If					\
    roll <>				\
    If					\
      ." FAILED, Incorrect Result"	\
      reset-stack			\
      -1 Left-Loop !		\ flag that loop is left prematurely
      -1				\ leave TRUE
    Else				\
      0					\ leave FALSE
    Endif				\
  Endif				        \
;

: Check-Stack  ( sentinel -- ; check stack for corruption )
  $0bad1 <>				\ check stack corruption
  If
    ." FAILED, stack corrupted"
    reset-stack				\ reset data stack
  Else
    ." PASSED."
  Endif
;

: )Ds
  cr ." Testing Line# : " base @ decimal test-line# @ . base !
  depth stack3 @ - dup			\ post execution elements
  stack3 @ stack1 @ -			\ pre execution elements
  <>					\ give up if not the same
  If					\
    ." FAILED, Number of elements not the same"
    reset-stack
  Else
    0 left-loop !		 \ turn loop exit flag off
    0 swap				 \
    ?Do					 \ check result against expected
      i check-element
      If  leave  Endif
      -1
    +Loop
    left-loop @ 0=
    If  check-stack  Endif
  Endif
;

\ decimal


\ Test it out .. J Lee -- 05/10/90

\
\ A word is executed surrounded by a formal stack definition.
\
\ DS( specified_inputs --) <WORD> (-- required_outputs )DS
\

\ decimal

