
\ 19981027 MSD001 Altered to run under MPE's XC6 systems.  (Not flagged.)
\          MSD001 Also, fixed a stack bug in INCREMENT.

\ Program to simulate the orbits of the earth, moon and sun about each other.
\ Original author: Brad Eckert

decimal
also f-pack

f# 1.520e11     fconstant sun2earth     \ meters
f# 3.844e8      fconstant earth2moon
f# 5.979e24     fconstant earthmass     \ kg
f# 1.991e30     fconstant sunmass
f# 7.35e22      fconstant moonmass

f# 3.0e4        fconstant earthv0       \ starting velocity
f# 1.0e3        fconstant moonv0

f# 1.0e3        fconstant stepsize      \ time between snapshots in seconds
f# 0.0e0        fconstant fzero
f# 6.67e-11     fconstant grav          \ gravitational constant

f# 1.0e0        fconstant fone          \ floating point value "one"

f# 1.0e8        fconstant f>dscale

false value singlestep?                 \ used to display slowly

: 'position     ( 'body -- 'position )  ;
: 'velocity     ( 'body -- 'velocity )  2 floats + ;
: 'mass         ( 'body -- 'mass )      4 floats + ;
: 'oldpos       ( 'body -- 'oldpos )    5 floats + ;
: 'radius       ( 'body -- 'radius )    5 floats 2 cells + + ;
: 'color        ( 'body -- 'color )     5 floats 3 cells + + ;


2variable eons          \ how old is this universe (in seconds) ?
20000 value displaystep \ seconds between simulation print-outs

: displayable?  ( -- f )
\ return T every displaystep-th second (20000 => 5 hours)
        eons 2@  stepsize f>d d+  2dup eons 2!
        drop displaystep mod 0= ;

: .eons         ( -- )
\ display the time in years, days, hours, minutes and seconds
        eons 2@ ( d)    \ seconds since start of universe
        [ 60 60 *  24 *  365 * ] literal um/mod 4 .r ."  years, "
        [ 60 60 *  24 * ]        literal   /mod 3 .r ."  days, "
        [ 60 60 *  ]             literal   /mod 2 .r ."  hours, "
        60 /mod 2 .r ."  minutes and " 2 .r ."  seconds." ;


\ ================= VECTOR ARITHMETIC ======================

: v@            ( 'v -- fx fy )
\ fetch the x,y components of vector
        dup >r f@
        r> 1 floats + f@ ;

: v!            ( fx fy 'v -- )
\ store the x,y components of vector
        dup >r 1 floats + f!
        r> f! ;

: vnegate       ( va -- -va )
        fswap fnegate fswap fnegate ;

: v+            ( va vb -- vc )
\ add two vectors
        frot f+                 ( ax bx cy )
        fswap frot f+           ( cy cx )
        fswap ;

: v-            ( va vb -- va-vb )
\ subtract two vectors
        vnegate v+ ;

: vdup          ( v -- v v )
        fover fover ;

: fsquare       ( f -- f^2 )
        fdup f* ;

: distance      ( x y -- dist )
        fsquare fswap fsquare f+ fsqr ;

: v*            ( va scalar -- vb )
        fdup frot f*            ( x s sy )
        frot frot f*            ( sy sx )
        fswap ;

: v/    ( va scalar -- vb )
        fone fswap f/ v* ;


\ ================= MOTION CALCULATIONS ======================

udata
fvariable massmass
cdata

: getmasses     ( 'body1 'body2 -- )
\ get the mass of each body and multiply them together
        >r 'mass f@
        r> 'mass f@
        f* massmass f! ;

: attraction    ( 'body1 'body2 -- fx fy )
\ calculate attraction between body1 and body2
\ using mass and inverse-square law: f = G*m1*m2/r^2
        2dup getmasses
        swap >r v@  r> v@   v-  \ vec1-vec2
        vdup distance           ( x y r )
        massmass f@  grav f*    ( x y r mmg )
        fover f/ fover f/
        fswap f/                ( x y force/r )
        v* ;                    \ square it ( z/r * 1/r/r )

: increment     ( vdelta 'va -- )
\ add a differential amount to a vector
        >r                                                                      \ MSD001
        stepsize v*             \ delta = delta * stepsize
        r@ v@ v+  r> v! ;       \ v = v + delta                                 \ MSD001

: motion        ( force 'body -- )
\ update the motion parameters of a body given an acting force
        >r
        r@ 'mass f@ v/          \ a = f/m
        r@ 'velocity increment  \ update velocity
        r@ 'velocity v@
        r> 'position increment  \ update position
        ;

: startup       ( mass position velocity 'body -- )
\ initial parameters, scalar quantities.
\ position is on the X axis, velocity is in the Y direction.
        >r
        fzero fswap r@ 'velocity v!
        fzero       r@ 'position v!
        r> 'mass f! ;

5 8 * 4 cells +   buffer:   earth
5 8 * 4 cells +   buffer:   moon
5 8 * 4 cells +   buffer:   sun

: TabOut        ( -- )
\ Tab out to next column stop.
        9 emit ;

: .F            ( f -- )
        f>dscale f* f>d
        swap over dabs
        <# # # # # # # # # [char] . hold # # # rot sign #>
        over c@ [char] - <> if space then
        type ;

: .XY           ( x y scale -- )
\ print normalized XY coordinates
        v/ fswap tabout .f tabout .f ;

: .moonXY       ( -- )
\ position of moon relative to earth
        moon v@  earth v@  v-
        earth2moon .xy ;

: .earthXY      ( -- )
\ position of earth relative to sun
        earth v@  sun v@  v-
        sun2earth .xy ;

variable DispCount
       0 DispCount !

: display       ( -- )
        1 DispCount +!
        DispCount @ s>d <# # # # # # # # #>
        cr type tabout
        .earthXY .moonXY ;

: orbit
\ run the simulation
        0, eons 2!                      \ Let there be Light...
        earthmass                       \ set earth in motion
        sun2earth
        earthv0
        earth startup
        moonmass                        \ set moon in motion
        sun2earth earth2moon f+
        moonv0 earthv0 f+
        moon startup
        sunmass                         \ sun is stationary
        fzero fzero sun startup
        begin   displayable?
                if      DISPLAY
                        key? singlestep? or
                        if      key 27 <>
                        else    true
                        then
                else    true
                then
        while   earth moon attraction
                earth sun attraction v+
                earth motion            \ move the earth
                moon earth attraction
                moon sun attraction v+
                moon motion             \ move the moon
        repeat ;

: goslow        ( -- )
        true to singlestep? ;

