\ TESTTASK - A multi-tasker test program.

\ 19980415 MSD003 Invoke the new APPSEE word in APPSTART after the call to
\          MSD003 APPLED so that the numbers are displayed automatically.
\ 19980415 MSD002 Make APPSEE invoke SEE-EVENT-WORD instead of doing its own
\          MSD002 display of the incremented numbers.  Thus, it will now start
\          MSD002 a task rather than running its own task zero loop.
\ 19980414 MSD001 Print out the values of the sending task number and the value
\          MSD001 returned by SELF if the first test in INCREMENTER-TASK-WORD
\          MSD001 fails.  Likewise, print out the two values of INCVALUE (one
\          MSD001 in the variable and the other in the message) if the second
\          MSD001 test fails.

((
        Copyright (c) 1988-1998
        MicroProcessor Engineering
        133 Hill Lane
        Southampton SO15 5AF
        England

        tel: +44 1703 631441
        fax: +44 1703 339691
        net: mpe@mpeltd.demon.co.uk
             tech-support@mpeltd.demon.co.uk
        web: www.mpeltd.demon.co.uk
))

0
1+ dup  constant        controller-task
1+ dup  constant        incrementer-task
1+ dup  constant        start-task
1+ dup  constant        middle-task
1+ dup  constant        end-task
1+ dup  constant        see-task
1+ dup  constant        led-task
drop

100     constant        controller-wait
100     constant        start-wait
45      constant        see-wait
90      constant        numbers-wait
100     constant        apprun-wait

variable led-wait-value #50 led-wait-value !
: led-wait led-wait-value ! ;

hex

\ --------------------------------------------------------------------------

: say                   \ msg$ -- ; Displays a message in a "standard" manner.
  cr
  count type
;

\ --------------------------------------------------------------------------

variable NextValue 0 NextValue !

: start-task-word       \ -- ; Sends a stream of numbered messages to MIDDLE.
  "" START active" say
  begin
    NextValue @ middle-task send-message
    1 NextValue +!
    start-wait wait
  again
;

\ --------------------------------------------------------------------------

: middle-task-word      \ -- ; Takes from START and forwards to MIDDLE.
  "" MIDDLE active" say
  begin
    get-message
    drop ( task number )
    ( message ) end-task send-message
  again
;

\ --------------------------------------------------------------------------

: end-task-word         \ -- ; Validates messages from MIDDLE.
  "" END active" say
  begin
    get-message
    drop ( task number )
    drop ( message )
  again
;

\ --------------------------------------------------------------------------

variable IncValue 0 IncValue !

: incrementer-task-word \ -- ; Validates messages sent to itself.
  "" INCREMENTER active" say
  IncValue @ self send-message
  begin
    get-message
    dup self <>                                                                 \ MSD001
    if
      cr ." INCREMENTER: task ( " . ." ) <> self ( " self . ." )"               \ MSD001
    else                                                                        \ MSD001
      drop                                                                      \ MSD001
    endif
    dup IncValue @ <>                                                           \ MSD001
    if
      cr ." INCREMENTER: value ( " . ." ) <> value@ ( " IncValue @ . ." )"      \ MSD001
    else                                                                        \ MSD001
      drop                                                                      \ MSD001
    endif
    pause
    1 IncValue +!
    IncValue @ self send-message
  again
;

\ --------------------------------------------------------------------------

variable LEDing 0 LEDing !

: appled                \ -- ; Toggles the LED task.
  LEDing @ not
  if
    "" Flashing LED" say
    led-task restart
    LEDing on
  else
    "" Blanking LED" say
    led-task halt
    LEDing off
  endif
;

\ --------------------------------------------------------------------------

\ These next few words comprise a task and event handler for
\ veiwing the numbers used by the multi-tasking examples.  Note
\ that this task can be toggled on and off by invoking whatever
\ will fire off the event handler, SEE-EVENT-WORD.

: see-task-word         \ -- ; Continuously displays task variables.
  "" SEE active" say
  stop
  begin
    cr nextvalue ? incvalue ?
    see-wait wait
  again
;


variable Seeing 0 Seeing !

: see-event-word        \ -- ; Event handler that toggles SEE-TASK-WORD.
  clr-event-run
  Seeing @ not
  if
    see-task restart
    Seeing on
  else
    see-task halt
    Seeing off
  endif
;

: see-isr               \ -- ; Attaches event-handler to NMI line.
  see-task set-event
;

\ --------------------------------------------------------------------------

: led-task-word         \ -- ; Continuously toggles the EVB LED.
  led-init
  stop
  "" LED active" say
  begin
    led-on  led-wait-value @ wait
    led-off led-wait-value @ wait
  again
;

\ --------------------------------------------------------------------------

variable TestTaskRunning 0 TestTaskRunning !

: controller-task-word  \ -- ; This starts and stops all the other tasks.
  "" CONTROLLER active" say
  TestTaskRunning on
  "" Activating END" say           end-task activate
  "" Activating MIDDLE" say        middle-task activate
  "" Activating START" say         start-task activate
  "" Activating INCREMENTER" say   incrementer-task activate
  "" Activating SEE" say           see-task activate
  "" Activating LED" say           led-task activate
  begin
    controller-wait wait
    TestTaskRunning @ not
  until
  "" Halting LED" say              led-task halt
  "" Halting SEE" say              see-task halt
  "" Halting INCREMENTER" say      incrementer-task halt
  "" Halting START" say            start-task halt
  "" Halting MIDDLE" say           middle-task halt
  "" Halting END" say              end-task halt
  "" CONTROLLER exits" say
  cr cr
  stop
;

\ --------------------------------------------------------------------------

: appstat               \ -- ; This just displays the task numbers
  cr
  cr controller-task  . ." CONTROLLER  "
  cr incrementer-task . ." INCREMENTER "
  cr start-task       . ." START       "
  cr middle-task      . ." MIDDLE      "
  cr end-task         . ." END         "
  cr see-task         . ." SEE         "
  cr led-task         . ." LED         "
  cr
  cr
;

\ --------------------------------------------------------------------------

: appassign             \ -- ;
  assign start-task-word        start-task       to-task
  assign middle-task-word       middle-task      to-task
  assign end-task-word          end-task         to-task
  assign incrementer-task-word  incrementer-task to-task
  assign see-task-word          see-task         to-task
  assign see-event-word         see-task         to-event
  assign controller-task-word   controller-task  to-task
  assign led-task-word          led-task         to-task
;

\ --------------------------------------------------------------------------

: appsee                \ -- ; View the multi-tasking counters.
  see-event-word                                                                \ MSD002
((                                                                              \ MSD002
  cr
  begin
    cr nextvalue ? incvalue ?
    see-wait wait
    key?
  until
  key drop
  cr
))                                                                              \ MSD002
;

\ --------------------------------------------------------------------------

: appstart              \ -- ; Start off multi-tasking examples.
  appassign
  multi
  cr
  "" Activating CONTROLLER" say
  controller-task activate
  apprun-wait wait
  appstat
  appled
  appsee                                                                        \ MSD003
;

\ --------------------------------------------------------------------------

: appstop               \ -- ; Stops the multi-tasking examples.
  TestTaskRunning off
;

\ --------------------------------------------------------------------------

variable number 0 number !

: appnum                \ -- ; Displays loads of numbers.
  cr
  begin
    number @ s>d <# # # # # # # # # #> type ."   "
    1 number +!
    numbers-wait wait
    key?
  until
  key drop
  cr
;

\ --------------------------------------------------------------------------

: testtask appstart ;   \ -- ;
: run testtask ;        \ -- ;

: testtaskturnkey       \ -- ;
  (init)
  init-ser
  init-multi
  testtask
  abort
;

decimal

