\ TESTHEAP - A heap manager test program.

((
        Copyright (c) 1988-1998
        MicroProcessor Engineering
        133 Hill Lane
        Southampton SO15 5AF
        England

        tel: +44 1703 631441
        fax: +44 1703 339691
        net: mpe@mpeltd.demon.co.uk
             tech-support@mpeltd.demon.co.uk
        web: www.mpeltd.demon.co.uk
))

\ 19990315 SJC001 Fixed sign bug in random number generation. Previously the
\                 definition RAND1..N was actually returning a random number
\                 <N including negative numbers.

[undefined] PFW-Version    [if] false constant PFW-Version    [then]
[undefined] WantInitHeap   [if] true  constant WantInitHeap   [then]
[undefined] WantKeyAtStart [if] false constant WantKeyAtStart [then]
[undefined] WantDumpAtEnd  [if] false constant WantDumpAtEnd  [then]
[undefined] DumpStart      [if] #0    constant DumpStart      [then]
[undefined] DumpEnd        [if] #0    constant DumpEnd        [then]
[undefined] Turnkey?       [if] false constant Turnkey?       [then]
[undefined] WantPause      [if] true  constant WantPause      [then]
[undefined] LoadFactor%    [if] #120  constant LoadFactor%    [then]
[undefined] BlockSlots     [if] #16   constant BlockSlots     [then]

[undefined] MaxBlockSize
[if]
  PFW-Version
  [if]
    $200000 constant MaxBlockSize
  [else]
    SIZEOFHEAP LoadFactor% * #100 /     \ Total heap size to allocate
    2 *                                 \ *Average* block size MAXBLOCKSIZE/2
    BlockSlots /                        \ Take into account the number of blocks
          constant MaxBlockSize
  [then]
[then]

\ ==========================================================================

8   constant ERROR_NOT_ENOUGH_MEMORY

$08 constant BackSpaceKey
$1B constant EscapeKey

\ ==========================================================================

[undefined] null                [if] 0 constant null            [then]
[undefined] heapok?             [if] : heapok? true ;           [then]
[undefined] idata               [if] : idata ;                  [then]
[undefined] udata               [if] : udata ;                  [then]
[undefined] cdata               [if] : cdata ;                  [then]
[undefined] host-compilation    [if] : host-compilation ;       [then]
[undefined] target-compilation  [if] : target-compilation ;     [then]
[undefined] .dword              [if] : .dword .lword ;          [then]

[undefined] WINGETLASTERROR
[if]
  : WINGETLASTERROR ERROR_NOT_ENOUGH_MEMORY ;
[then]

\ ==========================================================================

host-compilation

: Digits        { num | digs -- max_number_of_digits }
  0 -> digs
  begin
    num 0>
  while
    num base @ / -> num
    digs 1+ -> digs
  repeat
  digs dup 0= if drop 1 endif
;

target-compilation

MaxBlockSize Digits constant MaxBlockSizeDigits
BlockSlots   Digits constant BlockSlotsDigits

\ ==========================================================================

PFW-Version
[if]

  1 cells field-type BlockSize
  1 cells field-type RelAddress

  struct  BlockRec
          BlockSize  .Size
          RelAddress .Addr
  end-struct

[else]

  : OFFSET  ( Comp:  n "<spaces>name"  --  ; Run:  a-addr1  --  a-addr2  )
    create  ,
    does>  @ +
  ;

  0
  dup offset .SIZE  cell+
  dup offset .ADDR  cell+
  constant BLOCKREC

[then]

\ --------------------------------------------------------------------------

: BlockArray    \ n -- ; [child] n -- addr ; 1D array
  create
    BlockRec *          \ calculate size
\ TSE     here over erase     \ clear memory TSE \
    allot               \ allocate n "records"
  does>
    swap BlockRec * +
;

\ --------------------------------------------------------------------------

\ TSE idata TSE \
udata
BlockSlots BlockArray Blocks
cdata

\ ==========================================================================

variable RandSeed
       0 RandSeed !

\ #2053   constant RandA
\ #13849  constant RandC
\ #65536  constant RandM

#24298  constant RandA
#99991  constant RandC
#199017 constant RandM

: Rand          \ -- randomnumber ; Returns next in a sequence of "random" nums
  RandSeed @
  RandA * RandC + RandM mod
  dup RandSeed !
  abs                                                           \ sjc001
;

: Rand1..n      \ n -- randomnumber1..n ; Returns a "random" number 1..n
  dup Rand swap mod 1+
  over RandM >
  if                    \ Scale up because RandM is too small for this "N".
    swap RandM /
    ?dup if * endif
  else
    nip
  endif
;

: Randomise     \ -- ;
  [ PFW-Version ]
  [if]
    here WinGetSystemTime
    here #14 + w@         \ -- milliseconds
  [else]
    0
  [then]
  RandSeed !
;

: RandTest      \ -- ;
  begin
    key? 0=
  while
    9 Rand1..n ascii A 1- + emit
  repeat
  key drop
;

\ ==========================================================================

variable WantVerbose true WantVerbose !

: $V.           \ addr$ -- ;    Print string if verbose mode
  WantVerbose @ if $. else drop endif
;

: Verbose@      \ -- Verbosity ;
  WantVerbose @
;

: ToggleVerbosity \ -- ;
  Verbose@ not WantVerbose !
  cr cr ." Verbosity is "
  Verbose@ if ." NORMAL" else ." TERSE" endif
  cr
;

\ ==========================================================================

variable ActionNumber
       0 ActionNumber !

: IncActionNumber \ -- ;
  ActionNumber @ 1+ ActionNumber !
;

: SayActionNumber \ WantCRLF -- ;
  ActionNumber @
  #100 mod 0=  Verbose@  or
  if
    if cr endif
    decimal ActionNumber @ 0 <# # # # # # # # # # # #> type ."  "
  else
    drop
  endif
;

\ ==========================================================================

PFW-Version
[if]

  : U.Dec       \ number -- ;
    decimal u.
  ;

  : $U.Hex      \ number -- ;
    ." $" hex u.
  ;

  : U.Both      \ number -- ;
    dup U.Dec ." ( " $U.Hex ." ) "
  ;

[then]

\ ==========================================================================

: heapabort     \ GLEnum  -- ; Aborts after printing GetLastError()
  ?dup
  if
    cr ." GetLastError() returned " U.Both
  endif
  cr ." Last action was " FALSE SayActionNumber cr
  abort
;

\ ==========================================================================

MaxBlockSize $100 + constant PatternBlockSize

variable Pattern null Pattern !

: PatternFill   ( -- )
  Pattern @ null =
  if
    PatternBlockSize allocate swap Pattern !
    0<> if cr ." BAD PATTERN FILL!" WinGetLastError heapabort endif
    0 Pattern @ PatternBlockSize
    bounds
    do
      dup i c! 1+
    loop
    drop
  endif
;

: PatternFree   ( -- )
  Pattern @ null <>
  if
    Pattern @ free
    0<> if cr ." BAD PATTERN FREE!" WinGetLastError heapabort endif
    null Pattern !
  endif
;

: PatternOffset ( Address -- Offset )
  dup
  if
    begin
      dup $F and 0=
    while
      4 rshift
    repeat
  endif
  $FF and
;

: BlockFill     { This | -- }
  This .Addr @ null <>
  This .Size @ 0>
  and if
    Pattern @  This .Addr @ PatternOffset +
    This .Addr @
    This .Size @
    move
  endif
;

: Difference    { Addr1 Addr2 Size | nSize cOffset -- cOffset }
  0 -> cOffset
  Size -> nSize
  begin
    nSize 0 <= if cOffset exit endif
    Addr1 cOffset + c@
    Addr2 cOffset + c@
    <> if cOffset exit endif
    nSize 1- -> nSize
    cOffset 1+ -> cOffset
  again
;

: (BlockTest)   { cOffset Address Size | -- }
  Address null <>
  Size 0>
  and if
    Pattern @ cOffset +  Address  Size  s=
    not if
      cr ." BAD COMPARE of "
      Pattern @ cOffset +   dup $U.Hex ."  with "
      Address               dup $U.Hex ."  size "
      Size                  dup U.Both ."  offset "
      Difference                U.Both
      WinGetLastError heapabort
    endif
  endif
;

: BlockTest     { This | -- }
  This .Addr @ PatternOffset
  This .Addr @
  This .Size @
  (BlockTest)
;

: MovedTest     { OldAddr OldSize This | SizeToTest -- }
  This .Size @ -> SizeToTest
  SizeToTest OldSize >
  if
    OldSize -> SizeToTest
  endif
  OldAddr PatternOffset  This .Addr @  SizeToTest
  (BlockTest)
;

\ ==========================================================================

: SayBlockSize  \ #bytes -- ;
  Verbose@
  if
    decimal 0 <#
    MaxBlockSizeDigits 0 do # loop
    #> type
  else
    drop
  endif
;

: SayBlock      { This | -- }
  Verbose@
  if
    This .Size @ SayBlockSize
    This .Addr @ null <>
    This .Size @ 0>
    and if
      ."  (" This .Addr @ size SayBlockSize ." ) "
    endif
    ." @ " This .Addr @ .dword
  endif
;

: SayBlockNo    { BlockNumber | -- }
  Verbose@
  if
    BlockNumber ." [" decimal 0 <#
    BlockSlotsDigits 0 do # loop
    #> type ." ] "
  endif
;

\ ==========================================================================

: DoAllocate    { This NewSize | -- }

  "" ALLO " $V. NewSize SayBlockSize ""  > " $V.

  NewSize allocate 0<>
  if
    WinGetLastError dup ERROR_NOT_ENOUGH_MEMORY =
    if
      2drop null
    else
      cr ." BAD ALLOCATE!" WinGetLastError heapabort
    endif
  endif

  dup null <>
  if
    This .Addr !
    NewSize This .Size !
    This SayBlock
  else
    This .Addr !
    0 This .Size !
    "" FAILED!" $V.
  endif

;

\ --------------------------------------------------------------------------

: DoResize      { This NewSize | -- }
  "" RESZ " $V. This SayBlock ""  > " $V.
  NewSize SayBlockSize "" , " $V.
  This .Addr @ NewSize resize 0<>
  if
    WinGetLastError dup ERROR_NOT_ENOUGH_MEMORY =
    if
      2drop null
    else
      cr ." BAD RESIZE!" WinGetLastError heapabort
    endif
  endif
  dup null <>
  if
    This .Addr !
    NewSize This .Size !
    This SayBlock
  else
    drop "" FAILED!" $V.
  endif
;

: DoFree        { This | -- }
  "" FREE " $V. This SayBlock
  This .Addr @ free
  0<> if cr ." BAD FREE!" WinGetLastError heapabort endif
  null This .Addr !
  0 This .Size !
;

: DoSize        { This | -- }
  "" SIZE " $V. This SayBlock
  This .Addr @ null <>
  This .Size @ 0>
  and if
    This .Addr @ size
    Verbose@ if ."  = " dup . endif
    This .Size @ < if cr ." BAD SIZE!" WinGetLastError heapabort endif
  endif
;

\ ==========================================================================

: DoBlock       { BlockNumber | This OldSize NewSize OldAddr -- }
  BlockNumber Blocks -> This
  This .Size @ -> OldSize
  MaxBlockSize Rand1..n -> NewSize
  OldSize 0=
  if
    This NewSize DoAllocate
    This BlockFill
    This BlockTest
  else
    3 Rand1..n
    case
      1 of
        This BlockTest
        This DoFree
      endof
      2 of
        This BlockTest
        This DoSize
      endof
      3 of
        This BlockTest
        This .Addr @ -> OldAddr
        This NewSize DoResize
        OldAddr OldSize This MovedTest          \ Moved data intact?
        This BlockFill
        This BlockTest
      endof
    endcase
  endif
;

\ ==========================================================================

: DisplayTotals { | TotalBlocks TotalSize -- }
  Pattern @ null =
  if
    0 -> TotalBlocks
    0 -> TotalSize
  else
    1 -> TotalBlocks
    PatternBlockSize -> TotalSize
  endif
  BlockSlots 0
  do
    i Blocks .Size @ 0>
    i Blocks .Addr @ null <>
    and if
      1   addr TotalBlocks  +!
      i Blocks .Size @   addr TotalSize  +!
    endif
  loop
  cr
  cr TotalSize U.Dec ."  bytes in " TotalBlocks U.Dec ."  blocks" cr
  TotalBlocks 0>
  if
    ." Average is " TotalSize TotalBlocks / U.Dec ."  bytes per block" cr
  endif
;

\ --------------------------------------------------------------------------

PFW-Version
[if]

  1 cells field-type DWORD

  struct  MemoryStatus
      DWORD .dwLength         \ sizeof(MEMORYSTATUS)
      DWORD .dwMemoryLoad     \ percent of memory in use
      DWORD .dwTotalPhys      \ bytes of physical memory
      DWORD .dwAvailPhys      \ free physical memory bytes
      DWORD .dwTotalPageFile  \ bytes of paging file
      DWORD .dwAvailPageFile  \ free bytes of paging file
      DWORD .dwTotalVirtual   \ user bytes of address space
      DWORD .dwAvailVirtual   \ free user bytes
  end-struct

  1 0 libfunction: GlobalMemoryStatus

  open-libs

  : DisplayStatus { | Buff[ MemoryStatus ] -- }
    MemoryStatus Buff[ .dwLength !
    Buff[ rel>abs GlobalMemoryStatus
    decimal
    cr
    Buff[ .dwMemoryLoad    @ ." MemoryLoad    " U.Dec ." %"      cr
    Buff[ .dwTotalPhys     @ ." TotalPhys     " U.Dec ."  bytes" cr
    Buff[ .dwAvailPhys     @ ." AvailPhys     " U.Dec ."  bytes" cr
    Buff[ .dwTotalPageFile @ ." TotalPageFile " U.Dec ."  bytes" cr
    Buff[ .dwAvailPageFile @ ." AvailPageFile " U.Dec ."  bytes" cr
    Buff[ .dwTotalVirtual  @ ." TotalVirtual  " U.Dec ."  bytes" cr
    Buff[ .dwAvailVirtual  @ ." AvailVirtual  " U.Dec ."  bytes" cr
  ;

[else]

  : DISPLAYSTATUS  (  --  )
    .heap
  ;

[then]

\ ==========================================================================

: DumpHeap      \ -- ;
  [ PFW-Version not ]
  [if]
    startofheap sizeofheap dump
  [then]
;

: CheckHeap     \ -- ;
  HeapOK? not
  if
    [ WantDumpAtEnd ] [if] DumpHeap [then]
    0 heapabort
  endif
;

: CheckDepth    \ ExpectedDepth -- ;
  depth 1-              \ Minus one to allow for EXPECTEDDEPTH on the stack.
  swap over <>
  if
    cr ." BAD DEPTH ( " . ." ); STACK CORRUPT?"
    0 heapabort
  else
    drop
  endif
;

variable StartStackDepth

: RunTest       \ -- ;
  depth StartStackDepth !
  begin
    key? 0<>
    if
      key
      dup EscapeKey = if drop exit endif
      BackSpaceKey = if ToggleVerbosity endif
      DisplayTotals DisplayStatus
      key drop
    endif
    IncActionNumber TRUE SayActionNumber
    ActionNumber @ DumpStart DumpEnd within? if DumpHeap endif
    CheckHeap
    BlockSlots Rand1..n 1-
    dup SayBlockNo
    DoBlock
    StartStackDepth @ CheckDepth
    [ PFW-Version not ]
    [if]
      [ tasking?  WantPause  and ] [if] pause [then]
    [then]
  again
;

: InitBlockArray \ -- ;
  BlockSlots 0
  do
    null i Blocks .Addr !
    0    i Blocks .Size !
  loop
;

: FreeAll       \ -- ;
  BlockSlots 0
  do
    i Blocks .Size @ 0>
    i Blocks .Addr @ null <>
    and if
      IncActionNumber TRUE SayActionNumber
      i SayBlockNo
      i Blocks BlockTest
      i Blocks DoFree
    endif
  loop
;

: TestHeap      \ -- ;
  [ PFW-Version ]
  [if]
    open-libs
    (config)
    Randomise
  [else]
    [ WantInitHeap ] [if] init-heap [then]
    [defined] led-wait [if] 1 led-wait [then]
  [then]
  InitBlockArray
  PatternFill
  DisplayTotals DisplayStatus
  [ WantKeyAtStart ]  [if] key drop [then]
  begin
    RunTest
    DisplayTotals DisplayStatus
    FreeAll
    DisplayTotals DisplayStatus
    key EscapeKey <>
   while repeat
  PatternFree
  DisplayTotals DisplayStatus
  [ Turnkey? ]    [if] bye [then]
  [ PFW-Version ] [if] close-libs [then]
;
: Run TestHeap ;

variable initial-seed

: RunSeeded     \ seed  -- ;
  dup initial-seed !
  randseed !
  run
;

\ --------------------------------------------------------------------------

Turnkey?
[if]

  [defined] WantCoolBar   [if] WantCoolBar   off [then]
  [defined] WantStatusBar [if] WantStatusBar off [then]

  assign Run to-do loader

  "" TESTHEAP.INI" config$ $move
  autosave off

  save   testHEAP
  bind/d testHEAP

[then]

