\ Generic Version 6 XC, TARGEND.FTH, Target side message driver

((
Copyright (c) 1988-2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
fax: +44 (0)2380 339691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

Change History
==============
19990331 SFP008 Fixed 2* and 4* in 16/32 bit selection
19981211 SFP007 Added Harvard conversion and 16/32 bit alternatives
19980311 MSD006 Changed the invocation of SETI into an EI to make it more
         MSD006 processor non-specific.
19980305 MSD005 Rename SEND-MESSAGE as SEND-TO-HOST to avoid a redefintion
         MSD005 when the multi-tasker is included.  (It seems a good idea
         MSD005 to remove redefinition messages wherever possible.)
19980304 MSD004 In RUN-UMBILICAL, replaced C! with ! in BASE setting code.
19980222 SFP003 Converted C@-CMND>T to use SEND-BYTE
                converted SEND-CELL to need no acknowledgements
19980222 SFP002 Added type command to host
                Converted emit to host to send byte
19980212 MSD001 Add copy-block to command set.  (Message 0x0C.)
))


\ ********************************************************************
\ If suitable primitives exist use SYNONYM to use the primitives
\ as WAIT-BYTE and SEND-BYTE. Otherwise define WAIT-BYTE and SEND-BYTE
\ in a similar fashion to KEY and EMIT respectively.
\ ********************************************************************

[undefined] cells [if]          \ SFP008
16bit? [if]
compiler
: cells		2*  ;
target
[then]
32bit? [if]
compiler
: cells		4*  ;
target
[then]
[then]


\ ******************
\ Message primitives
\ ******************

: wait-ack      \ -- ; wait for response from host
  wait-byte ACK! <>             \ get response and check
  IF  message-passer  ENDIF  ;  \   Error!, go restart

: send&ack      \ n -- ; send character, wait for acknowldge
  send-byte                     \ send byte
  wait-ack  ;                   \ get response and check

: recv&ack      \ -- b ; receive character with acknowledge
  wait-byte  ACK! send-byte  ;
((
: recv&ack      \ -- b ; receive character with acknowledge
  wait-byte  dup send-byte
;
))

Hex

16bit? [if]
: send-cell     \ n --
  0 $0100 um/mod  send-byte	\ bml
                  send-byte	\ bl
;

: receive-cell  \ -- n
  recv&ack $0100 *		\ bml
  recv&ack        +		\ bl
;
[then]

32bit? [if]
: send-cell     \ n --
  0 $01000000 um/mod  send-byte	\ bh
  0 $010000 um/mod    send-byte	\ bmh
  0 $0100 um/mod      send-byte	\ bml
                      send-byte	\ bl
;

: receive-cell  \ -- n
  recv&ack  $01000000 *		\ bh
  recv&ack  $010000 *  +	\ bmh
  recv&ack  $0100 *  +		\ bml
  recv&ack         +		\ bl
;
[then]

decimal


\ depth send-stack receive-stack

[undefined] depth [if]
16bit? [if]
: depth         \ -- n ; return no. of items on stack before DEPTH
  s0 @ sp@ - 2- 2/  ;
[then]
32bit? [if]
: depth         \ -- n ; return no. of items on stack before DEPTH
  s0 @ sp@ - 4- 4/  ;
[then]
[then]

: send-stack    \ sys -- ; send stack to host
  depth ( 0 max )                       \ get depth, clip underflow
  dup send-cell  ?dup                   \ tell host what's coming
  if  0 swap 1-                         \ massage for -1 +LOOP
      do  i pick send-cell  -1 +loop    \ copy it - bottom first
  endif  ;

\ standard version
: receive-stack \ -- sys ; receive stack from host
  s0 @ sp!                      \ clear stack
  receive-cell 0                \ get depth
  ?do  receive-cell  loop	\ get stack - bottom first
;

(( \ uses less return stack, may be smaller with VFX
: receive-stack \ -- sys ; receive stack from host
  s0 @ sp!                      \ clear stack
  receive-cell			\ get depth
  begin
    ?dup
   while
    receive-cell  swap 1-
  repeat
;
))


\ ************************
\ Target to Host messaging
\ ************************

: UFkey?	\ -- t/f
  key?-cmnd>h send&ack  receive-cell
; 

: UFkey		\ -- char
  key-cmnd>h send&ack  receive-cell
;

: UFemit	\ char -- ; send character
  emit-cmnd>h send&ack  send-byte
;

harvard? [if]
: UFtypec	\ addr len -- ; works from code memory
  type-cmnd>h send&ack
  $0FF and                      \ clip for byte length
  dup send-byte                 \ send length
  bounds
  ?do  i c@c send-byte  loop	\ send buffer
;

: UFtype	\ addr len -- ; works from data memory
  bounds ?do  i c@ UFemit  loop
;
[else]
: UFtype	\ addr len --
  type-cmnd>h send&ack
  $0FF and                      \ clip for byte length
  dup send-byte                 \ send length
  bounds
  ?do  i c@ send-byte  loop	\ send buffer
;
[then]
((
: UFtype	\ addr len -- ; works from data memory
\ This construct uses MUCH less return stack space than DO ... LOOP
  begin
    ?dup
   while
    over c@ UFemit
    swap 1+  swap 1-
  repeat
  drop
;
))

: UFcr		\ --
  $0D UFemit  $0A UFemit
;

cdata
create UFconsole	\ -- addr ; OUT managed by upper driver
  ' UFkey ,		\ -- char ; receive char
  ' UFkey? ,		\ -- flag ; check receive char
  ' UFemit ,		\ -- char ; display char
  ' UFtype ,		\ caddr len -- ; display string
  ' UFcr ,		\ -- ; display new line
harvard? [if]
  ' UFtypec ,		\ caddr len -- ; display string from CDATA space
[then]


\ ************************
\ Host to target messaging
\ ************************

: tcmd1         \ ?? -- ?? ; execute command
  receive-stack                         \ receive host stack
  execute                               \ XT is top of stack
  OK! send-byte                         \ permit target messages
  send-stack                            \ send results back
;

16bit? [if]
: tcmd6         \ -- ; w fetch operation
  receive-cell @ send-cell  ;

: tcmd7         \ -- ; w store operation
  receive-cell receive-cell !  ;
[then]

32bit? [if]
: tcmd6         \ -- ; w fetch operation
  receive-cell w@ send-cell  ;

: tcmd7         \ -- ; w store operation
  receive-cell receive-cell w!  ;
[then]

: tcmd8         \ -- ; c fetch operation
  receive-cell c@ send-byte  ;
  
: tcmd9         \ -- ; c store operation
  receive-cell receive-cell c!  ;

32bit? [if]
: tcmd10        \ -- ; l fetch operation
  receive-cell @ send-cell  ;
 
: tcmd11        \ -- ; l store operation
  receive-cell receive-cell !  ;
[then]

: tcmd12        \ -- ; block store operation from host to target
  receive-cell receive-cell     	\ -- dest-addr len
  bounds
  ?do  recv&ack i c!  loop
;

harvard? [if]
: tcmd15 	\ -- ; code byte fetch from target
  receive-cell c@c send-byte
;

16bit? [if]
: tcmd13	\ -- ; code word fetch from target
  receive-cell @c send-cell  ;
[then]
32bit? [if]
: tcmd13	\ -- ; code word fetch from target
  receive-cell w@c send-cell  ;

: tcmd17 	\ -- ; code long fetch from target
  receive-cell @c send-cell
;
[then]
[then]

: bad-cmd       \ -- used for a bad command
  ERROR! send-byte  message-passer
;

l: commands>t
  ' tcmd1               ,       \ 1 = execute
  ' receive-cell        ,       \ 2 = push
  ' send-cell           ,       \ 3 = pop
  ' receive-stack       ,       \ 4 = receive stack
  ' send-stack          ,       \ 5 = send stack
  ' tcmd6               ,       \ 6 = w fetch
  ' tcmd7               ,       \ 7 = w store
  ' tcmd8               ,       \ 8 = c fetch
  ' tcmd9               ,       \ 9 = c store
16bit? [if]
  ' bad-cmd             ,       \ 10 = l fetch
  ' bad-cmd             ,       \ 11 = l store
[then]
32bit? [if]
  ' tcmd10              ,       \ 10 = l fetch
  ' tcmd11              ,       \ 11 = l store
[then]
  ' tcmd12              ,       \ 12 = receive data
harvard? [if]
  ' tcmd13		,	\ 13 = code word fetch from target
  ' bad-cmd		,	\ 14 = code word store to target
  ' tcmd15		, 	\ 15 = code byte fetch from target
  ' bad-cmd		,	\ 16 = code byte store to target
16bit? [if]
  ' bad-cmd		,	\ 17 = code long word fetch from target
  ' bad-cmd		,	\ 18 = code long word store to target
[then]
32bit? [if]
  ' tcmd17		,	\ 17 = code long word fetch from target
  ' bad-cmd		,	\ 18 = code long word store to target
[then]
[then]


\ **************************
\ Main Umbilical Forth words
\ **************************

: Message-Passer        \ -- ; Umbilical Forth message loop with multi-tasking
  r0 @ rp!				\ no exit from here
\ [char] G send-byte
\ begin ( wait-byte send-byte ) [char] A send-byte again
  BEGIN
[ tasking? ] [if]			\ Only run scheduler if requested
    Multi? if				\ is the multi-tasker running?
      Begin  Pause Wait-Byte?  Until	\ schedule until ready
    Endif
[then]
    wait-byte				\ get command byte
    1- dup #host-messages u> if		\ if invalid command (range 1..12, 1..18)
      drop ERROR!			\ shout at host
    else
      ACK! send-byte			\ acknowledge valid command
      cells commands>t + @c execute	\ execute relevant word
      wait-ack				\ wait for host acknowledgement
      depth max-depth u> if		\ if stack underflow/overflow
        s0 @ sp!  SUFL!			\ reset stack and tell host
      else
        OK!				\ else say OK
      endif
    endif
    send-byte				\ send command completion status
  AGAIN
;

: run-umbilical \ -- ; run Umbilical Forth
  r> drop                               \ no exit
  rp@ r0 !  sp@ s0 !			\ initialise user vars

  UFconsole opvec !  UFconsole ipvec !
[defined] handler [if]
  handler off				\ no CATCH defined yet
[then]
[ init-idata? ] [if]
  init-idata				\ perform default initialisation
[then]
  init-XTL                              \ set up XTL link
[ spi? ] [if]				\ and the serial link if XTL is SPI
  init-ser
[then]

  #10 base !                            \ set base

[ tasking? ] [if]
  init-multi                            \ initialise multi-tasker
[then]

  message-passer                        \ run message passer
;

