\ XMODEMTXRX.FTH - XMODEM file upload/download tools

((
Copyright (c) 2001, 2002, 2004
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20040504 MPE004 Simplified by rounding lengths to 128 byte units.
20040112 SFP003 Added ^Y equate. Reordered some code.
20021014 SFP002 Modified for COMMON directory.
20010717 EJB000 First release
))


\ ==================
\ *! xmodemtxrx
\ *T XMODEM Receiver and Transmitter
\ ==================

\ *S Introduction
\ *P This file implements the XMODEM 128-byte protocol in both 
\ ** directions. Use with AIDE requires AIDE release 2.500
\ ** upwards.

\ *P No test code is provided for this file as the system has
\ ** been tested by comparison of transferred binary files.

\ *S Words in XMODEMTXRX.FTH

decimal


\ ================
\ *N Configuration
\ ================

[undefined] XmodemTx? [if]
1 equ XmodemTx?	\ -- n
\ *G Non-zero to compile transmit code
[then]

[undefined] XmodemRx? [if]
1 equ XmodemRx?	\ -- n
\ *G Non-zero to compile receive code.
[then]


\ ==========================
\ *N Constants and variables
\ ==========================

\ Characters
$001 equ START	$001 equ SOH
$004 equ EOT
$006 equ ACK
$015 equ NAK
$018 equ CAN
$019 equ ^Y				\ AIDE: Xmodem to host
$01A equ ^Z				\ AIDE: Xmodem from host

\ Error codes
$0101 equ blkerror
\ *G A block number error has occurred.
$0103 equ noreply
\ *G There was no reply within one second.
$0104 equ crcerror
\ *G Bad CRC or checksum.
$0105 equ overflow
\ *G Too many blocks were sent.

$010 equ maxerrs	\ -- n
\ *G Maximum number of errors before transfer is aborted.

\ Variables
variable BLK#                           \ Current Block No.
variable CUR-ADDRESS                    \ Current host address
variable LAST-BLOCK                     \ last block we can receive
variable rxstatus			\ receiver status
variable rxerrcnt			\ counter for rx errors
#128 Buffer: x-buffer	\ -- addr
\ *G Holds a 128 byte Xmodem data block.


\ ==============
\ *N Common code
\ ==============

: init-blks		\ addr #bytes -- #blks
\ *G Given the size of an image, return the number of complete
\ ** 128 byte blocks, set the variable CUR-ADDRESS to ADDR and
\ ** set the variable BLK# to 1.
  swap cur-address !			\ set starting data address
  #127 +  #-128 and  7 rshift		\ calculate number of blocks
  1 blk# !				\ 1 is first Xmodem block
;


\ ======================
\ *N XMODEM transmission
\ ======================

XmodemTx? [if]

: (To-Buffer)	\ -- ; copy 128 bytes to X-BUFFER
\ *G Move 128 bytes from the memory pointed to by CUR-ADDRESS
\ ** into X-BUFFER. This is the default action of TO-BUFFER.
\ ** The variable CUR-ADDRESS is set by BIN-DOWN and friends.
  CUR-ADDRESS @ X-buffer #128 cmove     \ Move 128 bytes from
  #128 cur-address +!  ;                \ memory pointed to by
                                        \ CUR-ADDRESS into buff

Defer To-Buffer	\ --
\ *G Copy the next 128 bytes to transmit to X-BUFFER. They
\ ** are then transmitted from X-BUFFER. You can change this
\ ** action as required by your application.
\ ** The default action is (TO-BUFFER).
  assign (to-buffer) to-do to-buffer

: ?Ack          \ -- ; wait for char, abort if not ACK
\ *G Wait for a character and terminate the transfer and abort
\ ** if the character is not an ACK.
  ser-key ACK <>                        \ wait for response
  IF  EOT ser-emit                      \ flag end of XMODEM
      1 abort" File error"		\ MUST be ACK
  ENDIF  ;

: Send-Block	\ Blk# -- ; transmit a block
\ *G Transmit the 128 byte contents of X-BUFFER to the host.
  SOH ser-emit                          \ Start of header
  dup ser-emit                          \ Block Number
  not ser-emit                          \ 1's complement Block#
  0
  X-buffer #128 bounds DO               \ Send 128 byte Block
    i c@ dup ser-emit			\ transmit it
    + $0FF and				\ Add to Check-Sum
  LOOP
  ser-emit                              \ Send Check-Sum
  ?Ack  ;                               \ wait response

: Bin-Down	\ addr #bytes -- ; transfer memory to host
\ *G Download (transmit) the given block of memory to the host
\ ** using the XMODEM 128 byte block protocol.
\ ** On entry, the variable CUR-ADDRESS is set to ADDR on entry
\ ** and #bytes is rounded up to a 128 byte unit.
  cr cr ." Start XMODEM receiver on host" cr
  ^Y ser-emit                           \ trigger for AIDE/PowerTerm
  init-blks				\ -- #blks
  begin					\ Wait for receiver ...
    case  ser-key
      NAK     of  1  endof		\ simple crc
      ascii C of  1  endof		\ crc-16
        0
    endcase
  until
  1+ 1                                  \ block numbers from 1
  ?DO  to-buffer                        \ Transfer mem to buffer
       i send-block                     \ Send this block
       1 BLK# +!                        \ Count blocks sent
  LOOP
  EOT ser-emit  ?Ack                    \ terminate XMODEM
  cr ." File transfer successful" cr
;

[then]


\ ===================
\ *N XMODEM reception
\ ===================

XmodemRX? [if]

[undefined] ser-flush [if]
: ser-flush	\ -- ; flush the link input
\ *G Flush all input character from the host/target link.
  begin
    ser-key?
  while
    ser-key drop
  repeat
;
[then]

: send-ack	\ -- ; send ACK
\ *G Transmit an ACK character.
  ACK ser-emit
;

: send-nak	\ -- ; Transmit a NAK character
\ *G Transmit a NAK character.
  NAK ser-emit
;

: send-can	\ -- ; send CAN character
\ *G Transmit a CAN character.
  CAN ser-emit
;

: Get-BlockData	\ -- cksum ; get block from host
\ *G Receive a 128 byte XMODEM data block from the host.
  0
  X-buffer #128 bounds do		\ Get 128 byte Block
    ser-key dup i c!			\ save byte to buffer
    + $0FF and                          \ Add to Check-Sum
  loop
;

: toomanyerrs?	\ -- T|F ; true if too many errors
\ *G Return true if too many comms errors have occurred.
  1 rxerrcnt +!
  rxerrcnt @ maxerrs >
;

: (From-Buffer)	\ -- ; copy 128 bytes from X-BUFFER
\ *G Move 128 bytes from X-BUFFER into the memory pointed to
\ ** by CUR-ADDRESS
  X-buffer CUR-ADDRESS @ #128 cmove     \ Move 128 bytes to
  #128 cur-address +!  ;                \ memory pointed to by
                                        \ CUR-ADDRESS from buff

Defer From-Buffer	\ -- ; copy 128 bytes from X-BUFFER
\ *G Move 128 bytes from X-BUFFER into the memory pointed to
\ ** by CUR-ADDRESS. CUR-ADDRESS is set up by BIN-UP and friends.
\ ** The default action is (FROM-BUFFER). You can modify the
\ ** action to suit your own application.
  assign (from-buffer) to-do from-buffer

\ variable xnum

: Get-Block	\ -- ;
\ *G Receive an XMODEM 128 byte data block from the host, processing
\ ** the header and checksum data.
  ser-key ser-key			\ get block and inverse
  over + $0FF <> if
    toomanyerrs? if
      send-can send-can
      blkerror rxstatus !  ( 1 xnum ! )
    else
      send-nak ( 2 xnum ! )
    endif
    send-can
    blkerror rxstatus !  ( 3 xnum ! )
    drop				\ lose incoming block number
  else
    Blk# @ $00FF and <> if		\ correct block ??
      toomanyerrs? if
        send-can send-can
        blkerror rxstatus !  ( 4 xnum ! )
      else
        send-nak  ( 5 xnum ! )
      endif
    else
      Get-BlockData			\ go get the data
      ser-key  =			\ get single cksum byte
      if
        Blk# @ last-block @ > if
          send-can send-can		\ 2 CAN's aborts transfer
          overflow rxstatus !		\ no space left in buffer
	  ( 6 xnum ! )
        else
          From-Buffer			\ copy data to memory
          send-ack			\ all OK send ACK
          1 Blk# +!			\ bump for next block
          rxerrcnt off			\ clear error count again
        endif
      else
        toomanyerrs? if
          send-can send-can		\ 2 CAN's aborts transfer
          crcerror rxstatus !  ( 7 xnum ! )
        else
          send-nak  ( 8 xnum ! )
        endif
      endif      
    endif
  endif
;

: waitforresponse	\ -- ; wait for host to respond
\ *G Wait for the host to respond for up to 1 second. If the
\ ** host does not respond, a NOREPLY status is set in the 
\ ** variable RXSTATUS. This word requires the ticker interrupt
\ ** to be running.
  ticks #1000 +
  begin
[ tasking? ] [if] pause [then]
    dup ticks - 0<			\ timer expired?
    ser-key? or				\ or a character received
  until
  drop
  ser-key? 0=
  if  noreply rxstatus !  endif		\ flag no response
;

: Bin-Up	\ addr len -- status ; status 0 = GOOD
\ *G Upload (receive) a block of data of the given size into memory
\ ** using the XMODEM 128 byte block protocol. An error status is
\ ** returned, 0 indicating success.
\ ** On entry, the variable CUR-ADDRESS is set to ADDR on entry
\ ** and len is rounded up to a 128 byte unit.
\ ** Note that an error return of $0105 indicates the the file 
\ ** being sent is larger than the LEN input parameter.
  cr ." Start Xmodem transmitter on host" cr
  ^Z ser-emit				\ Trigger for AIDE/PowerTerm
  rxstatus off				\ start with good status
  rxerrcnt off				\ and no errors
  ser-flush				\ flush input

  init-blks last-block !		\ find last block

  NAK ser-emit				\ NAK requests single byte checksum
  #40 0 do				\ wait for remote startup
    #1000 ms				\ wait for a second
    ser-key? ?leave			\ done if a reply
    NAK ser-emit			\ then send NAK if still no key
  loop

  begin					\ outer loop handles blocks
    waitforresponse				\ check if we got a response
    rxstatus @ 0= if
      case  ser-key				\ get response
        EOT of  EOT rxstatus !  send-ack  endof	\ all finished
        CAN of  CAN rxstatus !  send-ack  endof	\ remote cancelled
        SOH of  Get-block                 endof	\ start of data block
						\ ignore noise characters
      endcase
    endif
    rxstatus @				\ loop until all done
  until

  rxstatus @ dup EOT = if
    drop 0				\ returns 0 for good transfer
  else
    cr ." Upload failed, code = "
    dup .dword				\ else a fail code
  endif
;

: RecvXmodem	\ addr len -- len' status ; status=0=good
\ *G Upload (receive) a block of data of the given size into memory
\ ** using the XMODEM 128 byte block protocol. The number of bytes
\ ** correctly received and an error status are returned, 0
\ ** indicating success. See BIN-UP above for more details.
  over >r  bin-up  cur-address @ r> - swap
;

[then]


\ ======
\ *> ### 
\ ======

