\ MPE ROM PowerForth 16 bit software floating point package

((
Copyright (C) 1983-2000
MicroProcessor Engineering Ltd.
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

32 bit mantissa, 16 bit exponent

20000713 SFP001 Renamed SINT to F>S and DINT to F>D
))


\ constants and variables

decimal

 26214 26214 -3 FCONSTANT %.1
     0 16384  1 FCONSTANT %1
     0 20480  4 FCONSTANT %10

 31872 19073 34 FCONSTANT %10^10
 32611 28147 -33 FCONSTANT %10^-10

( -- addr )
Variable places         \ holds max number digits after d.p.
  8 places !            \ default to 8 digits

internal

Variable exp
Variable lo1    Variable lo2
Variable hi1    Variable hi2
Variable *sn
Variable /sn
Variable +sn1
Variable +sn2
Variable exp1
Variable %exp
Variable %sgn
Variable %n
fvariable %stk1
fvariable dum


\ d>>n norm-r norm-l ..........................................

: d>>n          \ d1 n -- d1>>n
  0 ?Do  d>>1  Loop  ;

: norm-r        \ mant -- mant'
  1 exp +!  d>>1  ;

: norm-l        \ mant -- mant'
  -1 exp +!  d<<1  ;


\ dnorm - normalise d-word ........................ \  20/11/89

: dnorm         \ dn exp -- fn
  exp !  2dup or                  	( check zero )
  If
    dup 0< >r  dabs           		( save sign )
    Begin
      norm-l dup         		( shift Until -ve )
      0<
    Until
    norm-r
    r> ?dnegate
    exp @                     		( recover sign & exp )
  Else  0
  Endif
;

external


\ d>f s>d s>f

: d>f           \ d -- fn
  #31 dnorm  ;

: s>f           \ s -- fn
  s>d d>f  ;


\ fnegate ?fnegate fabs fsign


: fnegate	\ fn -- -fn
  >r dnegate r>  ;

: ?fnegate	\ fn n -- fn/-fn
  0< If fnegate Endif  ;

: fabs		\ fn -- |fn|
  >r dabs r>  ;

: fsign		\ fn -- |fn| flag ; true If negative
  over 0< >r  fabs  r>  ;


\ Double integer multiply

internal

: d*		\ d1 d2 -- q1
  hi2 ! lo2 !  hi1 ! lo1 !
        lo1 @ lo2 @               um*
         0    lo2 @ hi1 @         um* d+
              lo1 @ hi2 @         um* d+
               0    hi1 @ hi2 @   um* d+  ;

: d!		\ d --
  dum 2!  ;

: d@		\ -- d
  dum 2@  ;

external


\ f* 

: f*		\ f1 f2 -- f3
  fsign *sn !  1+ exp !  d!       \ save sign & exp 1
  fsign *sn +!  exp +!  d@        \ save sign & exp 2
  d* 2dup or                      \ 64 bit multply = 0 ?
  If
    Begin
        dup 16384 and 0=
      While
        -1 exp +!  q<<1
    Repeat
    d!  nip  d@ rot 0< negate 0 d+  ( round up )
    dup 0<
    If  norm-r  Endif
    *sn @ -1 =
    If  dnegate  Endif       ( recover sign )
    exp @                     ( recover exponent )
  Else
    fdrop drop
    0 0 0
  Endif
;

: f/		\ f1 f2 -- f3
  fsign /sn !  negate exp !         	\ negate divisor & save divisor.exp
  2dup or 0= abort" divide by 0 error"  \ check mantissa
  dnegate d!			        \ save divisor.mant
  fsign /sn +!  exp +!			\ make dividend +ve, save dividend.exp
  >r >r 0 0 r> r>                       \ -- dquot dividend ; 0 quotient to stack
  31 0                                  \ shift 31 times
  Do
    d@ d+ dup 0<
    If    >r >r  d<<1  r> r> d@ d-
    Else  >r >r  1 <-s drop  r> r>
    Endif
    d<<1
  Loop
  2drop  exp @ 1+ dnorm  /sn @ -1 =
  If  fnegate  Endif
;

: f+		\ f1 f2 -- f3
  2 pick 2 pick or 0=	( f1 non zero )
  if  fdrop  exit  endif

  fsign +sn1 !  exp !  >r >r
  2 pick 2 pick or 0=
  if
    fdrop
    r> r>  +sn1 @ ?dnegate  exp @
    exit
  endif
   
  fsign +sn2 !  exp1 !
  exp @  exp1 @ - dup
  If
    dup 0>
    If  d>>n
    Else  r> r>  rot abs d>>n  >r >r
    Endif
  Else
    drop
  Endif

  exp @ exp1 @ max exp !  r> r>  +sn1 @ +sn2 @ -
  If
    d-
  Else
    d+  1 exp +!  0 s-> 0 d+  dup 0<
    If  norm-r  Endif
  Endif

  +sn2 @ ?dnegate  exp @ dnorm
;

: f-		\ f1 f2 -- f3
  fnegate f+  ;

: f>d		\ fn -- d
  fsign >r  dup 0<
  If
    fdrop  0 0  r> drop
  Else
    31 swap - dup 0>
    If
      d>>n
    Else
      abort" number too large in F>D"
    Endif
    r> ?dnegate
  Endif  ;

: fint		\ f1 -- f2
  dup 32 <
  If  f>d d>f  Endif  ;

: f>s		\ fn -- n
  f>d ?dup
  If  1+ abort" number too large in F>S"  Endif  ;

  
\ f-compares

( f1 -- flag )
: f0<   drop nip 0<  ;
: f0>   drop  dup 0< invert >r  or 0<> r> and  ;
: f0=   drop or 0=  ;
: f0<>  drop or 0<>  ;

( f1 f2 -- flag )
: f=    f- f0=  ;
: f<    f- f0<  ;
: f>    f- f0>  ;
: f<>	f- f0<>  ;

\ fmax fmin fliteral

: fmax                  \ f1 f2 -- f3
  fover fover f<  If  fswap  Endif  fdrop  ;

: fmin                  \ f1 f2 -- f3
  fover fover f>  If  fswap  Endif  fdrop  ;

: fliteral              \ fn --
  3 #literal  ;  immediate


\ round

( rounds least signIficant eight bits to 0 If higher 8 bits )
( are all 0s or all 1s                                      )
: round		\ f1 -- f2
  fsign >r >r  swap dup 224 and 0=
  If  -256 and  Endif
  dup 224 and  224 - 0=
  If
    -256 and 256 + dup 0=
    If  swap 1+
    Else  swap
    Endif
  Else
    swap
  Endif
  r> over 0<
  If  swap 0 2 um/mod nip  swap 1+  Endif
  r> ?fnegate
;


\ floating point output

internal

: n#                    \ d1 n -- d2 ; converts n digits
  0 ?Do  #  Loop  ;

: ?10pwr        \ exp(2) -- exp(2) exp(10)
  dup 0< >r  dup abs 301 1000 */mod nip
  r> If  negate 1- Endif ;

: oprep		\ fn -- d
  fsign %sgn !
  ?10pwr  dup %exp !  8 swap -  ?dup
  If
    dup 0<
    If  abs 0 Do  %.1 f* Loop
    Else  0 Do  %10 f* Loop
    Endif
  Endif
  round  fdup %1 f<
  If  %10 f*  -1 %exp +!  Endif
  f>d  
;

: .exp		\ -- ; If -10<e<10 then print a "0" in front of exponent  )
  [char] e emit   %exp @  dup 0<
  If  [char] -  Else  [char] +  Endif  emit
  abs  0 <# # #s #> type  ;

external

: e. 		\ dn exp --
  oprep
  <#  8 n#   ascii . hold   #
  %sgn @
  If  ascii -  Else  bl  Endif  hold
  #>
  8 places @ min  0 max +  8 -  type
  .exp   ;

internal
: (f.)          \ fn -- ; display number
  oprep
  <#  8 %exp @ -  0 max  dup >r  ?dup
  If  n#  ascii . hold  Endif
  #s  %sgn @
  If  ascii -  Else  bl  Endif
  hold
  #>
  r> places @ - 0 max -
  dup 0
  Do  2dup + 1- c@  ascii 0 <> ?leave
      1-
  Loop
  2dup + 1- c@  ascii . =
  If  1-  Endif
  type  ;
external

: f.		\ fn --
  dup -10 27 within?
  If  (f.)
  Else  e.
  Endif  ;


\ Floating point input

internal

: fcheck        \ -- am lm ae le flag ; true If f.p. number
  pad count 2dup ascii E scan dup >r    \ ap lp ae le --
  tuck >r >r - r> r>                    \ am lm ae le --
  dup If 1 /string Endif                \ am lm ae' le' --
  pad count ascii . scan  dup >r
  If  ascii , swap c!  Else  drop  Endif
  r> r> or  ;

: mnum          \ am lm -- d 2 | 0 ; uses fchecked string
  2dup ascii , scan nip 0=              \ no , in string ?
  If  2dup + ascii , swap c!            \ add , to end of string
      1+                                \ increment string length
  Endif                                 \ n.b. mnum must be last scanner
  (integer?) dup 1 =
  If  drop s>d 2  Endif  ;

: enum          \ ae le -- n 1 | 0 ; str as above
  dup
  If  (integer?) dup 2 =
      If  2drop drop 0  Endif
  Else  2drop  0 1
  Endif  ;

: fixexp	\ mant exp -- mant' exp'
  >r d>f r>  10 /mod  swap >r  ?dup
  If  dup 0<
      If  abs 0
          Do  %10^-10 f*  Loop
      Else  0
            Do %10^10 f*  Loop
      Endif
  Endif
  r> ?dup
  If  dup 0<
      If  abs 0
          Do  %.1 f*  Loop
      Else  0
            Do  %10 f*  Loop
      Endif
  Endif  ;

external


\ fnumber? 

: fnumber?	\ addr -- 0/.../mant exp 3
  pad over c@ 1+ cmove                  \ copy string to pad
  pad count upper  fcheck               \ is it floating point?
  If  enum                              \ yes, find exponent
      If  -rot mnum                     \ ok, find mantissa
          If  rot dpl @  dup 0>         \ ok, allow for dpl
              If  -  Else  drop  Endif
              fixexp  3                 \ fix exponent, Done
          Else  drop 0
          Endif
      Else  2drop 0
      Endif
      dup #l !
  Else  2drop 2drop  pad integer?       \ otherwise try integer
  Endif  ;


\ reals integers

: reals		\ -- ; allow f.p. input
  assign fnumber? to-Do number?  ;      \ switch in fnumber?

: integers                              \ -- ; no f.p. input
  assign integer? to-Do number?  ;      \ switch in integer?


((
\ (f#) f#in f#

: (f#)		\ addr -- fn 3
  case  fnumber?
    0  of  0  enDof
    1  of  s>f 3  enDof
    2  of  d>f 3  enDof
  end-case  ;

: f#in     	\ -- fn t/f
  query bl word (f#)  ;

: f#       	\ -- f | ; or compiles it [ state smart ] )
  bl word (f#) ?undef  state @
  If  [compile] fliteral  Endif  ;  immediate
))


\ finvert ?finvert fseparate ffrac

: finvert	\ f1 -- 1/f1
  1.0 fswap f/  ;                       \ invert input

: ?finvert      \ f1 n -- f1 | 1/f1 ; invert If n <> 0
  If  finvert  Endif  ;

: fseparate     \ f1 f2 -- rem(f1/f2) quot(f1/f2)
  fover fover f/  frot frot fabs fswap fsign >r fswap
  8 pick 8 pick 8 pick fabs fint f* f-
  r> ?fnegate  fswap  ;

: ffrac         \ f1 f2 -- rem(f1/f2)
  fseparate fdrop  ;


\ magic constants

decimal
( -- fn )
 -8192 25735 18 Fconstant %sinmax
 15402 27213  2 Fconstant %1/lg2
 19775 19728 -1 Fconstant %lg2
 7572  23637  1 Fconstant %1/ln2
  3068 22713  0 Fconstant %ln2
   0  25600   8 Fconstant %alnmax
 23781 19338  0 Fconstant %s1
-11264 27348 -1 Fconstant %s2
     0 25344  7 Fconstant %algmax
  3066 22713  0 Fconstant %d8
 -5035 28461 -1 Fconstant %d9

 3.14159265     Fconstant pi
 60755 25735 1  Fconstant pi/2
 60756 25735 0  Fconstant pi/4
 49590 20860 -1 Fconstant 1/pi
 49590 20860 0  Fconstant 2/pi
 31129 23170 0  Fconstant %sqr.5
 180.0          Fconstant %180


\ fsqr by Heron's formula 
-short-branches
: fsqr		\ f1 -- f2
  2 pick 2 pick or                 ( check 0 )
  If  over 0<                             ( check -ve )
      abort" square root of negative no.!"
      >r 0                                ( .5 <=f<=1 )
      fdup %s1 f*  %s2 f+               ( hart est y0 )
      fover fover f/ f+
      fswap fover f/ fswap 2-  f+
      r@ 1 and  dup r> + >r
      If  %sqr.5 f*  Endif
      r> 2/ +
  Endif  ;
+short-branches


\ deg>rad rad>deg

((
variable DEGREES/RADIANS                \ degrees=-1, radians=0
  0 degrees/radians !

: DEGREES?	\ -- flag ; in degrees?
  degrees/radians @
;

: RADIANS?  	\ -- flag ; in Radians?
  degrees? not
;

: DEGREES  	\ -- ; switch to degrees
  -1 degrees/radians !
;

: RADIANS	\ -- ; switch to radians
  0 degrees/radians !
;
))
: deg>rad	\ n1 -- n2
  %180 f/ pi f*  ;

: rad>deg	\ n1 -- n2
  %180 f* pi f/  ;


\ (sin) 

internal

-1.6666666e-1                8.3333315e-3
-1.98409e-4                  2.7526e-6
-2.39e-8
5 farray sinarray

\ f1 -- f2
: (sin)
  fdup fdup f* fdup 0 sinarray f*
  4 1  Do
    i sinarray f+ fover f*
  Loop
  4 sinarray f+ f* 1.0 f+ f*  ;


\ sine functions

: ?tr-err	\ flag --
  abort" overflow in trig. function"  ;

: sinfn		\ f1 -- f2
  fdup 1/pi f*            ( ang ang/pi --- )
  fint fdup f>s %n !     ( ang int[ang/pi] --- )
  pi f* f-                ( ang-pi*int[ang/pi] - )
  fdup pi/2 f>
  If  pi fswap f-  Endif
  (sin)
  %n @ 1 and
  If  fnegate  Endif
;


\ sine cosine

: (fsin)        \ f1 -- f2
  fsign >r
  fdup %sinmax f> ?tr-err
  sinfn
  r> ?fnegate
;

external

: fsin		\ f1 -- f2
\  Degrees? if  deg>rad  then
  (fsin)
;

: fcos		\ f1 -- f2
\  Degrees? if  deg>rad  then
  pi/2 f+  (fsin)
;


\ tan

internal

\ n -- fn
3.3333140e-1                 1.3339240e-1
5.3374060e-2                 2.4565089e-2
2.9005250e-3                 9.5168091e-3
6 farray tanarray

: (tan)		\ f1 -- f2
  fdup fdup f* fdup 0 tanarray f*
  5 1
  Do
    i tanarray f+ fover f*
  Loop
  5 tanarray f+ f* 1.0 f+ f*
;

external

: ftan		\ f1 -- f2
\  degrees? if  deg>rad  then
  fsign >r
  fdup %sinmax f> ?tr-err
  fdup 2/pi f* fint fswap fover pi/2 f* f-
  fdup pi/4 f>
  If  pi/2 f- (tan) fswap 1.0 f+
  Else  (tan) fswap
  Endif
  f>s 1 and
  If  finvert  fnegate  Endif
  r> ?fnegate
;


\ arcfn

internal

\ n -- fn
1.5707963e0                  -2.1459880e-1
8.8978987e-2                 -5.0174305e-2
3.0891881e-2                 -1.7088126e-2
6.6700901e-3                 -1.2624911e-3
8 farray asinarray

: arcfn		\ f1 -- f2
  fdup 0 asinarray f*
  7 1
  Do
    i asinarray f+ fover f*
  Loop
  7 asinarray f+
  fswap 1.0 fswap f- fsqr f*
;


\ atfn 

\ n -- fn
-3.3333145e-1                1.9993551e-1
-1.4208899e-1                1.0656264e-1
-7.5289640e-2                4.2909614e-2
-1.6165737e-2                2.8662257e-3
8 farray atanarray

external

: atfn		\ f1 -- f2
  fdup fdup f* fdup 0 atanarray f*
  7 1
  Do
    i atanarray f+ fover f*
  Loop
  7 atanarray f+ f* 1.0 f+ f*  
;


\ fasin facos fatan 

: fasin		\ f1 -- f2
  fsign >r  arcfn pi/2 fswap f-
  r> ?fnegate
\  degrees? if  rad>deg  then
;

: facos		\ f1 -- f2
  fsign >r  arcfn
  r> If  pi fswap f-  Endif
\  degrees? if  rad>deg  then
;

: fatan		\ f1 -- f2
  fsign >r  fdup 1.0 f>
  If  finvert  atfn pi/2 fswap f-
  Else  atfn
  Endif
  r> ?fnegate
\  degrees? if  rad>deg  then
;


\ ln log
internal
9.9999642e-1                 -4.9987412e-1
3.3179903e-1                 -2.4073381e-1
1.6765407e-1                 -9.5329390e-2
3.6088494e-2                 -6.4535442e-3
8 farray logarray       \ n -- fn
external

: fln		\ f1 -- f2
  fdup f0> 0=
  abort" invalid argument to log function"
  1 swap 1- %n !  1.0 f-
  fdup 0 logarray f*
  7 1
  Do
    i logarray f+ fover f*
  Loop
  7 logarray f+ f* %n @ s>f  %d8 f* f+
;

: flog		\ f1 -- f2
  fln %d9 f*  ;


\ 10^x

internal
\ n -- fn
1.1512928e0                  6.6273088e-1
2.5439357e-1                 7.2951737e-2
1.7421120e-2                 2.5549180e-3
9.3264267e-4
7 farray 10array

: (10^x)	\ f1 -- f2
  fdup 0 10array f*
  6 1
  Do
    i 10array f+ fover f*
  Loop
  6 10array f+ f*
  1.0 f+  fdup f*  ;


\ e^x

\ n -- fn
-1.0000000e0    4.9999992e-1    -1.6666530e-1
4.1657348e-2    -8.3013598e-3   1.3298820e-3
-1.4131610e-4
7 farray e-array

: (e^x)		\ f1 -- f2
  fdup 0 e-array f*
  6 1
  Do
    i e-array f+ fover f*
  Loop
  6 e-array f+ f*
  1.0 f+  finvert  ;

external

: fe^x		\ f1 -- f2
  fsign >r
  fdup %alnmax f> abort" overflow in fe^x"
  fdup  %1/ln2 f*  fdup f>s %n !  fint %ln2 f* f-
  (e^x)  %n @ +
  r> ?finvert
;

: f10^x		\ f1 -- f2
  fsign >r
  fdup %algmax f> abort" overflow in f10^x"
  fdup %1/lg2 f*  fdup f>s %n !  fint %lg2 f* f-
  (10^x)  %n @ +
  r> ?finvert  ;


\ fx^n 

-short-branches
: fx^n		\ x-real n-integer -- fn
  ?dup
  If
    dup 0< >r  abs >r  1.0 fswap  r>
    Begin
        dup 0>
      While
        dup 2 / dup 2* rot -
        If  >r fswap fover f* fswap r>  Endif
        >r fdup f* r>
    Repeat
    drop fdrop  r> ?finvert
  Else
    fdrop  1.0
  Endif
;
+short-branches

\ fx^y

-short-branches
: fx^y		\ x-real y-real -- fn
  fdup f0<>
  If
    fdup fdup fint f- %stk1 f!
    f>s >r fdup r>  fx^n fswap
    %stk1 f@ fdup f0<>
    If
      fover f0< abort" result of fx^y is complex"
      fswap  fsign >r
      flog f* f10^x
      r> ?finvert
      f*
    Else
      fdrop fdrop
    Endif
  Else
    fdrop fdrop  1.0
  Endif
;
+short-branches


\ ***********************************************
\ ANS extensions not in original Forth 83 package
\ ***********************************************

\ ------------------------------------------------------------------------------
\       rounding
\ ------------------------------------------------------------------------------

f# 0.5 fconstant %ROUNDPOINT

: FLOOR		\ r1 -- r2
  fdup fint                             \ input, input rounded to zero
  fover f0>                             \ input was positive?
  if
    fnip                                \   positive integral fine
  else
    fswap fover f- f0<>                 \   negative, but input not an integral?
    if  %1 f-  then                     \     round towards -ve infinity
  then
;

: FROUND	\ r1 -- r2
  fdup fint                             \ input, input rounded to zero
  fover fover f- fabs %roundpoint f<    \ input fractional part closer to integral?
  if
    fnip                                \   just return integral
  else
    fswap f0>                           \   input positive?
    if
      %1                                \     round towards +ve infinity
    else
      %1 fnegate                        \     round towards -ve infinity
    then
    f+
  then
;


\ ------------------------------------------------------------------------------
\       miscellaneous
\ ------------------------------------------------------------------------------

: FALIGNED	\ addr -- f-addr
;

: FALIGN	\ --
;

: REPRESENT	\ r c-addr u -- n flag1 flag2
  >r >r  oprep s>d                      \ convert to a double
  <#  9 n#  #>                          \ generate string with max. digits
  r> min                                \ use input string size if smaller
  r> swap cmove                         \ copy string to buffer
  %exp @ 1+                             \ return exponent
  %sgn @                                \ return sign
  true                                  \ conversion always works
;

