\ MPE PowerForth v6 software floating point package for 32 bit targets

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change History
==============
20021112 SFP003 DOCGENned it.
20000713 SFP002 Renamed DINT to F>D for consistency. F>D is the ANS word.
		The original F>D was just a synonym. Similarly SINT was
		renamed to F>S
20000706 SFP001 Applied improvements supplied by Hiden Analytical.
		Added minor performance optimisations.
		Made F/ and DINT return clipped values on overflow.
19981029 MSD000 Brought the common stuff out of all the various versions of
         MSD000 SOFTFP.FTH into a single (that is, *this*) file.

32-bit mantissa and 32-bit exponent, combined FP and data stack
))


\ =========
\ *! softfp
\ =========
\ *T Software Floating Point

\ *S Introduction
\ *P Although most embedded applications only require integer
\ ** arithmetic, some do require floating-point. Therefore software
\ ** floating-point is supplied with the cross-compiler and the
\ ** target Forth. The target floating point wordset is not fully
\ ** ANS compliant, but satisfies the needs of embedded systems
\ ** without undue complexity. The Forth data stack and the floating
\ ** point stack are the same. The floating point data storage
\ ** format is not IEEE format, but is optimised for performance
\ ** on small controllers. If you need a separate floating point
\ ** stack or IEEE format storage, please contact MPE. Any
\ ** variations in the implementation will be documented in the
\ ** target specific section of the manual.

\ *P The cross-compiler has a more limited floating-point support
\ ** than the target, this means that some words are avaliable 
\ ** within colon definitions, but not outside them.

\ *S Source code
\ *P The source code is in two sets of files, one for 32 bit
\ ** Forth targets, the other for 16 bit targets. The files are:
\ *E   COMMON\SFP32HI    32 bit primitives
\ **   COMMON\SFP32COM   32 bit high level code
\ **   COMMON\SFP16HI    16 bit primitives
\ **   COMMON\SFP16COM   16 bit high level code
\ *P These files use no assembler definitions. Some targets have
\ ** code versions of the primitives, and these will be found in
\ ** the CPU specific code directory. A significant increase in
\ ** performance can be obtained by using the code files.

\ *S Entering floating-point numbers
\ *P Floating-point numbers can be entered in two forms,
\ ** 1.234 and 0.1234e1. Floating-point numbers are compiled as
\ ** literal numbers when in a colon definition and placed on the
\ ** cross-compiler's stack when outside a definition.

\ *S The form of floating-point numbers
\ *P A floating-point number is placed on the Forth data stack. In
\ ** the Forth literature, this is referred to as a combined floating 
\ ** point and data stack.
\ ** For 32 bit targets, a floating point number consists of two
\ ** 32-bit numbers, one for the mantissa and one for the exponent.
\ ** For 16 bit targets, it consists of a 32-bit double mantissa
\ ** and a single 16-bit exponent. The mantissa is normalised. 
\ ** The exponent is on the top of the stack.
\ ** Note that for 16 bit targets, number conversion is affected
\ ** by the cross-compiler directives HOST-MATH and TARGET-MATH.
\ ** HOST-MATH leaves double numbers and floats in 32-bit form,
\ ** whereas TARGET-MATH leaves them in 16-bit form.

\ *S Creating variables
\ *P To create a variable, use FVARIABLE. FVARIABLE works in the
\ ** same way as VARIABLE. For example, to create a floating-point
\ ** variable called VAR1 you code:
\ *C   FVARIABLE VAR1
\ *P When VAR1 is used, it returns the address of the floating-point
\ ** number.

\ *S Accessing variables
\ *P Two words are used to access floating-point variables, F@ and F!.
\ ** These are analogous to @ and !.

\ *S Creating constants
\ *P To create a floating-point constant, use FCONSTANT. FCONSTANT
\ ** is analogous to CONSTANT. For example, to generate a
\ ** floating-point constant called CON1 with a value of 1.234,
\ ** you enter:
\ *C   1.234 FCONSTANT CON1
\ *P When CON1 is executed, it returns 1.234 on the Forth stack.

\ *S Using the supplied words
\ *P The supplied words split into several groups:
\ *B sines, cosines and tangents
\ *B arc sines, cosines and tangents
\ *B arithmetic functions
\ *B logarithms
\ *B powers
\ *B displaying floating-point numbers
\ *B inputting floating-point numbers

\ *P The following functions only exist as target words so you
\ ** cannot use them in calculations in your source code when
\ ** outside a colon definition.

\ *N Calculating sines, cosines and tangents
\ *P To calculate sine, cosine and tangent, use FSIN, FCOS and
\ ** FTAN respectively. Angles are expressed in radians.

\ *N Calculating arc sines, cosines and tangents
\ *P To calculate arc sine, cosine and tangent, use FASIN, FACOS
\ *P and FATAN respectively. They return an angle in radians.

\ *N Calculating logarithms
\ *P Two words are supplied to calculate logarithms, FLOG and FLN.
\ ** FLOG calculates a logarithm to base 10 (decimal).
\ ** FLN calculates a logarithm to base e. Both take a
\ ** floating-point number in the range from 0 to Einf.

\ *N Calculating powers
\ *P Three power functions are supplied:
\ *C   FE^X F10^X X^Y

\ *S Degrees or radians
\ *P The angular measurement used in the trigonometric functions
\ ** are in radians. To convert between degrees and radians use
\ ** RAD>DEG or DEG>RAD. RAD>DEG converts an angle from radians
\ ** to degrees. DEG>RAD converts an angle from degrees to radians.

\ *S Displaying floating-point numbers
\ *P Two words are available for displaying floating-point numbers,
\ ** F. and E.. The word F. takes a floating-point number from
\ ** the stack and displays it in the form xxxx.xxxxx or x.xxxxxEyy
\ ** depending on the size of the number. The word E. displays the
\ ** number in the latter form.

\ *S Changes from v6.0 to v6.1
\ *P Renamed DINT to F>D for consistency. F>D is the ANS word.
\ ** The original F>D was just a synonym. Similarly SINT was 
\ ** renamed to F>S.

\ *P The word FLOATS that enabled floating point number conversion
\ ** has been renamed to REALS to avoid a name conflict with the
\ ** ANS word of the same name.

\ *P The F-PACK vocabulary has been removed as no one liked it,
\ ** and it could be considered contrary to the ANS Forth specification.
\ ** If you wish to retain the F-PACK vocabulary, add the following lines
\ ** before and after the compilation of the floating point code:
\ *E   only forth definitions         \ *** added ***
\ **   vocabulary f-pack              \ *** added ***
\ **   also f-pack definition         \ *** added ***
\ **   include %CommonDir%\Sfp32Hi    \ primitives
\ **   include %CommonDir%\Sfp32Com   \ common high level code
\ **   previous definitions           \ *** added ***

\ *P The code enabling floating point to work in degrees or
\ ** radians has been commented out for ANS compatibility. All
\ ** trig functions now operate in radians. The commented out
\ ** code may be uncommented if you need backward compatibility.

\ *N 32 bit targets: software floating point
\ *P Overhauled 32 bit software floating point and incorporated
\ ** improvements contributed by Hiden Analytical. These include
\ ** more complete special case detection, faster high level code,
\ ** and more accurate number input and output.

\ *P Removed all use of global variables except PLACES to make
\ ** the floating point code usable in interrupt routines and 
\ ** in multitasked systems. If the output routines are to be
\ ** multitasked, change the definition of PLACES from:
\ *C   VARIABLE PLACES  8 PLACES !
\ *P to:
\ *C   CELL +USER PLACES
\ *P and remember to initialise PLACES before using the floating
\ ** point output routines.

\ *P Many words that are only useful as factors have been made
\ ** headerless to save target memory space.

\ *N 16 bit targets: software floating point
\ *P Note that the 16 bit floating point pack is not re-entrant.
\ ** If you need to use the floating point pack in a multitasking
\ ** system, you should convert the global variables to USER
\ ** variables. The word +USER can be used
\ *C   <size> +USER <name>
\ *P to define a USER variable of a given size (normally a CELL)
\ ** at the next free offset in the USER area. Only PLACES will
\ ** need initialisation.

\ *S Glossary

decimal


\ -----------------------------------
\ *N Basic stack and memory operators
\ -----------------------------------

: F!		\ r addr --
\ *G Stores r at addr
  2!
;

: F@		\ addr -- r
\ *G Fetches r from addr.
  2@
;

: F,		\ r --
\ *G Lays a real number into the dictionary, reserving 8 bytes.
  here  8 allot  f!
;

: FDUP  	\ r -- r r
\ *G Floating point equivalent of DUP.
  2dup
;

: FOVER		\ r1 r2 -- r1 r2 r1
\ *G Floating point equivalent of OVER.
  2over
;

: FROT		\ r1 r2 r3 -- r2 r3 r1
\ *G Floating point equivalent of ROT.
  5 roll 5 roll
;

: FPICK		\ fu..f0 u -- fu..f0 fu
\ *G Floating point equivalent of PICK.
  2* 1+ dup >r pick  r> pick
;

: FROLL		\ f1 f2 f3 --  f2 f3 f1
\ *G Floating point equivalent of ROLL.
  2* 1+ dup >r roll  r> roll
;

: FSWAP		\ r1 r2 -- r2 r1
\ *G Floating point equivalent of SWAP.
  2swap
;

: FDROP		\ r --
\ *G Floating point equivalent of DROP.
  2drop
;

: FNIP		\ r1 r2 -- r2
\ *G Floating point equivalent of NIP.
  fswap fdrop
;

compiler	\ internal use only!
: f>r		2>r  ;
: fr>		2r>  ;
target


\ define some primitives as compiler macros
optimised? [if]
compiler
: F!		\ r f-addr --
  2!
;

: F@		\ f-addr -- r
  2@
;

: FDUP  	\ r -- r r
  2dup
;

: FDROP		\ r --
  2drop
;

: FNIP		\ r1 r2 -- r2
  2swap 2drop
;
target
((
: FOVER		\ r1 r2 -- r1 r2 r1
  2over
;

: FSWAP		\ r1 r2 -- r2 r1
  2swap
;
))
[then]


\ --------------------------------
\ *N Floating point defining words
\ --------------------------------

: FVARIABLE	\ "<spaces>name" -- ; Run: -- f-addr
\ *G Use in the form: FVARIABLE <name> to create a variable
\ ** that will hold a floating point number.
  create
    0 , 0 ,
  does>
;

: FCONSTANT	\ r "<spaces>name" -- ; Run: -- r
\ *G Use in the form: <float> FCONSTANT <name> to create a constant
\ ** that will return a floating point number.
  create
    , ,
  does>
    f@
;

: FARRAY	\ "<spaces>name" fn-1..f0 n -- ; Run: n -- rn
\ *G Use in the form: n FARRAY <name> to create a variable    
\ ** that will hold a default floating point number. When the 
\ ** array name is executed, the index i is used to retun the 
\ ** address of the i'th 0 zero-based element in the array.   
\ ** For example, 5 FARRAY TEST will set up 5 array elements  
\ ** each containing 0, and then f n TEST F! will store f in  
\ ** the nth element, and n TEST F@ will fetch it.
  create
    0 do , ,  loop
  does>
    swap 2 cells * +  f@
;


\ -------------------
\ *N Type conversions
\ -------------------

internal

[undefined] norm [if]
: NORM		\ n exp -- f
\ *G Normalise a single integer and a single exponent to produce a
\ ** floating point number. INTERNAL.
  over if			        \ -- n exp ; check zero
    swap				\ -- exp n
    dup >r  abs		                \ -- exp n' ; R: -- n ; save sign, make positive
    dup 0< if				\ -- exp n' ; $80000000 is a special case
      r> drop
    else				\ -- n' ; R: -- n
      begin
        swap 1- swap			\ decrement exponent
        1 lshift  dup 0<		\ shift until -ve
      until
      swap 1+ swap  1 rshift		\ step back
      r> ?negate			\ restore sign
    endif
    swap				\ -- n' exp
  else
    drop 0				\ -- 0 0
  then
;
[then]

: DNORM		\ d exp -- fn ; normalise a 64 bit double
\ *G Normalise a double integer and a single exponent to produce a
\ ** floating point number. INTERNAL.
  >R  2DUP or 0= IF
    R> DROP				\ return 0 mantissa \ 0 exp
  ELSE
    TUCK DABS				\ save sign, make mantissa poitive
    BEGIN
      DUP $40000000 AND 0=
     WHILE
      2DUP D+				\ Double it, ie shift right
      R> 1- >R				\ adjust exponent
    REPEAT
    NIP  SWAP 0< ?NEGATE  R>		\ discard ls, apply sign, merge exp
  THEN
;

external

: FSIGN		\ fn -- |fn| flag ; true if negative
\ *G Return the absolute value of fn and a flag which is true
\ ** if fn is negative.
  swap dup abs -rot 0<
;

: F>S		\ fn -- n
\ *G Converts a float to a single integer.
\ ** Note that F>S truncates the number towards zero
\ ** according to the ANS specification. If |fn| is greater
\ ** than maxint, +/-maxint is returned.
  fsign >r  dup 0< if
    r> drop  fdrop 0
  else
    31 swap -  dup 0> if
      rshift
    else
\     1 abort" number too large in F>S"
      2drop  $7FFFFFFF			\ return max int
    then
    r>
    if  negate  then
  then
;

: F>D		\ fn -- d
\ *G Converts a float to a single integer.
\ ** Note that F>D truncates the number towards zero
\ ** according to the ANS specification. If |fn| is greater
\ ** than dmaxint, +/-dmaxint is returned.
  fsign >r  dup 0< if
    r> drop  fdrop 0 0
  else
    63 swap -  dup 0> if
      0 -rot d>>n
    else
\     1 abort" number too large in F>D"
      2drop  $FFFFFFFF $7FFFFFFF	\ max dint
    then
    r>
    if  dnegate  then
  then
;

: FINT		\ f1 -- f2
\ *G Chop the number towards zero to produce a floating point
\ ** representation of an integer.
  dup #32 <
  if  f>s s>f  then
;

: S>F		\ n -- fn
\ *G Converts a single integer to a float.
  #31 norm
;

: D>F		\ d -- fn
\ *G Converts a double integer to a float.
  #63 dnorm
;


\ -------------
\ *N Arithmetic
\ -------------

: FNEGATE	\ r1 -- r2
\ *G Floating point negate.
  swap negate swap
;

: ?FNEGATE	\ fn n -- fn|-fn 
\ *G If n is negative, negate fn.
  0< if  fnegate  then
;

: FABS		\ fn -- |fn|
\ *G Floating point absolute.
  swap abs swap
;

: F*		\ r1 r2 -- r3
\ *G Floating point multiply.
  fsign >r  fswap fsign r> + >r		\ -- |r2| |r1| ; R: -- sign
  rot + 1+ -rot				\ -- exp mant2 mant1 ; R: -- sign
  um* 2dup or if			\ -- exp ud ; R: -- sign ; 64 bit multply <> 0 ?
    begin
      dup  $40000000 and 0=		\ normalised?
     while
      rot 1- -rot  d<<1			\ 1 step
    repeat
    swap 0< if  1+  then		\ round up on discarded low half
    dup 0<                              \ check for overflow
    if  swap 1+ swap  1 rshift  then	\ undo last step
    r> -1 =				\ see if product -ve
    if  negate  then                    \ make it -ve
    swap				\ recover exponent
  else
    2drop drop  r> drop  0 0		\ return 0
  then
;

: F/		\ r1 r2 -- r3
\ *G Floating point divide.
  fsign >r  fswap fsign r> + >r fswap	\ -- |r1| |r2| ; R: -- sign
  over 0= if				\ Check r2 mantissa
    fdrop fdrop				\ discard r1, r2
    $40000000 $7FFF			\ very large positive float
  else					\ -- mant1 exp1 mant2 exp2 ; R: -- sign
    rot swap - -rot			\ -- exp mant1 mant2 ; R: -- sign ; subtract exponents
    >r  0 swap				\ -- exp dividend ; R: -- mant2 sign ; dividend double prec.
    dup r@ u< 0= if			\ avoid divide overflow
      rot 1+ -rot			\   update exp
      d>>1				\   update dividend
    then
    r@ um/mod                           \ -- exp urem uquot ; R: -- mant2 sign ; do division
    swap r> 2/ u> if			\ -- exp uquot ; R: -- sign
      1+				\ if rem>div/2 round up
    then
    dup 0< if                           \ check for overflow
      1 rshift  swap 1+			\  and correct
    else
      swap
    endif
    1- norm				\ normalise
  endif
  r> -1 =			        \ check sign
  if  fnegate  then
;

: F+		\ r1 r2 -- r3
\ *G Floating point addition.
  over 0= if				\ check for r2=0
    2drop  exit
  endif

  3 pick 0= if				\ check for r1=0
    2swap 2drop  exit
  endif

  fsign >r
  fswap fsign >r fswap			\ -- m1 e1 m2 e2 ; R: -- s1 s2
  rot 2dup - >r				\ -- m1 m2 e2 e1 ; R: -- e2-e1 s1 s2
  max -rot  r@ if			\ -- esum m1 m2
    r@ 0> if				\ reduce smaller mantissa
      swap r> rshift swap		\ -- esum m1' m2 ; R: -- s1 s2
    else
      r> abs rshift			\ -- esum m1 m2' ; R: -- s1 s2
    then
  else
    r> drop				\ -- esum m1 m2  ; R: -- s1 s2
  then
  r> r>					\ -- esum m1 m2 s2 s1 ; R: --
  over <> if				\ -- esum m1 m2 s2 ; if signs differ
    >r  negate +			\ -- esum m1-m2 ; R: -- s2
  else
    >r  +				\ -- esum m1+m2 ; R: -- s2
    swap 1+ swap			\ scale exponent
    0 s-> +  dup 0< if			\ and mantissa, check nomalisation
      swap 1+ swap  1 rshift
    endif
  endif					\ esum msum ; R: -- s2
  r>					\ apply sign
  if  negate  endif
  swap norm				\ and normalise
;

: F-		\ r1 r2 -- r3
\ *G Floating point subtraction.
  fnegate f+
;

: FSEPARATE	\ f1 f2 -- f3 f4
\ *G Leave the signed integer quotient f4 and remainder f3 when
\ ** f1 is divided by f2. The remainder has the same sign as the
\ ** dividend.
  fover fover f/
  frot frot fabs
  fswap fsign >r fswap
  2 fpick fabs fint f* f-
  r>
  if  fnegate  then
  fswap
;

: FFRAC		\ f1 f2 -- f3
\ *G Leave the fractional remainder from the division f1/f2. The
\ ** remainder takes the sign of the dividend.
  fseparate fdrop
;


\ -----------------------
\ *N Relational operators
\ -----------------------

: F0<		\ f1 -- flag
\ *G Floating point 0<.
  drop 0<
;

: F0>		\ f1 -- flag
\ *G Floating point 0>.
  drop 0>
;

: F0=  		\ f1 -- flag 
\ *G Floating point 0=.
  drop 0=
;

: F0<>  	\ f1 -- flag 
\ *G Floating point 0<>.
  drop 0<>
;

optimised? [if]
compiler
: F0<		\ f1 -- flag
  drop 0<
;

: F0>		\ f1 -- flag
  drop 0>
;

: F0=  		\ f1 -- flag 
  drop 0=
;

: F0<>  	\ f1 -- flag 
  drop 0<>
;
target
[then]

: F=  		\ f1  f2 -- flag 
\ *G Floating point =.
  f- f0=
;

: F<  		\ r1  r2 -- flag 
\ *G Floating point <.
  f- f0<
;

: F>  		\ f1  f2 -- flag 
\ *G Floating point >.
  f- f0>
;

: FMAX  	\ r1 r2 -- r1|r2 
\ *G Floating point MAX.
  fover fover f<
  if  fswap  then
  fdrop
;

: FMIN  	\ r1 r2 -- r1|r2 
\ *G Floating point MIN.
  fover fover f>
  if  fswap  endif
  fdrop
;

\ -----------
\ *N Rounding
\ -----------

f# 0.5 fconstant %ROUNDPOINT
f# 1.0 fconstant %ONE
\ *G Floating point 1.0.

: FLOOR  	\ r1 -- r2 
\ *G Floored round towards -infinity.
  fdup fint                             \ input, input rounded to zero
  fover f0> if				\ input was positive?
    fnip                                \   positive integral fine
  else
    fswap fover f- f0<>                 \   negative, but input not an integral?
    if  %one f-  then                   \     round towards -ve infinity
  then
;

: FROUND  	\ r1 --  r2 
\ *G Round the number to nearest or even.
  fdup fint                             \ input, input rounded to zero
  fover fover f- fabs %roundpoint f< if	\ input fractional part closer to integral?
    fnip                                \   just return integral
  else
    fswap f0> if                        \   input positive?
      %one                              \     round towards +ve infinity
    else
      %one fnegate                      \     round towards -ve infinity
    then
    f+
  then
;


\ ----------------
\ *N Miscellaneous
\ ----------------

: FALIGNED	\ addr -- f-addr 
\ *G Aligns the address to accept an 8-byte float.
  7 + -8 and
;
optimised? [if]
compiler
: FALIGNED	\ addr -- f-addr 
  7 + -8 and
;
target
[then]

: FALIGN  	\ --
\ *G Aligns the dictionary to accept an 8-byte float.
  here dup faligned swap - allot
;

: FDEPTH  	\ -- +n
\ *G Returns the number of floats on the stack. 
  depth 2/
;

: FLOAT+  	\ f-addr1 -- f-addr2 
\ *G Increments addr by 8, the size of a float.
  8 +
;

: FLOATS  	\ n1 -- n2 
\ *G Returns n2, the size of n1 floats.
  8 *
;
optimised? [if]
compiler
: FLOAT+  	\ f-addr1 -- f-addr2 
  8 +
;

: FLOATS  	\ n1 -- n2 
  8 *
;
target
[then]


\ ------------------------
\ *N Floating point output
\ ------------------------

1 s>f 10 s>f f/ fconstant %.1
\ *G Floating point 0.1.
          1 s>f fconstant %1
\ *G Floating point 1.0.
         10 s>f fconstant %10
\ *G Floating point 10.0.
  1250000000 34 fconstant %10^10
\ *G Floating point 10^10.
 1844674407 -33 fconstant %10^-10
\ *G Floating point 10^-10.
F# 1.0E256 FCONSTANT %10^256
\ *G Floating point 10^256.
F# 1.0E-1 FCONSTANT %10E-1
\ *G Floating point 10^-1.
F# 1.0E-10 FCONSTANT %10E-10
\ *G Floating point 10^-10.
F# 1.0E-256 FCONSTANT %10^-256
\ *G Floating point 10^-256.

F# 1.0E15 F# 1.0E14 F# 1.0E13 F# 1.0E12 F# 1.0E11
F# 1.0E10 F# 1.0E09 F# 1.0E08 F# 1.0E07 F# 1.0E06
F# 1.0E05 F# 1.0E04 F# 1.0E03 F# 1.0E02 F# 1.0E01
F# 1.0e00
16 FARRAY POWERS-OF-10E1
\ *G An array of 16 powers of ten starting at 10^0
\ ** in steps of 1.

F# 1.0E256 F# 1.0E240 F# 1.0E224 F# 1.0E208 
F# 1.0E192 F# 1.0E176 F# 1.0E160 F# 1.0E144
F# 1.0E128 F# 1.0E112 F# 1.0E96  F# 1.0E80 
F# 1.0E64  F# 1.0E48  F# 1.0E32  F# 1.0E16 
F# 1.0E00 
17 FARRAY POWERS-OF-10E16
\ *G An array of 17 powers of ten starting at 10^0
\ ** in steps of 16.

F# 1.0E-15 F# 1.0E-14 F# 1.0E-13 F# 1.0E-12 F# 1.0E-11
F# 1.0E-10 F# 1.0E-09 F# 1.0E-08 F# 1.0E-07 F# 1.0E-06
F# 1.0E-05 F# 1.0E-04 F# 1.0E-03 F# 1.0E-02 F# 1.0E-01
F# 1.0e-00
16 FARRAY POWERS-OF-10E-1
\ *G An array of 16 powers of ten starting at 10^0
\ ** in steps of -1.

F# 1.0E-256 F# 1.0E-240 F# 1.0E-224 F# 1.0E-208 
F# 1.0E-192 F# 1.0E-176 F# 1.0E-160 F# 1.0E-144
F# 1.0E-128 F# 1.0E-112 F# 1.0E-96  F# 1.0E-80 
F# 1.0E-64	F# 1.0E-48 	F# 1.0E-32 	F# 1.0E-16 
F# 1.0E00 
17 FARRAY POWERS-OF-10E-16								
\ *G An array of 17 powers of ten starting at 10^0
\ ** in steps of -16.

: RAISE_POWER	\ mant exp -- mant' exp'
\ *G Raise the power in preparation for number formatting.
  DUP >R  #15 AND ?DUP
  IF  POWERS-OF-10E1 F*  THEN
  R@ 4 rshift  #15 AND ?DUP
  IF  #16 MIN POWERS-OF-10E16 F*  THEN
  R> 8 rshift  0
  ?DO  %10^256 F*  LOOP
;

: SINK_FRACTION	\ mant exp -- mant' exp'
\ *G Reduce the power in preparation for number formatting.
  ABS DUP >R  #15 AND ?DUP
  IF  POWERS-OF-10E-1 F*  THEN
  R@ 4 rshift #15 AND ?DUP
  IF  POWERS-OF-10E-16 F*  THEN
  R> 8 rshift  0
  ?DO  %10^-256 F*  LOOP
;

variable places  8 places !	\ -- addr
\ *G Number of digits output after the decimal point.

: ROUND  	\ f1 -- f2 
\ *G Rounds least significant eight bits to 0 if higher 2 bits
\ ** are all 0s or all 1s.
  fsign >r >r
  dup 224 and 0=
  if  -256 and  then
  dup 224 and  224 =
  if  -256 and 256 +  then
  r> over 0<
  if  swap 2/  swap 1+  then
  r>
  if  fnegate  then
;

internal
: ?10PWR  	\ exp[2] -- exp[2] exp[10] 
\ *G Generate the power of ten corresponding to the power of two. INTERNAL.
  dup 0< >r
  dup abs 301 1000 */mod nip		\ l10(2)~=.3010
  r> if  negate 1-  then
;

: SIGFIGS	\ fn n -- d dec_exponent
\ *G From fn, generate a double number corresponding to n significant digits
\ ** and a decimal exponent. INTERNAL.
  >R ?10PWR R>
  SWAP DUP >R  - ?DUP IF
    DUP 0< IF
      SINK_FRACTION
    ELSE
      RAISE_POWER
    THEN
  ENDIF
  R>
;

: op-prepare	\ fn -- d exp sign
\ *G From fn, generate a double number corresponding to n significant digits,
\ ** a decimal exponent and a sign indicator (nz=negative). INTERNAL.
  FSIGN >r
  8 SIGFIGS >r
  ROUND  F>D
  r> r>
;

: .EXP  	\ exp --
\ *G Display the exponent. INTERNAL.
  [char] E emit  dup 0<			\ if -10<e<10 then print a "0" in front of exponent
  if  [char] -  else  [char] +  then
  emit
  abs  0 <# # #s #> type
;

: N#  		\ d n -- d'
\ *G Convert n digits. INTERNAL.
  0 ?do # loop
;
external

: E.  		\ n exp --
\ *G Print the f.p. number on the stack in exponential form,
\ ** x.xxxxxEyy.
  op-prepare swap >r >r			\ -- d ; R: -- sign exp
  <#  8 n#  [char] . hold  #  r>
      if  [char] -  else  bl  then
      hold
  #>
  8 places @ min 0 max + 8 - type
  r> .exp
;

: REPRESENT	\ r c-addr u -- n flag1 flag2 
\ *G Assume that the floating number is of the form +/-0.xxxxEyy.
\ ** Place the significand xxxxx at c-addr with a maximum of u digits.
\ ** Return n the signed integer version of yy. Return flag1 true
\ ** if f is negative, and return flag2 true if the results are
\ ** valid. In this implementation all errors are handled by
\ ** exceptions, and so flag2 is always true.
  2swap op-prepare >r >r                \ convert to a double
  <#  9 n#  #>                          \ generate string with max. digits
  rot min                               \ use input string size if smaller
  rot swap cmove                        \ copy string to buffer
  r> 1+                                 \ return exponent
  r>                                    \ return sign
  true                                  \ conversion always works
;

: F.  		\ f --
\ *G Print the f.p. number in free format, xxxx.yyyy, if
\ ** possible. Otherwise display using the x.xxxxEyy format.
  dup -10 27 within? 0= if
    e.  exit
  endif
  op-prepare swap >r -rot		\ -- sign d ; r: -- exp
  <#  8 r> - 0 max dup >r ?dup
      if  n#  [char] . hold  then
      #s  rot
      if  [char] -  else  bl  then
      hold
  #>
  r> dup if
    places @ -  0 max -
    dup 0
    do
      2dup + 1- c@  [char] 0 <> ?leave
      1-
    loop
  else
    drop
  then
  2dup +  1- c@  [char] . =
  if  1-  then
  type
;


\ -----------------------
\ *N Floating point input
\ -----------------------

: FLITERAL	\ Comp: r -- ; Run: -- r 
\ *G Compiles a float as a literal into the current definition.
\ ** At execution time, a float is returned. For example,
\ ** [ %PI F2* ] FLITERAL will compile 2PI as a floating point
\ ** literal. Note that FLITERAL is immediate.
  swap  [compile] literal  [compile] literal
;
IMMEDIATE

internal

: CONVERT-EXP	\ c-addr --
\ *G If the character at c-addr is 'D' convert it to 'E'. INTERNAL.
  count [char] D scan if
    [char] E swap c!
  else
    drop
  then
;

: CONVERT-FPCHAR	\ c-addr --
\ *G Convert the f.p. char '.' to the double char ',' for
\ ** conversion. INTERNAL.
  count [char] . scan if
    [char] , swap c!
  else
    drop
  then
;

: ALL-BLANKS?	\ c-addr len -- flag
\ *G Return true if string is all blanks (spaces). INTERNAL.
  true -rot
  bounds ?Do
    i c@ bl <>
    if  drop false leave  then
  Loop
;

: FCHECK	\ -- am lm ae le e-flag .-flag 
\ *G Check the input string at PAD, returning the separated
\ ** mantissa and exponent flags. The e-flag is returned true
\ ** if the string contained an exponent indicator 'E' and
\ ** the .-flag is returned true if a '.' was found. INTERNAL.
  pad count 2dup [char] E scan dup >r   \ ap lp ae le --
  tuck >r >r - r> r>                    \ am lm ae le --
  dup if  1 /string  then               \ am lm ae' le' --
  pad count [char] . scan nip
  r> swap
;

: MNUM		\ c-addr u -- d 2 | 0
\ *G Convert the mantissa string to a double number and 2. If
\ ** conversion fails, just return 0. INTERNAL.
  (integer?) dup 1 =
  if  drop s>d 2  then
;

: ENUM		\ c-addr u -- n 1 | 0 ; str as above 
\ *G Convert the mantissa string to a single number and 1. If
\ ** conversion fails, just return 0. INTERNAL.
  dup if
    (integer?)  dup 2 =
    if  2drop drop 0  then
  else
    2drop  0 1
  then
;

: *10^X		\ float dec_exponent -- float'
\ *G Generate float' = float *10^dec_exp. INTERNAL.
  DUP 0< IF
    SINK_FRACTION
  ELSE
    RAISE_POWER
  THEN
;

: FIXEXP     \ dmant exp -- mant' exp'
\ *G Convert a double integer mantissa and a single integer
\ ** exponent into a floating point number. INTERNAL.
  >R drop s>f R>  
  *10^X
;
external

-short-branches
: FNUMBER?	\ addr --  0/.../mant exp 2
\ *G Behaves like the integer version of NUMBER? except that if
\ ** the number is in F.P. format and BASE is decimal, a floating
\ ** point conversion is attempted. If conversion is successful,
\ ** the floating point number is left on the float stack and
\ ** the result code is 2.
  pad over c@ 1+ cmove                  \ copy string to pad
  fcheck drop if                        \ valid f.p. number?
    pad convert-fpchar                  \   convert f.p. char ready for conversion to a double
    enum if                             \   convert exponent ok?
      -rot  over c@ [char] , <> ?undef  \     must be at least one digit for d.p.
      mnum if                           \     convert mantissa ok?
        rot dpl @  dup 0>
	if  -  else  drop  then
        fixexp  2
      else
        drop 0
      then
    else
      2drop 0
    then
    dup #l !  -1 dpl !
  else
    2drop 2drop  pad integer?           \ try to convert as integer
  then
;
+short-branches

: >FLOAT	\ c-addr u -- r true|false 
\ *G Try to convert the string at c-addr/u to a floating point number.
\ ** If conversion is successful, flag is returned true, and a floating
\ ** number is returned on the float stack, otherwise just flag=0 is returned.
  dup pad c! pad 1+ swap cmove          \ copy string to pad
  pad count upper                       \ make string upper case
  pad convert-exp                       \ make sure exponent char is an 'E'
  fcheck or if                          \ valid f.p. number?
    pad convert-fpchar                  \   convert f.p. char ready for conversion to a double
    enum if                             \   convert exponent ok?
      -rot mnum if                      \     convert mantissa ok?
        rot dpl @  dup 0>
	if  -  else  drop  then
        fixexp  true
      else
        drop 0
      then
    else
      2drop 0
    then
    dup #l !  -1 dpl !
  else
    2drop
    case  2dup 2>r (integer?)		\   try to convert integer to f.p.
      0 of  2r@ all-blanks? dup		\    all blanks is a special case returning f.p. 0
            if  0 0 rot  then
	endof
      1 of  s>f true     endof		\    convert single integer to float
      2 of  2drop false  endof		\    can't do doubles
    endcase
    2r> 2drop
  then
;

: (F#)  	\ addr -- fn 2 | 0
\ *G The primitive for F# and F#IN below.
  fnumber?
  case
    0  of  0      endof
    1  of  s>f 2  endof
    2  of  dpl @ -1 <>                  \ see if double
           if  drop s>f  then           \ covert to fp
           2
       endof
  endcase
;

: F#IN  	\ -- fn 2 | 0
\ *G Attempts to convert a token from the input stream to a
\ ** floating-point number. Numbers in integer format will be
\ ** converted to floating-point. An indicator (0 or 2/3) is
\ ** returned in the same way as an indicator is returned by
\ ** FNUMBER?.
  query bl word (f#)
;

: F#  		\ -- [f] ; or compiles it [ state smart ]
\ *G If interpreting, takes text from the input stream and,
\ ** if possible converts it to a f.p. number on the stack.
\ ** Numbers in integer format will be converted to floating-point.
\ ** If compiling, the converted number is compiled.
  bl word (f#) ?undef  state @
  if  [compile] fliteral  then
;
IMMEDIATE


also forth definitions

: REALS  	\ -- ; allow f.p input 
\ *G Switch NUMBER? to permit floating point input using FNUMBER?.
\ ** This action can be reversed by INTEGERS. Both REALS and INTEGERS
\ ** are in the FORTH vocabulary.
  assign fnumber? to-do number?         \ switch fnumber?
;

: INTEGERS  	\ -- ; no f.p input
\ *G Switch NUMBER? to restore integer only input.
  assign integer? to-do number?         \ switch in integer?
;

previous definitions


\ -------------------------
\ *N Trigonmetric functions
\ -------------------------
\ *P N.B. All angles are in radians.

 1518500249 0 fconstant %SQR.5
 1267358949 0 fconstant %S1
1792332800 -1 fconstant %S2
 1686629716 2 fconstant PI
 1686629715 1 fconstant PI/2
 1686629716 0 fconstant PI/4
1367130550 -1 fconstant 1/PI
      180 s>f fconstant %180
1686626304 18 fconstant %SINMAX
 1488522234 0 fconstant %D8
1865280597 -1 fconstant %D9
 1677721600 8 fconstant %ALNMAX
 1549082004 1 fconstant %1/LN2
 1488522236 0 fconstant %LN2
 1660944384 7 fconstant %ALGMAX
 1783446570 2 fconstant %1/LG2
1292913983 -1 fconstant %LG2
1367130550 0 fconstant 2/PI

: DEG>RAD	\ n1 -- n2 
\ *G Convert degrees to radians.
  %180 f/ pi f*
;

: RAD>DEG	\ n1 -- n2  
\ *G convert radians to degrees.
  %180 f* pi f/
;


internal
: ?TR-ERR	\ x --
  abort" Overflow in trig. function"
;

f# -1.6666666E-1
f# 8.3333315E-3
f# -1.98409E-4
f# 2.7526E-6
f# -2.39E-8
5 farray SINARRAY

: (SIN)  	\ f1 -- f2 
  fdup fdup f* fdup 0 sinarray f*
  4 1
  do  i sinarray f+ fover f*  loop
  4 sinarray f+ f* %1 f+ f*
;

: SINFN		\ f1 -- f2
  fdup 1/pi f*                          \ ang ang/pi --
  fint fdup f>s >r			\ ang int[ang/pi] --
  pi f* f-                              \ ang-pi*int[ang/pi] --
  fdup pi/2 f>
  if  pi fswap f-  then
  (sin)
  r> 1 and
  if  fnegate  then
;

: (FSIN)	\ f1 -- f2 
  fsign  >r
  fdup %sinmax f> ?tr-err
  sinfn
  r> ?fnegate
;
external

: FSIN		\ f1 -- f2 
\ *G f2=sin(f1).
  (FSIN)
;

: FCOS		\ f1 -- f2 
\ *G f2=cos(f1).
  pi/2 f+  (fsin)
;

internal
f# 3.3333140E-1
f# 1.3339240E-1
f# 5.3374060E-2
f# 2.4565089E-2
f# 2.9005250E-3
f# 9.5168091E-3
6 farray TANARRAY

: (TAN)		\ f1 -- f2 
  fdup fdup f* fdup 0 tanarray f*
  5 1
  do  i tanarray f+ fover f*  loop
  5 tanarray f+ f* %1 f+ f*
;
external

: FTAN		\ f1 -- f2 
\ *G f2=tan(f1).
  fsign >r
  fdup %sinmax f> ?tr-err
  fdup 2/pi f* fint fswap fover pi/2 f* f-
  fdup pi/4 f> if
    pi/2 f- (tan) fswap %1 f+
  else
    (tan) fswap
  then
  f>s 1 and
  if  %1 fswap f/ fnegate  then
  r>
  if  fnegate  then
;

internal

f# 1.5707963E0
f# -2.1459880E-1
f# 8.8978987E-2
f# -5.0174305E-2
f# 3.0891881E-2
f# -1.7088126E-2
f# 6.6700901E-3
f# -1.2624911E-3
8 farray ASINARRAY

: ARCFN  	\ f1 -- f2 
  fdup 0 asinarray f*
  7 1
  do  i asinarray f+ fover f*  loop
  7 asinarray f+
  fswap %1 fswap f- fsqr f*
;

f# -3.3333145E-1
f# 1.9993551E-1
f# -1.4208899E-1
f# 1.0656264E-1
f# -7.5289640E-2
f# 4.2909614E-2
f# -1.6165737E-2
f# 2.8662257E-3
8 farray ATANARRAY

: ATFN		\ f1 -- f2 
  fdup fdup f* fdup 0 atanarray f*
  7 1
  do  i atanarray f+ fover f*  loop
  7 atanarray f+ f* %1 f+ f*
;

external

: FASIN  	\ f1 -- f2 
\ *G f2=arcsin(f1).
  fsign >r
  arcfn pi/2 fswap f-
  r> if  fnegate  then
;

: FACOS  	\ f1 -- f2 
\ *G f2=arccos(f1).
  fsign >r
  arcfn
  r> if  pi fswap f-  then
;

: FATAN  	\ f1 -- f2 
\ *G f2=arctan(f1).
  fsign >r
  fdup %1 f> if
    %1 fswap f/  atfn pi/2 fswap f-
  else
    atfn
  then
  r> if  fnegate  then
;


\ ----------------------------------
\ *N Power and logarithmic functions
\ ----------------------------------

internal
f# 9.9999642E-1
f# -4.9987412E-1
f# 3.3179903E-1
f# -2.4073381E-1
f# 1.6765407E-1
f# -9.5329390E-2
f# 3.6088494E-2
f# -6.4535442E-3
8 farray LOGARRAY
external

: FLN  		\ f1 -- f2 
\ *G Take the logarithm of f1 to base e and return the result.
  fdup f0> 0=
  abort" Invalid argument to FLN/FLOG"
  1 swap 1- >r  %1 f-
  fdup 0 logarray f*
  7 1
  do  i logarray f+ fover f*  loop
  7 logarray f+ f*  r> s>f  %d8 f* f+
;

: FLOG  	\ f1 -- f2 
\ *G Take the logarithm of f1 to base 10 and return the result.
  fln %d9 f*
;

internal

f# 1.1512928E0
f# 6.6273088E-1
f# 2.5439357E-1
f# 7.2951737E-2
f# 1.7421120E-2
f# 2.5549180E-3
f# 9.3264267E-4
7 farray 10ARRAY

: (10^X)  	\ f1 -- f2 
  fdup 0 10array f*
  6 1
  do  i 10array f+ fover f*  loop
  6 10array f+ f*
  %1 f+  fdup f*
;

f# -1.0000000E0
f# 4.9999992E-1
f# -1.6666530E-1
f# 4.1657348E-2
f# -8.3013598E-3
f# 1.3298820E-3
f# -1.4131610E-4
7 farray E-ARRAY

: (E^X)  	\ f1 -- f2 
  fdup 0 e-array f*
  6 1
  do
    i e-array f+ fover f*
  loop
  6 e-array f+ f*
  %1 f+ %1 fswap  f/
;

external

: FE^X  	\ f1 -- f2
\ *G f2=e^f1.
  fsign >r
  fdup %alnmax f> abort" Overflow in FE^X"
  fdup  %1/ln2 f*
  fdup f>s >r
  fint %ln2 f* f-
  (e^x)
  r> +
  r> if  %1 fswap f/  then
;

: F10^X  	\ f1 -- f2 
\ *G f2=10^f1
  fsign >r
  fdup %algmax f> abort" overflow in f10^x"
  fdup %1/lg2 f*
  fdup f>s >r
  fint %lg2 f* f-
  (10^x)
  r> +
  r> if  %1 fswap f/  then
;

: FX^N  	\ x-real n-integer -- fx^n
\ *G fx^n=x^n where x is a float and n is an integer.
  dup 0= if
    drop fdrop  %1  exit
  then
  dup >r
  abs >r  %1 fswap  r>
  begin
    dup 0>
  while
    dup 2 / dup 2* rot -
    if
      >r fswap fover f* fswap r>
    then
    >r fdup f* r>
  repeat
  drop fdrop
  r> 0<
  if  %1 fswap f/  then
;

: FX^Y  	\ x-real y-real -- fn 
\ *G fn=X^Y where Y and Y are both floats.
  fdup f0= if
    fdrop fdrop  %1  exit
  then
  fdup fdup fint f- f>r
  f>s >r fdup r>
  fx^n fswap
  fr> fdup f0<> if
    fover f0< abort" Result of FX^Y is complex"
    fswap fdup f0< >r
    fabs flog f*  f10^x
    r> if  %1 fswap f/  then
    f*
  else
    fdrop fdrop
  then
;

: FSQR  	\ f1 -- f2 ; FSQR by Heron's formula 
\ *G F2=sqrt(f1) by Heron's formula.
  over 0= if				\ check 0
    exit
  then
  over 0<				\ check -ve
  abort" square root of negative no.!"
  >r 0                                	\ .5 <=f<=1
  fdup %s1 f*  %s2 f+			\ HART est Y0
  fover fover f/ f+
  fswap fover f/ fswap 2-  f+
  r@ 1 and  dup r> + >r
  if  %sqr.5 f*  then
  r> 2/ +
;


\ ======
\ *> ###
\ ======

