\ TSEPOS 194 9 25 \
\ =======================================================================
( *> GENIO                                                              )
( **                                                                    )
( *N File Device                                                        )
( **                                                                    )
( *P This Generic IO Device operates on a disk file for input           )
( ** and/or output.                                                     )
( **                                                                    )
( *P In order to create a device use the FILEDEV: definition given      )
( ** later. FILEDEV: is compatible with ProForth 2.                     )
( **                                                                    )
( *P When opening a file device the parameters to OPEN-GEN have the     )
( ** following meaning:                                                 )
( *E    ADDR            Address of string for filename.                 )
( **    LEN             Length of string at ADDR.                       )
( **    ATTRIBS         Open flags. These match the ANS r/o r/w etc.    )
\ =======================================================================

                MODULE  GENIO-FILE

struct filedev-sid      \ -- len ; defines the SID of a file device
  gen-sid +                             \ reuse field names of GEN-SID
  int filedev-fhandle                   \ FILE handle
  int filedev-scratch
  int filedev-scratch2
end-struct

: >zaddr                \ addr len zaddr -- ; Copy string as 0 terminated
  2dup + 0 swap c!                      \ store 0 terminator
  swap move                             \ and move string data
;

: filedev-open          { caddr len fam sid | zname[ MAX_PATH ] -- sid ior }
  caddr len zname[ >zaddr               \ Make 0 terminated name string
  zname[ fam NULL NULL                  \ open or create file
  OPEN_ALWAYS NULL NULL CreateFile
  dup sid filedev-fhandle !             \ store handle in private data
  INVALID_HANDLE_VALUE =                \ set ior
  sid swap
;

: filedev-close         \ sid -- ior
  dup filedev-fhandle @ >r              \ get fhandle and keep
  INVALID_HANDLE_VALUE swap             \ write default (INVALID) back
  filedev-fhandle !
  r> Close-File
;

: filedev-read          \ addr len sid -- ior
  filedev-fhandle @ read-file nip
;

: filedev-write         \ addr len sid -- ior
  filedev-fhandle @ write-file
;

: filedev-key           \ sid -- char
  dup dup filedev-scratch 1 rot filedev-read drop
  filedev-scratch c@
;

: filedev-key?          \ sid -- flag
  filedev-fhandle @ dup >r
  file-size drop        \ ud1 --
  r> file-position drop \ ud1 ud2 --
  d= 0=
;

: filedev-ekey          \ sid -- echar
  1 abort" EKEY operation not supported for FILEDEV (GENERIC-IO)"
;

: filedev-ekey?         \ sid -- flag
  1 abort" EKEY? operation not supported for FILEDEV (GENERIC-IO)"
;

: filedev-accept        \ addr len sid -- len'
  1 abort" ACCEPT operation not supported for FILEDEV (GENERIC-IO)"
;

: filedev-emit?         \ sid -- flag
  drop TRUE
;

: filedev-type          \ addr len sid --
  filedev-write drop
;

: filedev-emit          \ char sid --
  swap over filedev-scratch c!  \ sid --
  dup filedev-scratch 1 rot filedev-type
;

: filedev-cr            \ sid --
  #13 over filedev-emit
  #10 over filedev-emit
  drop out off
;

: filedev-lf            \ sid --
  #10 swap filedev-emit
;

: filedev-ff            \ sid -- ; page/cls on display devices
  #14 swap filedev-emit
;

: filedev-bs            \ sid -- ; destructive on display devices
  #8 swap filedev-emit
;

: filedev-bell          \ sid -- ; audible beeper
  1 abort" BELL operation not supported for FILEDEV (GENERIC-IO)"
;

: filedev-setpos        \ x y mode sid -- ior
  1 abort" SETPOS operation not supported for FILEDEV (GENERIC-IO)"
;

: filedev-getpos        \ mode sid -- x y ior
  1 abort" GETPOS operation not supported for FILEDEV (GENERIC-IO)"
;

: filedev-ioctl         \ addr len fn sid -- ior
  1 abort" IOCTL operation not supported for FILEDEV (GENERIC-IO)"
;

: filedev-flushOP       \ sid -- ior
  drop 0
;

: filedev-RFU           \ --
  1 abort" RFU operation not supported for FILEDEV (GENERIC-IO)"
;

: filedev-init          \ addr len sid -- ior
  1 abort" INIT operation not supported for FILEDEV (GENERIC-IO)"
;

: filedev-term          \ sid -- ior
  1 abort" TERM operation not supported for FILEDEV (GENERIC-IO)"
;

: filedev-config        \ sid -- ior ; produces a dialog
  1 abort" CONFIG operation not supported for FILEDEV (GENERIC-IO)"
;

create filedev-vectors
  ' filedev-open ,
  ' filedev-close ,
  ' filedev-read ,
  ' filedev-write ,
  ' filedev-Key ,
  ' filedev-Key? ,
  ' filedev-Ekey ,
  ' filedev-EKey? ,
  ' filedev-Accept ,
  ' filedev-Emit ,
  ' filedev-Emit? ,
  ' filedev-Type ,
  ' filedev-CR ,
  ' filedev-LF ,
  ' filedev-FF ,
  ' filedev-BS ,
  ' filedev-Bell ,
  ' filedev-setpos ,
  ' filedev-Getpos ,
  ' filedev-Ioctl ,
  ' filedev-FlushOP  ,
  ' filedev-RFU ,
  ' filedev-Init ,
  ' filedev-Term ,
  ' filedev-Config ,

\ =======================================================================
( *H Device Creation                                                    )
\ =======================================================================

: filedev:        \ "name" -- ; Exec: -- sid
( *G Create a File based Generic IO device in the dictionary.           )
  create
    -1 ,  filedev-vectors ,             \ handle, vector
    INVALID_HANDLE_VALUE ,              \ file-handle
    0 , 0 ,                             \ scratch
;



                EXPORT filedev:
                END-MODULE

\ =======================================================================
( *> ###                                                                )
\ =======================================================================

