\ hex converter code for PC PowerForth


hex

: $move         \ src dest --
  over c@ 1+ move
;

create ip-buffer  0100 allot

: ed-$          \ buffer -- len ; input counted string to buffer
  ip-buffer 0100 expect
  ip-buffer span @ rot place
  span @
;

: (action-of)
  r> dup 2+ >r @ >body @  ;

: action-of
  state @
  IF    compile (action-of)
  ELSE  ' >body @
  ENDIF
;  IMMEDIATE


HEX                                                             

010  constant line-length               \ no of bytes in a line
0200 constant buffer-length             \ length of the file buffer

variable old-op                         \ previous op-handle
variable old-ip                         \ previous ip-handle

pcb bin-source                          \ the source file name
pcb hex-dest                            \ the destination file name

create load-buffer  buffer-length allot \ the file buffer

create $bin-name                        \ string for name of bin file
  40 allot  $bin-name off                       \ default
create $hex-name                        \ string for name of hex file
  40 allot  "" TARGET.HEX" $hex-name $move      \ default
create addr$                            \ string to base address
  20 allot  "" 0000" addr$ $move                \ default

variable checksum                       \ checksum so far
variable length
variable record                                               
variable target-address       
variable host-address         

variable record#
                                                                
: bin-ascii     \ 8b -- char ; convert nibble to ascii char
  0f and  dup 9 >               \ if 0..9
  if
    037 +
  else
    030 +                       \ if a..f
  endif  
;                
                                                                
: byte-ascii            \ byte -- lsc msc ; convert byte to ascii
  dup bin-ascii
  swap 010 /  bin-ascii  
;         
                                                                
: sum                   \ data -- ; advance checksum
  checksum +!  
;                                  

: byte-target           \ byte -- ; send one byte
  0ff and dup sum                                 
  byte-ascii emit emit  
;           
                                                                
: send-header           \ -- ; send line header
  ascii : emit
  checksum off  
;                
                                                                
: record-length         \ -- ; send record length
  length @ byte-target  
;                         
                                                                
: start-address         \ -- ; send start address
  target-address @ 0 0100 um/mod                  
  byte-target  byte-target  
;                     
                                                                
: type-of-record        \ -- ; send type (line/end)
  record @ byte-target  
;                        
                                                                
: data                  \ -- ; send data
  host-address @ length @ +   host-address @      
  do
    i c@ byte-target                    \ send each byte in turn
  loop  
;                 
                                                                
: send-check            \ -- ; send checksum
  checksum @ negate byte-target                   
  checksum off  
;                                 
                                                                
: send-record           \ -- ; send one record
  send-header                   \ the header
  record-length                 \ the length
  start-address                 \ the address
  type-of-record                \ the type
  data                          \ the data
  send-check                    \ the checksum
  cr                            \ ready for next line
;        
                                                                
: send-last-record      \ --; send final record
  length off  target-address off  1 record !      
  send-header
  record-length
  start-address              
  type-of-record
  send-check
  cr
  record off  
; 

create dot  ascii . c,

: .record#              \ -- ; screen progress record
\  dot 1 old-op @ write-path 2drop 
\  ." record: " record# ?                \ get record#
;
                                                                
: send-blocks           \ -- ; send the data
  record# off                           \ start at record 0
  begin
    load-buffer buffer-length bin-source handle read-path   \ read some  binary
    0= swap  0<>  and                   \ see if we read ok
   while 
    load-buffer host-address !          \ while there is source file
    line-length length !                \ set the line length
    buffer-length line-length / 0       \ divide into lines per buffer
    do                                  \ for each record in a buffer
      1 record# +!                      \ increment count
      .record#                          \ show progress
      send-record                       \ send the record
      line-length host-address +!       \ increment the address
      line-length target-address +!     \ and this address
    loop                                                  
  repeat                                \ next buffer
;                                                     

\ -----End of Page-----
\ -- R.G.G -- 9/1/92 --  Intel hex down loader

: open-source           \ -- t/f ; source opened
  $bin-name bin-source set-pathname     \ set filename
  bin-source open-path-pcb dup          \ try to open file
  if
    cr ."  Can't open source file"      \ say so
  endif
  0=
;

: create-dest           \ -- t/f ; file created
  hex-dest create-path-pcb              \ try to create
  0=                                    \ true for success
;

: open-dest             \ -- t/f ; true:- create the file
  $hex-name hex-dest set-pathname               \ set filename
  hex-dest open-path-pcb                        \ try to open file      
  0=                                            \ if it exists          
  hex-dest handle close-path                    \ try to close file  
  0= and                                        \ if can do both
  if
     cr ."  Hex file " hex-dest .pcb ."  already exists."     
     cr ."  Overwrite it (y/n)?"                \ should we overwrite  
     y/n?                                       \ yes to create file
  else
     -1                                         \ may create the file         
  endif                                                         
;

hex

: down-load         \ -- ; download the file
  open-source                   \ open source file
  if
    open-dest                   \ test destination file
    if
      create-dest               \ create it
      if
        op-handle @ old-op !
        hex-dest handle op-handle !             \ switch to file i/o
        send-blocks                             \ actually send data
        bin-source handle close-path drop       \ close source file
        send-last-record                        \ send last record
        hex-dest handle close-path drop         \ close hex file
        old-op @ op-handle !
      endif
    endif
  endif
;                                      

: addr$>number          \ -- t/f ; make address number
  00                                            \ dummy fail flag
\  -1 addr$ c+!                                 \ decrement count to just number
  base @ >r hex
  addr$ number?                                 \ try to convert
  r> base !
  case
    00 of cr ."  Invalid number"        endof   \ bad number
    01 of      target-address ! drop -1 endof   \ ok - single
    02 of drop target-address ! drop -1 endof   \ ok - double
  endcase
;

decimal

\ headers

: (bin2hex)             \ -- ; do the whole thing
  cr ." MPE image to hex file converter version 1.100 January 1999"
  cr
  cr ."  Image file name:          "            \ prompt
  $bin-name ed-$                                \ get input file name
  if
    cr ."  Hex file name:            "          \ next prompt
    $hex-name ed-$                              \ get string
    if
      cr ."  Image load address (hex): "        \ next prompt
      addr$ ed-$                                \ get address as string
      if
        addr$>number                            \ convert to number
        if                                      \ if valid number
          down-load                             \ then finally download file
        endif
      endif
    endif
  endif
;

: binhex        \ --
  ansi-drivers
  (bin2hex)
  bye
;

assign binhex to-do loader
save binhex
bye

