\ screen to text file conversion utility

pto

This code released to convert screen files to text files.

(c) MicroProcessor Engineering Ltd., 1991
133 Hill Lane
Southampton
SO1 5AF

Phone: (+44) 703 631441
fax:   (+44) 703 339691

This code to be compiled on PC PowerForth or Modular Forth v3.4x

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  variables, etc

create source-buff  64 allot		\ source buffer

pcb DEST-file				\ destination file

variable fth-page#			\ page number
variable fth-line#			\ line number

variable #blank-lines			\ send a blank line?

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  bye-bye   

: bye-bye		\ -- ; quit application
  0 cooked			\ reset i/p stream
  1 cooked			\ reset o/p stream
  f#04c				\ bye out
\  .s abort			\ debug only
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  close-files

: close-files		\ -- ; close all files
  dest-file handle close-path		\ close this one
  drop					\ dont care
  screen-pcb handle close-path		\ close this one
  drop					\ dont care
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  write-dest

: write-dest           \ len -- ; write to o/p file
  source-buff swap dest-file handle write-path	\ write code
  nip 0<>					\ if error..
  if
    cr ." Error writing destination file: aborting"	\ say so
    close-files					\ close files
    bye-bye					\ quit
  endif
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  read-source

: read-source           \ -- t/f ; read block from source file
  source-buff 64 screen-pcb handle read-path	\ get one line
  0<> swap 64 <> or				\ if an error..
  dup
  if
    cr ." End of source file: closing"		\ say so
    close-files					\ close files
    drop					\ tidy up
    bye-bye					\ quit
  endif
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  open-source open-files

: open-source		\ -- ; try to open source file
  "" .SCR" screen-pcb set-ext
  screen-pcb open-path-pcb
  if
    cr ." Can't open source file: aborting"
    cr ." Use SCR2FTH /? for help"
    bye-bye
  endif
;

: open-files            \ -- ; open source and dest files
  open-source
  dest-file create-path-pcb
  if cr ." Can't open destination file: aborting" bye-bye endif
;

\ Try to open source and create destination, abort if cant

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  send-line?

: send-line?		\ len -- len t/f ; write if first blank line
  dup 0=			\ is length zero?
  if
    1 #blank-lines +!		\ increment count
    #blank-lines @
    1 =
    if -1  else 0  endif	\ only write if #1
  else
    #blank-lines off		\ write if non-blank
    -1
  endif
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  write-line

: write-line		\ -- ; write a line to dest. file
  source-buff 64 -trailing nip		\ actual number of chars
  send-line?				\ should we send or not
  if
    write-dest				\ write text
    13 source-buff c! 10 source-buff 1+ c!	\ make CR
    2 write-dest			\ write CR
  else
    drop				\ else tidy up
  endif
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  page-break

: page-break		\ -- ; write a page break
  12 source-buff c!		\ make FF
  1 write-dest			\ write FF
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  tidy-file

: tidy-file		\ -- ; manage counters, and inform user
  1 fth-line# +!			\ increment line#
  fth-line# @ 16 =			\ if end of page
  if
    1 fth-page# +!			\ increment page number
    fth-line# off			\ reset line counter
    fth-page# @ 1 <>			\ if not page 0/1
    if page-break endif			\ do a page-break
    13 emit ." Screen/page " fth-page# ?	\  say 'next page'
  endif
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  convert

: convert		\ -- ; do the actual conversion
  open-files			\ try to open source
  fth-page# off fth-line# off	\ reset counters
  begin
    read-source			\ get some source
    0=				\ while successful
  while
    write-line			\ write to destination
    tidy-file			\ do page breaks, etc
  repeat			\ and the next
  close-files			\ close the files
  bye-bye			\ and quit
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  signon

: signon		\ -- ; say who we are
  ." Screen File to text file converter, v2.20"
  cr ." (c) MicroProcessor Engineering Ltd, 1990"
  cr cr
;

\ Say what app is running, including version number

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  .help ?help

: .help			\ -- ; show help screen
     ." Usage: SCR2FTH <source> <dest>
  cr
  cr ." <Source> is the name of a screen file to be converted to a text file."
  cr ." <Dest> is the name of the destination file."
  cr
  cr ." If no destination file is supplied, CONVERT.FTH will be generated."
  cr ." If no source file extension is supplied, .SCR will be used."
  cr ." If no destination extension is supplied, .FTH will be used."
  cr
;

: ?help			\ addr -- ; do they want help?
  "" /?" $=					\ have asked for help?
  if
    .help					\ show help screen
    bye-bye					\ quit
  endif
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  get-source get-dest

: get-source		\ -- ; get the name of the source file
  bl word dup screen-pcb set-pathname	\ get and make name of source
  ?help					\ do they want help
  open-source				\ try to open it
;

: get-dest		\ -- ; get name of dest. file
  bl word  dup c@ 0=				\ get dest.
  if						\ if none ...
    ." No destination file supplied: using CONVERT.FTH"
    cr drop					\ use default
    "" CONVERT.FTH" dest-file set-pathname	\ set default
  else						\ if supplied ...
    dest-file set-pathname			\ or supplied
    "" .FTH" dest-file set-ext			\ set .FTH
  endif
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  info

: info			\ -- ;  show what is about to happen
  ." Converting " screen-pcb .pcb	\ first message
  ."  to " dest-file .pcb
  cr cr					\ tidy screen
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  it

 : it                    \ -- ; the whole thing
  signon	 		\ what it is
  get-source	 		\ get the source name
  get-dest	 		\ get name of dest file
  info				\ show what is about to happen
  convert	 		\ do it
;

\ -----End of Page-----
\ -- R.G.G -- 10/12/91 --  assignments

assign it to-do loader          \ it will be done
assign noop to-do cls           \ with no cls

save scr2fth			\ save to disc

bye                             \ ready to try it
