\ SPLITIMG.FTH - split image file into separate files

((
Copyright (c) 2001
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20010922 SFP001 First release
))

max_path 1+ buffer: ifile	\ -- addr ; input file name
  z" prog.img" ifile zmove
max_path buffer: ofile		\ -- addr ; output file name
  z" prog.evn" ofile zmove

0 value img		\ -- addr ; points to memory image file
0 value imglen		\ -- len ; size of memory image

8 constant #fids		\ -- n ; 8 for 64 bit buses
#fids cells buffer: fids	\ -- addr ; table of fileids
  
: (LoadImage)	\ c-addr u -- addr len ior
\ *G Open a file and read it into a memory buffer. ior is returned
\ ** false for success. For failure, the returned value addr/len is
\ ** meaningless.
  2dup ifile zplace
  r/o Open-File ?dup
  if  exit   endif
\ -- fileid

  dup file-size dup if			\ -- fileid dfilesize ior
    >r
    rot close-file drop	
    r>  exit				\ -- dfilesize ior
  endif
  2drop					\ discard ior and upper size

  dup aligned allocate if		\ -- fileid size addr ior ; force size to cells
    nip
    2 pick Close-File drop  exit	\ -- fileid size ior
  then
\ -- fileid size addr

  dup >r				\ stash addr
  swap rot dup >r                      	\ -- addr size fileid ; r: -- addr fileid
  read-file  r> close-file drop		\ -- size ior ; r: -- addr
  r> -rot				\ -- addr size ior
;

: LoadImage	\ c-addr u -- 
\ *G Open a file and read it into a memory buffer.
\ ** The image buffer address and size are in the
\ ** VALUEs IMG and IMGLEN.
  (LoadImage) if
    2drop 0 0 
    cr ." Error loading image file"
  endif
  -> imglen  -> img
;

: ?noImage	\ -- flag
\ *G Returns true if no image file has been loaded
\ ** or the image file is of zero length.
  img imglen or 0=
  dup if
    cr ." No valid image loaded"
  endif
;

: hasExt?	\ c-addr len -- n
\ *G N is distance back to first '.'. Any '\' character causes
\ ** the search to be abandoned.
  tuck  +  swap	1+			\ ^last+1 len+1
  1 ?do
    dup i - c@				\ ^last+1 char
    dup [char] \ =
    if  2drop 0  unloop  exit  endif	\ no extension
    [char] . =
    if  drop  i  unloop  exit  endif	\ '.' found
  loop
  drop  0				\ no '\' or '.'
;

4 constant extSize	\ -- n ; size of extension

create Ext64/16
  ", .X01.X23.X45.X67"
create Ext64/8
  ", .XB0.XB1.XB2.XB3.XB4.XB5.XB6.XB7"
create Ext32/16
  ", .L01.L23"
create Ext32/8
  ", .LB0.LB1.LB2.LB3"
create Ext16/8
  ", .EVN.ODD"
create Ext24/8
  ", .1ST.2ND.3RD"

: CloseOpFiles	\ n -- ; close n output files
  0 ?do
    fids i cells + dup @ ?dup
    if  close-file drop  endif
    0 swap !
  loop
;

: MakeOpFiles	\ table n --
  fids #fids cells erase		\ wipe FID table
\ *G Make n output files using extension table
  0 ?do
    0 ofile c!				\ null output file
    ifile zcount 2dup hasExt? -		\ remove extension if any
    ofile zplace			\ set o/p file
    dup 1+  i extSize * +  extsize	\ addr/len of extension
    ofile appendz			\ add to file
    ofile zcount r/w create-file if
      drop  i CloseOpFiles
      cr ." Cannot create o/p file "
      ofile zcount type
      -1 throw
    endif
    fids i cells + !			\ stash fileid
  loop
  drop
;

: writefid	\ addr len n --
\ *G write addr len to the nth file in the FID table.
\ ** Throw on error.
  cells fids + @ write-file throw
;

: Write64/16	\ -- ; write image to output files
  ?noImage if  exit  endif
  Ext64/16 4 MakeOpFiles
  img imglen bounds ?do
    i     2 0 writefid
    i 2 + 2 1 writefid
    i 4 + 2 2 writefid
    i 6 + 2 3 writefid
  8 +loop
  4 CloseOpFiles
;

: Write64/8	\ -- ; write image to output files
  ?noImage if  exit  endif
  Ext64/8 8 MakeOpFiles
  img imglen bounds ?do
    i     1 0 writefid
    i 1 + 1 1 writefid
    i 2 + 1 2 writefid
    i 3 + 1 3 writefid
    i 4 + 1 4 writefid
    i 5 + 1 5 writefid
    i 6 + 1 6 writefid
    i 7 + 1 7 writefid
  8 +loop
  8 CloseOpFiles
;

: Write32/16	\ -- ; write image to output files
  ?noImage if  exit  endif
  Ext32/16 2 MakeOpFiles
  img imglen bounds ?do
    i    2 0 writefid
    i 2+ 2 1 writefid
  4 +loop
  2 CloseOpFiles
;

: Write32/8	\ -- ; write image to output files
  ?noImage if  exit  endif
  Ext32/8 4 MakeOpFiles
  img imglen bounds ?do
    i     1 0 writefid
    i 1 + 1 1 writefid
    i 2 + 1 2 writefid
    i 3 + 1 3 writefid
  4 +loop
  4 CloseOpFiles
;

: Write16/8	\ -- ; write image to output files
  ?noImage if  exit  endif
  Ext16/8 2 MakeOpFiles
  img imglen bounds ?do
    i    1 0 writefid
    i 1+ 1 1 writefid
  2 +loop
  2 CloseOpFiles
;

: Write24/8	\ -- ; write image to output files
  ?noImage if  exit  endif
  Ext24/8 3 MakeOpFiles
  img imglen bounds ?do
    i     1 0 writefid
    i 1 + 1 1 writefid
    i 2 + 1 2 writefid
  4 +loop
  3 CloseOpFiles
;

: freeImage	\ -- ; release image memory
  img if
    img free drop
    0 -> imglen
  endif
;
' freeImage AtCold

: Help		\ -- ; console help message
  cr ." Use in the form:"
  cr ."   SPLITIMG /f <InputImageFileName> /optype"
  cr ." where /optype is one of the following:"
  cr ."   /64/16  convert 64 bit data to four 16 bit images"
  cr ."           wth the extensions .X01 .X23 .X45 .X67"
  cr ."   /64/8   convert 64 bit data to eight 8 bit images"
  cr ."           wth the extensions .XB0 .XB1 .. .XB7"
  cr ."   /32/16  convert 32 bit data to two 16 bit images"
  cr ."           wth the extensions .L01 .L23"
  cr ."   /32/8   convert 32 bit data to four 8 bit images"
  cr ."           wth the extensions .LB0 .LB1 .B2 .B3"
  cr ."   /16/8   convert 16 bit data to two 8 bit images"
  cr ."           with the extensions .EVN .ODD"
  cr ."   /24/8   convert 24 bit data to three 8 bit images"
  cr ."           with the extensions .1ST .2ND .3RD"
  cr
  cr ."   /H      these notes"
  cr ."   /?      these notes"
  cr
;

vocabulary switches
also switches definitions

: /f		\ "<filename>" --
  GetPathSpec LoadImage			\ set and load input file
;

\ --
: /64/16	Write64/16  ;
: /64/8		Write64/8  ;
: /32/16	Write32/16  ;
: /32/8		Write32/8  ;
: /16/8		Write16/8  ;
: /24/8		Write24/8  ;

previous definitions

: WinMain	\ hinst hpint lpstrcmd nshow -- flag
  cr ." SPLITIMG v1.0  copyright (c) 2001 MicroProcessor Engineering Ltd."
  drop nip nip			\ just leave pointer to command line
  ?dup if
    zcount bl scan		\ step over SPLITIMG itself
    2dup bl skip nip 0=
    if  help  endif
    only switches		\ only process switches
    ['] evaluate catch if	\ stop errors entering Forth
      cr ." Error processing command line"
    then
  else
    help
  then
  0
;
assign WinMain to-do EntryPoint

